Package: actuar
Type: Package
Title: Actuarial Functions and Heavy Tailed Distributions
Version: 3.3-7
Date: 2026-03-02
Authors@R: c(person("Vincent", "Goulet", role = c("cre", "aut"),
 	            email = "vincent.goulet@act.ulaval.ca",
		    comment = c(ORCID = "0000-0002-9315-5719")),
	     person("Sébastien", "Auclair", role = "ctb"),
	     person("Jérémy", "Déraspe", role = "ctb"),
	     person("Christophe", "Dutang", role = "aut",
	            email = "dutang@ceremade.dauphine.fr",
                    comment = c(ORCID = "0000-0001-6732-1501")),
	     person("Walter", "Garcia-Fontes", role = "ctb",
	            email = "walter.garcia@upf.edu"),
	     person("Nicholas", "Langevin", role = "ctb"),
	     person("Xavier", "Milhaud", role = "ctb"),
	     person("Tommy", "Ouellet", role = "ctb"),
	     person("Alexandre", "Parent", role = "ctb"),
	     person("Mathieu", "Pigeon", role = "aut",
	            email = "pigeon.mathieu.2@uqam.ca"),
	     person("Louis-Philippe", "Pouliot", role = "ctb"),
     	     person("Jeffrey A.", "Ryan", role = "aut",
	            email = "jeff.a.ryan@gmail.com",
		    comment = "Package API"),
	     person("Robert", "Gentleman", role = "aut",
	            comment = "Parts of the R to C interface"),
	     person("Ross", "Ihaka", role = "aut",
	            comment = "Parts of the R to C interface"),
	     person(given = "R Core Team", role = "aut",
	            comment = "Parts of the R to C interface"),
	     person(given = "R Foundation", role = "aut",
	            comment = "Parts of the R to C interface"))
Description: Functions and data sets for actuarial science:
  modeling of loss distributions; risk theory and ruin theory;
  simulation of compound models, discrete mixtures and compound
  hierarchical models; credibility theory. Support for many additional
  probability distributions to model insurance loss size and
  frequency: 23 continuous heavy tailed distributions; the
  Poisson-inverse Gaussian discrete distribution; zero-truncated and
  zero-modified extensions of the standard discrete distributions.
  Support for phase-type distributions commonly used to compute ruin
  probabilities. Main reference: <doi:10.18637/jss.v025.i07>.
  Implementation of the Feller-Pareto family of distributions:
  <doi:10.18637/jss.v103.i06>.
Depends: R (>= 4.1.0)
Imports: stats, graphics, expint
LinkingTo: expint
Suggests: MASS
License: GPL (>= 2)
URL: https://gitlab.com/vigou3/actuar
BugReports: https://gitlab.com/vigou3/actuar/-/issues
Encoding: UTF-8
LazyData: yes
Classification/MSC-2010: 62P05, 91B30, 62G32
NeedsCompilation: yes
Packaged: 2026-03-02 22:54:40 UTC; vincent
Author: Vincent Goulet [cre, aut] (ORCID:
    <https://orcid.org/0000-0002-9315-5719>),
  Sébastien Auclair [ctb],
  Jérémy Déraspe [ctb],
  Christophe Dutang [aut] (ORCID:
    <https://orcid.org/0000-0001-6732-1501>),
  Walter Garcia-Fontes [ctb],
  Nicholas Langevin [ctb],
  Xavier Milhaud [ctb],
  Tommy Ouellet [ctb],
  Alexandre Parent [ctb],
  Mathieu Pigeon [aut],
  Louis-Philippe Pouliot [ctb],
  Jeffrey A. Ryan [aut] (Package API),
  Robert Gentleman [aut] (Parts of the R to C interface),
  Ross Ihaka [aut] (Parts of the R to C interface),
  R Core Team [aut] (Parts of the R to C interface),
  R Foundation [aut] (Parts of the R to C interface)
Maintainer: Vincent Goulet <vincent.goulet@act.ulaval.ca>
Repository: CRAN
Date/Publication: 2026-03-02 23:50:09 UTC

### C code
useDynLib(actuar, .registration = TRUE, .fixes = "C_")

### Imports
import(expint)                          # for C code
import(stats, graphics)
importFrom(utils, head, tail)

### Exports
export(
    ## Credibility theory
    cm,
    ## Simulation of insurance data
    rcompound, rcomppois, rmixture,
    rcomphierarc, simul, severity, unroll,
    ## Risk theory
    aggregateDist, CTE, TVaR, discretize, discretise, VaR, adjCoef, ruin,
    ## One parameter distributions
    dinvexp, pinvexp, qinvexp, rinvexp, minvexp, levinvexp,
    mexp, levexp, mgfexp,
    dlogarithmic, plogarithmic, qlogarithmic, rlogarithmic,
    dztpois, pztpois, qztpois, rztpois,
    dztgeom, pztgeom, qztgeom, rztgeom,
    ## Two parameter distributions
    munif, levunif, mgfunif,
    mnorm, mgfnorm,
    mbeta, levbeta,
    mgamma, levgamma, mgfgamma,
    mchisq, levchisq, mgfchisq,
    dinvgamma, pinvgamma, qinvgamma, rinvgamma, minvgamma, levinvgamma, mgfinvgamma,
    dinvparalogis, pinvparalogis, qinvparalogis, rinvparalogis, minvparalogis, levinvparalogis,
    dinvpareto, pinvpareto, qinvpareto, rinvpareto, minvpareto, levinvpareto,
    dinvweibull, pinvweibull, qinvweibull, rinvweibull, minvweibull, levinvweibull,
    dlgompertz, plgompertz, qlgompertz, rlgompertz, mlgompertz, levlgompertz, # aliases
    dlgamma, plgamma, qlgamma, rlgamma, mlgamma, levlgamma,
    dllogis, pllogis, qllogis, rllogis, mllogis, levllogis,
    mlnorm, levlnorm,
    dparalogis, pparalogis, qparalogis, rparalogis, mparalogis, levparalogis,
    dpareto, ppareto, qpareto, rpareto, mpareto, levpareto,
    dpareto1, ppareto1, qpareto1, rpareto1, mpareto1, levpareto1,
    mweibull, levweibull,
    ## minvGauss, levinvGauss, mgfinvGauss, [defunct v3.0-0]
    dgumbel, pgumbel, qgumbel, rgumbel, mgumbel, mgfgumbel,
    dinvgauss, pinvgauss, qinvgauss, rinvgauss, minvgauss, levinvgauss, mgfinvgauss,
    dztnbinom, pztnbinom, qztnbinom, rztnbinom,
    dztbinom, pztbinom, qztbinom, rztbinom,
    dzmlogarithmic, pzmlogarithmic, qzmlogarithmic, rzmlogarithmic,
    dzmpois, pzmpois, qzmpois, rzmpois,
    dzmgeom, pzmgeom, qzmgeom, rzmgeom,
    dpoisinvgauss, ppoisinvgauss, qpoisinvgauss, rpoisinvgauss,
    dpig, ppig, qpig, rpig, # aliases
    ## Three parameter distributions
    dburr, pburr, qburr, rburr, mburr, levburr,
    dgenpareto, pgenpareto, qgenpareto, rgenpareto, mgenpareto, levgenpareto,
    dinvburr, pinvburr, qinvburr, rinvburr, minvburr, levinvburr,
    dinvtrgamma, pinvtrgamma, qinvtrgamma, rinvtrgamma, minvtrgamma, levinvtrgamma,
    dtrgamma, ptrgamma, qtrgamma, rtrgamma, mtrgamma, levtrgamma,
    dpareto2, ppareto2, qpareto2, rpareto2, mpareto2, levpareto2,
    dpareto3, ppareto3, qpareto3, rpareto3, mpareto3, levpareto3,
    dzmnbinom, pzmnbinom, qzmnbinom, rzmnbinom,
    dzmbinom, pzmbinom, qzmbinom, rzmbinom,
    ## Four parameter distributions
    dgenbeta, pgenbeta, qgenbeta, rgenbeta, mgenbeta, levgenbeta,
    dtrbeta, ptrbeta, qtrbeta, rtrbeta, mtrbeta, levtrbeta,
    dpearson6, ppearson6, qpearson6, rpearson6, mpearson6, levpearson6, #aliases
    dpareto4, ppareto4, qpareto4, rpareto4, mpareto4, levpareto4,
    ## Five parameter distributions
    dfpareto, pfpareto, qfpareto, rfpareto, mfpareto, levfpareto,
    ## Phase-type distributions
    dphtype, pphtype, rphtype, mphtype, mgfphtype,
    ## Loss distributions
    grouped.data, ogive, emm, mde, elev, coverage, var, sd
)

### Methods
S3method("[", grouped.data)
S3method("[<-", grouped.data)

S3method(aggregate, portfolio)

S3method(CTE, aggregateDist)

S3method(diff, aggregateDist)

S3method(elev, default)
S3method(elev, grouped.data)

S3method(emm, default)
S3method(emm, grouped.data)

S3method(frequency, portfolio)

S3method(hist, grouped.data)

S3method(knots, ogive)
S3method(knots, elev)

S3method(mean, aggregateDist)
S3method(mean, grouped.data)

S3method(sd, default)
S3method(sd, grouped.data)
S3method(var, default)
S3method(var, grouped.data)

S3method(ogive, default)
S3method(ogive, grouped.data)

S3method(plot, adjCoef)
S3method(plot, aggregateDist)
S3method(plot, elev)
S3method(plot, ogive)
S3method(plot, ruin)

S3method(predict, bstraub)
S3method(predict, cm)
S3method(predict, hache)
S3method(predict, hierarc)
S3method(predict, bayes)

S3method(print, aggregateDist)
S3method(print, elev)
S3method(print, cm)
S3method(print, mde)
S3method(print, ogive)
S3method(print, summary.ogive)
S3method(print, portfolio)
S3method(print, summary.aggregateDist)
S3method(print, summary.cm)

S3method(quantile, aggregateDist)
S3method(quantile, grouped.data)

S3method(severity, default)
S3method(severity, portfolio)

S3method(summary, aggregateDist)
S3method(summary, cm)
S3method(summary, grouped.data)
S3method(summary, elev)
S3method(summary, ogive)

S3method(VaR, aggregateDist)

S3method(weights, portfolio)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Demo of the loss distributions facilities provided by actuar
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

require(actuar)
require(graphics)


### A utility function to create graphs for probability laws
showgraphs <- function(fun, par, what = c("d", "p", "m", "lev"), xlim)
{
    dist <- switch(fun,
                   trbeta = "TRANSFORMED BETA DISTRIBUTION",
                   genpareto = "GENERALIZED PARETO DISTRIBUTION",
                   burr = "BURR DISTRIBUTION",
                   invburr = "INVERSE BURR DISTRIBUTION",
                   pareto = "PARETO DISTRIBUTION",
                   invpareto = "INVERSE PARETO DISTRIBUTION",
                   llogis = "LOGLOGISTIC DISTRIBUTION",
                   paralogis = "PARALOGISTIC DISTRIBUTION",
                   invparalogis = "INVERSE PARALOGISTIC DISTRIBUTION",
                   trgamma = "TRANSFORMED GAMMA DISTRIBUTION",
                   invtrgamma = "INVERSE TRANSFORMED GAMMA DISTRIBUTION",
                   invgamma = "INVERSE GAMMA DISTRIBUTION",
                   weibull = "WEIBULL DISTRIBUTION",
                   invweibull = "INVERSE WEIBULL DISTRIBUTION",
                   invexp = "INVERSE EXPONENTIAL DISTRIBUTION",
                   pareto1 = "SINGLE PARAMETER PARETO DISTRIBUTION",
                   lgamma = "LOGGAMMA DISTRIBUTION",
                   genbeta = "GENERALIZED BETA DISTRIBUTION",
                   phtype = "PHASE-TYPE DISTRIBUTION",
                   gamma = "GAMMA DISTRIBUTION",
                   exp = "EXPONENTIAL DISTRIBUTION",
                   chisq = "CHI-SQUARE DISTRIBUTION",
                   lnorm = "LOGNORMAL DISTRIBUTION",
                   invgauss = "INVERSE GAUSSIAN DISTRIBUTION",
                   norm = "NORMAL DISTRIBUTION",
                   beta = "BETA DISTRIBUTION",
                   unif = "UNIFORM DISTRIBUTION")

    if (missing(xlim))
    {
        qf <- match.fun(paste("q", fun, sep = ""))
        formals(qf)[names(par)] <- par
        xlim <- c(0, qf(0.999))
    }
    k <- seq.int(4)
    limit <- seq(0, xlim[2], len = 10)
    mfrow = c(ceiling(length(what) / 2), 2)
    op <- par(mfrow = mfrow, oma = c(0, 0, 2, 0))

    for (t in what)
    {
        f   <- match.fun(paste(t, fun, sep = ""))
        formals(f)[names(par)] <- par

        main <- switch(t,
                       "d" = "Probability Density Function",
                       "p" = "Cumulative Distribution Function",
                       "m" = "Raw Moments",
                       "lev" = "Limited Expected Value Function",
                       "mgf" = "Moment Generating Function")

        if (t == "m")
            plot(k, f(k), type = "l", col = 4, lwd = 2, main = main)
        else if (t == "lev")
            plot(limit, f(limit), type = "l", col = 4, lwd = 2, main = main)
        else if (t == "mgf")
            curve(f(x), xlim = c(0, 2), col = 4, lwd = 2, main = main)
        else
            curve(f(x), xlim = xlim, col = 4, lwd = 2, main = main)
        title(main = dist, outer = TRUE)
    }
    par(op)
}

###
### DATA SETS
###

## The package includes the individual dental claims and grouped
## dental claims data sets often referred to in Klugman, Panjer &
## Willmot (1998, 2004)
data(dental); dental
data(gdental); gdental


###
### PROBABILITY LAWS
###

## Illustration of the new probability laws functions provided by the
## package.

## TRANSFORMED BETA FAMILY

## Transformed beta distribution
showgraphs("trbeta", list(shape1 = 3, shape2 = 4, shape3 = 5, scale = 10))

## Generalized Pareto distribution
showgraphs("genpareto", list(shape1 = 10, shape2 = 4, scale = 10))

## Burr distribution
showgraphs("burr", list(shape1 = 3, shape2 = 4, scale = 10))

## Inverse Burr distribution
showgraphs("invburr", list(shape1 = 3, shape2 = 6, scale = 10))

## Pareto distribution
showgraphs("pareto", list(shape = 10, scale = 10))

## Inverse Pareto distribution
showgraphs("invpareto", list(shape = 4, scale = 1), what = c("d", "p"))

## Loglogistic distribution
showgraphs("llogis", list(shape = 6, scale = 10))

## Paralogistic distribution
showgraphs("paralogis", list(shape = 3, scale = 10))

## Inverse paralogistic distribution
showgraphs("invparalogis", list(shape = 6, scale = 10))

## TRANSFORMED GAMMA FAMILY

## Transformed gamma distribution
showgraphs("trgamma", list(shape1 = 3, shape2 = 1, scale = 10))

## Inverse transformed gamma distribution
showgraphs("invtrgamma", list(shape1 = 3, shape2 = 2, scale = 10))

## Inverse gamma distribution
showgraphs("invgamma", list(shape = 6, scale = 10))

## Weibull distribution ('mweibull' and 'levweibull')
showgraphs("weibull", list(shape = 1.5, scale = 10))

## Inverse Weibull distribution
showgraphs("invweibull", list(shape = 6, scale = 10))

## Inverse exponential distribution
showgraphs("invexp", list(rate = 1), what = c("d", "p"))

## OTHER DISTRIBUTIONS

## Single parameter Pareto distribution
showgraphs("pareto1", list(shape = 5, min = 10), xlim = c(0, 50))

## Loggamma distribution
showgraphs("lgamma", list(shapelog = 2, ratelog = 5))

## Generalized beta distribution
showgraphs("genbeta", list(shape1 = 1, shape2 = 2, shape3 = 3, scale = 2))

## Phase-type distribution
showgraphs("phtype", list(prob = c(0.5614, 0.4386), rates = matrix(c(-8.64, 0.101, 1.997, -1.095), 2, 2)), what = c("d", "p", "m", "mgf"), xlim = c(0.001, 5))

## DISTRIBUTIONS ALREADY IN R

## Gamma distribution
showgraphs("gamma", list(shape = 3, rate = 5), what = c("m", "lev", "mgf"))

## Chi-square distribution
showgraphs("chisq", list(df = 3), what = c("m", "lev", "mgf"))

## Exponential distribution
showgraphs("exp", list(rate = 5), what = c("m", "lev", "mgf"))

## Lognormal distribution
showgraphs("lnorm", list(meanlog = 1, sdlog = 1), what = c("m", "lev"))

## Inverse gaussian distribution (from package SuppDists)
showgraphs("invgauss", list(nu = 1, lambda = 10), what = c("m", "lev", "mgf"), xlim = c(0, 10))

## Normal distribution
showgraphs("norm", list(mean = 0, sd = 1), what = c("m", "mgf"))

## Beta distribution
showgraphs("beta", list(shape1 = 1, shape2 = 2), what = c("m", "lev"))

## Uniform distribution
showgraphs("unif", list(min = 0, max = 1), what = c("m", "lev", "mgf"))


###
### GROUPED DATA MANIPULATION
###

## Creation of grouped data objects
x <- grouped.data(groups = c(0, 25, 50, 100, 150, 250, 500),
                  line1 = c(30, 31, 57, 42, 65, 84),
                  line2 = c(26, 33, 31, 19, 16, 11))
x

## Extraction and replacement: only "[" and "[<-" are officially
## supported.
x[, 1]                                  # group boundaries
x[1]                                    # notice the difference
x[, -1]                                 # group frequencies
x[1:3,]                                 # first 3 groups
x[1, 2] <- 22; x                        # frequency replacement
x[1, 1] <- c(0, 20); x                  # boundary replacement

## Mean, variance and standard deviation for grouped data objects.
mean(x)
var(x)
sd(x)

## In the sequel, only the first frequencies column is considered.
x <- x[, -3]

## Function 'hist' handles individual data only. We provide a method
## for grouped data.
hist(x)

## Function 'ogive' returns a function to compute the ogive of grouped
## data in any point, much like 'ecdf' does for individual data.
## Methods also exist to extract the group boundaries ('knots') and
## to plot the ogive.
Fnt <- ogive(x)
summary(Fnt)
knots(Fnt)                              # group boundaries
Fnt(knots(Fnt))                         # ogive at group boundaries
plot(Fnt)                               # plot of the ogive

## The method of 'quantile' for grouped data objects computes linearly
## smoothed quantiles, that is the inverse of the ogive in various
## points.
quantile(x)
Fnt(quantile(x))

## The method of 'summary' for grouped data objects returns the
## quantiles and the mean in a single object.
summary(x)

###
### EMPIRICAL MOMENTS CALCULATION
###

## Function 'emm' computes the k-th empirical moment of a sample,
## whether it is individual or grouped data.
emm(dental)                             # == mean(dental)
emm(gdental)                            # == mean(gdental)
emm(dental, order = 1:3)                # first three moments
emm(gdental, order = 1:3)               # idem

## Function 'elev' is similar to 'ecdf' and 'ogive' in that it returns
## a function to compute the empirical limited expected value (first
## limited moment) for any limit. There are methods for individual and
## grouped data.
lev <- elev(dental)
lev(knots(lev))                         # ELEV at data points
plot(lev, type = "o", pch = 19)         # plot of the ELEV function

lev <- elev(gdental)
lev(knots(lev))                         # ELEV at data points
plot(lev, type = "o", pch = 19)         # plot of the ELEV function


###
### MINIMUM DISTANCE ESTIMATION
###

## Maximum likelihood estimation (for individual data) is well covered
## by 'fitdistr' in package MASS. We provide function 'mde' to fit
## models using three distance minimization techniques: Cramer-von
## Mises (for individual and grouped data), chi-square and layer
## average severity (both grouped data only). Usage (and inner
## working) is very similar to 'fitdistr'.
mde(dental, pexp, start = list(rate = 1/200), measure = "CvM")
mde(gdental, pexp, start = list(rate = 1/200), measure = "CvM")
mde(gdental, pexp, start = list(rate = 1/200), measure = "chi-square")
mde(gdental, levexp, start = list(rate = 1/200), measure = "LAS")


###
### COVERAGE MODIFICATIONS
###

## Function 'coverage' is useful to obtain the probability density
## function (pdf) or cumulative distribution function (cdf) of a loss
## random variable under coverage modifications.
f <- coverage(dgamma, pgamma, deductible = 1, limit = 7)
curve(dgamma(x, 3), xlim = c(0, 10), ylim = c(0, 0.3))    # original
curve(f(x, 3), xlim = c(0.01, 5.99), col = 4, add = TRUE) # modified

x <- rgamma(1000, 3, 1)                 # sample of claim amounts
x <- pmin(x, 7)[x > 1] - 1              # deductible and limit

library(MASS)                           # for ML estimation
m <- mean(x)                            # empirical mean
v <- var(x)                             # empirical variance
(p <- fitdistr(x, f, start = list(shape = m^2/v, rate = m/v))$estimate ) # MLE
hist(x + 1, breaks = 0:10, prob = TRUE)  # histogram of observed data
curve(dgamma(x, p[1], p[2]), add = TRUE) # fit of underlying distribution

par(op)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Demo of the portfolio simulation facilities provided by actuar
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

require(actuar)


## A simple Compound Poisson model: S_t = C_1 + ... + C_{N_t}, with
## N_t ~ Poisson(10), C ~ Lognormal(log(1500) - 1, 1). The names of
## the components serve no purpose here but are required.
pf <- rcomphierarc(list(y = 10), model.freq = expression(y = rpois(10)),
                   model.sev = expression(y = rlnorm(log(1500) - 1, 1)))
pf                                      # print method
aggregate(pf)                           # aggregate claim amounts
frequency(pf)                           # frequencies
severity(pf)                            # individual claim amounts
severity(pf, splitcol = 10)             # last period separate


## Simple (continuous) mixture of models: S_t|Theta ~ Poisson(Theta),
## Theta ~ Gamma(2, 1). Any names can be used in the model.
pf <- rcomphierarc(list(Theta = 1, S = 10),
                   model.freq = expression(Theta = rgamma(2, 1),
                                           S = rpois(Theta)))
aggregate(pf)                           # actual data
frequency(pf)                           # same, here

## Model with with mixtures for both frequency and severity.
pf <- rcomphierarc(list(entity = 10, year = 5),
            model.freq = expression(entity = rgamma(2, 1),
                year = rpois(entity)),
            model.sev = expression(entity = rnorm(5, 1),
                year = rlnorm(entity, 1)))
pf
aggregate(pf)
frequency(pf)

## Same model as above, but with weights incorporated into the model.
## The string "weights" should appear in the model specification
## wherever weights are to be used.
wit <- runif(10, 2, 10)
(wit <- runif(50, rep(0.5 * wit, each = 5), rep(1.5 * wit, each = 5)))
(pf <- rcomphierarc(list(entity = 10, year = 5),
                    model.freq = expression(entity = rgamma(2, 1),
                                            year = rpois(weights * entity)),
                    model.sev = expression(entity = rnorm(5, 1),
                                           year = rlnorm(entity, 1)),
                    weights = wit))
weights(pf)                             # extraction of weights

## Three level hierarchical model (sector, unit, contract). Claim
## severity varies only by sector and unit. The number of "nodes" at
## each level is different.
nodes <- list(sector = 2, unit = c(3, 4),
              contract = c(10, 5, 8, 5, 7, 11, 4), year = 6)
mf <- expression(sector = rexp(2),
                 unit = rgamma(sector, 0.1),
                 contract = rgamma(unit, 1),
                 year = rpois(weights * contract))
ms <- expression(sector = rnorm(2, sqrt(0.1)),
                 unit = rnorm(sector, 1),
                 contract = NULL,
                 year = rlnorm(unit, 1))
wijkt <- runif(50, 2, 10)
wijkt <- runif(300, rep(0.5 * wijkt, each = 6), rep(1.5 * wijkt, each = 6))
pf <- rcomphierarc(nodes, model.freq = mf, model.sev = ms, weights = wijkt)
frequency(pf)
weights(pf)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Demo of the credibility theory facilities provided by actuar
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

require(actuar)


## The package provides the famous data set of Hachemeister (1975) as
## a matrix of 5 lines (one for each state) and 25 columns (the state
## number, 12 periods of ratios, 12 periods of corresponding weights).
data(hachemeister)
hachemeister

## Fitting of a Buhlmann model to the Hachemeister data set using
## function 'cm'. The interface of the function is similar to 'lm'.
fit <- cm(~state, hachemeister, ratios = ratio.1:ratio.12)
fit                          # print method
summary(fit)                 # more information
fit$means                    # (weighted) averages
fit$weights                  # total weights
fit$unbiased                 # unbiased variance estimators
predict(fit)                 # credibility premiums

## Fitting of a Buhlmann-Straub model require weights. Here, iterative
## estimators of the variance components are used.
fit <- cm(~state, hachemeister, ratios = ratio.1:ratio.12,
          weights = weight.1:weight.12, method = "iterative")
summary(fit)
predict(fit)

## Simulation of a three level hierarchical portfolio.
nodes <- list(sector = 2, unit = c(3, 4),
              contract = c(10, 5, 8, 5, 7, 11, 4), year = 6)
mf <- expression(sector = rexp(2),
                 unit = rgamma(sector, 0.1),
                 contract = rgamma(unit, 1),
                 year = rpois(weights * contract))
ms <- expression(sector = rnorm(2, sqrt(0.1)),
                 unit = rnorm(sector, 1),
                 contract = NULL,
                 year = rlnorm(unit, 1))
wijkt <- runif(50, 2, 10)
wijkt <- runif(300, rep(0.5 * wijkt, each = 6), rep(1.5 * wijkt, each = 6))
pf <- simul(nodes, model.freq = mf, model.sev = ms, weights = wijkt)

## Fitting of a hierarchical model to the portfolio simulated above.
DB <- cbind(weights(pf, prefix = "weight."),
            aggregate(pf, classif = FALSE) / weights(pf, classif = FALSE))

fit <- cm(~sector + sector:unit + sector:unit:contract,
          data = DB, ratios = year.1:year.6,
          weights = weight.year.1:weight.year.6)
fit
predict(fit)                          # credibility premiums
predict(fit, levels = "unit")         # unit credibility premiums only
summary(fit)                          # portfolio summary
summary(fit, levels = "unit")         # unit portfolio summary only

## Fitting of Hachemeister regression model with intercept at time origin.
fit <- cm(~state, hachemeister, ratios = ratio.1:ratio.12,
          weights = weight.1:weight.12, regformula = ~time,
          regdata = data.frame(time = 1:12))
summary(fit, newdata = data.frame(time = 13)) # 'newdata' is the future value of regressor
predict(fit, newdata = data.frame(time = 13))

## Position the intercept at the barycenter of time.
fit <- cm(~state, hachemeister, ratios = ratio.1:ratio.12,
          weights = weight.1:weight.12, regformula = ~time,
          regdata = data.frame(time = 1:12), adj.intercept = TRUE)
summary(fit, newdata = data.frame(time = 13))

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}invweibull functions to compute
### characteristics of the Inverse Weibull distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = exp(-(x/scale)^shape), x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvweibull <- function (x, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "dinvweibull", x, shape, scale, log)

pinvweibull <- function(q, shape, rate = 1, scale = 1/rate,
                        lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvweibull", q, shape, scale, lower.tail, log.p)

qinvweibull <- function(p, shape, rate = 1, scale = 1/rate,
                        lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qinvweibull", p, shape, scale, lower.tail, log.p)

rinvweibull <- function(n, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rinvweibull", n, shape, scale)

minvweibull <- function(order, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "minvweibull", order, shape, scale, FALSE)

levinvweibull <- function(limit, shape, rate = 1, scale = 1/rate,
                          order = 1)
    .External(C_actuar_do_dpq, "levinvweibull", limit, shape, scale, order, FALSE)

## Aliases
dlgompertz <- dinvweibull
plgompertz <- pinvweibull
qlgompertz <- qinvweibull
rlgompertz <- rinvweibull
mlgompertz <- minvweibull
levlgompertz <- levinvweibull

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev}weibull functions to compute raw and
### limited moments for the Weibull distribution (as defined in R).
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

mweibull <- function(order, shape, scale = 1)
    .External(C_actuar_do_dpq, "mweibull", order, shape, scale, FALSE)

levweibull <- function(limit, shape, scale = 1, order = 1)
    .External(C_actuar_do_dpq, "levweibull", limit, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}burr functions to compute
### characteristics of the Burr distribution. The version used in
### these functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - (1/(1 + (x/scale)^shape2))^shape1, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dburr <- function (x, shape1, shape2, rate = 1, scale = 1/rate,
                   log = FALSE)
    .External(C_actuar_do_dpq, "dburr", x, shape1, shape2, scale, log)

pburr <- function(q, shape1, shape2, rate = 1, scale = 1/rate,
                  lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pburr", q, shape1, shape2, scale,
              lower.tail, log.p)

qburr <- function(p, shape1, shape2, rate = 1, scale = 1/rate,
                  lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qburr", p, shape1, shape2, scale,
              lower.tail, log.p)

rburr <- function(n, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rburr", n, shape1, shape2, scale)

mburr <- function(order, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mburr", order, shape1, shape2, scale, FALSE)

levburr <- function(limit, shape1, shape2, rate = 1, scale = 1/rate,
                    order = 1)
    .External(C_actuar_do_dpq, "levburr", limit, shape1, shape2, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev,mgf}unif functions to compute raw and
### limited moments, and the moment generating function for the
### Uniform distribution (as defined in R).
###
### <http://en.wikipedia.org/wiki/Uniform_distribution_%28continuous%29>
###
### AUTHORS: Christophe Dutang, Vincent Goulet <vincent.goulet@act.ulaval.ca>

munif <- function(order, min = 0, max = 1)
    .External(C_actuar_do_dpq, "munif", order, min, max, FALSE)

levunif <- function(limit, min = 0, max =1, order = 1)
    .External(C_actuar_do_dpq, "levunif", limit, min, max, order, FALSE)

mgfunif <- function(t, min = 0, max = 1, log = FALSE)
    .External(C_actuar_do_dpq, "mgfunif", t, min, max, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Display all values of a matrix of vectors by 'unrolling' the
### object vertically or horizontally.
###
### AUTHORS: Louis-Philippe Pouliot,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

unroll <- function(x, bycol = FALSE, drop = TRUE)
{
    dx <- dim(x)

    if (length(dx) > 2L)
        stop(sprintf("%s must be a vector or a matrix", sQuote("x")))

    if (length(dx) < 2L)
        x <- rbind(x, deparse.level = 0L)

    fun <- function(x) if (identical(x, NA)) NA else length(x)
    frequencies <- array(sapply(x, fun), dim = dim(x))

    if (bycol)
    {
        lengths <- colSums(frequencies, na.rm = TRUE)
        mat <- matrix(NA, max(lengths), ncol(x), dimnames = dimnames(x))
        for (i in seq_len(ncol(x)))
            if (0L < (lengthi <- lengths[i]))
                mat[seq_len(lengthi), i] <- unlist(x[!is.na(x[, i]), i])
    }
    else
    {
        lengths <- rowSums(frequencies, na.rm = TRUE)
        mat <- matrix(NA, nrow(x), max(lengths),
                      dimnames = list(rownames(x), NULL))
        for (i in seq_len(nrow(x)))
            if (0L < (lengthi <- lengths[i]))
                mat[i, seq_len(lengthi)] <- unlist(x[i, !is.na(x[i, ])])
    }
    mat[, , drop = drop]
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}lgamma functions to compute
### characteristics of the Loggamma distribution. The version used in
### these functions has cumulative distribution function
###
###   Pr[X <= x] = pgamma(log(x), shape = shapelog, rate = ratelog).
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dlgamma <- function(x, shapelog, ratelog, log = FALSE)
    .External(C_actuar_do_dpq, "dlgamma", x, shapelog, ratelog, log)

plgamma <- function(q, shapelog, ratelog, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "plgamma", q, shapelog, ratelog, lower.tail, log.p)

qlgamma <- function(p, shapelog, ratelog, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qlgamma", p, shapelog, ratelog, lower.tail, log.p)

rlgamma <- function(n, shapelog, ratelog)
    .External(C_actuar_do_random, "rlgamma", n, shapelog, ratelog)

mlgamma <- function(order, shapelog, ratelog)
    .External(C_actuar_do_dpq, "mlgamma", order, shapelog, ratelog, FALSE)

levlgamma <- function(limit, shapelog, ratelog, order = 1)
    .External(C_actuar_do_dpq, "levlgamma", limit, shapelog, ratelog, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of a generic function for the Value at Risk. Currently,
### there is no default method for this function. The only method is
### for objects of class 'aggregateDist'.
###
### AUTHORS: Tommy Ouellet, Vincent Goulet <vincent.goulet@act.ulaval.ca>

VaR <- function(x, ...) UseMethod("VaR")

VaR.aggregateDist <- function(x, conf.level = c(0.9, 0.95, 0.99),
                              smooth = FALSE, names = TRUE, ...)
    quantile.aggregateDist(x, conf.level, smooth, names, ...)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Hierarchical credibility model calculations
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Louis-Philippe Pouliot, Tommy Ouellet.

hierarc <- function(ratios, weights, classification,
                    method = c("Buhlmann-Gisler", "Ohlsson", "iterative"),
                    tol = sqrt(.Machine$double.eps), maxit = 100,
                    echo = FALSE)
{
    ## === HANDS ON THE DATA ===
    ##
    ## Arguments 'ratios' and 'weights' must be matrices of real
    ## numbers, whereas 'classification' must be a matrix of integers
    ## giving the affiliation of each entity in the portfolio.
    nlevels <- ncol(classification)              # number of levels
    nlevels1p <- nlevels + 1L                    # frequently used

    ## To symmetrize further calculations, bind a column of ones
    ## representing the affiliation to the global portfolio.
    classification <- cbind(pf = 1L, classification)

    ## If weights are not specified, use equal weights.
    if (missing(weights))
    {
        if (any(is.na(ratios)))
            stop("missing ratios not allowed when weights are not supplied")
        array(1, dim(ratios))       # matrix of ones
    }

    ## Sanity check if weights and ratios correspond.
    if (!identical(which(is.na(ratios)), which(is.na(weights))))
        stop("missing values are not in the same positions in 'weights' and in 'ratios'")

    ## === NUMBER OF NODES AND SPLITTING FACTORS ===
    ##
    ## Future computation of per level summaries will require a set of
    ## factors based on the number of nodes in each level. An example
    ## will best explain what is achieved here: suppose there are two
    ## sectors; sector 1 has 3 units; sector 2 has 2 units. To make
    ## per sector summaries, the following factors can be used to
    ## split the unit data: 1 1 1 2 2.
    ##
    ## Generating such factors first requires to know the number of
    ## nodes at each level in a format identical to the 'nodes'
    ## argument of simul(). [In the previous example, the number of
    ## nodes would be 'list(2, c(3, 2))'.] Then, the factors are
    ## obtained by repeating a sequence the same length as the number
    ## of nodes at one level [2] according to the number of nodes at
    ## the level below [c(3, 2)].
    ##
    ## 0. Initialization
    fnodes <- nnodes <- vector("list", nlevels)

    ## 1. Calculation of the number of nodes: the main idea is to
    ## create a unique factor for each node using interaction()
    ## recursively on the columns of 'classification'. We can do
    ## something simpler for the lowest level (the entities), since we
    ## know the combinations of indexes to all be different at this
    ## level.
    fx <- vector("list", nlevels1p)
    fx[[nlevels1p]] <- factor(classification[, nlevels1p]) # entity level

    for (i in nlevels:1L)
    {
        ## Function 'interaction' expects its arguments separately or
        ## as a list, hence the lapply() below.
        fx[[i]] <- as.integer(interaction(lapply(seq.int(i),
                                                 function(j) classification[, j]),
                                          drop = TRUE))
        ## 'as.vector' below is used to get rid of names
        nnodes[[i]] <- as.vector(sapply(split(fx[[i + 1]], fx[[i]]),
                                        function(x) length(unique(x))))
    }

    ## 2. Generation of the factors. Following the rule described
    ## above, this could simply be
    ##
    ##   fnodes <- lapply(nnodes, function(x) rep(seq_along(x), x))
    ##
    ## However, this will not work if rows of data are not sorted per
    ## level. (In the example above, if the data of unit 3 of sector 1
    ## is at the bottom of the matrix of data, then the factors need
    ## to be 1 1 2 2 1.)
    ##
    ## The solution is actually simple: converting the entity level
    ## factors ('fx[[nlevels]]') to integers will assure that any
    ## summary made using these factors will be sorted. This done, it
    ## is possible to use the command above for the upper levels.
    fnodes[[nlevels]] <- as.integer(fx[[nlevels]])
    fnodes[-nlevels] <- lapply(nnodes[-nlevels],
                               function(x) rep(seq_along(x), x))

    ## === PER ENTITY SUMMARIES ===
    ##
    ## Individual weighted averages. It could happen that an entity
    ## has no observations, for example when applying the model on
    ## claim amounts. In such a situation, put the total weight of the
    ## entity and the weighted average both equal to zero. That way,
    ## the premium will be equal to the credibility weighted average,
    ## as it should, but the entity will otherwise have no
    ## contribution in the calculations.
    weights.s <- rowSums(weights, na.rm = TRUE)
    ratios.w <- ifelse(weights.s > 0, rowSums(weights * ratios, na.rm = TRUE) / weights.s, 0)

    ## === EFFECTIVE NUMBER OF NODES ===
    ##
    ## Given the possibility to have whole levels with no data, as
    ## explained above, it is necessary to count the *effective*
    ## number of nodes in each level, that is the number of nodes with
    ## data. This comes this late since it relies on 'weights.s'.
    ##
    ## Object 'eff.nnodes' is in every respect equivalent to 'nnodes'
    ## except that each element of the list is a vector of the number of
    ## non "empty" nodes for each classification of the level
    ## above.
    eff.nnodes <- vector("list", nlevels)
    w <- weights.s
    for (i in nlevels:1L)
    {
        eff.nnodes[[i]] <- tapply(w, fnodes[[i]], function(x) sum(x > 0))
        w <- tapply(w, fnodes[[i]], sum) # running totals
    }

    ## === DENOMINATORS OF VARIANCE ESTIMATORS ===
    ##
    ## The denominators for all the variance estimators never
    ## change. The denominator at one level is equal to the total
    ## number of nodes at that level minus the total number of nodes
    ## at the level above. At the lowest level (the denominator of
    ## s^2), this is
    ##
    ##   number of (non NA) ratios - (effective) number of entities.
    ##
    ## The number of (non missing) ratios is not included in
    ## 'eff.nnodes'.  For the portfolio level, the denominator is
    ##
    ##   (effective) number of "sectors" - 1
    ##
    ## The 1 neither is included in 'eff.nnodes'.
    denoms <- diff(c(1L, sapply(eff.nnodes, sum), sum(!is.na(ratios))))

    ## Final sanity checks
    if (any(!denoms))
        stop("there must be at least two nodes at every level")
    if (ncol(ratios) < 2L)
        stop("there must be at least one node with more than one period of experience")

    ## === ESTIMATION OF s^2 ===
    s2 <-  sum(weights * (ratios - ratios.w)^2, na.rm = TRUE) /
        denoms[nlevels1p]

    ## === ESTIMATION OF THE OTHER VARIANCE COMPONENTS ===
    ##
    ## Create vectors to hold values to be computed at each level
    ## (from portfolio to entity), namely: the total node weights, the
    ## node weighted averages, the between variances and the node
    ## credibility factors.
    ##
    ## Only credibility factors are not computed for the portfolio
    ## level, hence this list is one shorter than the others.
    tweights <- vector("list", nlevels1p)       # total level weights
    wmeans <- vector("list", nlevels1p)         # weighted averages
    b <- c(numeric(nlevels), s2)                # variance estimators
    cred <- vector("list", nlevels)             # credibility factors

    ## Values already computed at the entity level.
    tweights[[nlevels1p]] <- as.vector(weights.s);
    wmeans[[nlevels1p]] <- as.vector(ratios.w);

    ## The unbiased variance estimators are evaluated first as they will
    ## be used as starting values for the iterative part below.
    ##
    ## At the entity level: node weight is given by the natural
    ## weight, weighted averages use the natural weights.
    ##
    ## Level above the entity: node weight is the sum of the natural
    ## weights at the level below, weighted averages use the natural
    ## weights.
    ##
    ## All upper levels: node weight is the sum of the credibility
    ## factors at the level below, weighted averages use credibility
    ## factors from previous level.
    ##
    ## Buhlmann-Gisler estimators truncate the per node variance
    ## estimates to 0 before taking the mean, whereas the Ohlsson
    ## estimators do not make any truncation.
    method <- match.arg(method)

    if (method == "Buhlmann-Gisler")
        bexp <- expression(b[i] <- mean(pmax(ifelse(ci != 0, bi/ci, 0), 0), na.rm = TRUE))
    else                                # Ohlsson
        bexp <- expression(b[i] <- sum(bi, na.rm = TRUE) / sum(ci, na.rm = TRUE))

    for (i in nlevels:1L)
    {
        ## Total weight of the level as per the rule above.
        tweights[[i]] <- as.vector(tapply(tweights[[i + 1L]], fnodes[[i]], sum))

        ## Calculation of the weighted averages of the level. Before
        ## the between variance is estimated, these use the total
        ## weights calculated above.
        wmeans[[i]] <-
            ifelse(tweights[[i]] > 0,
                   as.vector(tapply(tweights[[i + 1L]] * wmeans[[i + 1L]],
                                    fnodes[[i]],
                                    sum) / tweights[[i]]),
                   0)

        ## Latest non-zero between variance estimate -- the one used
        ## in the estimator and in the credibility factors.
        between <- b[b != 0][1L]

        ## Calculation of the per node variance estimate.
        bi <- as.vector(tapply(tweights[[i + 1L]] *
                                (wmeans[[i + 1L]] - wmeans[[i]][fnodes[[i]]])^2,
                                fnodes[[i]],
                                sum)) -
                                    (eff.nnodes[[i]] - 1) * between
        ci <- tweights[[i]] -
            as.vector(tapply(tweights[[i + 1L]]^2, fnodes[[i]], sum)) / tweights[[i]]

        ## The final estimate is the average of all the per node estimates.
        eval(bexp)

        ## Calculation of the credibility factors. If these are
        ## non-zero, the total weights for the current level are
        ## replaced by the sum of the credibility factors and the
        ## weighted averages are recomputed with these new weights.
        #if (max(bu[i], 0))           # don't compute negative factors!
        if (b[i])
        {
            cred[[i]] <- 1/(1 + between/(b[i] * tweights[[i + 1L]]))
            tweights[[i]] <- as.vector(tapply(cred[[i]], fnodes[[i]], sum))
            wmeans[[i]] <-
                ifelse(tweights[[i]] > 0,
                       as.vector(tapply(cred[[i]] * wmeans[[i + 1L]],
                                        fnodes[[i]],
                                        sum) / tweights[[i]]),
                       0)
        }
        else
            cred[[i]] <- numeric(sum(nnodes[[i]]))
    }

    ## Iterative estimation of the structure parameters.
    ##
    ## At the entity level: total weight is the sum of the natural
    ## weights, weighted averages use the natural weights and between
    ## variance is s^2.
    ##
    ## All upper levels: total weight is the sum of the credibility
    ## factors of the level below, weighted averages use credibility
    ## factors, between variance estimated recursively and credibility
    ## factor use total weight of the level, between variance of the
    ## level below (hence the within variance) and between variance of
    ## the current level.
    if (method == "iterative")
    {
        b <- pmax(b, 0)            # truncation for starting values
        if (any(head(b, -1L) > 0)) # at least one non-zero starting value
            .External(C_actuar_do_hierarc, cred, tweights, wmeans, fnodes, denoms,
                      b, tol, maxit, echo)
    }

    ## Results
    structure(list(means = wmeans,
                   weights = tweights,
                   unbiased = if (method != "iterative") b,
                   iterative = if (method == "iterative") b,
                   cred = cred,
                   nodes = nnodes,
                   classification = classification[, -1L],
                   ordering = fnodes),
              class = "hierarc",
              model = "hierarchical")
}

predict.hierarc <- function(object, levels = NULL, newdata, ...)
{
    ## The credibility premium of a node at one level is equal to
    ##
    ##   p + z * (m - p)
    ##
    ## where 'p' is the credibility premium of the level above (or the
    ## collective premium for the portfolio), 'z' is the credibility
    ## factor of the node, and 'm' is the weighted average of the
    ## node.
    fnodes <- object$ordering
    cred <- object$cred
    means <- object$means
    nlevels <- length(object$nodes)
    level.names <- names(object$nodes)

    if (is.null(levels))
        levels <- seq_len(nlevels)
    if (any(is.na(levels)) || !is.numeric(levels))
        stop("invalid level number")
    n <- max(levels)

    res <- vector("list", n)

    ## First level credibility premiums
    res[[1L]] <- means[[1L]] + cred[[1L]] * (means[[2L]] - means[[1L]])

    for (i in seq(2, length.out = n - 1))
    {
        p <- res[[i - 1]][fnodes[[i]]]
        res[[i]] <- p + cred[[i]] * (means[[i + 1]] - p)
    }

    structure(res[levels], names = level.names[levels], ...)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}ztnbinom functions to compute
### characteristics of the Zero Truncated Negative Binomial
### distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Second
### Edition, Wiley, 2004.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dztnbinom <- function (x, size, prob, log = FALSE)
    .External(C_actuar_do_dpq, "dztnbinom", x, size, prob, log)

pztnbinom <- function(q, size, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pztnbinom", q, size, prob, lower.tail, log.p)

qztnbinom <- function(p, size, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qztnbinom", p, size, prob, lower.tail, log.p)

rztnbinom <- function(n, size, prob)
    .External(C_actuar_do_random, "rztnbinom", n, size, prob)

## not exported; for internal use in panjer()
pgfztnbinom <- function(x, size, prob)
    expm1(-size * log1p(x * (prob - 1)))/expm1(-size * log(prob))

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Use one of five methods to compute the aggregate claim amount
### distribution of a portfolio over a period given a frequency and a
### severity model or the true moments of the distribution.
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Louis-Philippe Pouliot

aggregateDist <-
    function(method = c("recursive", "convolution", "normal", "npower", "simulation"),
             model.freq = NULL, model.sev = NULL, p0 = NULL, x.scale = 1,
             convolve = 0, moments, nb.simul, ...,
             tol = 1e-06, maxit = 500, echo = FALSE)
{
    Call <- match.call()

    ## The method used essentially tells which function should be
    ## called for the calculation of the aggregate claims
    ## distribution.
    method <- match.arg(method)

    if (method == "normal")
    {
        ## An error message is issued if the number of moments listed
        ## is not appropriate for the method. However it is the user's
        ## responsability to list the moments in the correct order
        ## since the vector is not required to be named.
        if (missing(moments) || length(moments) < 2)
            stop(sprintf("%s must supply the mean and variance of the distribution",
                         sQuote("moments")))
        FUN <- normal(moments[1], moments[2])
        comment(FUN) <- "Normal approximation"
    }

    else if (method == "npower")
    {
        if (missing(moments) || length(moments) < 3)
            stop(sprintf("%s must supply the mean, variance and skewness of the distribution",
                         sQuote("moments")))
        FUN <- npower(moments[1], moments[2], moments[3])
        comment(FUN) <- "Normal Power approximation"
    }
    else if (method == "simulation")
    {
        if (missing(nb.simul))
            stop(sprintf("%s must supply the number of simulations",
                         sQuote("nb.simul")))
        if (is.null(names(model.freq)) && is.null(names(model.sev)))
            stop(sprintf("expressions in %s and %s must be named",
                         sQuote("model.freq"), sQuote("model.sev")))
        FUN <- simS(nb.simul, model.freq = model.freq, model.sev = model.sev)
        comment(FUN) <- "Approximation by simulation"
    }
    else
    {
        ## "recursive" and "convolution" cases. Both require a
        ## discrete distribution of claim amounts, that is a vector of
        ## probabilities in argument 'model.sev'.
        if (!is.numeric(model.sev))
            stop(sprintf("%s must be a vector of probabilities",
                         sQuote("model.sev")))

        ## Recursive method uses a model for the frequency distribution.
        if (method == "recursive")
        {
            if (is.null(model.freq) || !is.character(model.freq))
                stop("frequency distribution must be supplied as a character string")
            dist <- match.arg(tolower(model.freq),
                              c("poisson",
                                "geometric",
                                "negative binomial",
                                "binomial",
                                "logarithmic",
                                "zero-truncated poisson",
                                "zero-truncated geometric",
                                "zero-truncated negative binomial",
                                "zero-truncated binomial",
                                "zero-modified logarithmic",
                                "zero-modified poisson",
                                "zero-modified geometric",
                                "zero-modified negative binomial",
                                "zero-modified binomial"))
            FUN <- panjer(fx = model.sev, dist = dist, p0 = p0,
                          x.scale = x.scale, ..., convolve = convolve,
                          tol = tol, maxit = maxit, echo = echo)
            comment(FUN) <- "Recursive method approximation"
        }

        ## Convolution method requires a vector of probabilites in
        ## argument 'model.freq'.
        else if (method == "convolution")
        {
            if (!is.numeric(model.freq))
                stop(sprintf("%s must be a vector of probabilities",
                             sQuote("model.freq")))
            FUN <- exact(fx = model.sev, pn = model.freq, x.scale = x.scale)
            comment(FUN) <- "Exact calculation (convolutions)"
        }
        else
            stop("internal error")
    }

    ## Return cumulative distribution function
    class(FUN) <- c("aggregateDist", class(FUN))
    attr(FUN, "call") <- Call
    FUN
}

print.aggregateDist <- function(x, ...)
{
    cat("\nAggregate Claim Amount Distribution\n")
    cat("  ", label <- comment(x), "\n\n", sep = "")

    cat("Call:\n")
    print(attr(x, "call"), ...)
    cat("\n")

    if (label %in% c("Exact calculation (convolutions)",
                     "Recursive method approximation",
                     "Approximation by simulation"))
    {
        n <- length(get("x", envir = environment(x)))
        cat("Data:  (", n, "obs. )\n")
        numform <- function(x) paste(formatC(x, digits = 4, width = 5), collapse = ", ")
        i1 <- 1L:min(3L, n)
        i2 <- if (n >= 4L)
            max(4L, n - 1L):n
        else integer()
        xx <- eval(expression(x), envir = environment(x))
        cat(" x[1:", n, "] = ", numform(xx[i1]), if (n > 3L)
        ", ", if (n > 5L)
        " ..., ", numform(xx[i2]), "\n", sep = "")
        cat("\n")
    }
    if (label %in% c("Normal approximation",
                     "Normal Power approximation"))
        cat(attr(x, "source"), "\n")
    invisible(x)
}

plot.aggregateDist <- function(x, xlim,
                               ylab = expression(F[S](x)),
                               main = "Aggregate Claim Amount Distribution",
                               sub = comment(x), ...)
{
    ## Function plot() is used for the step cdfs and function curve()
    ## in the continuous cases.
    if ("stepfun" %in% class(x))
    {
        ## Method for class 'ecdf' will most probably be used.
        NextMethod(main = main, ylab = ylab, ...)
    }
    else
    {
        ## Limits for the x-axis are supplied if none are given
        ## in argument.
        if (missing(xlim))
        {
            mean <- get("mean", envir = environment(x))
            sd <- sqrt(get("variance", envir = environment(x)))
            xlim <- c(mean - 3 * sd, mean + 3 * sd)
        }
        curve(x, main = main, ylab = ylab, xlim = xlim, ylim = c(0, 1), ...)
    }
    mtext(sub, line = 0.5)
}

summary.aggregateDist <- function(object, ...)
    structure(object, class = c("summary.aggregateDist", class(object)), ...)

print.summary.aggregateDist <- function(x, ...)
{
    cat(ifelse(comment(x) %in%
               c("Normal approximation", "Normal Power approximation"),
               "Aggregate Claim Amount CDF:\n",
               "Aggregate Claim Amount Empirical CDF:\n"))
    q <- quantile(x, p = c(0.25, 0.5, 0.75))
    expectation <- mean(x)

    if (comment(x) %in% c("Normal approximation", "Normal Power approximation"))
    {
        min <- 0
        max <- NA
    }
    else
    {
        max <- tail(eval(expression(x), environment(x)), 1)
        min <- head(eval(expression(x), environment(x)), 1)
    }
    res <- c(min, q[c(1, 2)], expectation, q[3], max)
    names(res) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.")
    print(res, ...)
    invisible(x)
}

mean.aggregateDist <- function(x, ...)
{
    label <- comment(x)

    ## Simply return the value of the true mean given in argument in
    ## the case of the Normal and Normal Power approximations.
    if (label %in%
        c("Normal approximation", "Normal Power approximation"))
        return(get("mean", envir = environment(x)))

    ## For the recursive, exact and simulation methods, compute the
    ## mean from the stepwise cdf using the pmf saved in the
    ## environment of the object.
    drop(crossprod(get("x", envir = environment(x)),
                   get("fs", envir = environment(x))))
}

diff.aggregateDist <- function(x, ...)
{
    label <- comment(x)

    ## The 'diff' method is defined for the recursive, exact and
    ## simulation methods only.
    if (label == "Normal approximation" || label == "Normal Power approximation")
        stop("function not defined for approximating distributions")

    ## The probability vector is already stored in the environment of
    ## the "aggregateDist" object.
    get("fs", environment(x))
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Conditional Tail Expectation for objects of class 'aggregateDist'.
###
### AUTHORS: Tommy Ouellet, Vincent Goulet <vincent.goulet@act.ulaval.ca>

CTE <- function(x, ...) UseMethod("CTE")

CTE.aggregateDist <- function(x, conf.level = c(0.9, 0.95, 0.99),
                              names = TRUE, ...)
{
    chkDots(...)                        # method does not use '...'

    label <- comment(x)

    ## Normal approximation; an exact formula is available
    if (label == "Normal approximation")
    {
        m <- get("mean", environment(x))
        sd <- sqrt(get("variance", environment(x)))
        res <- m + sd * dnorm(qnorm(conf.level)) / (1 - conf.level)
    }
    ## Normal Power approximation; explicit formula in Castaner,
    ## Claramunt and Marmol (2013)
    else if (label == "Normal Power approximation")
    {
        m <- get("mean", envir = environment(x))
        sd <- sqrt(get("variance", envir = environment(x)))
        sk <- get("skewness", envir = environment(x))

        q <- qnorm(conf.level)
        res <- m + sd * dnorm(q) * (1 + sk * q/6) / (1 - conf.level)
    }
    ## Recursive method, simulation and convolutions; each yield a
    ## step function that can be used to make calculations.
    else
    {
        val <- get("x", envir = environment(x))
        prob <- get("fs", envir = environment(x))
        f2 <- function(a)
        {
            pos <- val > VaR(x, a)
            drop(crossprod(val[pos], prob[pos])) / sum(prob[pos])
        }
        res <- sapply(conf.level, f2)
    }

    if (names)
    {
        dig <- max(2, getOption("digits"))
        names(res) <- formatC(paste(100 * conf.level, "%", sep = ""),
                              format = "fg", width = 1, digits = dig)
    }
    res
}

TVaR <- CTE

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}invparalogis functions to compute
### characteristics of the Inverse Paralogistic distribution. The
### version used in these functions has cumulative distribution
### function
###
###   Pr[X <= x] = (u/(1 + u))^shape, u = (x/scale)^shape, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvparalogis <- function (x, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "dinvparalogis", x, shape, scale, log)

pinvparalogis <- function(q, shape, rate = 1, scale = 1/rate,
                          lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvparalogis", q, shape, scale, lower.tail, log.p)

qinvparalogis <- function(p, shape, rate = 1, scale = 1/rate,
                          lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qinvparalogis", p, shape, scale, lower.tail, log.p)

rinvparalogis <- function(n, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rinvparalogis", n, shape, scale)

minvparalogis <- function(order, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_dpq, "minvparalogis", order, shape, scale, FALSE)

levinvparalogis <- function(limit, shape, rate = 1, scale = 1/rate,
                            order = 1)
     .External(C_actuar_do_dpq, "levinvparalogis", limit, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Ogive for grouped data.
###
### A default method exists for either a vector of individual data, or
### two vectors of group boundaries and group frequencies. It first
### creates a grouped data object using 'grouped.data' and then calls
### a utility function to create the ogive.
###
### For the definition of the ogive, see Klugman, Panjer & Willmot,
### Loss Models, Wiley, 1998.
###
### More details on the admissible arguments for the default method
### are to be found in ./grouped.data.R.
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Mathieu Pigeon
###
### CREDITS: Arguments, 'breaks', 'nclass' and their treatment taken
### from R function hist().

ogive <- function(x, ...) UseMethod("ogive")

ogive.default <- function(x, y = NULL,
                          breaks = "Sturges", nclass = NULL, ...)
{
    chkDots(...)           # method does not use '...'
    Call <- match.call()
    if (exists(".Generic", inherits = FALSE))
        Call[[1]] <- as.name(.Generic)

    ## Avoid using calling 'hist' with 'nclass' specified.
    if (!missing(breaks))
    {
        if (!missing(nclass))
            warning(sprintf("%s not used when %s is specified",
                            sQuote("nclass"), sQuote("breaks")))
    }
    else if (!is.null(nclass) && length(nclass) == 1L)
        breaks <- nclass

    ## Create the "grouped.data" object.
    x <- if (is.null(y))   # one argument: individual data
             grouped.data(x, breaks = breaks)
         else              # two arguments: boundaries and frequencies
             grouped.data(x, y)

    ## Group frequencies in the second column of the data frame; group
    ## boundaries in the environment of 'x'.
    y <- x[, 2L]
    x <- eval(expression(cj), envir = environment(x))

    ## Create an object of class 'ogive'.
    res <- .ogiveFUN(x, y)
    attr(res, "call") <- Call
    res
}

ogive.grouped.data <- function(x, ...)
{
    chkDots(...)                      # method does not use '...'
    Call <- match.call()
    if (exists(".Generic", inherits = FALSE))
        Call[[1]] <- as.name(.Generic)

    ## We keep the first frequencies column only; group boundaries are
    ## in the environment of 'x'
    y <- x[, 2L]
    x <- eval(expression(cj), envir = environment(x))

    ## Create an object of class 'ogive'.
    res <- .ogiveFUN(x, y)
    attr(res, "call") <- Call
    res
}

.ogiveFUN <- function(x, y)
{
    FUN <- approxfun(x, cumsum(c(0, y)) / sum(y), yleft = 0, yright = 1,
                     method = "linear", ties = "ordered")
    class(FUN) <- c("ogive", class(FUN))
    FUN
}

### Essentially identical to stats:::print.ecdf.
print.ogive <- function(x, digits = getOption("digits") - 2, ...)
{
    ## Utility function
    numform <- function(x) paste(formatC(x, digits = digits), collapse = ", ")

    ## The rest is adapted from stats::ecdf
    cat("Ogive for grouped data \nCall: ")
    print(attr(x, "call"), ...)
    nc <- length(xxc <- get("x", envir = environment(x)))
    nn <- length(xxn <- get("y", envir = environment(x)))
    i1 <- 1L:min(3L, nc)
    i2 <- if (nc >= 4L) max(4L, nc - 1L):nc else integer(0)
    i3 <- 1L:min(3L, nn)
    i4 <- if (nn >= 4L) max(4L, nn - 1L):nn else integer(0)
    cat("    x = ", numform(xxc[i1]), if (nc > 3L) ", ",
        if (nc > 5L) " ..., ", numform(xxc[i2]), "\n", sep = "")
    cat(" F(x) = ", numform(xxn[i3]), if (nn > 3L) ", ",
        if (nn > 5L) " ..., ", numform(xxn[i4]), "\n", sep = "")
    invisible(x)
}

### Essentially identical to stats:::summary.ecdf.
summary.ogive <- function (object, ...)
{
    n <- length(eval(expression(x), envir = environment(object)))
    header <- paste("Ogive:	 ", n,
                    "unique values with summary\n")
    structure(summary(knots(object), ...),
              header = header, class = "summary.ogive")
}

### Identical to stats:::print.summary.ecdf.
print.summary.ogive <- function(x, ...)
{
    cat(attr(x, "header"))
    y <- x; attr(y, "header") <- NULL; class(y) <- "summaryDefault"
    print(y, ...)
    invisible(x)
}

### Identical to stats:::knots.stepfun.
knots.ogive <- function(Fn, ...)
    eval(expression(x), envir = environment(Fn))

plot.ogive <- function(x, main = NULL, xlab = "x", ylab = "F(x)", ...)
{
    if (missing(main))
        main <- {
            cl <- attr(x, "call")
            deparse(if (!is.null(cl)) cl else sys.call())
        }

    kn <- knots(x)
    Fn <- x(kn)
    plot(kn, Fn, ..., type = "o", pch = 16,
         main = main, xlab = xlab, ylab = ylab)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev,mgf}gamma functions to compute raw and
### limited moments, and the moment generating function for
### the Gamma distribution (as defined in R)
###
### See Chapter 17 of Johnson & Kotz, Continuous univariate
### distributions, volume 1, Wiley, 1970
###
### AUTHORS: Mathieu Pigeon, Christophe Dutang,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

mgamma <- function(order, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mgamma", order, shape, scale, FALSE)

levgamma <- function(limit, shape, rate = 1, scale = 1/rate, order = 1)
    .External(C_actuar_do_dpq, "levgamma", limit, shape, scale, order, FALSE)

mgfgamma <- function(t, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "mgfgamma", t, shape, scale, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}trbeta functions to compute
### characteristics of the Transformed Beta distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = Pr[Y <= (x/scale)^shape2 / (1 + (x/scale)^shape2)],
###
### where Y has a Beta distribution with parameters shape3 and shape1.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dtrbeta <-
    function (x, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              log = FALSE)
    .External(C_actuar_do_dpq, "dtrbeta", x, shape1, shape2, shape3, scale, log)

ptrbeta <-
    function (q, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ptrbeta", q, shape1, shape2, shape3, scale,
              lower.tail, log.p)

qtrbeta <-
    function (p, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qtrbeta", p, shape1, shape2, shape3, scale,
              lower.tail, log.p)

rtrbeta <-
    function (n, shape1, shape2, shape3, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rtrbeta", n, shape1, shape2, shape3, scale)

mtrbeta <-
    function (order, shape1, shape2, shape3, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mtrbeta", order, shape1, shape2, shape3, scale, FALSE)

levtrbeta <-
    function (limit, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              order = 1)
    .External(C_actuar_do_dpq, "levtrbeta", limit, shape1, shape2, shape3, scale, order, FALSE)


## Aliases
dpearson6 <- dtrbeta
ppearson6 <- ptrbeta
qpearson6 <- qtrbeta
rpearson6 <- rtrbeta
mpearson6 <- mtrbeta
levpearson6 <- levtrbeta

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Display all values of a matrix of vectors by 'unrolling' the
### object vertically or horizontally.
###
### AUTHORS: Louis-Philippe Pouliot,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

### New generic
severity <- function(x, ...) UseMethod("severity")

### Default method. Currently identical to 'unroll' by lack of a
### better alternative. This default method is never called in the
### package.
severity.default <- function(x, bycol = FALSE, drop = TRUE, ...)
{
    chkDots(...)                        # method does not use '...'
    unroll(x, bycol, drop)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev}beta functions to compute raw and limited
### moments for the Beta distribution (as defined in R). The
### noncentral beta distribution is _not_ supported.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

mbeta <- function(order, shape1, shape2)
    .External(C_actuar_do_dpq, "mbeta", order, shape1, shape2, FALSE)

levbeta <- function(limit, shape1, shape2, order = 1)
    .External(C_actuar_do_dpq, "levbeta", limit, shape1, shape2, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}llogis functions to compute
### characteristics of the loglogistic distribution. The version used
### in these functions has cumulative distribution function
###
###   Pr[X <= x] = v/(1 + v), v = (x/scale)^shape, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dllogis <- function (x, shape, rate = 1, scale = 1/rate, log = FALSE)
     .External(C_actuar_do_dpq, "dllogis", x, shape, scale, log)

pllogis <- function(q, shape, rate = 1, scale = 1/rate,
                    lower.tail = TRUE, log.p = FALSE)
     .External(C_actuar_do_dpq, "pllogis", q, shape, scale, lower.tail, log.p)

qllogis <- function(p, shape, rate = 1, scale = 1/rate,
                    lower.tail = TRUE, log.p = FALSE)
     .External(C_actuar_do_dpq, "qllogis", p, shape, scale, lower.tail, log.p)

rllogis <- function(n, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_random, "rllogis", n, shape, scale)

mllogis <- function(order, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_dpq, "mllogis", order, shape, scale, FALSE)

levllogis <- function(limit, shape, rate = 1, scale = 1/rate,
                      order = 1)
     .External(C_actuar_do_dpq, "levllogis", limit, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}invtrgamma functions to compute
### characteristics of the Inverse Transformed Gamma distribution. The
### version used in these functions has cumulative distribution
### function
###
###   Pr[X <= x] = 1 - pgamma((x/scale)^shape2, shape1, scale = 1)
###
### or, equivalently,
###
###   Pr[X <= x] = 1 - pgamma(x^shape2, shape1, scale = scale^shape2).
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvtrgamma <- function (x, shape1, shape2, rate = 1, scale = 1/rate,
                         log = FALSE)
    .External(C_actuar_do_dpq, "dinvtrgamma", x, shape1, shape2, scale, log)

pinvtrgamma <- function(q, shape1, shape2, rate = 1, scale = 1/rate,
                        lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvtrgamma", q, shape1, shape2, scale,
              lower.tail, log.p)

qinvtrgamma <- function(p, shape1, shape2, rate = 1, scale = 1/rate,
                        lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qinvtrgamma", p, shape1, shape2, scale,
              lower.tail, log.p)

rinvtrgamma <- function(n, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rinvtrgamma", n, shape1, shape2, scale)

minvtrgamma <- function(order, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "minvtrgamma", order, shape1, shape2, scale, FALSE)

levinvtrgamma <- function(limit, shape1, shape2, rate = 1, scale = 1/rate,
                          order = 1)
    .External(C_actuar_do_dpq, "levinvtrgamma", limit, shape1, shape2, scale,
              order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Minimum distance estimation.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

mde <- function(x, fun, start, measure = c("CvM", "chi-square", "LAS"),
                weights = NULL, ...)
{
    ## General form of the function to minimize.
    myfn <- function(parm, x, weights, ...)
    {
        y <- G(parm, x, ...) - Gn(x)
        drop(crossprod(weights * y, y))
    }

    ## Extract call; used to build the call to optim().
    Call <- match.call(expand.dots = TRUE)

    ## Argument checking
    if (missing(start) || !is.list(start))
        stop(sprintf("%s must be a named list", sQuote("start")))
    if (missing(fun) || !(is.function(fun)))
        stop(sprintf("%s must be supplied as a function", sQuote("fun")))
    grouped <- inherits(x, "grouped.data")
    if (!(is.numeric(x) || grouped))
        stop(sprintf("%s must be a numeric vector or an object of class %s",
                     sQuote("x"), dQuote("grouped.data")))

    ## Make sure that any argument of 'fun' specified in '...' is held
    ## fixed.
    dots <- names(list(...))
    dots <- dots[!is.element(dots, c("upper", "lower"))]
    start <- start[!is.element(names(start), dots)]

    ## Adapt 'fun' to our needs; taken from MASS::fitdistr.
    nm <- names(start)
    f <- formals(fun)
    args <- names(f)
    m <- match(nm, args)
    if (any(is.na(m)))
        stop(sprintf("%s specifies names which are not arguments to %s",
                     sQuote("start"), sQuote("fun")))
    formals(fun) <- c(f[c(1, m)], f[-c(1, m)]) # reorder arguments
    fn <- function(parm, x, ...) fun(x, parm, ...)
    if ((l <- length(nm)) > 1)
        body(fn) <- parse(text = paste("fun(x,", paste("parm[", 1:l, "]", collapse = ", "), ")"))

    measure <- match.arg(measure)

    ## Cramer-von Mises. Use the true and empirical cdf for individual
    ## data, or the true cdf and the ogive for grouped data.
    if (measure == "CvM")
    {
        G <- fn
        Gn <- if (grouped) ogive(x) else ecdf(x)
        if (is.null(weights)) weights <- 1
        Call$x <- knots(Gn)
        Call$par <- start
    }

    ## Modified Chi-square.
    if (measure == "chi-square")
    {
        if (!grouped)
            stop(sprintf("%s measure requires an object of class %s",
                         dQuote("chi-square"), dQuote("grouped.data")))
        if (any((nj <- x[, 2]) == 0))
            stop("frequency must be larger than 0 in all groups")
        og <- ogive(x)
        x <- knots(og)
        n <- sum(nj)
        G <- function(...) n * diff(fn(...))
        Gn <- function(...) n * diff(og(...))
        if (is.null(weights)) weights <- 1/nj
        Call$x <- x
        Call$par <- start
    }

    ## Layer average severity.
    if (measure == "LAS")
    {
        if (!grouped)
            stop(sprintf("%s measure requires an object of class %s",
                         dQuote("LAS"), dQuote("grouped.data")))
        e <- elev(x)
        x <- knots(e)
        G <- function(...) diff(fn(...))
        Gn <- function(...) diff(e(...))
        if (is.null(weights)) weights <- 1
        Call$x <- x
        Call$par <- start
    }

    ## optim() call
    Call[[1]] <- as.name("optim")
    Call$fun <- Call$start <- Call$measure <- NULL
    Call$fn <- myfn
    Call$weights <- weights
    Call$hessian <- FALSE
    if (is.null(Call$method))
    {
        if (any(c("lower", "upper") %in% names(Call)))
            Call$method <- "L-BFGS-B"
        else if (length(start) > 1)
            Call$method <- "BFGS"
        else
            Call$method <- "Nelder-Mead"
    }
    res <- eval(Call)

    ## Return result
    if (res$convergence > 0)
        stop("optimization failed")
    structure(list(estimate = res$par, distance = res$value),
              class = c("mde","list"))
}

print.mde <- function(x, digits = getOption("digits"), ...)
{
    ans1 <- format(x$estimate, digits = digits)
    ans1 <- sapply(ans1, function(x) paste("", x))
    nm1 <- names(ans1)
    nm1 <- paste(substring("      ", 1L, (nchar(ans1) - nchar(nm1)) %/% 2),
                 nm1)
    nm1 <- paste(nm1,
                 substring("      ", 1L, (nchar(ans1) - nchar(nm1)) %/% 2 + 1))
    names(ans1) <- nm1

    ans2 <- format(x$distance, digits = digits)
    ans2 <- sapply(ans2, function(x) paste("", x))
    nm2 <- "distance"
    nm2 <- paste(substring("      ", 1L, (nchar(ans2) - nchar(nm2)) %/% 2),
                 nm2)
    nm2 <- paste(nm2,
                 substring("      ", 1L, (nchar(ans2) - nchar(nm2)) %/% 2))
    names(ans2) <- nm2

    print(ans1, quote = FALSE)
    cat("\n")
    print(ans2, quote = FALSE)
    invisible(x)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Tests of functions for continuous and discrete probability
### distributions.
###
### Despite the name of the file, the tests are for [dpqrm,lev]<dist>
### functions (for continuous distributions):
###
###   d<dist>: density or probability mass
###   p<dist>: cumulative distribution
###   q<dist>: quantile
###   r<dist>: random number generation
###   m<dist>: moment
###   lev<dist>: limited moment
###
### Distributions are classified and sorted as in appendix A and
### appendix B of the 'distributions' package vignette.
###
### Inspired by (and some parts taken from) `tests/d-p-q-r-tests.R` in
### R sources.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca> with
###         indirect help from the R Core Team

## Load the package
library(actuar)
library(expint)                         # for gammainc

## Define a "local" version of the otherwise non-exported function
## 'betaint'.
betaint <- actuar:::betaint

## No warnings, unless explicitly asserted via tools::assertWarning.
options(warn = 2)
assertWarning <- tools::assertWarning

## Special values and utilities. Taken from `tests/d-p-q-r-tests.R`.
Meps <- .Machine$double.eps
xMax <- .Machine$double.xmax
xMin <- .Machine$double.xmin
All.eq <- function(x, y)
{
    all.equal.numeric(x, y, tolerance = 64 * .Machine$double.eps,
                      scale = max(0, mean(abs(x), na.rm = TRUE)))
}

if(!interactive())
    set.seed(123)

###
### CONTINUOUS DISTRIBUTIONS
###

##
## FELLER-PARETO AND PARETO II, III, IV DISTRIBUTIONS
##

## When reasonable, we also test consistency with the special cases
## min = 0:
##
##   Feller-Pareto -> Transformated beta
##   Pareto IV     -> Burr
##   Pareto III    -> Loglogistic
##   Pareto II     -> Pareto

## Density: first check that functions return 0 when scale = Inf, and
## when x = scale = Inf.
stopifnot(exprs = {
    dfpareto(c(42, Inf), min = 1, shape1 = 2, shape2 = 3, shape3 = 4, scale = Inf) == c(0, 0)
    dpareto4(c(42, Inf), min = 1, shape1 = 2, shape2 = 3, scale = Inf) == c(0, 0)
    dpareto3(c(42, Inf), min = 1, shape  = 3, scale = Inf) == c(0, 0)
    dpareto2(c(42, Inf), min = 1, shape  = 2, scale = Inf) == c(0, 0)
})

## Next test density functions for an array of standard values.
nshpar <- 3                     # (maximum) number of shape parameters
min <- round(rnorm(30, 2), 2)
shpar <- replicate(30, rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
for (i in seq_along(min))
{
    m <- min[i]
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, t)
    for (s in scpar)
    {
        x <- rfpareto(100, min = m, shape1 = a, shape2 = g, shape3 = t, scale = s)
        y <- (x - m)/s
        u <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            all.equal(d1 <- dfpareto(x, min = m,
                                     shape1 = a, shape2 = g, shape3 = t,
                                     scale = s),
                      d2 <- dfpareto(y, min = 0,
                                     shape1 = a, shape2 = g, shape3 = t,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      dtrbeta(y,
                              shape1 = a, shape2 = g, shape3 = t,
                              scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      g * y^(g*t - 1)/(s * Be * (1 + y^g)^(a + t)),
                      tolerance = 1e-10)
            all.equal(d1,
                      g * u^t * (1 - u)^a/((x - m) * Be),
                      tolerance = 1e-10)
        })
        x <- rpareto4(100, min = m, shape1 = a, shape2 = g, scale = s)
        y <- (x - m)/s
        u <- 1/(1 + y^g)
        stopifnot(exprs = {
            all.equal(d1 <- dpareto4(x, min = m,
                                     shape1 = a, shape2 = g,
                                     scale = s),
                      d2 <- dpareto4(y, min = 0,
                                     shape1 = a, shape2 = g,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      dburr(y,
                            shape1 = a, shape2 = g,
                            scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      a * g * y^(g - 1)/(s * (1 + y^g)^(a + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      a * g * u^a * (1 - u)/(x - m),
                      tolerance = 1e-10)
        })
        x <- rpareto3(100, min = m, shape = g, scale = s)
        y <- (x - m)/s
        u <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            all.equal(d1 <- dpareto3(x, min = m,
                                     shape = g,
                                     scale = s),
                      d2 <- dpareto3(y, min = 0,
                                     shape = g,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      dllogis(y,
                              shape = g,
                              scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      g * y^(g - 1)/(s * (1 + y^g)^2),
                      tolerance = 1e-10)
            all.equal(d1,
                      g * u * (1 - u)/(x - m),
                      tolerance = 1e-10)
        })
        x <- rpareto2(100, min = m, shape = a, scale = s)
        y <- (x - m)/s
        u <- 1/(1 + y)
        stopifnot(exprs = {
            all.equal(d1 <- dpareto2(x, min = m,
                                     shape = a,
                                     scale = s),
                      d2 <- dpareto2(y, min = 0,
                                     shape = a,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      dpareto(y,
                              shape = a,
                              scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      a/(s * (1 + y)^(a + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      a * u^a * (1 - u)/(x - m),
                      tolerance = 1e-10)
        })
    }
}

## Tests on the cumulative distribution function.
##
## Note: when shape1 = shape3 = 1, the underlying beta distribution is
## a uniform. Therefore, pfpareto(x, min, 1, shape2, 1, scale) should
## return the value of u = v/(1 + v), v = ((x - min)/scale)^shape2.
##
## x = 2/Meps = 2^53 (with min = 0, shape2 = scale = 1) is the value
## where the cdf would jump to 1 if we weren't using the trick to
## compute the cdf with pbeta(1 - u, ..., lower = FALSE).
scLrg <- 1e300 * c(0.5, 1, 2)
m <- rnorm(1)
stopifnot(exprs = {
    pfpareto(Inf,         min = 10,   1, 2, 3, scale = xMax) == 1
    pfpareto(2^53,        min = 0,    1, 1, 1, scale = 1) != 1
    pfpareto(2^53 + xMax, min = xMax, 1, 1, 1, scale = 1) != 1
    all.equal(pfpareto(xMin + m, min = m, 1, 1, 1, scale = 1), xMin)
    all.equal(y <- pfpareto(1e300 + m, min = m,
                            shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                            shape3 = 1,
                            scale = scLrg, log = TRUE),
              ptrbeta(1e300,
                      shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                      shape3 = 1,
                      scale = scLrg, log = TRUE))
    all.equal(y,
              c(pbeta(c(2/3, 1/2), 1, 3, lower.tail = TRUE,  log = TRUE),
                pbeta(2/3, 3, 1, lower.tail = FALSE, log = TRUE),
                pbeta(c(4/5, 1/2), 1, 3, lower.tail = TRUE,  log = TRUE),
                pbeta(4/5, 3, 1, lower.tail = FALSE, log = TRUE)))
})
stopifnot(exprs = {
    ppareto4(Inf,         min = 10,   1, 3, scale = xMax) == 1
    ppareto4(2^53,        min = 0,    1, 1, scale = 1) != 1
    ppareto4(2^53 + xMax, min = xMax, 1, 1, scale = 1) != 1
    all.equal(ppareto4(xMin + m, min = m, 1, 1, scale = 1), xMin)
    all.equal(y <- ppareto4(1e300 + m, min = m,
                            shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                            scale = scLrg, log = TRUE),
              pburr(1e300,
                    shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                    scale = scLrg, log = TRUE))
    all.equal(y,
              c(log1p(-c(1/3, 1/2, 2/3)^3),
                log1p(-c(1/5, 1/2, 4/5)^3)))
})
stopifnot(exprs = {
    ppareto3(Inf,         min = 10,   3, scale = xMax) == 1
    ppareto3(2^53,        min = 0,    1, scale = 1) != 1
    ppareto3(2^53 + xMax, min = xMax, 1, scale = 1) != 1
    all.equal(ppareto3(xMin + m, min = m, 1, scale = 1), xMin)
    all.equal(y <- ppareto3(1e300 + m, min = m,
                            shape = rep(c(1, 2), each = length(scLrg)),
                            scale = scLrg, log = TRUE),
              pllogis (1e300,
                       shape = rep(c(1, 2), each = length(scLrg)),
                       scale = scLrg, log = TRUE))
    all.equal(y,
              c(log(c(2/3, 1/2, 1/3)),
                log(c(4/5, 1/2, 1/5))))
})
stopifnot(exprs = {
    ppareto2(Inf,         min = 10,   3, scale = xMax) == 1
    ppareto2(2^53,        min = 0,    1, scale = 1) != 1
    ppareto2(2^53 + xMax, min = xMax, 1, scale = 1) != 1
    all.equal(ppareto2(xMin + m, min = m, 1, scale = 1), xMin)
    all.equal(y <- ppareto2(1e300 + m, min = m,
                            shape = 3,
                            scale = scLrg, log = TRUE),
              ppareto (1e300,
                       shape = 3,
                       scale = scLrg, log = TRUE))
    all.equal(y,
              c(log1p(-c(1/3, 1/2, 2/3)^3)))
})

## Also check that distribution functions return 0 when scale = Inf.
stopifnot(exprs = {
    pfpareto(x, min = m, shape1 = a, shape2 = g, shape3 = t, scale = Inf) == 0
    ppareto4(x, min = m, shape1 = a, shape2 = g, scale = Inf) == 0
    ppareto3(x, min = m, shape  = g, scale = Inf) == 0
    ppareto2(x, min = m, shape  = a, scale = Inf) == 0
})

## Tests for first three (positive) moments
##
## Simulation of new parameters ensuring that the first three moments
## exist.
set.seed(123)                   # reset the seed
nshpar <- 3                     # (maximum) number of shape parameters
min <- round(rnorm(30, 2), 2)
shpar <- replicate(30, c(3, 3, 0) + rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
for (i in seq_along(min))
{
    m <- min[i]
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, t)
    Ga <- gamma(a)
    for (s in scpar)
    {
        stopifnot(exprs = {
            All.eq(mfpareto(1, min = m,
                            shape1 = a, shape2 = g, shape3 = t,
                            scale = s),
                   m * (Be + (s/m) * beta(t + 1/g, a - 1/g))/Be)
            All.eq(mfpareto(2, min = m,
                            shape1 = a, shape2 = g, shape3 = t,
                            scale = s),
                   m^2 * (Be + 2 * (s/m) * beta(t + 1/g, a - 1/g)
                       + (s/m)^2 * beta(t + 2/g, a - 2/g))/Be)
            All.eq(mfpareto(3, min = m,
                            shape1 = a, shape2 = g, shape3 = t,
                            scale = s),
                   m^3 * (Be + 3 * (s/m) * beta(t + 1/g, a - 1/g)
                       + 3 * (s/m)^2 * beta(t + 2/g, a - 2/g)
                       + (s/m)^3 * beta(t + 3/g, a - 3/g))/Be)
        })
        stopifnot(exprs = {
            All.eq(mpareto4(1, min = m,
                            shape1 = a, shape2 = g,
                            scale = s),
                   m * (Ga + (s/m) * gamma(1 + 1/g) * gamma(a - 1/g))/Ga)
            All.eq(mpareto4(2, min = m,
                            shape1 = a, shape2 = g,
                            scale = s),
                   m^2 * (Ga + 2 * (s/m) * gamma(1 + 1/g) * gamma(a - 1/g)
                       + (s/m)^2 * gamma(1 + 2/g) * gamma(a - 2/g))/Ga)
            All.eq(mpareto4(3, min = m,
                            shape1 = a, shape2 = g,
                            scale = s),
                   m^3 * (Ga + 3 * (s/m) * gamma(1 + 1/g) * gamma(a - 1/g)
                       + 3 * (s/m)^2 * gamma(1 + 2/g) * gamma(a - 2/g)
                       + (s/m)^3 * gamma(1 + 3/g) * gamma(a - 3/g))/Ga)
        })
        stopifnot(exprs = {
            All.eq(mpareto3(1, min = m,
                            shape = g,
                            scale = s),
                   m * (1 + (s/m) * gamma(1 + 1/g) * gamma(1 - 1/g)))
            All.eq(mpareto3(2, min = m,
                            shape = g,
                            scale = s),
                   m^2 * (1 + 2 * (s/m) * gamma(1 + 1/g) * gamma(1 - 1/g)
                       + (s/m)^2 * gamma(1 + 2/g) * gamma(1 - 2/g)))
            All.eq(mpareto3(3, min = m,
                            shape = g,
                            scale = s),
                   m^3 * (1 + 3 * (s/m) * gamma(1 + 1/g) * gamma(1 - 1/g)
                       + 3 * (s/m)^2 * gamma(1 + 2/g) * gamma(1 - 2/g)
                       + (s/m)^3 * gamma(1 + 3/g) * gamma(1 - 3/g)))
        })
        stopifnot(exprs = {
            All.eq(mpareto2(1, min = m,
                            shape = a,
                            scale = s),
                   m * (Ga + (s/m) * gamma(1 + 1) * gamma(a - 1))/Ga)
            All.eq(mpareto2(2, min = m,
                            shape = a,
                            scale = s),
                   m^2 * (Ga + 2 * (s/m) * gamma(1 + 1) * gamma(a - 1)
                       + (s/m)^2 * gamma(1 + 2) * gamma(a - 2))/Ga)
            All.eq(mpareto2(3, min = m,
                            shape = a,
                            scale = s),
                   m^3 * (Ga + 3 * (s/m) * gamma(1 + 1) * gamma(a - 1)
                       + 3 * (s/m)^2 * gamma(1 + 2) * gamma(a - 2)
                       + (s/m)^3 * gamma(1 + 3) * gamma(a - 3))/Ga)
        })
    }
}

## Tests for first three limited moments
##
## Limits are taken from quantiles of each distribution.
q <- c(0.25, 0.50, 0.75, 0.9, 0.95)
for (i in seq_along(min))
{
    m <- min[i]
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Ga <- gamma(a)
    Gt <- gamma(t)
    for (s in scpar)
    {
        limit <- qfpareto(q, min = m,
                          shape1 = a, shape2 = g, shape3 = t,
                          scale = s)
        y <- (limit - m)/s
        u <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            All.eq(levfpareto(limit, order = 1, min = m,
                              shape1 = a, shape2 = g, shape3 = t,
                              scale = s),
                   m * (betaint(u, t, a) + (s/m) * betaint(u, t + 1/g, a - 1/g))/(Ga * Gt) +
                   limit * pbeta(u, t, a, lower = FALSE))
            All.eq(levfpareto(limit, order = 2, min = m,
                              shape1 = a, shape2 = g, shape3 = t,
                              scale = s),
                   m^2 * (betaint(u, t, a) + 2 * (s/m) * betaint(u, t + 1/g, a - 1/g)
                       + (s/m)^2 * betaint(u, t + 2/g, a - 2/g))/(Ga * Gt) +
                   limit^2 * pbeta(u, t, a, lower = FALSE))
            All.eq(levfpareto(limit, order = 3, min = m,
                              shape1 = a, shape2 = g, shape3 = t,
                              scale = s),
                   m^3 * (betaint(u, t, a) + 3 * (s/m) * betaint(u, t + 1/g, a - 1/g)
                       + 3 * (s/m)^2 * betaint(u, t + 2/g, a - 2/g)
                       + (s/m)^3 * betaint(u, t + 3/g, a - 3/g))/(Ga * Gt) +
                   limit^3 * pbeta(u, t, a, lower = FALSE))
        })
        limit <- qpareto4(q, min = m,
                          shape1 = a, shape2 = g,
                          scale = s)
        y <- (limit - m)/s
        u <- 1/(1 + y^g)
        u1m <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            All.eq(levpareto4(limit, order = 1, min = m,
                              shape1 = a, shape2 = g,
                              scale = s),
                   m * (betaint(u1m, 1, a) + (s/m) * betaint(u1m, 1 + 1/g, a - 1/g))/Ga +
                   limit * u^a)
            All.eq(levpareto4(limit, order = 2, min = m,
                              shape1 = a, shape2 = g,
                              scale = s),
                   m^2 * (betaint(u1m, 1, a) + 2 * (s/m) * betaint(u1m, 1 + 1/g, a - 1/g)
                       + (s/m)^2 * betaint(u1m, 1 + 2/g, a - 2/g))/Ga +
                   limit^2 * u^a)
            All.eq(levpareto4(limit, order = 3, min = m,
                              shape1 = a, shape2 = g,
                              scale = s),
                   m^3 * (betaint(u1m, 1, a) + 3 * (s/m) * betaint(u1m, 1 + 1/g, a - 1/g)
                       + 3 * (s/m)^2 * betaint(u1m, 1 + 2/g, a - 2/g)
                       + (s/m)^3 * betaint(u1m, 1 + 3/g, a - 3/g))/Ga +
                   limit^3 * u^a)
        })
        limit <- qpareto3(q, min = m,
                          shape = g,
                          scale = s)
        y <- (limit - m)/s
        u <- 1/(1 + y^(-g))
        u1m <- 1/(1 + y^g)
        stopifnot(exprs = {
            All.eq(levpareto3(limit, order = 1, min = m,
                              shape = g,
                              scale = s),
                   m * (u + (s/m) * betaint(u, 1 + 1/g, 1 - 1/g)) +
                   limit * u1m)
            All.eq(levpareto3(limit, order = 2, min = m,
                              shape = g,
                              scale = s),
                   m^2 * (u + 2 * (s/m) * betaint(u, 1 + 1/g, 1 - 1/g)
                       + (s/m)^2 * betaint(u, 1 + 2/g, 1 - 2/g)) +
                   limit^2 * u1m)
            All.eq(levpareto3(limit, order = 3, min = m,
                              shape = g,
                              scale = s),
                   m^3 * (u + 3 * (s/m) * betaint(u, 1 + 1/g, 1 - 1/g)
                       + 3 * (s/m)^2 * betaint(u, 1 + 2/g, 1 - 2/g)
                       + (s/m)^3 * betaint(u, 1 + 3/g, 1 - 3/g)) +
                   limit^3 * u1m)
        })
        limit <- qpareto2(q, min = m,
                          shape = a,
                          scale = s)
        y <- (limit - m)/s
        u <- 1/(1 + y)
        u1m <- 1/(1 + y^(-1))
        stopifnot(exprs = {
            All.eq(levpareto2(limit, order = 1, min = m,
                              shape = a,
                              scale = s),
                   m * (betaint(u1m, 1, a) + (s/m) * betaint(u1m, 1 + 1, a - 1))/Ga +
                   limit * u^a)
            All.eq(levpareto2(limit, order = 2, min = m,
                              shape = a,
                              scale = s),
                   m^2 * (betaint(u1m, 1, a) + 2 * (s/m) * betaint(u1m, 1 + 1, a - 1)
                       + (s/m)^2 * betaint(u1m, 1 + 2, a - 2))/Ga +
                   limit^2 * u^a)
            All.eq(levpareto2(limit, order = 3, min = m,
                              shape = a,
                              scale = s),
                   m^3 * (betaint(u1m, 1, a) + 3 * (s/m) * betaint(u1m, 1 + 1, a - 1)
                       + 3 * (s/m)^2 * betaint(u1m, 1 + 2, a - 2)
                       + (s/m)^3 * betaint(u1m, 1 + 3, a - 3))/Ga +
                   limit^3 * u^a)
        })
    }
}

##
## TRANSFORMED BETA FAMILY
##

## Density: first check that functions return 0 when scale = Inf, and
## when x = scale = Inf.
stopifnot(exprs = {
    dtrbeta      (c(42, Inf), shape1 = 2, shape2 = 3, shape3 = 4, scale = Inf) == c(0, 0)
    dburr        (c(42, Inf), shape1 = 2, shape2 = 3, scale = Inf) == c(0, 0)
    dllogis      (c(42, Inf), shape  = 3, scale = Inf) == c(0, 0)
    dparalogis   (c(42, Inf), shape  = 2, scale = Inf) == c(0, 0)
    dgenpareto   (c(42, Inf), shape1 = 2, shape2 = 4, scale = Inf) == c(0, 0)
    dpareto      (c(42, Inf), shape  = 2, scale = Inf) == c(0, 0)
    dinvburr     (c(42, Inf), shape1 = 4, shape2 = 3, scale = Inf) == c(0, 0)
    dinvpareto   (c(42, Inf), shape  = 4, scale = Inf) == c(0, 0)
    dinvparalogis(c(42, Inf), shape  = 4, scale = Inf) == c(0, 0)
})

## Next test density functions for an array of standard values.
set.seed(123)                   # reset the seed
nshpar <- 3                     # (maximum) number of shape parameters
shpar <- replicate(30, rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, t)
    for (s in scpar)
    {
        x <- rtrbeta(100, shape1 = a, shape2 = g, shape3 = t, scale = s)
        y <- x/s
        u <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            all.equal(d1 <- dtrbeta(x, shape1 = a, shape2 = g, shape3 = t,
                                    scale = s),
                      d2 <- dtrbeta(y, shape1 = a, shape2 = g, shape3 = t,
                                    scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      g * y^(g*t - 1)/(s * Be * (1 + y^g)^(a + t)),
                      tolerance = 1e-10)
            all.equal(d1,
                      g * u^t * (1 - u)^a/(x * Be),
                      tolerance = 1e-10)
        })
        x <- rburr(100, shape1 = a, shape2 = g, scale = s)
        y <- x/s
        u <- 1/(1 + y^g)
        stopifnot(exprs = {
            all.equal(d1 <- dburr(x, shape1 = a, shape2 = g,
                                  scale = s),
                      d2 <- dburr(y, shape1 = a, shape2 = g,
                                  scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      a * g * y^(g - 1)/(s * (1 + y^g)^(a + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      a * g * u^a * (1 - u)/x,
                      tolerance = 1e-10)
        })
        x <- rllogis(100, shape = g, scale = s)
        y <- x/s
        u <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            all.equal(d1 <- dllogis(x, shape = g,
                                    scale = s),
                      d2 <- dllogis(y, shape = g,
                                    scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      g * y^(g - 1)/(s * (1 + y^g)^2),
                      tolerance = 1e-10)
            all.equal(d1,
                      g * u * (1 - u)/x,
                      tolerance = 1e-10)
        })
        x <- rparalogis(100, shape = a, scale = s)
        y <- x/s
        u <- 1/(1 + y^a)
        stopifnot(exprs = {
            all.equal(d1 <- dparalogis(x, shape = a,
                                       scale = s),
                      d2 <- dparalogis(y, shape = a,
                                       scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      a^2 * y^(a - 1)/(s * (1 + y^a)^(a + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      a^2 * u^a * (1 - u)/x,
                      tolerance = 1e-10)
        })
        x <- rgenpareto(100, shape1 = a, shape2 = t, scale = s)
        y <- x/s
        u <- 1/(1 + y^(-1))
        stopifnot(exprs = {
            all.equal(d1 <- dgenpareto(x, shape1 = a, shape2 = t,
                                       scale = s),
                      d2 <- dgenpareto(y, shape1 = a, shape2 = t,
                                       scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      y^(t - 1)/(s * Be * (1 + y)^(a + t)),
                      tolerance = 1e-10)
            all.equal(d1,
                      u^t * (1 - u)^a/(x * Be),
                      tolerance = 1e-10)
        })
        x <- rpareto(100, shape = a, scale = s)
        y <- x/s
        u <- 1/(1 + y)
        stopifnot(exprs = {
            all.equal(d1 <- dpareto(x, shape = a,
                                    scale = s),
                      d2 <- dpareto(y, shape = a,
                                    scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      a/(s * (1 + y)^(a + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      a * u^a * (1 - u)/x,
                      tolerance = 1e-10)
        })
        x <- rpareto1(100, min = s, shape = a)
        stopifnot(exprs = {
            all.equal(d1 <- dpareto1(x, min = s, shape = a),
                      a * s^a/(x^(a + 1)),
                      tolerance = 1e-10)
        })
        x <- rinvburr(100, shape1 = t, shape2 = g, scale = s)
        y <- x/s
        u <- 1/(1 + y^(-g))
        stopifnot(exprs = {
            all.equal(d1 <- dinvburr(x, shape1 = t, shape2 = g,
                                     scale = s),
                      d2 <- dinvburr(y, shape1 = t, shape2 = g,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      t * g * y^(g*t - 1)/(s * (1 + y^g)^(t + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      t * g * u^t * (1 - u)/x,
                      tolerance = 1e-10)
        })
        x <- rinvpareto(100, shape = t, scale = s)
        y <- x/s
        u <- 1/(1 + y^(-1))
        stopifnot(exprs = {
            all.equal(d1 <- dinvpareto(x, shape = t,
                                 scale = s),
                      d2 <- dinvpareto(y, shape = t,
                                       scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      t * y^(t - 1)/(s * (1 + y)^(t + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      t * u^t * (1 - u)/x,
                      tolerance = 1e-10)
        })
        x <- rinvparalogis(100, shape = t, scale = s)
        y <- x/s
        u <- 1/(1 + y^(-t))
        stopifnot(exprs = {
            all.equal(d1 <- dinvparalogis(x, shape = t,
                                          scale = s),
                      d2 <- dinvparalogis(y, shape = t,
                                          scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      t^2 * y^(t^2 - 1)/(s * (1 + y^t)^(t + 1)),
                      tolerance = 1e-10)
            all.equal(d1,
                      t^2 * u^t * (1 - u)/x,
                      tolerance = 1e-10)
        })
    }
}

## Tests on the cumulative distribution function.
##
## Note: when shape1 = shape3 = 1, the underlying beta distribution is
## a uniform. Therefore, ptrbeta(x, 1, shape2, 1, scale) should return
## the value of u = v/(1 + v), v = (x/scale)^shape2.
##
## x = 2/Meps = 2^53 (with, shape2 = scale = 1) is the value where the
## cdf would jump to 1 if we weren't using the trick to compute the
## cdf with pbeta(1 - u, ..., lower = FALSE).
scLrg <- 1e300 * c(0.5, 1, 2)
stopifnot(exprs = {
    ptrbeta(Inf,  1, 2, 3, scale = xMax) == 1
    ptrbeta(2^53, 1, 1, 1, scale = 1) != 1
    all.equal(ptrbeta(xMin, 1, 1, 1, scale = 1), xMin)
    all.equal(ptrbeta(1e300,
                      shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                      shape3 = 1,
                      scale = scLrg, log = TRUE),
              c(pbeta(c(2/3, 1/2), 1, 3, lower.tail = TRUE,  log = TRUE),
                pbeta(2/3, 3, 1, lower.tail = FALSE, log = TRUE),
                pbeta(c(4/5, 1/2), 1, 3, lower.tail = TRUE,  log = TRUE),
                pbeta(4/5, 3, 1, lower.tail = FALSE, log = TRUE)))
})
stopifnot(exprs = {
    pburr(Inf,  1, 3, scale = xMax) == 1
    pburr(2^53, 1, 1, scale = 1) != 1
    all.equal(pburr(xMin, 1, 1, scale = 1), xMin)
    all.equal(pburr(1e300,
                    shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                    scale = scLrg, log = TRUE),
              c(log1p(-c(1/3, 1/2, 2/3)^3),
                log1p(-c(1/5, 1/2, 4/5)^3)))
})
stopifnot(exprs = {
    pllogis(Inf,  3, scale = xMax) == 1
    pllogis(2^53, 1, scale = 1) != 1
    all.equal(pllogis(xMin, 1, scale = 1), xMin)
    all.equal(pllogis(1e300,
                      shape = rep(c(1, 2), each = length(scLrg)),
                      scale = scLrg, log = TRUE),
              c(log(c(2/3, 1/2, 1/3)),
                log(c(4/5, 1/2, 1/5))))
})
stopifnot(exprs = {
    pparalogis(Inf,  3, scale = xMax) == 1
    pparalogis(2^53, 1, scale = 1) != 1
    all.equal(pparalogis(xMin, 1, scale = 1), xMin)
    all.equal(pparalogis(1e300,
                         shape = rep(c(2, 3), each = length(scLrg)),
                         scale = scLrg, log = TRUE),
              c(log1p(-c(1/5, 1/2, 4/5)^2),
                log1p(-c(1/9, 1/2, 8/9)^3)))
})
stopifnot(exprs = {
    pgenpareto(Inf,  1, 3, scale = xMax) == 1
    pgenpareto(2^53, 1, 1, scale = 1) != 1
    all.equal(pgenpareto(xMin, 1, 1, scale = 1), xMin)
    all.equal(pgenpareto(1e300,
                         shape1 = 3, shape2 = 1,
                         scale = scLrg, log = TRUE),
              c(pbeta(c(2/3, 1/2), 1, 3, lower.tail = TRUE,  log = TRUE),
                pbeta(2/3, 3, 1, lower.tail = FALSE, log = TRUE)))
})
stopifnot(exprs = {
    ppareto(Inf,  3, scale = xMax) == 1
    ppareto(2^53, 1, scale = 1) != 1
    all.equal(ppareto(xMin, 1, scale = 1), xMin)
    all.equal(ppareto(1e300,
                      shape = 3,
                      scale = scLrg, log = TRUE),
              c(log1p(-c(1/3, 1/2, 2/3)^3)))
})
stopifnot(exprs = {
    ppareto1(Inf,  3, min = xMax) == 1
    ppareto1(2^53, 1, min = 1) != 1
    all.equal(ppareto1(xMin, 1, min = 1), xMin)
    all.equal(ppareto1(1e300,
                      shape = 3,
                      min = 1e300 * c(0.001, 0.1, 0.5), log = TRUE),
              c(log1p(-c(0.001, 0.1, 0.5)^3)))
})
stopifnot(exprs = {
    pinvburr(Inf,  1, 3, scale = xMax) == 1
    pinvburr(2^53, 1, 1, scale = 1) != 1
    all.equal(pinvburr(xMin, 1, 1, scale = 1), xMin)
    all.equal(pinvburr(1e300,
                       shape1 = 3, shape2 = rep(c(1, 2), each = length(scLrg)),
                       scale = scLrg, log = TRUE),
              c(log(c(2/3, 1/2, 1/3)^3),
                log(c(4/5, 1/2, 1/5)^3)))
})
stopifnot(exprs = {
    pinvpareto(Inf,  3, scale = xMax) == 1
    pinvpareto(2^53, 1, scale = 1) != 1
    all.equal(pinvpareto(xMin, 1, scale = 1), xMin)
    all.equal(pinvpareto(1e300,
                         shape = 3,
                         scale = scLrg, log = TRUE),
              c(log(c(2/3, 1/2, 1/3)^3)))
})
stopifnot(exprs = {
    pinvparalogis(Inf,  3, scale = xMax) == 1
    pinvparalogis(2^53, 1, scale = 1) != 1
    all.equal(pinvparalogis(xMin, 1, scale = 1), xMin)
    all.equal(pinvparalogis(1e300,
                            shape = rep(c(2, 3), each = length(scLrg)),
                            scale = scLrg, log = TRUE),
              c(log(c(4/5, 1/2, 1/5)^2),
                log(c(8/9, 1/2, 1/9)^3)))
})

## Also check that distribution functions return 0 when scale = Inf.
stopifnot(exprs = {
    ptrbeta      (x, shape1 = a, shape2 = g, shape3 = t, scale = Inf) == 0
    pburr        (x, shape1 = a, shape2 = g, scale = Inf) == 0
    pllogis      (x, shape  = g, scale = Inf) == 0
    pparalogis   (x, shape  = a, scale = Inf) == 0
    pgenpareto   (x, shape1 = a, shape2 = t, scale = Inf) == 0
    ppareto      (x, shape  = a, scale = Inf) == 0
    pinvburr     (x, shape1 = t, shape2 = g, scale = Inf) == 0
    pinvpareto   (x, shape  = t, scale = Inf) == 0
    pinvparalogis(x, shape  = t, scale = Inf) == 0
})

## Tests for first three positive moments and first two negative
## moments.
##
## Simulation of new parameters ensuring that said moments exist.
set.seed(123)                   # reset the seed
nshpar <- 3                     # (maximum) number of shape parameters
shpar <- replicate(30, c(3, 3, 3) + rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
k <- c(-2, -1, 1, 2, 3)         # orders
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, t)
    Ga <- gamma(a)
    for (s in scpar)
    {
        stopifnot(exprs = {
            All.eq(mtrbeta(k, shape1 = a, shape2 = g, shape3 = t, scale = s),
                   s^k * beta(t + k/g, a - k/g)/Be)
            All.eq(mburr(k, shape1 = a, shape2 = g, scale = s),
                   s^k * gamma(1 + k/g) * gamma(a - k/g)/Ga)
            All.eq(mllogis(k, shape = g, scale = s),
                   s^k * gamma(1 + k/g) * gamma(1 - k/g))
            All.eq(mparalogis(k, shape = a, scale = s),
                   s^k * gamma(1 + k/a) * gamma(a - k/a)/Ga)
            All.eq(mgenpareto(k, shape1 = a, shape2 = t, scale = s),
                   s^k * beta(t + k, a - k)/Be)
            All.eq(mpareto(k[k > -1], shape = a, scale = s),
                   s^k[k > -1] * gamma(1 + k[k > -1]) * gamma(a - k[k > -1])/Ga)
            All.eq(mpareto1(k, shape = a, min = s),
                   s^k * a/(a - k))
            All.eq(minvburr(k, shape1 = a, shape2 = g, scale = s),
                   s^k * gamma(a + k/g) * gamma(1 - k/g)/Ga)
            All.eq(minvpareto(k[k < 1], shape = a, scale = s),
                s^k[k < 1] * gamma(a + k[k < 1]) * gamma(1 - k[k < 1])/Ga)
            All.eq(minvparalogis(k, shape = a, scale = s),
                   s^k * gamma(a + k/a) * gamma(1 - k/a)/Ga)
        })
    }
}

## Tests for first three positive limited moments and first two
## negative limited moments.
##
## Limits are taken from quantiles of each distribution.
order <- c(-2, -1, 1, 2, 3)             # orders
q <- c(0.25, 0.50, 0.75, 0.9, 0.95)     # quantiles
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Ga <- gamma(a)
    Gt <- gamma(t)
    for (s in scpar)
    {
        limit <- qtrbeta(q, shape1 = a, shape2 = g, shape3 = t, scale = s)
        y <- limit/s
        u <- 1/(1 + y^(-g))
        for (k in order)
            stopifnot(exprs = {
                All.eq(levtrbeta(limit, order = k, shape1 = a, shape2 = g, shape3 = t, scale = s),
                       s^k * betaint(u, t + k/g, a - k/g)/(Ga * Gt) +
                       limit^k * pbeta(u, t, a, lower = FALSE))
            })
        limit <- qburr(q, shape1 = a, shape2 = g, scale = s)
        y <- limit/s
        u <- 1/(1 + y^g)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levburr(limit, order = k, shape1 = a, shape2 = g, scale = s),
                       s^k * betaint(1 - u, 1 + k/g, a - k/g)/Ga +
                       limit^k * u^a)
            })
        limit <- qllogis(q, shape = g, scale = s)
        y <- limit/s
        u <- 1/(1 + y^(-g))
        for (k in order)
            stopifnot(exprs = {
                All.eq(levllogis(limit, order = k, shape = g, scale = s),
                       s^k * betaint(u, 1 + k/g, 1 - k/g) +
                       limit^k * (1 - u))
            })
        limit <- qparalogis(q, shape = a, scale = s)
        y <- limit/s
        u <- 1/(1 + y^a)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levparalogis(limit, order = k, shape = a, scale = s),
                       s^k * betaint(1 - u, 1 + k/a, a - k/a)/Ga +
                       limit^k * u^a)
            })
        limit <- qgenpareto(q, shape1 = a, shape2 = t, scale = s)
        y <- limit/s
        u <- 1/(1 + y^(-1))
        for (k in order)
            stopifnot(exprs = {
                All.eq(levgenpareto(limit, order = k, shape1 = a, shape2 = t, scale = s),
                       s^k * betaint(u, t + k, a - k)/(Ga * Gt) +
                       limit^k * pbeta(u, t, a, lower = FALSE))
            })
        limit <- qpareto(q, shape = a, scale = s)
        y <- limit/s
        u <- 1/(1 + y)
        for (k in order[order > -1])
            stopifnot(exprs = {
                All.eq(levpareto(limit, order = k, shape = a, scale = s),
                       s^k * betaint(1 - u, 1 + k, a - k)/Ga +
                       limit^k * u^a)
        })
        limit <- qpareto1(q, shape = a, min = s)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levpareto1(limit, order = k, shape = a, min = s),
                       s^k * a/(a - k) - k * s^a/((a - k) * limit^(a - k)))
            })
        limit <- qinvburr(q, shape1 = a, shape2 = g, scale = s)
        y <- limit/s
        u <- 1/(1 + y^(-g))
        for (k in order)
            stopifnot(exprs = {
                All.eq(levinvburr(limit, order = k, shape1 = a, shape2 = g, scale = s),
                       s^k * betaint(u, a + k/g, 1 - k/g)/Ga +
                       limit^k * (1 - u^a))
            })
        limit <- qinvpareto(q, shape = a, scale = s)
        y <- limit/s
        u <- 1/(1 + y^(-1))
        for (k in order[order < 1])
            stopifnot(exprs = {
                All.eq(levinvpareto(limit, order = k, shape = a, scale = s),
                       s^k * a *
                       sapply(u,
                              function(upper)
                                  integrate(function(x) x^(a+k-1) * (1-x)^(-k),
                                            lower = 0, upper = upper)$value) +
                       limit^k * (1 - u^a))
            })
        limit <- qinvparalogis(q, shape = a, scale = s)
        y <- limit/s
        u <- 1/(1 + y^(-a))
        for (k in order)
            stopifnot(exprs = {
                All.eq(levinvparalogis(limit, order = k, shape = a, scale = s),
                       s^k * betaint(u, a + k/a, 1 - k/a)/Ga +
                       limit^k * (1 - u^a))
        })
    }
}

##
## TRANSFORMED GAMMA AND INVERSE TRANSFORMED GAMMA FAMILIES
##

## Density: first check that functions return 0 when scale = Inf, and
## when x = scale = Inf (transformed gamma), or when scale = 0 and
## when x = scale = 0 (inverse distributions).
stopifnot(exprs = {
    dtrgamma   (c(42, Inf), shape1 = 2, shape2 = 3, scale = Inf) == c(0, 0)
    dinvtrgamma(c(42, 0), shape1 = 2, shape2 = 3, scale = 0) == c(0, 0)
    dinvgamma  (c(42, 0), shape  = 2, scale = 0) == c(0, 0)
    dinvweibull(c(42, 0), shape = 3, scale = 0) == c(0, 0)
    dinvexp    (c(42, 0), scale = 0) == c(0, 0)
})

## Tests on the density
set.seed(123)                   # reset the seed
nshpar <- 2                     # (maximum) number of shape parameters
shpar <- replicate(30, rgamma(nshpar, 5), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; t <- shpar[[c(i, 2)]]
    Ga <- gamma(a)
    for (s in scpar)
    {
        x <- rtrgamma(100, shape1 = a, shape2 = t, scale = s)
        y <- x/s
        u <- y^t
        stopifnot(exprs = {
            all.equal(d1 <- dtrgamma(x, shape1 = a, shape2 = t,
                                    scale = s),
                      d2 <- dtrgamma(y, shape1 = a, shape2 = t,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      t/(Ga * s^(a * t)) * x^(a * t - 1) * exp(-u),
                      tolerance = 1e-10)
            all.equal(d1,
                      t/(Ga * x) * u^a * exp(-u),
                      tolerance = 1e-10)
        })
        x <- rinvtrgamma(100, shape1 = a, shape2 = t, scale = s)
        y <- x/s
        u <- y^(-t)
        stopifnot(exprs = {
            all.equal(d1 <- dinvtrgamma(x, shape1 = a, shape2 = t,
                                        scale = s),
                      d2 <- dinvtrgamma(y, shape1 = a, shape2 = t,
                                        scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      t * s^(a * t)/(Ga * x^(a * t + 1)) * exp(-u),
                      tolerance = 1e-10)
            all.equal(d1,
                      t/(Ga * x) * u^a * exp(-u),
                      tolerance = 1e-10)
        })
        x <- rinvgamma(100, shape = a, scale = s)
        y <- x/s
        u <- y^(-1)
        stopifnot(exprs = {
            all.equal(d1 <- dinvgamma(x, shape = a, scale = s),
                      d2 <- dinvgamma(y, shape = a, scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      s^a/(Ga * x^(a + 1)) * exp(-u),
                      tolerance = 1e-10)
            all.equal(d1,
                      1/(Ga * x) * u^a * exp(-u),
                      tolerance = 1e-10)
        })
        x <- rinvweibull(100, shape = t, scale = s)
        y <- x/s
        u <- y^(-t)
        stopifnot(exprs = {
            all.equal(d1 <- dinvweibull(x, shape = t, scale = s),
                      d2 <- dinvweibull(y, shape = t, scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      t * s^t/x^(t + 1) * exp(-u),
                      tolerance = 1e-10)
            all.equal(d1,
                      t/x * u * exp(-u),
                      tolerance = 1e-10)
        })
        x <- rinvexp(100, scale = s)
        y <- x/s
        u <- y^(-1)
        stopifnot(exprs = {
            all.equal(d1 <- dinvexp(x, scale = s),
                      d2 <- dinvexp(y, scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d2,
                      s/x^2 * exp(-u),
                      tolerance = 1e-10)
            all.equal(d1,
                      1/x * u * exp(-u),
                      tolerance = 1e-10)
        })
    }
}

## Tests on the cumulative distribution function.
scLrg <- c(2, 100, 1e300 * c(0.1, 1, 10, 100), 1e307, xMax, Inf)
stopifnot(exprs = {
    ptrgamma(Inf,  2, 3, scale = xMax) == 1
    ptrgamma(xMax, 2, 3, scale = xMax) == pgamma(1, 2, 1)
    ptrgamma(xMin, 2, 1, scale = 1)    == pgamma(xMin, 2, 1)
    all.equal(ptrgamma(1e300, shape1 = 2, shape2 = 1, scale = scLrg, log = TRUE),
              pgamma(c(5e299, 1e+298, 10, 1, 0.1, 0.01, 1e-7, 1e+300/xMax, 0),
                     2, 1, log = TRUE))
})
scLrg <- c(2, 100, 1e300 * c(0.1, 1, 10, 100), 1e307, xMax, 0)
stopifnot(exprs = {
    pinvtrgamma(Inf,  2, 3, scale = xMax) == 1
    pinvtrgamma(xMax, 2, 3, scale = xMax) == pgamma(1, 2, 1, lower = FALSE)
    pinvtrgamma(xMin, 2, 1, scale = 1)    == pgamma(1/xMin, 2, 1, lower = FALSE)
    all.equal(pinvtrgamma(1e300, shape1 = 2, shape2 = 1, scale = scLrg, log = TRUE),
              pgamma(c(2e-300, 1e-298, 0.1, 1, 10, 100, 1e+7, xMax/1e+300, 0),
                     2, 1, lower = FALSE, log = TRUE))
})
stopifnot(exprs = {
    pinvgamma(Inf,  2, scale = xMax) == 1
    pinvgamma(xMax, 2, scale = xMax) == pgamma(1, 2, 1, lower = FALSE)
    pinvgamma(xMin, 2, scale = 1)    == pgamma(1/xMin, 2, 1, lower = FALSE)
    all.equal(pinvgamma(1e300, shape = 2, scale = scLrg, log = TRUE),
              pgamma(c(2e-300, 1e-298, 0.1, 1, 10, 100, 1e+7, xMax/1e+300, 0),
                     2, 1, lower = FALSE, log = TRUE))
})
stopifnot(exprs = {
    pinvweibull(Inf,  3, scale = xMax) == 1
    pinvweibull(xMax, 3, scale = xMax) == exp(-1)
    pinvweibull(xMin, 1, scale = 1)    == exp(-1/xMin)
    all.equal(pinvweibull(1e300, shape = 1, scale = scLrg, log = TRUE),
              -c(2e-300, 1e-298, 0.1, 1, 10, 100, 1e+7, xMax/1e+300, 0))
})
stopifnot(exprs = {
    pinvexp(Inf,  3, scale = xMax) == 1
    pinvexp(xMax, 3, scale = xMax) == exp(-1)
    pinvexp(xMin, 1, scale = 1)    == exp(-1/xMin)
    all.equal(pinvexp(1e300, scale = scLrg, log = TRUE),
              -c(2e-300, 1e-298, 0.1, 1, 10, 100, 1e+7, xMax/1e+300, 0))
})

## Tests for first three positive moments and first two negative
## moments. (Including for the Gamma, Weibull and Exponential
## distributions of base R.)
##
## Simulation of new parameters ensuring that said moments exist.
set.seed(123)                   # reset the seed
nshpar <- 2                     # (maximum) number of shape parameters
shpar <- replicate(30, c(3, 3) + rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
k <- c(-2, -1, 1, 2, 3)         # orders
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; t <- shpar[[c(i, 2)]]
    Ga <- gamma(a)
    for (s in scpar)
    {
        stopifnot(exprs = {
            All.eq(mtrgamma(k, shape1 = a, shape2 = t, scale = s),
                   s^k * gamma(a + k/t)/Ga)
            All.eq(mgamma(k, shape = a, scale = s),
                   s^k * gamma(a + k)/Ga)
            All.eq(mweibull(k, shape = t, scale = s),
                   s^k * gamma(1 + k/t))
            All.eq(mexp(k[k > -1], rate = 1/s),
                   s^k[k > -1] * gamma(1 + k[k > -1]))
            All.eq(minvtrgamma(k, shape1 = a, shape2 = t, scale = s),
                   s^k * gamma(a - k/t)/Ga)
            All.eq(minvgamma(k, shape = a, scale = s),
                   s^k * gamma(a - k)/Ga)
            All.eq(minvweibull(k, shape = t, scale = s),
                   s^k * gamma(1 - k/t))
            All.eq(minvexp(k[k < 1], scale = s),
                   s^k[k < 1] * gamma(1 - k[k < 1]))
        })
    }
}

## Tests for first three positive limited moments and first two
## negative limited moments. (Including for the Gamma, Weibull and
## Exponential distributions of base R.)
##
## Limits are taken from quantiles of each distribution.
order <- c(-2, -1, 1, 2, 3)             # orders
q <- c(0.25, 0.50, 0.75, 0.9, 0.95)     # quantiles
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; t <- shpar[[c(i, 2)]]
    Ga <- gamma(a)
    for (s in scpar)
    {
        limit <- qtrgamma(q, shape1 = a, shape2 = t, scale = s)
        y <- limit/s
        u <- y^t
        for (k in order)
            stopifnot(exprs = {
                All.eq(levtrgamma(limit, order = k, shape1 = a, shape2 = t, scale = s),
                       s^k * gamma(a + k/t)/Ga * pgamma(u, a + k/t, scale = 1) +
                       limit^k * pgamma(u, a, scale = 1, lower = FALSE))
            })
        limit <- qgamma(q, shape = a, scale = s)
        y <- limit/s
        for (k in order)
            stopifnot(exprs = {
                All.eq(levgamma(limit, order = k, shape = a, scale = s),
                       s^k * gamma(a + k)/Ga * pgamma(y, a + k, scale = 1) +
                       limit^k * pgamma(y, a, scale = 1, lower = FALSE))
            })
        limit <- qweibull(q, shape = t, scale = s)
        y <- limit/s
        u <- y^t
        for (k in order)
            stopifnot(exprs = {
                All.eq(levweibull(limit, order = k, shape = t, scale = s),
                       s^k * gamma(1 + k/t) * pgamma(u, 1 + k/t, scale = 1) +
                       limit^k * pgamma(u, 1, scale = 1, lower = FALSE))
            })
        limit <- qexp(q, rate = 1/s)
        y <- limit/s
        for (k in order[order > -1])
            stopifnot(exprs = {
                All.eq(levexp(limit, order = k, rate = 1/s),
                       s^k * gamma(1 + k) * pgamma(y, 1 + k, scale = 1) +
                       limit^k * pgamma(y, 1, scale = 1, lower = FALSE))
            })
        limit <- qinvtrgamma(q, shape1 = a, shape2 = t, scale = s)
        y <- limit/s
        u <- y^(-t)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levinvtrgamma(limit, order = k, shape1 = a, shape2 = t, scale = s),
                       s^k * (gammainc(a - k/t, u)/Ga) +
                       limit^k * pgamma(u, a, scale = 1))
            })
        limit <- qinvgamma(q, shape = a, scale = s)
        y <- limit/s
        u <- y^(-1)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levinvgamma(limit, order = k, shape = a, scale = s),
                       s^k * (gammainc(a - k, u)/Ga) +
                       limit^k * pgamma(u, a, scale = 1))
            })
        limit <- qinvweibull(q, shape = t, scale = s)
        y <- limit/s
        u <- y^(-t)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levinvweibull(limit, order = k, shape = t, scale = s),
                       s^k * gammainc(1 - k/t, u) +
                       limit^k * (-expm1(-u)))
            })
        limit <- qinvexp(q, scale = s)
        y <- limit/s
        u <- y^(-1)
        for (k in order)
            stopifnot(exprs = {
                All.eq(levinvexp(limit, order = k, scale = s),
                       s^k * gammainc(1 - k, u) +
                       limit^k * (-expm1(-u)))
            })
    }
}

##
## OTHER DISTRIBUTIONS
##

## Distributions in this category are quite different, so let's treat
## them separately.

## LOGGAMMA

## Tests on the density.
set.seed(123)                   # reset the seed
stopifnot(exprs = {
    dlgamma(c(42, Inf), shapelog = 2, ratelog = 0) == c(0, 0)
})
assertWarning(stopifnot(exprs = {
    is.nan(dlgamma(c(0, 42, Inf), shapelog = 2, ratelog = Inf))
}))
x <- rlgamma(100, shapelog = 2, ratelog = 1)
for (a in round(rlnorm(30), 2))
{
    Ga <- gamma(a)
    for (r in round(rlnorm(30), 2))
	stopifnot(exprs = {
            All.eq(dlgamma(x, shapelog = a, ratelog = r),
                   r^a * (log(x))^(a - 1)/(Ga * x^(r + 1)))
        })
}

## Tests on the cumulative distribution function.
assertWarning(stopifnot(exprs = {
    is.nan(plgamma(Inf,   1, ratelog = Inf))
    is.nan(plgamma(Inf, Inf, ratelog = Inf))
}))
scLrg <- log(c(2, 100, 1e300 * c(0.1, 1, 10, 100), 1e307, xMax, Inf))
stopifnot(exprs = {
    plgamma(Inf,  2, ratelog = xMax) == 1
    plgamma(xMax, 2, ratelog = 0) == 0
    all.equal(plgamma(1e300, 2, ratelog = 1/scLrg, log = TRUE),
              pgamma(log(1e300), 2, scale = scLrg, log = TRUE))
})

## Tests for first three positive moments and first two negative
## moments.
set.seed(123)                   # reset the seed
k <- c(-2, -1, 1, 2, 3)         # orders
for (a in round(rlnorm(30), 2))
{
    Ga <- gamma(a)
    for (r in 3 + round(rlnorm(30), 2))
	stopifnot(exprs = {
            All.eq(mlgamma(k, shapelog = a, ratelog = r),
                   (1 - k/r)^(-a))
        })
}

## Tests for first three positive limited moments and first two
## negative limited moments.
order <- c(-2, -1, 1, 2, 3)             # orders
q <- c(0.25, 0.50, 0.75, 0.9, 0.95)     # quantiles
for (a in round(rlnorm(30), 2))
{
    Ga <- gamma(a)
    for (r in 3 + round(rlnorm(30), 2))
    {
        limit <- qlgamma(q, shapelog = a, ratelog = r)
        for (k in order)
        {
            u <- log(limit)
        stopifnot(exprs = {
            All.eq(levlgamma(limit, order = k, shapelog = a, ratelog = r),
                   (1 - k/r)^(-a) * pgamma((r - k) * u, a, scale = 1) +
                   limit^k * pgamma(r * u, a, scale = 1,lower = FALSE))
            })
        }
    }
}

## GUMBEL

## Tests on the density.
set.seed(123)                   # reset the seed
stopifnot(exprs = {
    dgumbel(c(1, 3, Inf), alpha = 2,   scale = Inf) == c(0, 0, 0)
    dgumbel(c(1, 2, 3),   alpha = 2,   scale = 0) == c(0, Inf, 0)
    dgumbel(c(-Inf, Inf), alpha = 1,   scale = 1) == c(0, 0)
    dgumbel(1,            alpha = Inf, scale = 1) == 0
})
assertWarning(stopifnot(exprs = {
    is.nan(dgumbel(Inf,  alpha = Inf,  scale =  1))
    is.nan(dgumbel(-Inf, alpha = -Inf, scale =  1))
    is.nan(dgumbel(Inf,  alpha = 1,    scale = -1))
    is.nan(dgumbel(1,    alpha = 1,    scale = -1))
    is.nan(dgumbel(1,    alpha = Inf,  scale = -1))
}))
x <- rgumbel(100, alpha = 2, scale = 5)
for (a in round(rlnorm(30), 2))
{
    Ga <- gamma(a)
    for (s in round(rlnorm(30), 2))
    {
        u <- (x - a)/s
	stopifnot(exprs = {
            All.eq(dgumbel(x, alpha = a, scale = s),
                   exp(-(u + exp(-u)))/s)
        })
    }
}

## Tests on the cumulative distribution function.
assertWarning(stopifnot(exprs = {
    is.nan(pgumbel(Inf,  alpha = Inf,  scale =  1))
    is.nan(pgumbel(-Inf, alpha = -Inf, scale =  1))
    is.nan(pgumbel(Inf,  alpha = 1,    scale = -1))
    is.nan(pgumbel(1,    alpha = 1,    scale = -1))
    is.nan(pgumbel(1,    alpha = Inf,  scale = -1))
}))
scLrg <- c(2, 100, 1e300 * c(0.1, 1, 10, 100), 1e307, xMax, Inf)
stopifnot(exprs = {
    pgumbel(c(-Inf, Inf),  2, scale = xMax) == c(0, 1)
    pgumbel(c(xMin, xMax), 2, scale = 0) == c(0, 1)
    all.equal(pgumbel(1e300, 0, scale = scLrg, log = TRUE),
              -exp(-c(5e299, 1e+298, 10, 1, 0.1, 0.01, 1e-7, 1e+300/xMax, 0)))
})

## Test the first two moments, the only ones implemented.
assertWarning(stopifnot(exprs = {
    is.nan(mgumbel(c(-2, -1, 3, 4), alpha = 2, scale = 5))
}))
stopifnot(exprs = {
    All.eq(mgumbel(1, alpha = 2, scale = 5),
           2 + 5 * 0.577215664901532860606512090082)
    All.eq(mgumbel(2, alpha = 2, scale = 5),
           pi^2 * 25/6 + (2 + 5 * 0.577215664901532860606512090082)^2)
})

## INVERSE GAUSSIAN

## Tests on the density.
set.seed(123)                   # reset the seed
stopifnot(exprs = {
    dinvgauss(c(1, 3, Inf),  mean = 2,   dispersion = Inf) == c(0, 0, 0)
    dinvgauss(c(0, 42, Inf), mean = 2,   dispersion = 0) == c(Inf, 0, 0)
    dinvgauss(c(0, Inf),     mean = 1,   dispersion = 1) == c(0, 0)
    dinvgauss(1,             mean = Inf, dispersion = 2) == dinvgamma(1, 0.5, scale = 0.25)
})
assertWarning(stopifnot(exprs = {
    is.nan(dinvgauss(-Inf, mean = -1,  dispersion =  1))
    is.nan(dinvgauss(Inf,  mean = 1,   dispersion = -1))
    is.nan(dinvgauss(1,    mean = 1,   dispersion = -1))
    is.nan(dinvgauss(1,    mean = Inf, dispersion = -1))
}))
x <- rinvgauss(100, mean = 2, dispersion = 5)
for (mu in round(rlnorm(30), 2))
{
    for (phi in round(rlnorm(30), 2))
	stopifnot(exprs = {
            All.eq(dinvgauss(x, mean = mu, dispersion = phi),
                   1/sqrt(2*pi*phi*x^3) * exp(-((x/mu - 1)^2)/(2*phi*x)))
        })
}

## Tests on the cumulative distribution function.
assertWarning(stopifnot(exprs = {
    is.nan(pinvgauss(-Inf, mean = -Inf, dispersion =  1))
    is.nan(pinvgauss(Inf,  mean = 1,    dispersion = -1))
    is.nan(pinvgauss(1,    mean = Inf,  dispersion = -1))
}))
x <- c(1:50, 10^c(3:10, 20, 50, 150, 250))
sqx <- sqrt(x)
stopifnot(exprs = {
    pinvgauss(c(0, Inf),  mean = 2, dispersion = xMax) == c(0, 1)
    pinvgauss(c(0, xMax), mean = xMax, dispersion = 0) == c(0, 1)
    all.equal(pinvgauss(x, 1, dispersion = 1, log = TRUE),
              log(pnorm(sqx - 1/sqx) + exp(2) * pnorm(-sqx - 1/sqx)))
})

## Tests on pinvgauss for a small (<1e-14) coefficient of variation
## squared (equal to mean * dispersion) where a gamma approximation is
## used (ported from statmod v1.4.29).
mu <- 1.5
phi <- 6e-15
cv2 <- mu * phi
q <- mu + seq.int(from = -10, to = 10, length.out = 11) * 1e-8
stopifnot(exprs = {
    identical(pinvgauss(q, mean = mu, dispersion = phi),
              exp(pgamma(q, shape = 1/cv2, scale = cv2 * mu, log.p = TRUE)))
    ## test commented out for CRAN
    ## identical(pinvgauss(q, mean = mu, dispersion = phi),
    ##           statmod::pinvgauss(q, mean = mu, dispersion = phi))
})

## Tests on qinvgauss for a small (<1e-8) coefficient of variation
## squared (equal to mean * dispersion) where a gamma approximation is
## used (ported from statmod v1.4.29).
mu <- 1.5
phi <- 6e-9
cv2 <- phi * mu
p <- seq.int(from = 0.05, to = 0.95, by = 0.1)
stopifnot(exprs = {
    identical(qinvgauss(p, mean = mu, dispersion = phi),
              mu * qgamma(log(p), shape = 1/cv2, scale = cv2, log.p = TRUE))
    ## test commented out for CRAN ('All.eq' instead of 'identical'
    ## here because the order of computations is slightly different
    ## from statmod)
    ## All.eq(qinvgauss(p, mean = mu, dispersion = phi),
    ##           statmod::qinvgauss(p, mean = mu, dispersion = phi))
})

## Tests for small value of 'shape'. Added for the patch in 4294e9c.
q <- runif(100)
stopifnot(exprs = {
    All.eq(q,
           pinvgauss(qinvgauss(q, 0.1, 1e-2), 0.1, 1e-2))
    All.eq(q,
           pinvgauss(qinvgauss(q, 0.1, 1e-6), 0.1, 1e-6))
})

## Tests for the random number generator after patches for large
## values of phi (or mu) in 7b22ec7 (ported from statmod v1.4.29).
## First, check that the new generator returns the same values as the
## old one for "safe" values.
n <- 1000
mu <- round(rlnorm(n), 2)
phi <- round(rlnorm(n), 2)
xold <- xnew <- numeric(n)
for (i in seq_len(n))
{
    set.seed(1)
    phi0 <- phi[i] * mu[i]
    y <- rnorm(1)^2
    x <- 1 + phi0/2 * (y - sqrt(4 * y/phi0 + y^2))
    xold[i] <- mu[i] * ifelse(runif(1) <= 1/(1 + x), x, 1/x)
    set.seed(1)
    xnew[i] <- rinvgauss(1, mean = mu[i], dispersion = phi[i])
}
stopifnot(All.eq(xold, xnew))

## Next, check that the new generator returns only positive values.
stopifnot(exprs = {
    all(rinvgauss(1000, mean = 5e8, dispersion = 1) >= 0)
    all(rinvgauss(1000, mean = 1, dispersion = 5e8) >= 0)
    all(rinvgauss(1000, mean = 5e5, dispersion = 5e3) >= 0)
})

## Tests for first three positive, integer moments.
set.seed(123)                   # reset the seed
k <- 1:3
for (mu in round(rlnorm(30), 2))
{
    for (phi in round(rlnorm(30), 2))
	stopifnot(exprs = {
            All.eq(minvgauss(k, mean = mu, dispersion = phi),
                   c(mu,
                     mu^2 * (1 + phi * mu),
                     mu^3 * (1 + 3 * phi * mu + 3 * (phi * mu)^2)))
        })
}

## Tests for limited expected value.
q <- c(0.25, 0.50, 0.75, 0.9, 0.95)     # quantiles
for (mu in round(rlnorm(30), 2))
{
    for (phi in round(rlnorm(30), 2))
    {
        limit <- qinvgauss(q, mean = mu, dispersion = phi)
        stopifnot(exprs = {
            All.eq(levinvgauss(limit, mean = mu, dispersion = phi),
                   mu * (pnorm((limit/mu - 1)/sqrt(phi * limit)) -
                         exp(2/phi/mu) * pnorm(-(limit/mu + 1)/sqrt(phi * limit))) +
                   limit * pinvgauss(limit, mean = mu, dispersion = phi, lower = FALSE))
        })
    }
}

## GENERALIZED BETA
set.seed(123)                   # reset the seed
stopifnot(exprs = {
    dgenbeta(c(0, 2.5, 5), shape1 = 0,   shape2 = 0,   shape3 = 3,   scale = 5) == c(Inf, 0, Inf)
    dgenbeta(c(0, 2.5, 5), shape1 = 0,   shape2 = 0,   shape3 = 0,   scale = 5) == c(Inf, 0, Inf)
    dgenbeta(c(0, 2.5, 5), shape1 = 0,   shape2 = 2,   shape3 = 0,   scale = 5) == c(Inf, 0, 0)
    dgenbeta(c(0, 2.5, 5), shape1 = 0,   shape2 = Inf, shape3 = 3,   scale = 5) == c(Inf, 0, 0)
    dgenbeta(c(0, 2.5, 5), shape1 = 1,   shape2 = Inf, shape3 = 3,   scale = 5) == c(Inf, 0, 0)
    dgenbeta(c(0, 2.5, 5), shape1 = Inf, shape2 = Inf, shape3 = 3,   scale = 5) == c(0, Inf, 0)
    dgenbeta(c(0, 2.5, 5), shape1 = Inf, shape2 = Inf, shape3 = Inf, scale = 5) == c(0, 0, Inf)
})
nshpar <- 3                     # number of shape parameters
shpar <- replicate(30, rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; b <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, b)
    for (s in scpar)
    {
        u <- rbeta(100, a, b)
        y <- u^(1/t)
        x <- s * y
        stopifnot(exprs = {
            all.equal(d1 <- dgenbeta(x, shape1 = a, shape2 = b, shape3 = t,
                                     scale = s),
                      d2 <- dgenbeta(y, shape1 = a, shape2 = b, shape3 = t,
                                     scale = 1)/s,
                      tolerance = 1e-10)
            all.equal(d1,
                      t * y^(a*t - 1) * (1 - y^t)^(b - 1)/(s * Be),
                      tolerance = 1e-10)
            all.equal(d1,
                      t * u^a * (1 - u)^(b - 1)/(x * Be),
                      tolerance = 1e-10)
        })
    }
}

## Tests on the cumulative distribution function.
scLrg <- 1e300 * c(0.5, 1, 2, 4)
stopifnot(exprs = {
    all.equal(pgenbeta(1e300,
                       shape1 = 3, shape2 = 1,
                       shape3 = rep(c(1, 2), each = length(scLrg)),
                       scale = scLrg, log = TRUE),
              c(0, pbeta(c(1, 1/2, 1/4), 3, 1, log = TRUE),
                0, pbeta(c(1, 1/4, 1/16), 3, 1, log = TRUE)))
})

## Tests for first three positive moments and first two negative
## moments.
##
## Simulation of new parameters ensuring that said moments exist.
set.seed(123)                   # reset the seed
nshpar <- 3                     # number of shape parameters
shpar <- replicate(30, sqrt(c(3, 0, 3)) + rlnorm(nshpar, 2), simplify = FALSE)
scpar <- rlnorm(30, 2)          # scale parameters
k <- c(-2, -1, 1, 2, 3)         # orders
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; b <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, b)
    for (s in scpar)
        stopifnot(exprs = {
            All.eq(mgenbeta(k, shape1 = a, shape2 = b, shape3 = t, scale = s),
                   s^k * beta(a + k/t, b)/Be)
        })
}

## Tests for first three positive limited moments and first two
## negative limited moments.
##
## Simulation of new parameters ensuring that said moments exist.
order <- c(-2, -1, 1, 2, 3)             # orders
q <- c(0.25, 0.50, 0.75, 0.9, 0.95)     # quantiles
for (i in seq_along(shpar))
{
    a <- shpar[[c(i, 1)]]; g <- shpar[[c(i, 2)]]; t <- shpar[[c(i, 3)]]
    Be <- beta(a, b)
    for (s in scpar)
    {
        limit <- qgenbeta(q, shape1 = a, shape2 = b, shape3 = t, scale = s)
        u <- (limit/s)^t
        for (k in order)
            stopifnot(exprs = {
                All.eq(levgenbeta(limit, order = k, shape1 = a, shape2 = b, shape3 = t, scale = s),
                       s^k * beta(a + k/t, b)/Be * pbeta(u, a + k/t, b) +
                       limit^k * pbeta(u, a, b, lower = FALSE))
            })
    }
}

##
## RANDOM NUMBERS (all continuous distributions)
##
set.seed(123)
n <- 20
m <- rnorm(1)

## Generate variates
Rfpareto <- rfpareto(n, min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)
Rpareto4 <- rpareto4(n, min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2)
Rpareto3 <- rpareto3(n, min = m,               shape  = 1.5,             scale = 2)
Rpareto2 <- rpareto2(n, min = m, shape  = 0.8,                           scale = 2)
Rtrbeta        <- rtrbeta      (n, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)
Rburr          <- rburr        (n, shape1 = 0.8, shape2 = 1.5,             scale = 2)
Rllogis        <- rllogis      (n, shape  = 1.5, scale = 2)
Rparalogis     <- rparalogis   (n, shape  = 0.8, scale = 2)
Rgenpareto     <- rgenpareto   (n, shape1 = 0.8, shape2 = 2, scale = 2)
Rpareto        <- rpareto      (n, shape  = 0.8, scale = 2)
Rpareto1       <- rpareto1     (n, shape  = 0.8, min = 2)
Rinvburr       <- rinvburr     (n, shape1 = 1.5, shape2 = 2, scale = 2)
Rinvpareto     <- rinvpareto   (n, shape  = 2,   scale = 2)
Rinvparalogis  <- rinvparalogis(n, shape  = 2,   scale = 2)
Rtrgamma       <- rtrgamma     (n, shape1 = 2, shape2 = 3, scale = 5)
Rinvtrgamma    <- rinvtrgamma  (n, shape1 = 2, shape2 = 3, scale = 5)
Rinvgamma      <- rinvgamma    (n, shape  = 2,             scale = 5)
Rinvweibull    <- rinvweibull  (n,             shape  = 3, scale = 5)
Rinvexp        <- rinvexp      (n,                         scale = 5)
Rlgamma <- rlgamma(n, shapelog = 1.5, ratelog = 5)
Rgumbel <- rgumbel(n, alpha = 2, scale = 5)
Rinvgauss <- rinvgauss(n, mean = 2, dispersion = 5)
Rgenbeta <- rgenbeta(n, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)

## Compute quantiles
Pfpareto <- pfpareto(Rfpareto, min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)
Ppareto4 <- ppareto4(Rpareto4, min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2)
Ppareto3 <- ppareto3(Rpareto3, min = m,               shape  = 1.5,             scale = 2)
Ppareto2 <- ppareto2(Rpareto2, min = m, shape  = 0.8,                           scale = 2)
Ptrbeta        <- ptrbeta      (Rtrbeta,       shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)
Pburr          <- pburr        (Rburr,         shape1 = 0.8, shape2 = 1.5,             scale = 2)
Pllogis        <- pllogis      (Rllogis,       shape  = 1.5, scale = 2)
Pparalogis     <- pparalogis   (Rparalogis,    shape  = 0.8, scale = 2)
Pgenpareto     <- pgenpareto   (Rgenpareto,    shape1 = 0.8, shape2 = 2, scale = 2)
Ppareto        <- ppareto      (Rpareto,       shape  = 0.8, scale = 2)
Ppareto1       <- ppareto1     (Rpareto1,      shape  = 0.8, min = 2)
Pinvburr       <- pinvburr     (Rinvburr,      shape1 = 1.5, shape2 = 2, scale = 2)
Pinvpareto     <- pinvpareto   (Rinvpareto,    shape  = 2,   scale = 2)
Pinvparalogis  <- pinvparalogis(Rinvparalogis, shape  = 2,   scale = 2)
Ptrgamma       <- ptrgamma     (Rtrgamma,      shape1 = 2, shape2 = 3, scale = 5)
Pinvtrgamma    <- pinvtrgamma  (Rinvtrgamma,   shape1 = 2, shape2 = 3, scale = 5)
Pinvgamma      <- pinvgamma    (Rinvgamma,     shape  = 2,             scale = 5)
Pinvweibull    <- pinvweibull  (Rinvweibull,               shape  = 3, scale = 5)
Pinvexp        <- pinvexp      (Rinvexp,                               scale = 5)
Plgamma <- plgamma(Rlgamma, shapelog = 1.5, ratelog = 5)
Pgumbel <- pgumbel(Rgumbel, alpha = 2, scale = 5)
Pinvgauss <- pinvgauss(Rinvgauss, mean = 2, dispersion = 5)
Pgenbeta <- pgenbeta(Rgenbeta, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)

## Just compute pdf
Dfpareto <- dfpareto(Rfpareto, min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)
Dpareto4 <- dpareto4(Rpareto4, min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2)
Dpareto3 <- dpareto3(Rpareto3, min = m,               shape  = 1.5,             scale = 2)
Dpareto2 <- dpareto2(Rpareto2, min = m, shape  = 0.8,                           scale = 2)
Dtrbeta        <- dtrbeta      (Rtrbeta,       shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)
Dburr          <- dburr        (Rburr,         shape1 = 0.8, shape2 = 1.5,             scale = 2)
Dllogis        <- dllogis      (Rllogis,       shape  = 1.5, scale = 2)
Dparalogis     <- dparalogis   (Rparalogis,    shape  = 0.8, scale = 2)
Dgenpareto     <- dgenpareto   (Rgenpareto,    shape1 = 0.8, shape2 = 2, scale = 2)
Dpareto        <- dpareto      (Rpareto,       shape  = 0.8, scale = 2)
Dpareto1       <- dpareto1     (Rpareto1,      shape  = 0.8, min = 2)
Dinvburr       <- dinvburr     (Rinvburr,      shape1 = 1.5, shape2 = 2, scale = 2)
Dinvpareto     <- dinvpareto   (Rinvpareto,    shape  = 2,   scale = 2)
Dinvparalogis  <- dinvparalogis(Rinvparalogis, shape  = 2,   scale = 2)
Dtrgamma       <- dtrgamma     (Rtrgamma,      shape1 = 2, shape2 = 3, scale = 5)
Dinvtrgamma    <- dinvtrgamma  (Rinvtrgamma,   shape1 = 2, shape2 = 3, scale = 5)
Dinvgamma      <- dinvgamma    (Rinvtrgamma,   shape  = 2,             scale = 5)
Dinvweibull    <- dinvweibull  (Rinvweibull,               shape  = 3, scale = 5)
Dinvexp        <- dinvexp      (Rinvexp,                               scale = 5)
Dlgamma <- dlgamma(Rlgamma, shapelog = 1.5, ratelog = 5)
Dgumbel <- dgumbel(Rgumbel, alpha = 2, scale = 5)
Dinvgauss <- dinvgauss(Rinvgauss, mean = 2, dispersion = 5)
Dgenbeta <- dgenbeta(Rgenbeta, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2)

## Check q<dist>(p<dist>(.)) identity
stopifnot(exprs = {
    All.eq(Rfpareto, qfpareto(Pfpareto, min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2))
    All.eq(Rpareto4, qpareto4(Ppareto4, min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2))
    All.eq(Rpareto3, qpareto3(Ppareto3, min = m,               shape  = 1.5,             scale = 2))
    All.eq(Rpareto2, qpareto2(Ppareto2, min = m, shape  = 0.8,                           scale = 2))
    All.eq(Rtrbeta,       qtrbeta      (Ptrbeta,       shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2))
    All.eq(Rburr,         qburr        (Pburr,         shape1 = 0.8, shape2 = 1.5,             scale = 2))
    All.eq(Rllogis,       qllogis      (Pllogis,       shape  = 1.5, scale = 2))
    All.eq(Rparalogis,    qparalogis   (Pparalogis,    shape  = 0.8, scale = 2))
    All.eq(Rgenpareto,    qgenpareto   (Pgenpareto,    shape1 = 0.8, shape2 = 2, scale = 2))
    All.eq(Rpareto,       qpareto      (Ppareto,       shape  = 0.8, scale = 2))
    All.eq(Rpareto1,      qpareto1     (Ppareto1,      shape  = 0.8, min = 2))
    All.eq(Rinvburr,      qinvburr     (Pinvburr,      shape1 = 1.5, shape2 = 2, scale = 2))
    All.eq(Rinvpareto,    qinvpareto   (Pinvpareto,    shape  = 2,   scale = 2))
    All.eq(Rinvparalogis, qinvparalogis(Pinvparalogis, shape  = 2,   scale = 2))
    All.eq(Rtrgamma,      qtrgamma     (Ptrgamma,      shape1 = 2, shape2 = 3, scale = 5))
    All.eq(Rinvtrgamma,   qinvtrgamma  (Pinvtrgamma,   shape1 = 2, shape2 = 3, scale = 5))
    All.eq(Rinvgamma,     qinvgamma    (Pinvgamma,     shape  = 2,             scale = 5))
    All.eq(Rinvweibull,   qinvweibull  (Pinvweibull,               shape  = 3, scale = 5))
    All.eq(Rinvexp,       qinvexp      (Pinvexp,                               scale = 5))
    All.eq(Rlgamma, qlgamma(Plgamma, shapelog = 1.5, ratelog = 5))
    All.eq(Rgumbel, qgumbel(Pgumbel, alpha = 2, scale = 5))
    All.eq(Rinvgauss, qinvgauss(Pinvgauss, mean = 2, dispersion = 5))
    All.eq(Rgenbeta, qgenbeta(Pgenbeta, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2))
})

## Check q<dist>(p<dist>(.)) identity for special cases
stopifnot(exprs = {
    All.eq(Rfpareto - m, qtrbeta(Pfpareto, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2))
    All.eq(Rpareto4 - m, qburr  (Ppareto4, shape1 = 0.8, shape2 = 1.5,             scale = 2))
    All.eq(Rpareto3 - m, qllogis(Ppareto3,               shape  = 1.5,             scale = 2))
    All.eq(Rpareto2 - m, qpareto(Ppareto2, shape  = 0.8,                           scale = 2))
})

## Check q<dist>(p<dist>(.)) identity with upper tail
stopifnot(exprs = {
    All.eq(Rfpareto, qfpareto(1 - Pfpareto, min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, lower = FALSE))
    All.eq(Rpareto4, qpareto4(1 - Ppareto4, min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2, lower = FALSE))
    All.eq(Rpareto3, qpareto3(1 - Ppareto3, min = m,               shape  = 1.5,             scale = 2, lower = FALSE))
    All.eq(Rpareto2, qpareto2(1 - Ppareto2, min = m, shape  = 0.8,                           scale = 2, lower = FALSE))
    All.eq(Rtrbeta,       qtrbeta      (1 - Ptrbeta,       shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, lower = FALSE))
    All.eq(Rburr,         qburr        (1 - Pburr,         shape1 = 0.8, shape2 = 1.5,             scale = 2, lower = FALSE))
    All.eq(Rllogis,       qllogis      (1 - Pllogis,       shape  = 1.5, scale = 2, lower = FALSE))
    All.eq(Rparalogis,    qparalogis   (1 - Pparalogis,    shape  = 0.8, scale = 2, lower = FALSE))
    All.eq(Rgenpareto,    qgenpareto   (1 - Pgenpareto,    shape1 = 0.8, shape2 = 2, scale = 2, lower = FALSE))
    All.eq(Rpareto,       qpareto      (1 - Ppareto,       shape  = 0.8, scale = 2, lower = FALSE))
    All.eq(Rpareto1,      qpareto1     (1 - Ppareto1,      shape  = 0.8, min = 2, lower = FALSE))
    All.eq(Rinvburr,      qinvburr     (1 - Pinvburr,      shape1 = 1.5, shape2 = 2, scale = 2, lower = FALSE))
    All.eq(Rinvpareto,    qinvpareto   (1 - Pinvpareto,    shape  = 2,   scale = 2, lower = FALSE))
    All.eq(Rinvparalogis, qinvparalogis(1 - Pinvparalogis, shape  = 2,   scale = 2, lower = FALSE))
    All.eq(Rtrgamma,      qtrgamma     (1 - Ptrgamma,      shape1 = 2, shape2 = 3, scale = 5, lower = FALSE))
    All.eq(Rinvtrgamma,   qinvtrgamma  (1 - Pinvtrgamma,   shape1 = 2, shape2 = 3, scale = 5, lower = FALSE))
    All.eq(Rinvgamma,     qinvgamma    (1 - Pinvgamma,     shape  = 2,             scale = 5, lower = FALSE))
    All.eq(Rinvweibull,   qinvweibull  (1 - Pinvweibull,               shape  = 3, scale = 5, lower = FALSE))
    All.eq(Rinvexp,       qinvexp      (1 - Pinvexp,                               scale = 5, lower = FALSE))
    All.eq(Rlgamma, qlgamma(1 - Plgamma, shapelog = 1.5, ratelog = 5, lower = FALSE))
    All.eq(Rgumbel, qgumbel(1 - Pgumbel, alpha = 2, scale = 5, lower = FALSE))
    All.eq(Rinvgauss, qinvgauss(1 - Pinvgauss, mean = 2, dispersion = 5, lower = FALSE))
    All.eq(Rgenbeta, qgenbeta(1 - Pgenbeta, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, lower = FALSE))
})

## Check q<dist>(p<dist>(., log), log) identity
stopifnot(exprs = {
    All.eq(Rfpareto, qfpareto(log(Pfpareto), min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, log = TRUE))
    All.eq(Rpareto4, qpareto4(log(Ppareto4), min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2, log = TRUE))
    All.eq(Rpareto3, qpareto3(log(Ppareto3), min = m,               shape  = 1.5,             scale = 2, log = TRUE))
    All.eq(Rpareto2, qpareto2(log(Ppareto2), min = m, shape  = 0.8,                           scale = 2, log = TRUE))
    All.eq(Rtrbeta,       qtrbeta      (log(Ptrbeta),       shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, log = TRUE))
    All.eq(Rburr,         qburr        (log(Pburr),         shape1 = 0.8, shape2 = 1.5,             scale = 2, log = TRUE))
    All.eq(Rllogis,       qllogis      (log(Pllogis),       shape  = 1.5, scale = 2, log = TRUE))
    All.eq(Rparalogis,    qparalogis   (log(Pparalogis),    shape  = 0.8, scale = 2, log = TRUE))
    All.eq(Rgenpareto,    qgenpareto   (log(Pgenpareto),    shape1 = 0.8, shape2 = 2, scale = 2, log = TRUE))
    All.eq(Rpareto,       qpareto      (log(Ppareto),       shape  = 0.8, scale = 2, log = TRUE))
    All.eq(Rpareto1,      qpareto1     (log(Ppareto1),      shape  = 0.8, min = 2, log = TRUE))
    All.eq(Rinvburr,      qinvburr     (log(Pinvburr),      shape1 = 1.5, shape2 = 2, scale = 2, log = TRUE))
    All.eq(Rinvpareto,    qinvpareto   (log(Pinvpareto),    shape  = 2,   scale = 2, log = TRUE))
    All.eq(Rinvparalogis, qinvparalogis(log(Pinvparalogis), shape  = 2,   scale = 2, log = TRUE))
    All.eq(Rtrgamma,      qtrgamma     (log(Ptrgamma),      shape1 = 2, shape2 = 3, scale = 5, log = TRUE))
    All.eq(Rinvtrgamma,   qinvtrgamma  (log(Pinvtrgamma),   shape1 = 2, shape2 = 3, scale = 5, log = TRUE))
    All.eq(Rinvgamma,     qinvgamma    (log(Pinvgamma),     shape  = 2,             scale = 5, log = TRUE))
    All.eq(Rinvweibull,   qinvweibull  (log(Pinvweibull),               shape  = 3, scale = 5, log = TRUE))
    All.eq(Rinvexp,       qinvexp      (log(Pinvexp),                               scale = 5, log = TRUE))
    All.eq(Rlgamma, qlgamma(log(Plgamma), shapelog = 1.5, ratelog = 5, log = TRUE))
    All.eq(Rgumbel, qgumbel(log(Pgumbel), alpha = 2, scale = 5, log = TRUE))
    All.eq(Rinvgauss, qinvgauss(log(Pinvgauss), mean = 2, dispersion = 5, log = TRUE))
    All.eq(Rgenbeta, qgenbeta(log(Pgenbeta), shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, log = TRUE))
})

## Check q<dist>(p<dist>(., log), log) identity with upper tail
stopifnot(exprs = {
    All.eq(Rfpareto, qfpareto(log1p(-Pfpareto), min = m, shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rpareto4, qpareto4(log1p(-Ppareto4), min = m, shape1 = 0.8, shape2 = 1.5,             scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rpareto3, qpareto3(log1p(-Ppareto3), min = m,               shape  = 1.5,             scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rpareto2, qpareto2(log1p(-Ppareto2), min = m, shape  = 0.8,                           scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rtrbeta,       qtrbeta      (log1p(-Ptrbeta),       shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rburr,         qburr        (log1p(-Pburr),         shape1 = 0.8, shape2 = 1.5,             scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rllogis,       qllogis      (log1p(-Pllogis),       shape  = 1.5, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rparalogis,    qparalogis   (log1p(-Pparalogis),    shape  = 0.8, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rgenpareto,    qgenpareto   (log1p(-Pgenpareto),    shape1 = 0.8, shape2 = 2, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rpareto,       qpareto      (log1p(-Ppareto),       shape  = 0.8, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rpareto1,      qpareto1     (log1p(-Ppareto1),      shape  = 0.8, min = 2, lower = FALSE, log = TRUE))
    All.eq(Rinvburr,      qinvburr     (log1p(-Pinvburr),      shape1 = 1.5, shape2 = 2, scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rinvpareto,    qinvpareto   (log1p(-Pinvpareto),    shape  = 2,   scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rinvparalogis, qinvparalogis(log1p(-Pinvparalogis), shape  = 2,   scale = 2, lower = FALSE, log = TRUE))
    All.eq(Rtrgamma,      qtrgamma     (log1p(-Ptrgamma),      shape1 = 2, shape2 = 3, scale = 5, lower = FALSE, log = TRUE))
    All.eq(Rinvtrgamma,   qinvtrgamma  (log1p(-Pinvtrgamma),   shape1 = 2, shape2 = 3, scale = 5, lower = FALSE, log = TRUE))
    All.eq(Rinvgamma,     qinvgamma    (log1p(-Pinvgamma),     shape  = 2,             scale = 5, lower = FALSE, log = TRUE))
    All.eq(Rinvweibull,   qinvweibull  (log1p(-Pinvweibull),               shape  = 3, scale = 5, lower = FALSE, log = TRUE))
    All.eq(Rinvexp,       qinvexp      (log1p(-Pinvexp),                               scale = 5, lower = FALSE, log = TRUE))
    All.eq(Rlgamma, qlgamma(log1p(-Plgamma), shapelog = 1.5, ratelog = 5, lower = FALSE, log = TRUE))
    All.eq(Rgumbel, qgumbel(log1p(-Pgumbel), alpha = 2, scale = 5, lower = FALSE, log = TRUE))
    All.eq(Rinvgauss, qinvgauss(log1p(-Pinvgauss), mean = 2, dispersion = 5, lower = FALSE, log = TRUE))
    All.eq(Rgenbeta, qgenbeta(log1p(-Pgenbeta), shape1 = 0.8, shape2 = 1.5, shape3 = 2, scale = 2, lower = FALSE, log = TRUE))
})


###
### DISCRETE DISTRIBUTIONS
###

## Reset seed
set.seed(123)

## Define a small function to compute probabilities for the (a, b, 1)
## family of discrete distributions using the recursive relation
##
##   p[k] = (a + b/k)p[k - 1], k = 2, 3, ...
##
## for a, b and p[1] given.
dab1 <- function(x, a, b, p1)
{
    x <- floor(x)
    if (x < 1)
        stop("recursive computations possible for x >= 2 only")
    for (k in seq(2, length.out = x - 1))
    {
        p2 <- (a + b/k) * p1
        p1 <- p2
    }
    p1
}

## ZERO-TRUNCATED (a, b, 1) CLASS

## Tests on the probability mass function:
##
## 1. probability is 0 at x = 0;
## 2. pmf satisfies the recursive relation
## 3. limiting cases have mass 1 at x = 1
## 4. logarithmic is a special case of ztnbinom
lambda <- rlnorm(30, 2)                 # Poisson parameters
r <- lambda                             # size for negative binomial
prob <- runif(30)                       # probs
size <- round(lambda)                   # size for binomial
x <- sapply(size, sample, size = 1)
stopifnot(exprs = {
    dztpois(0, lambda) == 0
    dztnbinom(0, r, prob) == 0
    dztgeom(0, prob) == 0
    dztbinom(0, size, prob) == 0
    dlogarithmic(0, prob) == 0
})
stopifnot(exprs = {
    All.eq(dztpois(x, lambda),
           mapply(dab1, x,
                  a = 0,
                  b = lambda,
                  p1 = lambda/(exp(lambda) - 1)))
    All.eq(dztnbinom(x, r, prob),
           mapply(dab1, x,
                  a = 1 - prob,
                  b = (r - 1) * (1 - prob),
                  p1 = r * prob^r * (1 - prob)/(1 - prob^r)))
    All.eq(dztgeom(x, prob),
           mapply(dab1, x,
                  a = 1 - prob,
                  b = 0,
                  p1 = prob))
    All.eq(dztbinom(x, size, prob),
           mapply(dab1, x,
                  a = -prob/(1 - prob),
                  b = (size + 1) * prob/(1 - prob),
                  p1 = size * prob * (1 - prob)^(size - 1)/(1 - (1 - prob)^size)))
    All.eq(dlogarithmic(x, prob),
           mapply(dab1, x,
                  a = prob,
                  b = -prob,
                  p1 = -prob/log1p(-prob)))
})
p01 <- as.numeric(x == 1)
stopifnot(exprs = {
    dztpois(x, 0)           == p01
    dztnbinom(x, r, 1)      == p01
    dztgeom(x, 1)           == p01
    dztbinom(x, 1,    prob) == p01
    dztbinom(x, size, 0)    == p01
    dlogarithmic(x, 0)      == p01
})
stopifnot(exprs = {
    dztnbinom(x, 0, prob) == dlogarithmic(x, 1 - prob)
})

## Tests on the cumulative distribution function.
##
## 1. cdf is equal to cumulative sum of pdf
## 2. limiting cases have mass 1 at x = 1
## 3. logarithmic is a special case of ztnbinom
for (l in lambda)
    stopifnot(exprs = {
        all.equal(cumsum(dztpois(0:20, l)),
                  pztpois(0:20, l),
                  tolerance = 1e-8)
    })
for (i in seq_along(r))
    stopifnot(exprs = {
        all.equal(cumsum(dztnbinom(0:20, r[i], prob[i])),
                  pztnbinom(0:20, r[i], prob[i]),
                  tolerance = 1e-8)
    })
for (i in seq_along(r))
    stopifnot(exprs = {
        all.equal(cumsum(dztgeom(0:20, prob[i])),
                  pztgeom(0:20, prob[i]),
                  tolerance = 1e-8)
    })
for (i in seq_along(size))
    stopifnot(exprs = {
        all.equal(cumsum(dztbinom(0:20, size[i], prob[i])),
                  pztbinom(0:20, size[i], prob[i]),
                  tolerance = 1e-8)
    })
for (p in prob)
    stopifnot(exprs = {
        all.equal(cumsum(dlogarithmic(0:20, p)),
                  plogarithmic(0:20, p),
                  tolerance = 1e-8)
    })
p01 <- as.numeric(x >= 1)
stopifnot(exprs = {
    pztpois(x, 0)           == p01
    pztnbinom(x, r, 1)      == p01
    pztgeom(x, 1)           == p01
    pztbinom(x, 1,    prob) == p01
    pztbinom(x, size, 0)    == p01
    plogarithmic(x, 0)      == p01
})
stopifnot(exprs = {
    All.eq(pztnbinom(x, 0, prob),
           plogarithmic(x, 1 - prob))
})

## Tests on the limiting cases of the quantile function.
p <- c(0, runif(20), 1)
prob <- 0.2
stopifnot(exprs = {
    qztpois(p,      lambda = 0, TRUE,  FALSE) == 1
    qztpois(p,      lambda = 0, FALSE, FALSE) == 1
    qztpois(log(p), lambda = 0, TRUE,  TRUE)  == 1
    qztpois(log(p), lambda = 0, FALSE, TRUE)  == 1

    qztnbinom(p,      5, prob = 1, TRUE,  FALSE) == 1
    qztnbinom(p,      5, prob = 1, FALSE, FALSE) == 1
    qztnbinom(log(p), 5, prob = 1, TRUE,  TRUE)  == 1
    qztnbinom(log(p), 5, prob = 1, FALSE, TRUE)  == 1
    qztnbinom(p,      size = 0, prob,     TRUE,  FALSE) ==
        qlogarithmic(p,         1 - prob, TRUE,  FALSE)
    qztnbinom(p,      size = 0, prob,     FALSE, FALSE) ==
        qlogarithmic(p,         1 - prob, FALSE, FALSE)
    qztnbinom(log(p), size = 0, prob,     TRUE,  TRUE)  ==
        qlogarithmic(log(p),    1 - prob, TRUE,  TRUE)
    qztnbinom(log(p), size = 0, prob,     FALSE, TRUE)  ==
        qlogarithmic(log(p),    1 - prob, FALSE, TRUE)

    qztgeom(p,      prob = 1, TRUE,  FALSE) == 1
    qztgeom(p,      prob = 1, FALSE, FALSE) == 1
    qztgeom(log(p), prob = 1, TRUE,  TRUE)  == 1
    qztgeom(log(p), prob = 1, FALSE, TRUE)  == 1

    qztbinom(p,      size = 1, 0.5, TRUE,  FALSE) == 1
    qztbinom(p,      size = 1, 0.5, FALSE, FALSE) == 1
    qztbinom(log(p), size = 1, 0.5, TRUE,  TRUE)  == 1
    qztbinom(log(p), size = 1, 0.5, FALSE, TRUE)  == 1
    qztbinom(p,      5, prob = 0, TRUE,  FALSE) == 1
    qztbinom(p,      5, prob = 0, FALSE, FALSE) == 1
    qztbinom(log(p), 5, prob = 0, TRUE,  TRUE)  == 1
    qztbinom(log(p), 5, prob = 0, FALSE, TRUE)  == 1

    qlogarithmic(p,      prob = 0, TRUE,  FALSE) == 1
    qlogarithmic(p,      prob = 0, FALSE, FALSE) == 1
    qlogarithmic(log(p), prob = 0, TRUE,  TRUE)  == 1
    qlogarithmic(log(p), prob = 0, FALSE, TRUE)  == 1
})

## Tests at the extremes of the quantile function.
p <- c(0, 1)
q <- c(1, Inf)
lower <- rep(c(TRUE, FALSE), each = 2)
stopifnot(exprs = {
    qztpois(p,      1, lower, FALSE) == q
    qztpois(log(p), 1, lower, TRUE)  == q

    qztnbinom(p,      5, 0.5, lower, FALSE) == q
    qztnbinom(log(p), 5, 0.5, lower, TRUE)  == q

    qztgeom(p,      0.5, lower, FALSE) == q
    qztgeom(log(p), 0.5, lower, TRUE)  == q

    qztbinom(p,      5, 0.5, lower, FALSE) == q
    qztbinom(log(p), 5, 0.5, lower, TRUE)  == q

    qlogarithmic(p,      0.5, lower, FALSE) == q
    qlogarithmic(log(p), 0.5, lower, TRUE)  == q
})

## ZERO-MODIFIED (a, b, 1) CLASS

## Tests on the probability mass function:
##
## 1. probability is p0 at x = 0
## 2. pmf satisfies the recursive relation
## 3. limiting cases have mass 1 at x = 1
## 4. zmlogarithmic is a special case of zmnbinom
## 5. zm* with p0 = 0 is equal to zt*
lambda <- rlnorm(30, 2)                 # Poisson parameters
r <- lambda                             # size for negative binomial
prob <- runif(30)                       # probs
size <- round(lambda)                   # size for binomial
p0 <- c(0, runif(28), 1)                # probs at 0
x <- sapply(size, sample, size = 1)
stopifnot(exprs = {
    dzmpois(0, lambda, p0)      == p0
    dzmnbinom(0, r, prob, p0)   == p0
    dzmgeom(0, prob, p0)        == p0
    dzmbinom(0, size, prob, p0) == p0
    dzmlogarithmic(0, prob, p0) == p0
})
stopifnot(exprs = {
    All.eq(dzmpois(x, lambda, p0),
           mapply(dab1, x,
                  a = 0,
                  b = lambda,
                  p1 = (1 - p0) *lambda/(exp(lambda) - 1)))
    All.eq(dzmnbinom(x, r, prob, p0),
           mapply(dab1, x,
                  a = 1 - prob,
                  b = (r - 1) * (1 - prob),
                  p1 = (1 - p0) * r * prob^r * (1 - prob)/(1 - prob^r)))
    All.eq(dzmgeom(x, prob, p0),
           mapply(dab1, x,
                  a = 1 - prob,
                  b = 0,
                  p1 = (1 - p0) * prob))
    All.eq(dzmbinom(x, size, prob, p0),
           mapply(dab1, x,
                  a = -prob/(1 - prob),
                  b = (size + 1) * prob/(1 - prob),
                  p1 = (1 - p0) * size * prob * (1 - prob)^(size - 1)/(1 - (1 - prob)^size)))
    All.eq(dzmlogarithmic(x, prob, p0),
           mapply(dab1, x,
                  a = prob,
                  b = -prob,
                  p1 = -(1 - p0) * prob/log1p(-prob)))
})
p01 <- p0 * (x == 0) + (1 - p0) * (x == 1)
stopifnot(exprs = {
    dzmpois(x, 0, p0)           == p01
    dzmnbinom(x, r, 1, p0)      == p01
    dzmgeom(x, 1, p0)           == p01
    dzmbinom(x, 1,    prob, p0) == p01
    dzmbinom(x, size, 0,    p0) == p01
    dzmlogarithmic(x, 0, p0)    == p01
})
stopifnot(exprs = {
    dzmnbinom(x, 0, prob, p0) == dzmlogarithmic(x, 1 - prob, p0)
})
stopifnot(exprs = {
    dzmpois(x, lambda, 0)      == dztpois(x, lambda)
    dzmnbinom(x, r, prob, 0)   == dztnbinom(x, r, prob)
    dzmgeom(x, prob, 0)        == dztgeom(x, prob)
    dzmbinom(x, size, prob, 0) == dztbinom(x, size, prob)
    dzmlogarithmic(x, prob, 0) == dlogarithmic(x, prob)
})

## Tests on cumulative distribution function.
##
## 1. cdf is equal to cumulative sum of pdf
## 2. limiting cases have mass 1 at x = 1
## 3. zmlogarithmic is a special case of zmnbinom
## 4. zm* with p0 = 0 is equal to zt*
for (i in seq_along(lambda))
    stopifnot(exprs = {
        all.equal(cumsum(dzmpois(0:20, lambda[i], p0 = p0[i])),
                  pzmpois(0:20, lambda[i], p0 = p0[i]),
                  tolerance = 1e-8)
    })
for (i in seq_along(r))
    stopifnot(exprs = {
        all.equal(cumsum(dzmnbinom(0:20, r[i], prob[i], p0[i])),
                  pzmnbinom(0:20, r[i], prob[i], p0[i]),
                  tolerance = 1e-8)
    })
for (i in seq_along(r))
    stopifnot(exprs = {
        all.equal(cumsum(dzmgeom(0:20, prob[i], p0[i])),
                  pzmgeom(0:20, prob[i], p0[i]),
                  tolerance = 1e-8)
    })
for (i in seq_along(size))
    stopifnot(exprs = {
        all.equal(cumsum(dzmbinom(0:20, size[i], prob[i], p0[i])),
                  pzmbinom(0:20, size[i], prob[i], p0[i]),
                  tolerance = 1e-8)
    })
for (i in seq_along(prob))
    stopifnot(exprs = {
        all.equal(cumsum(dzmlogarithmic(0:20, prob[i], p0[i])),
                  pzmlogarithmic(0:20, prob[i], p0[i]),
                  tolerance = 1e-8)
    })
p01 <- p0 * (x < 1) + (x >= 1)
stopifnot(exprs = {
    pzmpois(x, 0, p0)           == p01
    pzmnbinom(x, r, 1, p0)      == p01
    pzmgeom(x, 1, p0)           == p01
    pzmbinom(x, 1,    prob, p0) == p01
    pzmbinom(x, size, 0,    p0) == p01
    pzmlogarithmic(x, 0, p0)    == p01
})
stopifnot(exprs = {
    All.eq(pzmnbinom(x, 0, prob, p0),
           pzmlogarithmic(x, 1 - prob, p0))
})
stopifnot(exprs = {
    All.eq(pzmpois(x, lambda, 0),
           pztpois(x, lambda))
    All.eq(pzmnbinom(x, r, prob, 0),
           pztnbinom(x, r, prob))
    All.eq(pzmgeom(x, prob, 0),
           pztgeom(x, prob))
    All.eq(pzmbinom(x, size, prob, 0),
           pztbinom(x, size, prob))
    All.eq(pzmlogarithmic(0, prob, 0),
           plogarithmic(0,   prob))
})

## Tests on the limiting cases of the quantile function.
p <- c(0, runif(20), 1)
p0m <- 0.2
stopifnot(exprs = {
    qzmpois(p,      lambda = 0, p0m, TRUE,  FALSE) == 1 - (p <= p0m)
    qzmpois(p,      lambda = 0, p0m, FALSE, FALSE) == 1 - (1 - p <= p0m)
    qzmpois(log(p), lambda = 0, p0m, TRUE,  TRUE)  == 1 - (p <= p0m)
    qzmpois(log(p), lambda = 0, p0m, FALSE, TRUE)  == 1 - (1 - p <= p0m)

    qzmnbinom(p,      size = 5, 1, p0m, TRUE,  FALSE) == 1 - (p <= p0m)
    qzmnbinom(p,      size = 5, 1, p0m, FALSE, FALSE) == 1 - (1 - p <= p0m)
    qzmnbinom(log(p), size = 5, 1, p0m, TRUE,  TRUE)  == 1 - (p <= p0m)
    qzmnbinom(log(p), size = 5, 1, p0m, FALSE, TRUE)  == 1 - (1 - p <= p0m)
    qzmnbinom(p,      size = 0, prob,     p0m, TRUE,  FALSE) ==
        qzmlogarithmic(p,       1 - prob, p0m, TRUE,  FALSE)
    qzmnbinom(p,      size = 0, prob,     p0m, FALSE, FALSE) ==
        qzmlogarithmic(p,       1 - prob, p0m, FALSE, FALSE)
    qzmnbinom(log(p), size = 0, prob,     p0m, TRUE,  TRUE)  ==
        qzmlogarithmic(log(p),  1 - prob, p0m, TRUE,  TRUE)
    qzmnbinom(log(p), size = 0, prob,     p0m, FALSE, TRUE)  ==
        qzmlogarithmic(log(p),  1 - prob, p0m, FALSE, TRUE)

    qzmgeom(p,      1, p0m, TRUE,  FALSE) == 1 - (p <= p0m)
    qzmgeom(p,      1, p0m, FALSE, FALSE) == 1 - (1 - p <= p0m)
    qzmgeom(log(p), 1, p0m, TRUE,  TRUE)  == 1 - (p <= p0m)
    qzmgeom(log(p), 1, p0m, FALSE, TRUE)  == 1 - (1 - p <= p0m)

    qzmbinom(p,      size = 1, 0.5, p0m, TRUE,  FALSE) == 1 - (p <= p0m)
    qzmbinom(p,      size = 1, 0.5, p0m, FALSE, FALSE) == 1 - (1 - p <= p0m)
    qzmbinom(log(p), size = 1, 0.5, p0m, TRUE,  TRUE)  == 1 - (p <= p0m)
    qzmbinom(log(p), size = 1, 0.5, p0m, FALSE, TRUE)  == 1 - (1 - p <= p0m)
    qzmbinom(p,      5, prob = 0, p0m, TRUE,  FALSE) == 1 - (p <= p0m)
    qzmbinom(p,      5, prob = 0, p0m, FALSE, FALSE) == 1 - (1 - p <= p0m)
    qzmbinom(log(p), 5, prob = 0, p0m, TRUE,  TRUE)  == 1 - (p <= p0m)
    qzmbinom(log(p), 5, prob = 0, p0m, FALSE, TRUE)  == 1 - (1 - p <= p0m)

    qzmlogarithmic(p,      0, p0m, TRUE,  FALSE) == 1 - (p <= p0m)
    qzmlogarithmic(p,      0, p0m, FALSE, FALSE) == 1 - (1 - p <= p0m)
    qzmlogarithmic(log(p), 0, p0m, TRUE,  TRUE)  == 1 - (p <= p0m)
    qzmlogarithmic(log(p), 0, p0m, FALSE, TRUE)  == 1 - (1 - p <= p0m)
})

## Tests with p0 = 0 at 0
stopifnot(exprs = {
    qzmpois(0,    2, p0 = 0, TRUE,  FALSE)  == 1
    qzmpois(1,    2, p0 = 0, FALSE, FALSE)  == 1
    qzmpois(-Inf, 2, p0 = 0, TRUE,  TRUE)   == 1
    qzmpois(0,    2, p0 = 0, FALSE, TRUE)   == 1

    qzmnbinom(0,    7, prob = 0.8, p0 = 0, TRUE,  FALSE) == 1
    qzmnbinom(1,    7, prob = 0.8, p0 = 0, FALSE, FALSE) == 1
    qzmnbinom(-Inf, 7, prob = 0.8, p0 = 0, TRUE,  TRUE)  == 1
    qzmnbinom(0,    7, prob = 0.8, p0 = 0, FALSE, TRUE)  == 1

    qzmgeom(0,    0.8, p0 = 0, TRUE,  FALSE) == 1
    qzmgeom(1,    0.8, p0 = 0, FALSE, FALSE) == 1
    qzmgeom(-Inf, 0.8, p0 = 0, TRUE,  TRUE)  == 1
    qzmgeom(0,    0.8, p0 = 0, FALSE, TRUE)  == 1

    qzmbinom(0,    5, 0.5, p0 = 0, TRUE,  FALSE) == 1
    qzmbinom(1,    5, 0.5, p0 = 0, FALSE, FALSE) == 1
    qzmbinom(-Inf, 5, 0.5, p0 = 0, TRUE,  TRUE)  == 1
    qzmbinom(0,    5, 0.5, p0 = 0, FALSE, TRUE)  == 1

    qzmlogarithmic(0,    0.9, p0 = 0, TRUE,  FALSE) == 1
    qzmlogarithmic(1,    0.9, p0 = 0, FALSE, FALSE) == 1
    qzmlogarithmic(-Inf, 0.9, p0 = 0, TRUE,  TRUE)  == 1
    qzmlogarithmic(0,    0.9, p0 = 0, FALSE, TRUE)  == 1
})

## Tests with p0 = 1 ('lower.tail' and 'log.p' without effect)
stopifnot(exprs = {
    qzmpois       (p, 2,      p0 = 1) == 0
    qzmnbinom     (p, 7, 0.8, p0 = 1) == 0
    qzmgeom       (p, 0.8,    p0 = 1) == 0
    qzmbinom      (p, 5, 0.5, p0 = 1) == 0
    qzmlogarithmic(p, 0.9,    p0 = 1) == 0
})

## Tests at the extremes of the quantile function.
p <- c(0, 1)
q <- c(0, Inf)
stopifnot(exprs = {
    qzmpois(p,      lambda = 2, p0m, log.p = FALSE) == q
    qzmpois(log(p), lambda = 2, p0m, log.p = TRUE)  == q

    qzmbinom(p,      size = 5, prob = 0.5, p0m, log.p = FALSE) == q
    qzmbinom(log(p), size = 5, prob = 0.5, p0m, log.p = TRUE)  == q

    qzmgeom(p,      prob = 0.5, p0m, log.p = FALSE) == q
    qzmgeom(log(p), prob = 0.5, p0m, log.p = TRUE)  == q

    qzmnbinom(p,      size = 5, prob = 0.5, p0m, log.p = FALSE) == q
    qzmnbinom(log(p), size = 5, prob = 0.5, p0m, log.p = TRUE)  == q

    qzmlogarithmic(p,      prob = 0.5, p0m, log.p = FALSE) == q
    qzmlogarithmic(log(p), prob = 0.5, p0m, log.p = TRUE)  == q
})


## POISSON-INVERSE GAUSSIAN

## Reset seed
set.seed(123)

## Define a small function to compute probabilities for the PIG
## directly using the Bessel function.
dpigBK <- function(x, mu, phi)
{
    M_LN2 <- 0.693147180559945309417232121458
    M_SQRT_2dPI <- 0.225791352644727432363097614947

    phimu  <- phi * mu
    lphi <- log(phi)
    y <- x - 0.5

    logA = -lphi/2 - M_SQRT_2dPI
    logB = (M_LN2 + lphi + log1p(1/(2 * phimu * mu)))/2;

    exp(logA + 1/phimu - lfactorial(x) - y * logB) *
        besselK(exp(logB - lphi), y)
}

## Tests on the probability mass function.
mu <- rlnorm(30, 2)
phi <- rlnorm(30, 2)
x <- 0:100
for (i in seq_along(phi))
{
    stopifnot(exprs = {
        all.equal(dpoisinvgauss(x, mean = mu[i], dispersion = phi[i]),
                  dpigBK(x, mu[i], phi[i]))
        all.equal(dpoisinvgauss(x, mean = Inf, dispersion = phi[i]),
                  dpigBK(x, Inf, phi[i]))
    })
}

## Tests on cumulative distribution function.
for (i in seq_along(phi))
    stopifnot(exprs = {
        all.equal(cumsum(dpoisinvgauss(0:20, mu[i], phi[i])),
                  ppoisinvgauss(0:20, mu[i], phi[i]),
                  tolerance = 1e-8)
        all.equal(cumsum(dpoisinvgauss(0:20, Inf, phi[i])),
                  ppoisinvgauss(0:20, Inf, phi[i]),
                  tolerance = 1e-8)
    })

##
## RANDOM NUMBERS (all discrete distributions)
##
set.seed(123)
n <- 20

## Generate variates.
##
## For zero-modified distributions, we simulate two sets of values:
## one with p0m < p0 (suffix 'p0lt') and one with p0m > p0 (suffix
## 'p0gt').
Rztpois      <- sort(unique(
                rztpois(n, lambda = 12)))
Rztnbinom    <- sort(unique(
                rztnbinom(n, size = 7, prob = 0.01)))
Rztgeom      <- sort(unique(
                rztgeom(n, prob = pi/16)))
Rztbinom     <- sort(unique(
                rztbinom(n, size = 55, prob = pi/16)))
Rlogarithmic <- sort(unique(
                rlogarithmic(n, prob = 0.99)))
Rzmpoisp0lt        <- sort(unique(
                      rzmpois(n, lambda = 6, p0 = 0.001)))
Rzmpoisp0gt        <- sort(unique(
                      rzmpois(n, lambda = 6, p0 = 0.010)))
Rzmnbinomp0lt      <- sort(unique(
                      rzmnbinom(n, size = 7, prob = 0.8, p0 = 0.01)))
Rzmnbinomp0gt      <- sort(unique(
                      rzmnbinom(n, size = 7, prob = 0.8, p0 = 0.40)))
Rzmgeomp0lt        <- sort(unique(
                      rzmgeom(n, prob = pi/16, p0 = 0.01)))
Rzmgeomp0gt        <- sort(unique(
                      rzmgeom(n, prob = pi/16, p0 = 0.40)))
Rzmbinomp0lt       <- sort(unique(
                      rzmbinom(n, size = 12, prob = pi/16, p0 = 0.01)))
Rzmbinomp0gt       <- sort(unique(
                      rzmbinom(n, size = 12, prob = pi/16, p0 = 0.12)))
Rzmlogarithmicp0lt <- sort(unique(
                      rzmlogarithmic(n, prob = 0.99, p0 = 0.05)))
Rzmlogarithmicp0gt <- sort(unique(
                      rzmlogarithmic(n, prob = 0.99, p0 = 0.55)))
Rpoisinvgauss    <- sort(unique(
                    rpoisinvgauss(n, mean = 12,  dispersion = 0.1)))
RpoisinvgaussInf <- sort(unique(
                    rpoisinvgauss(n, mean = Inf, dispersion = 1.1)))

## Compute quantiles
Pztpois      <- pztpois     (Rztpois,      lambda = 12)
Pztnbinom    <- pztnbinom   (Rztnbinom,    size = 7, prob = 0.01)
Pztgeom      <- pztgeom     (Rztgeom,      prob = pi/16)
Pztbinom     <- pztbinom    (Rztbinom,     size = 55, prob = pi/16)
Plogarithmic <- plogarithmic(Rlogarithmic, prob = 0.99)
Pzmpoisp0lt        <- pzmpois       (Rzmpoisp0lt,        lambda = 6, p0 = 0.001)
Pzmpoisp0gt        <- pzmpois       (Rzmpoisp0gt,        lambda = 6, p0 = 0.010)
Pzmnbinomp0lt      <- pzmnbinom     (Rzmnbinomp0lt,      size = 7, prob = 0.8, p0 = 0.01)
Pzmnbinomp0gt      <- pzmnbinom     (Rzmnbinomp0gt,      size = 7, prob = 0.8, p0 = 0.40)
Pzmgeomp0lt        <- pzmgeom       (Rzmgeomp0lt,        prob = pi/16, p0 = 0.01)
Pzmgeomp0gt        <- pzmgeom       (Rzmgeomp0gt,        prob = pi/16, p0 = 0.40)
Pzmbinomp0lt       <- pzmbinom      (Rzmbinomp0lt,       size = 12, prob = pi/16, p0 = 0.01)
Pzmbinomp0gt       <- pzmbinom      (Rzmbinomp0gt,       size = 12, prob = pi/16, p0 = 0.12)
Pzmlogarithmicp0lt <- pzmlogarithmic(Rzmlogarithmicp0lt, prob = 0.99, p0 = 0.05)
Pzmlogarithmicp0gt <- pzmlogarithmic(Rzmlogarithmicp0gt, prob = 0.99, p0 = 0.55)
Ppoisinvgauss    <- ppoisinvgauss(Rpoisinvgauss,    mean = 12,  dispersion = 0.1)
PpoisinvgaussInf <- ppoisinvgauss(RpoisinvgaussInf, mean = Inf, dispersion = 1.1)

## Just compute pmf
Dztpois      <- dztpois     (Rztpois,      lambda = 12)
Dztnbinom    <- dztnbinom   (Rztnbinom,    size = 7, prob = 0.01)
Dztgeom      <- dztgeom     (Rztgeom,      prob = pi/16)
Dztbinom     <- dztbinom    (Rztbinom,     size = 55, prob = pi/16)
Dlogarithmic <- dlogarithmic(Rlogarithmic, prob = pi/16)
Dzmpoisp0lt        <- dzmpois       (Rzmpoisp0lt,        lambda = 6, p0 = 0.001)
Dzmpoisp0gt        <- dzmpois       (Rzmpoisp0gt,        lambda = 6, p0 = 0.010)
Dzmnbinomp0lt      <- dzmnbinom     (Rzmnbinomp0lt,      size = 7, prob = 0.8, p0 = 0.01)
Dzmnbinomp0gt      <- dzmnbinom     (Rzmnbinomp0gt,      size = 7, prob = 0.8, p0 = 0.40)
Dzmgeomp0lt        <- dzmgeom       (Rzmgeomp0lt,        prob = pi/16, p0 = 0.01)
Dzmgeomp0gt        <- dzmgeom       (Rzmgeomp0gt,        prob = pi/16, p0 = 0.40)
Dzmbinomp0lt       <- dzmbinom      (Rzmbinomp0lt,       size = 12, prob = pi/16, p0 = 0.01)
Dzmbinomp0gt       <- dzmbinom      (Rzmbinomp0gt,       size = 12, prob = pi/16, p0 = 0.12)
Dzmlogarithmicp0lt <- dzmlogarithmic(Rzmlogarithmicp0lt, prob = 0.99, p0 = 0.05)
Dzmlogarithmicp0gt <- dzmlogarithmic(Rzmlogarithmicp0gt, prob = 0.99, p0 = 0.55)
Dpoisinvgauss    <- dpoisinvgauss(Rpoisinvgauss,    mean = 12,  dispersion = 0.1)
DpoisinvgaussInf <- dpoisinvgauss(RpoisinvgaussInf, mean = Inf, dispersion = 1.1)

## Check q<dist>(p<dist>(.)) identity
ep <- 1e-7
f1 <- 1 - ep # = 0.9999999
stopifnot(exprs = {
    Rztpois      == qztpois     (f1 * Pztpois,      lambda = 12)
    Rztnbinom    == qztnbinom   (f1 * Pztnbinom,    size = 7, prob = 0.01)
    Rztgeom      == qztgeom     (f1 * Pztgeom,      prob = pi/16)
    Rztbinom     == qztbinom    (f1 * Pztbinom,     size = 55, prob = pi/16)
    Rlogarithmic == qlogarithmic(f1 * Plogarithmic, prob = 0.99)
    Rzmpoisp0lt        == qzmpois       (Pzmpoisp0lt,        lambda = 6, p0 = 0.001)
    Rzmpoisp0gt        == qzmpois       (Pzmpoisp0gt,        lambda = 6, p0 = 0.010)
    Rzmnbinomp0lt      == qzmnbinom     (Pzmnbinomp0lt,      size = 7, prob = 0.8, p0 = 0.01)
    Rzmnbinomp0gt      == qzmnbinom     (Pzmnbinomp0gt,      size = 7, prob = 0.8, p0 = 0.40)
    Rzmgeomp0lt        == qzmgeom       (Pzmgeomp0lt,        prob = pi/16, p0 = 0.01)
    Rzmgeomp0gt        == qzmgeom       (Pzmgeomp0gt,        prob = pi/16, p0 = 0.40)
    Rzmbinomp0lt       == qzmbinom      (Pzmbinomp0lt,       size = 12, prob = pi/16, p0 = 0.01)
    Rzmbinomp0gt       == qzmbinom      (Pzmbinomp0gt,       size = 12, prob = pi/16, p0 = 0.12)
    Rzmlogarithmicp0lt == qzmlogarithmic(Pzmlogarithmicp0lt, prob = 0.99, p0 = 0.05)
    Rzmlogarithmicp0gt == qzmlogarithmic(Pzmlogarithmicp0gt, prob = 0.99, p0 = 0.55)
    Rpoisinvgauss    == qpoisinvgauss(Ppoisinvgauss,    mean = 12,  dispersion = 0.1)
    RpoisinvgaussInf == qpoisinvgauss(PpoisinvgaussInf, mean = Inf, dispersion = 1.1)
})

## Check q<dist>(p<dist>(.)) identity with upper tail
p1 <- 1 + ep # 1.0000001
stopifnot(exprs = {
    Rztpois      == qztpois     (p1 - Pztpois,      lambda = 12, lower = FALSE)
    Rztnbinom    == qztnbinom   (p1 - Pztnbinom,    size = 7, prob = 0.01, lower = FALSE)
    Rztgeom      == qztgeom     (p1 - Pztgeom,      prob = pi/16, lower = FALSE)
    Rztbinom     == qztbinom    (p1 - Pztbinom,     size = 55, prob = pi/16, lower = FALSE)
    Rlogarithmic == qlogarithmic(p1 - Plogarithmic, prob = 0.99, lower = FALSE)
    Rzmpoisp0lt        == qzmpois       (1 - Pzmpoisp0lt,        lambda = 6, p0 = 0.001, lower = FALSE)
    Rzmpoisp0gt        == qzmpois       (1 - Pzmpoisp0gt,        lambda = 6, p0 = 0.010, lower = FALSE)
    Rzmnbinomp0lt      == qzmnbinom     (1 - Pzmnbinomp0lt,      size = 7, prob = 0.8, p0 = 0.01, lower = FALSE)
    Rzmnbinomp0gt      == qzmnbinom     (1 - Pzmnbinomp0gt,      size = 7, prob = 0.8, p0 = 0.40, lower = FALSE)
    Rzmgeomp0lt        == qzmgeom       (1 - Pzmgeomp0lt,        prob = pi/16, p0 = 0.01, lower = FALSE)
    Rzmgeomp0gt        == qzmgeom       (1 - Pzmgeomp0gt,        prob = pi/16, p0 = 0.40, lower = FALSE)
    Rzmbinomp0lt       == qzmbinom      (1 - Pzmbinomp0lt,       size = 12, prob = pi/16, p0 = 0.01, lower = FALSE)
    Rzmbinomp0gt       == qzmbinom      (1 - Pzmbinomp0gt,       size = 12, prob = pi/16, p0 = 0.12, lower = FALSE)
    Rzmlogarithmicp0lt == qzmlogarithmic(1 - Pzmlogarithmicp0lt, prob = 0.99, p0 = 0.05, lower = FALSE)
    Rzmlogarithmicp0gt == qzmlogarithmic(1 - Pzmlogarithmicp0gt, prob = 0.99, p0 = 0.55, lower = FALSE)
    Rpoisinvgauss    == qpoisinvgauss(1 - Ppoisinvgauss,    mean = 12,  dispersion = 0.1, lower = FALSE)
    RpoisinvgaussInf == qpoisinvgauss(1 - PpoisinvgaussInf, mean = Inf, dispersion = 1.1, lower = FALSE)
})

## Check q<dist>(p<dist>(., log), log) identity
stopifnot(exprs = {
    Rztpois      == qztpois     (log(Pztpois) - ep,      lambda = 12, log = TRUE)
    Rztnbinom    == qztnbinom   (log(Pztnbinom) - ep,    size = 7, prob = 0.01, log = TRUE)
    Rztgeom      == qztgeom     (log(Pztgeom) - ep,      prob = pi/16, log = TRUE)
    Rztbinom     == qztbinom    (log(Pztbinom) - ep,     size = 55, prob = pi/16, log = TRUE)
    Rlogarithmic == qlogarithmic(log(Plogarithmic) - ep, prob = 0.99, log = TRUE)
    Rzmpoisp0lt        == qzmpois       (log(Pzmpoisp0lt),        lambda = 6, p0 = 0.001, log = TRUE)
    Rzmpoisp0gt        == qzmpois       (log(Pzmpoisp0gt),        lambda = 6, p0 = 0.010, log = TRUE)
    Rzmnbinomp0lt      == qzmnbinom     (log(Pzmnbinomp0lt),      size = 7, prob = 0.8, p0 = 0.01, log = TRUE)
    Rzmnbinomp0gt      == qzmnbinom     (log(Pzmnbinomp0gt),      size = 7, prob = 0.8, p0 = 0.40, log = TRUE)
    Rzmgeomp0lt        == qzmgeom       (log(Pzmgeomp0lt),        prob = pi/16, p0 = 0.01, log = TRUE)
    Rzmgeomp0gt        == qzmgeom       (log(Pzmgeomp0gt),        prob = pi/16, p0 = 0.40, log = TRUE)
    Rzmbinomp0lt       == qzmbinom      (log(Pzmbinomp0lt),       size = 12, prob = pi/16, p0 = 0.01, log = TRUE)
    Rzmbinomp0gt       == qzmbinom      (log(Pzmbinomp0gt),       size = 12, prob = pi/16, p0 = 0.12, log = TRUE)
    Rzmlogarithmicp0lt == qzmlogarithmic(log(Pzmlogarithmicp0lt), prob = 0.99, p0 = 0.05, log = TRUE)
    Rzmlogarithmicp0gt == qzmlogarithmic(log(Pzmlogarithmicp0gt), prob = 0.99, p0 = 0.55, log = TRUE)
    Rpoisinvgauss    == qpoisinvgauss(log(Ppoisinvgauss),    mean = 12,  dispersion = 0.1, log = TRUE)
    RpoisinvgaussInf == qpoisinvgauss(log(PpoisinvgaussInf), mean = Inf, dispersion = 1.1, log = TRUE)
})

## Check q<dist>(p<dist>(., log), log) identity with upper tail
stopifnot(exprs = {
    Rztpois      == qztpois     (ep + log1p(-Pztpois),      lambda = 12, lower = FALSE, log = TRUE)
    Rztnbinom    == qztnbinom   (ep + log1p(-Pztnbinom),    size = 7, prob = 0.01, lower = FALSE, log = TRUE)
    Rztgeom      == qztgeom     (ep + log1p(-Pztgeom),      prob = pi/16, lower = FALSE, log = TRUE)
    Rztbinom     == qztbinom    (ep + log1p(-Pztbinom),     size = 55, prob = pi/16, lower = FALSE, log = TRUE)
    Rlogarithmic == qlogarithmic(ep + log1p(-Plogarithmic), prob = 0.99, lower = FALSE, log = TRUE)
    Rzmpoisp0lt        == qzmpois       (log1p(-Pzmpoisp0lt),        lambda = 6, p0 = 0.001, lower = FALSE, log = TRUE)
    Rzmpoisp0gt        == qzmpois       (log1p(-Pzmpoisp0gt),        lambda = 6, p0 = 0.010, lower = FALSE, log = TRUE)
    Rzmnbinomp0lt      == qzmnbinom     (log1p(-Pzmnbinomp0lt),      size = 7, prob = 0.8, p0 = 0.01, lower = FALSE, log = TRUE)
    Rzmnbinomp0gt      == qzmnbinom     (log1p(-Pzmnbinomp0gt),      size = 7, prob = 0.8, p0 = 0.40, lower = FALSE, log = TRUE)
    Rzmgeomp0lt        == qzmgeom       (log1p(-Pzmgeomp0lt),        prob = pi/16, p0 = 0.01, lower = FALSE, log = TRUE)
    Rzmgeomp0gt        == qzmgeom       (log1p(-Pzmgeomp0gt),        prob = pi/16, p0 = 0.40, lower = FALSE, log = TRUE)
    Rzmbinomp0lt       == qzmbinom      (log1p(-Pzmbinomp0lt),       size = 12, prob = pi/16, p0 = 0.01, lower = FALSE, log = TRUE)
    Rzmbinomp0gt       == qzmbinom      (log1p(-Pzmbinomp0gt),       size = 12, prob = pi/16, p0 = 0.12, lower = FALSE, log = TRUE)
    Rzmlogarithmicp0lt == qzmlogarithmic(log1p(-Pzmlogarithmicp0lt), prob = 0.99, p0 = 0.05, lower = FALSE, log = TRUE)
    Rzmlogarithmicp0gt == qzmlogarithmic(log1p(-Pzmlogarithmicp0gt), prob = 0.99, p0 = 0.55, lower = FALSE, log = TRUE)
    Rpoisinvgauss    == qpoisinvgauss(log1p(-Ppoisinvgauss),    mean = 12,  dispersion = 0.1, lower = FALSE, log = TRUE)
    RpoisinvgaussInf == qpoisinvgauss(log1p(-PpoisinvgaussInf), mean = Inf, dispersion = 1.1, lower = FALSE, log = TRUE)
})

## Check limiting cases
stopifnot(exprs = {
    rztpois     (1, lambda = 0)              == 1
    rztnbinom   (1, size = 7, prob = 1)      == 1
    rztgeom     (1, prob = 1)                == 1
    rztbinom    (1, size = 1,  prob = pi/16) == 1
    rztbinom    (1, size = 55, prob = 0)     == 1
    rlogarithmic(1, prob = 0)                == 1
    rzmpois       (1, lambda = 0,              0.5) %in% c(0, 1)
    rzmnbinom     (1, size = 7, prob = 1,      0.5) %in% c(0, 1)
    rzmgeom       (1, prob = 1,                0.5) %in% c(0, 1)
    rzmbinom      (1, size = 1,  prob = pi/16, 0.5) %in% c(0, 1)
    rzmbinom      (1, size = 55, prob = 0,     0.5) %in% c(0, 1)
    rzmlogarithmic(1, prob = 0,                0.5) %in% c(0, 1)
})

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Tests for the 'var' and 'sd' methods for individual and grouped
### data.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

## Load the package
library(actuar)

###
### Individual data
###

## Check that results are identical to stats::var and stats::sd, as it
## should be.
stopifnot(exprs = {
    identical(var(dental), stats::var(dental))
    identical(sd(dental), stats::sd(dental))
})

## Check correct handling of missing data (issue #5 fixed with
## 62b92eff).
x <- c(dental, NA)
stopifnot(exprs = {
    identical(var(x, na.rm = TRUE), stats::var(x, na.rm = TRUE))
    identical(sd(x, na.rm = TRUE), stats::sd(x, na.rm = TRUE))
})

###
### Grouped data
###

## Extract group boundaries and frequencies from a grouped data
## object.
cj <- gdental[, 1]
nj <- gdental[, 2]

## Compute variance and standard deviation by hand.
midpoints <- cj[-length(cj)] + diff(cj)/2
means <- drop(crossprod(nj, midpoints)/sum(nj))
v <- drop(crossprod(nj, (midpoints - means)^2)/(sum(nj) - 1))
s <- sqrt(v)
stopifnot(exprs = {
    all.equal(v, unname(var(gdental)))
    all.equal(s, unname(sd(gdental)))
})

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m}invpareto functions to compute
### characteristics of the Inverse Pareto distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = (x/(x + scale))^shape, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvpareto <- function(x, shape, scale, log = FALSE)
     .External(C_actuar_do_dpq, "dinvpareto", x, shape, scale, log)

pinvpareto <- function(q, shape, scale, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvpareto", q, shape, scale, lower.tail, log.p)

qinvpareto <- function(p, shape, scale, lower.tail = TRUE, log.p = FALSE)
     .External(C_actuar_do_dpq, "qinvpareto", p, shape, scale, lower.tail, log.p)

rinvpareto <- function(n, shape, scale)
     .External(C_actuar_do_random, "rinvpareto", n, shape, scale)

minvpareto <- function(order, shape, scale)
     .External(C_actuar_do_dpq, "minvpareto", order, shape, scale, FALSE)

levinvpareto <- function(limit, shape, scale, order = 1)
     .External(C_actuar_do_dpq, "levinvpareto", limit, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Main interface to credibility model fitting functions.
###
### AUTHORS: Louis-Philippe Pouliot, Tommy Ouellet,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>.

cm <- function(formula, data, ratios, weights, subset,
               regformula = NULL, regdata, adj.intercept = FALSE,
               method = c("Buhlmann-Gisler", "Ohlsson", "iterative"),
               likelihood, ...,
               tol = sqrt(.Machine$double.eps), maxit = 100,
               echo = FALSE)
{
    Call <- match.call(expand.dots = TRUE)

    ## Catch the pure bayesian special case.
    if (formula == "bayes")
    {
        if (missing(data) || length(data) == 0L)
            data <- NULL
        res <- bayes(data, likelihood, ...)
        class(res) <- c("cm", class(res))
        attr(res, "call") <- Call
        return(res)
    }

    ## === MODEL ANALYSIS ===
    ##
    ## Decompose the formula giving the portfolio structure. Attribute
    ## "order" gives the interaction level of each term in the
    ## formula. In hierarchical structures, each term should represent
    ## a different level, hence there should not be any duplicates in
    ## this attribute. The column names in 'data' containing the
    ## portfolio structure can be obtained from the rownames of the
    ## matrix in attribute "factors".
    ##
    ## Note that the very last level, the data, is not taken into
    ## account here.
    tf <- terms(formula)
    level.numbers <- attr(tf, "order")           # level IDs
    level.names <- rownames(attr(tf, "factors")) # level names
    nlevels <- length(level.names)               # number of levels

    ## Sanity checks
    ##
    ## 1. only hierarchical interactions are allowed in 'formula';
    ## 2. hierarchical regression models are not supported.
    ## 3. if 'ratios' is missing, all columns of 'data' are taken to
    ##    be ratios, so 'weights' should also be missing;
    ##
    if (any(duplicated(level.numbers)))
        stop(sprintf("unsupported interactions in %s",
                     sQuote("formula")))
    if (nlevels > 1 && !is.null(regformula))
        stop("hierarchical regression models not supported")
    if (missing(ratios) & !missing(weights))
        stop("ratios have to be supplied if weights are")

    ## === DATA EXTRACTION ===
    ##
    ## 'data' is split into three matrices: one for the portfolio
    ## structure, one for the ratios and one for the weights. They are
    ## obtained via calls to subset() built from this function's
    ## call. That way, arguments 'ratios', 'weights' and 'subset' are
    ## not evaluated before being passed to subset(). Argument
    ## matching is as follows:
    ##
    ##   Argument of cm()     Argument of subset()
    ##   ================     ====================
    ##     data                 x
    ##     ratios               select
    ##     weights              select
    ##     subset               subset
    ##
    ## Positions of the arguments that will be needed.
    m <- match(c("data", "ratios", "weights", "subset"), names(Call), 0)

    ## Extraction of the portfolio structure. Arguments 'data' and
    ## 'subset' are passed to subset().
    cl <- Call[c(1, m[c(1, 4)])]        # use data and subset only
    cl[[1]] <- as.name("subset")        # change function name
    names(cl)[2] <- "x"                 # argument matching
    cl$select <- level.names            # add argument 'select'
    levs <- eval(cl, parent.frame())    # extraction

    ## Object 'levs' is a data frame or matrix with as many colums as
    ## there are levels in the model (still notwithstanding the data
    ## level). Rows contain nodes identifiers which can be
    ## anything. For calculations, these identifiers are converted
    ## into simple subscripts (i, j, k, ...) as used in mathematical
    ## notation.
    ##
    ## Note that 'apply' will coerce to a matrix.
    ilevs <- apply(levs, 2, function(x) as.integer(factor(x)))

    ## Extraction of the ratios. If argument 'ratios' is missing, then
    ## use all columns of 'data' except those of the portfolio
    ## structure.
    cl$select <-
        if (missing(ratios))
            setdiff(colnames(data), level.names)
        else
            Call[[m[2]]]
    ratios <- as.matrix(eval(cl, parent.frame())) # ratios as matrix

    ## Creation of a weight matrix. Extract from data if argument
    ## 'weights' is specified, otherwise create a matrix of ones. For
    ## extraction, the only change from ratio extraction is the
    ## content of element "select" of the call.
    weights <-
        if (missing(weights))
        {
            if (any(is.na(ratios)))
                stop("missing ratios not allowed when weights are not supplied")
            array(1, dim(ratios))       # matrix of ones
        }
        else
        {
            cl$select <- Call[[m[3]]]
            as.matrix(eval(cl, parent.frame())) # weights as matrix
        }

    ## == DISPATCH TO APPROPRIATE CALCULATION FUNCTION ==
    ##
    ## Buhlmann-Straub models are handled by bstraub(), regression
    ## models by hache() and hierarchical models by hierarc().
    if (nlevels < 2)                    # one-dimensional model
    {
        ## One-dimensional models accept only "unbiased" and
        ## "iterative" for argument 'method'.
        method <- match.arg(method)
        if (method == "Buhlmann-Gisler" || method == "Ohlsson")
            method <- "unbiased"

        if (is.null(regformula))        # Buhlmann-Straub
        {
            res <- bstraub(ratios, weights, method = method,
                           tol = tol, maxit = maxit, echo = echo)
        }
        else                            # Hachemeister
        {
            ## If regression model is actually empty or has only an
            ## intercept, call bstraub().
            trf <- terms(regformula)
            res <-
                if (length(attr(trf, "factors")) == 0)
                {
                    warning("empty regression model; fitting with Buhlmann-Straub's model")
                    bstraub(ratios, weights, method = method,
                            tol = tol, maxit = maxit, echo = echo)
                }
                else
                    hache(ratios, weights, regformula, regdata,
                          adj.intercept = adj.intercept,
                          method = method, tol = tol,
                          maxit = maxit, echo = echo)
        }

        ## Add missing quantities to results.
        res$classification <- levs
        res$ordering <- list(seq_along(levs))
    }
    else                                # hierarchical model
    {
        ## Computations with auxiliary function.
        res <- hierarc(ratios, weights, classification = ilevs,
                       method = method, tol = tol, maxit = maxit,
                       echo = echo)

        ## Put back original level names into the object
        res$classification <- levs
    }

    ## Transfer level names to lists
    names(res$means) <- names(res$weights) <- c("portfolio", level.names)
    names(res$unbiased) <- if (!is.null(res$unbiased)) names(res$means)
    names(res$iterative) <- if (!is.null(res$iterative)) names(res$means)
    names(res$nodes) <- names(res$ordering) <- level.names
    if (is.list(res$cred))
        names(res$cred) <- level.names

    ## Results
    class(res) <- c("cm", class(res))
    attr(res, "call") <- Call
    res
}

predict.cm <- function(object, levels = NULL, newdata, ...)
{
    ## Convert the character 'levels' argument into numeric and pass
    ## to next method.
    level.names <- names(object$nodes)

    levels <-
        if (is.null(levels))
            seq_along(level.names)
        else
            pmatch(levels, level.names)
    if (any(is.na(levels)))
        stop("invalid level name")
    NextMethod()
}

print.cm <- function(x, ...)
{
    chkDots(...)                        # method does not use '...'
    nlevels <- length(x$nodes)
    level.names <- names(x$nodes)
    b <- if (is.null(x$iterative)) x$unbiased else x$iterative

    cat("Call:\n",
        paste(deparse(attr(x, "call"),
                      width.cutoff = getOption("deparse.cutoff")),
              sep = "\n", collapse = "\n"),
        "\n\n", sep = "")

    cat("Structure Parameters Estimators\n\n")
    cat("  Collective premium:", x$means[[1]], "\n")
    for (i in seq.int(nlevels))
    {
        if (i == 1L)
        {
            ## Treat the Hachemeister model separately since in this
            ## case the variance components vector is a list, with the
            ## first element a matrix. (Note that since a matrix with
            ## empty column names is printed to the screen, there will
            ## be a blank line in the display. Hence the inserted
            ## newline in the 'else' case.)
            if (attr(x, "model") == "regression")
            {
                m <- b[[1]]
                dimnames(m) <- list(c(paste("  Between", level.names[i], "variance: "),
                                      rep("", nrow(m) - 1)),
                                    rep("", ncol(m)))
                print(m)
            }
            else
                cat("\n  Between", level.names[i], "variance:",
                    b[i], "\n")
        }
        else
            cat("  Within ", level.names[i - 1],
                "/Between ", level.names[i], " variance: ",
                b[i], "\n", sep = "")
    }
    cat("  Within", level.names[nlevels], "variance:",
        b[[nlevels + 1]], "\n", fill = TRUE)
    invisible(x)
}

summary.cm <- function(object, levels = NULL, newdata, ...)
{
    nlevels <- length(object$nodes)

    if (nlevels == 1L)
    {
        ## Single level cases (Buhlmann-Straub and Hachemeister):
        ## return the object with the following modifications: put
        ## credibility factors into a list and add a list of the
        ## credibility premiums.
        object$premiums <- list(predict(object, newdata = newdata))
        object$cred <- list(object$cred)
        class(object) = c("summary.cm", class(object))
    }
    else
    {
        ## Multi-level case (hierarchical): select result of the
        ## appropriate level(s).
        plevs <-
            if (is.null(levels))
                seq_along(names(object$nodes))
            else
                pmatch(levels, names(object$nodes))
        if (any(is.na(plevs)))
            stop("invalid level name")

        object$premiums <- predict(object, levels) # new element
        object$means <- object$means[c(1, plevs + 1)]
        object$weights <- object$weights[c(1, plevs + 1)]
        object$unbiased <- object$unbiased[sort(unique(c(plevs, plevs + 1)))]
        object$iterative <- object$iterative[sort(unique(c(plevs, plevs + 1)))]
        object$cred <- object$cred[plevs]
        object$classification <- object$classification[, seq.int(max(plevs)), drop = FALSE]
        object$nodes <- object$nodes[plevs]
        class(object) <- c("summary.cm", class(object))
    }
    structure(object, ...)     # attach additional attributes in '...'
}

print.summary.cm <- function(x, ...)
{
    nlevels <- length(x$nodes)
    level.names <- names(x$nodes)
    NextMethod()                        # print.cm()
    cat("Detailed premiums\n\n")
    for (i in seq.int(nlevels))
    {
        ## Print a "section title" only if there is more than one
        ## level. (Provision introduced in v2.3.0; before the title
        ## was always printed.)
        if (nlevels > 1L)
            cat("  Level:", level.names[i], "\n")

        ## There are no level names in the linear Bayes case, so we
        ## skip this column in the results.
        if (is.null(level.names))
            levs <- NULL
        else
        {
            level.id <- match(level.names[i], colnames(x$classification))
            levs <- x$classification[, seq.int(level.id), drop = FALSE]
            m <- duplicated(levs)
        }

        if (attr(x, "model") == "regression")
        {
            ## Hachemeister model: results contain matrices
            y <- cbind(" ",
                       as.vector(format(x$means[[i + 1L]], ...)),
                       as.vector(apply(format(x$cred[[i]], ...), c(1L, 3L),
                                       paste, collapse = " ")),
                       as.vector(format(sapply(x$adj.models, coef), ...)),
                       " ")
            y[seq(1, nrow(y), dim(x$cred[[i]])[1]), c(1L, 5L)] <-
                c(levs[!m, , drop = FALSE], format(x$premiums[[i]], ...))
            colnames(y) <- c(colnames(levs),
                             "Indiv. coef.", "Cred. matrix",
                             "Adj. coef.", "Cred. premium")
        }
        else if (is.null(levs))
        {
            ## Linear Bayes model: simplified results with no level
            ## column
            y <- cbind(format(x$means[[i + 1L]], ...),
                       format(x$weights[[i + 1L]], ...),
                       format(x$cred[[i]], ...),
                       format(x$premiums[[i]], ...))
            colnames(y) <- c("Indiv. mean", "Weight",
                             "Cred. factor", "Bayes premium")
        }
        else
        {
            ## All other models
            y <- cbind(as.matrix(levs[!m, , drop = FALSE]),
                       format(x$means[[i + 1L]], ...),
                       format(x$weights[[i + 1L]], ...),
                       format(x$cred[[i]], ...),
                       format(x$premiums[[i]], ...))
            colnames(y) <- c(colnames(levs),
                             "Indiv. mean", "Weight",
                             "Cred. factor", "Cred. premium")
        }
        rownames(y) <- rep("   ", nrow(y))
        print(y, quote = FALSE, right = FALSE, ...)
        cat("\n")
    }
    invisible(x)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}single-parameter pareto
### functions. The single-parameter Pareto distribution used in these
### functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - (min/x)^shape, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dpareto1 <- function (x, shape, min, log = FALSE)
    .External(C_actuar_do_dpq, "dpareto1", x, shape, min, log)

ppareto1 <- function(q, shape, min, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ppareto1", q, shape, min, lower.tail, log.p)

qpareto1 <- function(p, shape, min, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qpareto1", p, shape, min, lower.tail, log.p)

rpareto1 <- function(n, shape, min)
    .External(C_actuar_do_random, "rpareto1", n, shape, min)

mpareto1 <- function(order, shape, min)
     .External(C_actuar_do_dpq, "mpareto1", order, shape, min, FALSE)

levpareto1 <- function(limit, shape, min, order = 1)
     .External(C_actuar_do_dpq, "levpareto1", limit, shape, min, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}pareto4 functions to compute
### characteristics of the Pareto (type) IV distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - (1/(1 + v))^shape1, x > min,
###
### where v = ((x - min)/scale)^shape2.
###
### See Arnold, B. C. (2015), Pareto Distributions, Second Edition,
### CRC Press.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dpareto4 <- function(x, min, shape1, shape2, rate = 1, scale = 1/rate,
                     log = FALSE)
    .External(C_actuar_do_dpq, "dpareto4", x, min, shape1, shape2, scale, log)

ppareto4 <- function(q, min, shape1, shape2, rate = 1, scale = 1/rate,
                     lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ppareto4", q, min, shape1, shape2, scale,
              lower.tail, log.p)

qpareto4 <- function(p, min, shape1, shape2, rate = 1, scale = 1/rate,
                     lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qpareto4", p, min, shape1, shape2, scale,
              lower.tail, log.p)

rpareto4 <- function(n, min, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rpareto4", n, min, shape1, shape2, scale)

mpareto4 <- function(order, min, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mpareto4", order, min, shape1, shape2, scale, FALSE)

levpareto4 <- function(limit, min, shape1, shape2, rate = 1, scale = 1/rate,
                       order = 1)
    .External(C_actuar_do_dpq, "levpareto4", limit, min, shape1, shape2, scale,
              order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Compute the adjustment coefficient in ruin theory, that is the
### smallest (strictly) positive root of the Lundberg equation
###
###      h(r) = E[e^(r X - r c W)] = 1,
###
### where X is the claim size random variable, W the inter-occurence
### time and c the premium rate.
###
### AUTHORS: Christophe Dutang, Vincent Goulet <vincent.goulet@act.ulaval.ca>

adjCoef <- function(mgf.claim, mgf.wait = mgfexp, premium.rate, upper.bound,
                    h, reinsurance = c("none", "proportional", "excess-of-loss"),
                    from, to, n = 101)
{
    reinsurance <- match.arg(reinsurance)

    ## Sanity check
    if (missing(mgf.claim) && missing(h))
        stop(sprintf("one of %s or %s is needed",
                     sQuote("mgf.claim"), sQuote("h")))

    ## === NO REINSURANCE CASE ===
    ##
    ## Moment generating functions are unidimensional, premium rate
    ## and adjustment coefficient are both single numeric values.
    if (reinsurance == "none")
    {
        ## For each of 'mgf.claim', 'mgf.wait' and 'h' (if needed): if
        ## the expression is only the name of a function (say f),
        ## build a call 'f(x)'. Otherwise, check that the expression
        ## is a function call containing an 'x'. Taken from 'curve'
        ## and 'discretize'.
        ##
        ## NOTE: argument 'h' will be used iff 'mgf.claim' is missing,
        ## thereby giving priority to 'mgf.claim'.
        if (missing(mgf.claim))
        {
            sh <- substitute(h)
            if (is.name(sh))
            {
                fcall <- paste(sh, "(x)")
                h1 <- function(x)
                    eval(parse(text = fcall),
                         envir = list(x = x),
                         enclos = parent.frame(2))
            }
            else
            {
                if (!(is.call(sh) && match("x", all.vars(sh), nomatch = 0L)))
                    stop(sprintf("%s must be a function or an expression containing %s",
                                 sQuote("h"), sQuote("x")))
                h1 <- function(x)
                    eval(sh,
                         envir = list(x = x),
                         enclos = parent.frame(2))
            }
        }
        else
        {
            smgfx <- substitute(mgf.claim)
            if (is.name(smgfx))
            {
                fcall <- paste(smgfx, "(x)")
                mgfx <- parse(text = fcall)
            }
            else
            {
                if (!(is.call(smgfx) && match("x", all.vars(smgfx), nomatch = 0L)))
                    stop(sprintf("%s must be a function or an expression containing %s",
                                 sQuote("mgf.claim"), sQuote("x")))
                mgfx <- smgfx
            }
            smgfw <- substitute(mgf.wait)
            if (is.name(smgfw))
            {
                fcall <- paste(smgfw, "(x)")
                mgfw <- parse(text = fcall)
            }
            else
            {
                if (!(is.call(smgfw) && match("x", all.vars(smgfw), nomatch = 0L)))
                    stop(sprintf("%s must be a function or an expression containing %s",
                                 sQuote("mgf.wait"), sQuote("x")))
                mgfw <- smgfw
            }
            h1 <- function(x)
                eval(mgfx) * eval(mgfw, list(x = -x * premium.rate))
        }

        f1 <- function(r) (h1(r) - 1)^2

        return(optimize(f1, c(0, upper.bound - .Machine$double.eps),
                        tol = sqrt(.Machine$double.eps))$minimum)
    }

    ## === WITH REINSURANCE CASES ===
    ##
    ## Claim amount moment generating function is a function of 'x'
    ## and the retention level 'y', inter-occurence time moment
    ## generating function is a function of 'x', premium rate and
    ## adjustment coefficient are both functions of the retention
    ## level 'y'.
    ##
    ## Do same as in the no reinsurance case for each of 'mgf.claim',
    ## 'mgf.wait' and 'h' (if needed) and also 'premium'. The first
    ## must be functions of 'x' and 'y', whereas the last one is a
    ## function of 'y' only.
    if (missing(mgf.claim))
    {
        sh <- substitute(h)
        if (is.name(sh))
        {
            fcall <- paste(sh, "(x, y)")
            h2 <- function(x, y)
                eval(parse(text = fcall),
                     envir = list(x = x, y = y),
                     enclos = parent.frame(2))
        }
        else
        {
            if (!(is.call(sh) && all(match(c("x", "y"), all.vars(sh), nomatch = 0L))))
                    stop(sprintf("%s must be a function or an expression containing %s and %s",
                                 sQuote("h"), sQuote("x"), sQuote("y")))
            h2 <- function(x, y)
                eval(sh,
                     envir = list(x = x, y = y),
                     enclos = parent.frame(2))
        }
    }
    else
    {
        if (!is.function(premium.rate))
            stop(sprintf("%s must be a function when using reinsurance",
                         sQuote("premium.rate")))

        smgfx <- substitute(mgf.claim)
        if (is.name(smgfx))
        {
            fcall <- paste(smgfx, "(x, y)")
            mgfx <- parse(text = fcall)
        }
        else
        {
            if (!(is.call(smgfx) && all(match(c("x", "y"), all.vars(smgfx), nomatch = 0L))))
                    stop(sprintf("%s must be a function or an expression containing %s and %s",
                                 sQuote("mgf.claim"), sQuote("x"), sQuote("y")))
            mgfx <- smgfx
        }
        smgfw <- substitute(mgf.wait)
        if (is.name(smgfw))
        {
            fcall <- paste(smgfw, "(x)")
            mgfw <- parse(text = fcall)
        }
        else
        {
            if (!(is.call(smgfw) && match("x", all.vars(smgfw), nomatch = 0L)))
                    stop(sprintf("%s must be a function or an expression containing %s",
                                 sQuote("mgf.wait"), sQuote("x")))
            mgfw <- smgfw
        }
        spremium <- substitute(premium.rate)
        if (is.name(spremium))
        {
            fcall <- paste(spremium, "(y)")
            premium.rate <- parse(text = fcall)
        }
        else
        {
            if (!(is.call(spremium) && match("y", all.vars(spremium), nomatch = 0L)))
                stop(sprintf("%s must be a function or an expression containing %s",
                             sQuote("premium.rate"), sQuote("y")))
            premium.rate <- spremium
        }

        h2 <- function(x, y)
            eval(mgfx) * eval(mgfw, list(x = -x * eval(premium.rate)))
    }

    f2 <- function(x, y) (h2(x, y) - 1)^2
    retention <- seq(from, to, length.out = n)

    ## Compute the adjustment coefficient for each retention level.
    ## The output of 'sapply' is a matrix with minima in the first
    ## line.
    ##
    ## The sapply() below passes the retention levels (argument 'y' of
    ## function 'f') to optimize(). Since the first two arguments ('f'
    ## and 'interval') of the latter function are specified, the
    ## retention levels end up in '...' and hence are considered as
    ## second argument of 'f'. *This requires R >= 2.6.0 to work since
    ## argument '...' comes much earlier in the definition of
    ## optimize().
    coef <- sapply(retention, optimize, f = f2,
                   interval = c(0, upper.bound-.Machine$double.eps),
                   tol = sqrt(.Machine$double.eps))[1L, ]

    ## Make a function from the (retention, coefficient) pairs
    ## computed above, joining the points by straight line segments.
    FUN <- approxfun(retention, coef, rule = 2, method = "linear")

    comment(FUN) <- paste(toupper(substring(reinsurance, 1L, 1L)),
                          substring(reinsurance, 2L),
                          " reinsurance",
                          sep = "", collapse = "")
    class(FUN) <- c("adjCoef", class(FUN))
    attr(FUN, "call") <- sys.call()
    FUN
}

plot.adjCoef <- function(x, xlab = "x", ylab = "R(x)",
                         main = "Adjustment Coefficient",
                         sub = comment(x), type = "l", add = FALSE, ...)
{
    xx <- eval(expression(x), envir = environment(x))
    yy <- eval(expression(y), envir = environment(x))
    if (add)
        lines(xx, yy, ..., main = main, xlab = xlab, ylab = ylab,
              type = type)
    else
        plot(xx, yy, ..., main = main, xlab = xlab, ylab = ylab,
             type = type)
    mtext(sub, line = 0.5)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Creation of grouped data objects.
###
### The function can create a grouped data object from two types of
### arguments.
###
### 1. Individual data. The call has at least two elements in '...'.
###    The first is then the vector of group boundaries and the others
###    are vectors (or a matrix) of group frequencies.
###
### 2. Group boundaries and frequencies. The call has one or more
###    elements in '...' and either 'breaks' or 'nclass' is provided
###    or 'group' is TRUE. In this case, elements of '...' are grouped
###    using graphics::hist automatically based on the first element
###    of '...', or with group boundaries 'breaks' if the latter is a
###    vector.
###
### For details on grouped data, see Klugman, Panjer & Willmot, Loss
### Models, Wiley, 1998.
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Mathieu Pigeon, Louis-Philippe Pouliot
###
### CREDITS: Manipulation and creation of names taken in part from R
### function data.frame(). Arguments, 'breaks', 'nclass' and their
### treatment taken from R function 'hist'.

grouped.data <- function(..., breaks = "Sturges",
                         include.lowest = TRUE, right = TRUE,
                         nclass = NULL, group = FALSE,
                         row.names = NULL, check.rows = FALSE,
                         check.names = TRUE)
{
    ## Utility function to format numbers.
    numform <- function(x, w)
        formatC(x, digits = 2, width = w, format = "fg")

    ## Keep the calls in '...' in object 'ox', and the evaluated
    ## elements in '...' in object 'x'.
    ox <- as.list(substitute(list(...)))[-1L]
    x <- list(...)
    xlen <- length(x)                   # number of arguments in '...'
    use.br <- !missing(breaks)          # 'breaks' specified

    ## If any elements of '...' are unnamed, set names based on the
    ## variable name provided in the function call (e.g. f(x) -> "x")
    ## or from the deparsed expression (e.g. f(1:3) -> "1:3").
    xnames <- names(x)
    if(length(xnames) != xlen)
	xnames <- character(xlen)
    no.xn <- !nzchar(xnames)
    if (any(no.xn))
    {
        for (i in which(no.xn))
            xnames[i] <- deparse(ox[[i]], nlines = 1L)[1L]
        names(x) <- xnames
    }

    ## Single argument implies individual data.
    if (xlen == 1L)
        group <- TRUE

    ## Avoid using calling 'hist' with 'nclass' specified.
    if (use.br)
    {
        if (!missing(nclass))
            warning(sprintf("%s not used when %s is specified",
                    sQuote("nclass"), sQuote("breaks")))
        if (!(missing(group) || group))
            warning(sprintf("%s ignored when %s is specified",
                    sQuote("group"), sQuote("breaks")))
        group <- TRUE
    }
    else if (!is.null(nclass) && length(nclass) == 1L)
    {
        breaks <- nclass
        if (!(missing(group) || group))
            warning(sprintf("%s ignored when %s is specified",
                    sQuote("group"), sQuote("nclass")))
        group <- TRUE
    }

    if (group)        # individual data in argument; group with 'hist'
    {
        ## Set group boudaries (and the first set of group
        ## frequencies) using the first argument in '...'.
        y <- hist(x[[1]], plot = FALSE, breaks = breaks,
                  include.lowest = include.lowest,
                  right = right)
        br <- y$breaks
        y <- y$counts

        ## If there are other vectors in '...', compute group
        ## frequencies using 'hist' with the group boundaries
        ## determined above. If 'breaks' were set automatically, there
        ## is a great risk of error, but we can't do much better.
        if (xlen > 1)
        {
            f <- function(x, br)
                hist(x, plot = FALSE, breaks = br,
                     include.lowest = include.lowest,
                     right = right)$counts
            y <- cbind(y, sapply(x[-1], f, br = br))
        }

        y <- as.data.frame(y)
        x <- as.data.frame(br)
        names(y) <- xnames
        xnames <- ""
        nx <- nrow(x)
    }
    else              # group boundaries and frequencies in argument
    {
        y <- as.data.frame(x[-1L])      # group frequencies
        x <- as.data.frame(x[[1L]])     # group boundaries
        nx <- nrow(x)

        ## There must be exactly one more group boundary than frequencies.
        if (nx - nrow(y) != 1L)
            stop("invalid number of group boundaries and frequencies")

        ## Replace missing frequencies by zeros.
        nax <- is.na(x)
        if (any(nax))
        {
            x[nax] <- 0
            warning("missing frequencies replaced by zeros")
        }
    }

    ## Return a data frame with formatted group boundaries in the
    ## first column.
    w <- max(nchar(x[-1L, ]))            # longest upper boundary
    xfmt <- paste(if (right) "(" else "[",
                  numform(x[-nx, ], -1), ", ", numform(x[-1L, ], w),
                  if (right) "]" else ")",
                  sep = "")
    res <- data.frame(xfmt, y, row.names = row.names, check.rows = check.rows,
                      check.names = check.names)
    names(res) <- c(xnames[1L], names(y))
    class(res) <- c("grouped.data", "data.frame")
    environment(res) <- new.env()
    assign("cj", unlist(x, use.names = FALSE), environment(res))
    attr(res, "right") <- right
    res
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}ztgeom functions to compute
### characteristics of the Zero Truncated Geometric distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dztgeom <- function(x, prob, log = FALSE)
    .External(C_actuar_do_dpq, "dztgeom", x, prob, log)

pztgeom <- function(q, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pztgeom", q, prob, lower.tail, log.p)

qztgeom <- function(p, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qztgeom", p, prob, lower.tail, log.p)

rztgeom <- function(n, prob)
    .External(C_actuar_do_random, "rztgeom", n, prob)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev,mgf}invgamma functions to compute
### characteristics of the Inverse Gamma distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - pgamma(scale/x, shape, scale = 1)
###
### or, equivalently,
###
###   Pr[X <= x] = 1 - pgamma(1/x, shape1, scale = 1/scale).
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Second
### Edition, Wiley, 2004 and
### <http://en.wikipedia.org/wiki/Inverse-gamma_distribution>
###
### AUTHORS:  Mathieu Pigeon, Christophe Dutang and
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvgamma <- function (x, shape, rate = 1, scale = 1/rate, log = FALSE)
     .External(C_actuar_do_dpq, "dinvgamma", x, shape, scale, log)

pinvgamma <- function(q, shape, rate = 1, scale = 1/rate,
                      lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvgamma", q, shape, scale, lower.tail, log.p)

qinvgamma <- function(p, shape, rate = 1, scale = 1/rate,
                      lower.tail = TRUE, log.p = FALSE)
     .External(C_actuar_do_dpq, "qinvgamma", p, shape, scale, lower.tail, log.p)

rinvgamma <- function(n, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_random, "rinvgamma", n, shape, scale)

minvgamma <- function(order, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_dpq, "minvgamma", order, shape, scale, FALSE)

levinvgamma <- function(limit, shape, rate = 1, scale = 1/rate,
                        order = 1)
     .External(C_actuar_do_dpq, "levinvgamma", limit, shape, scale, order, FALSE)

mgfinvgamma <- function(t, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "mgfinvgamma", t, shape, scale, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}zmbinom functions to compute
### characteristics of the Zero Modified Binomial distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dzmbinom <- function (x, size, prob, p0, log = FALSE)
    .External(C_actuar_do_dpq, "dzmbinom", x, size, prob, p0, log)

pzmbinom <- function(q, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pzmbinom", q, size, prob, p0, lower.tail, log.p)

qzmbinom <- function(p, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qzmbinom", p, size, prob, p0, lower.tail, log.p)

rzmbinom <- function(n, size, prob, p0)
    .External(C_actuar_do_random, "rzmbinom", n, size, prob, p0)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}pareto2 functions to compute
### characteristics of the Pareto (type) II distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - (1/(1 + v))^shape, x > min,
###
### where v = (x - min)/scale.
###
### See Arnold, B. C. (2015), Pareto Distributions, Second Edition,
### CRC Press.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dpareto2 <- function (x, min, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "dpareto2", x, min, shape, scale, log)

ppareto2 <- function (q, min, shape, rate = 1, scale = 1/rate, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ppareto2", q, min, shape, scale, lower.tail, log.p)

qpareto2 <- function (p, min, shape, rate = 1, scale = 1/rate, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qpareto2", p, min, shape, scale, lower.tail, log.p)

rpareto2 <- function(n, min, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rpareto2", n, min, shape, scale)

mpareto2 <- function(order, min, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_dpq, "mpareto2", order, min, shape, scale, FALSE)

levpareto2 <- function(limit, min, shape, rate = 1, scale = 1/rate, order = 1)
     .External(C_actuar_do_dpq, "levpareto2", limit, min, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Calulation of infinite time ruin probabilities in the models of
### Cramer-Lundberg and Sparre Andersen. In general, one has
###
###   psi(u) = pphtype(u, pi, Q, lower.tail = FALSE)
###
### for definitions of pi and Q that depend on the severity and waiting
### times models. An explicit solution exists when both severity and
### waiting times are exponential (no mixture).
###
### _Combinations_ of exponentials as defined in Dufresne & Gerber
### (1988) are NOT supported.
###
### References:
###
### Dufresne, F. and Gerber, H. U. (1988), "Three methods to calculate
### the probability of ruin", Astin Bulletin 19, p. 71-90.
###
### Asmussen, S. and Rolski, T. (1991), "Computational methods in risk
### theory: A matrix-algorithmic approach", Insurance: Mathematics and
### Economics 10, p. 259-274.
###
### AUTHORS: Christophe Dutang, Vincent Goulet <vincent.goulet@act.ulaval.ca>

ruin <- function(claims = c("exponential", "Erlang", "phase-type"), par.claims,
                 wait = c("exponential", "Erlang", "phase-type"), par.wait,
                 premium.rate = 1, tol = sqrt(.Machine$double.eps), maxit = 200L,
                 echo = FALSE)
{
    ## Sanity checks
    if (missing(par.claims) || !is.list(par.claims))
        stop(sprintf("%s must be a named list", sQuote("par.claims")))
    if (missing(par.wait) || !is.list(par.wait))
        stop(sprintf("%s must be a named list", sQuote("par.wait")))

    claims <- match.arg(claims)
    wait <- match.arg(wait)

    ## ==============================================
    ##  Extraction of the claims severity parameters
    choices <-
        switch(claims,
               exponential = c("rate", "weights"),
               Erlang = c("shape", "rate", "scale", "weights"),
               "phase-type" = c("prob", "rates"))
    i <- pmatch(names(par.claims), choices, nomatch = 0L, duplicates.ok = TRUE)
    if (all(i == 0L))
        stop(sprintf("parameters %s missing in %s",
                     paste(dQuote(choices), collapse = ", "),
                     sQuote("par.claims")))
    par.claims <- par.claims[i > 0L]    # keep relevant components
    p <- choices[i[i > 0L]]             # keep relevant names
    names(par.claims) <- p              # use full names

    if (claims == "exponential")
    {
        if ("rate" %in% p)
            rate <- par.claims$rate
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("rate"), sQuote("par.claims")))
        n <- length(rate)

        if ("weights" %in% p && n > 1L)
            prob <- rep_len(par.claims$weights, n)
        else if (n == 1L)
            prob <- 1
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("weights"), sQuote("par.claims")))

        rates <- diag(-rate, n)
    }
    else if (claims == "Erlang")
    {
        if ("shape" %in% p)
            shape <- par.claims$shape
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("shape"), sQuote("par.claims")))
        if ("rate" %in% p)
            rate <- par.claims$rate
        else if ("scale" %in% p)
            rate <- 1/par.claims$scale
        else
            stop(sprintf("parameter %s or %s missing in %s",
                         dQuote("rate"), dQuote("scale"), sQuote("par.claims")))
        if (length(shape) < length(rate))
            shape <- rep_len(shape, length(rate))
        else
            rate <-  rep_len(rate, length(shape))
        n <- sum(shape)

        if ("weights" %in% p && length(shape) > 1L)
        {
            prob <- numeric(n)
            prob[cumsum(c(1, head(shape, -1)))] <- par.claims$weights
        }
        else if (length(shape) == 1L)
            prob <- c(1, rep.int(0, n - 1))
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("weights"), sQuote("par.claims")))

        rates <- diag(rep.int(-rate, shape), n)
        if (n > 1 && any(shape > 1))
        {
            tmp <- -head(diag(rates), -1L)
            tmp[cumsum(head(shape, -1L))] <- 0 # insert 0s in "ll corners"
            rates[cbind(seq_len(n - 1), seq(2, len = n - 1))] <- tmp
        }
    }
    else                                # claims == "phase-type"
    {
        if ("prob" %in% p)
            prob <- par.claims$prob
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("prob"), sQuote("par.claims")))
        if ("rates" %in% p)
            rates <- par.claims$rates
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("rates"), sQuote("par.claims")))
        n <- length(prob)
        if (!(is.matrix(rates) && nrow(rates) == n))
            stop(sprintf("invalid parameters in %s", sQuote("par.claims")))
    }
    ## ==============================================

    ## =================================================
    ##  Extraction of the interarrival times parameters
    choices <-
        switch(wait,
               exponential = c("rate", "weights"),
               Erlang = c("shape", "rate", "scale", "weights"),
               "phase-type" = c("prob", "rates"))
    i <- pmatch(names(par.wait), choices, nomatch = 0L, duplicates.ok = TRUE)
    if (all(i == 0L))
        stop(sprintf("parameters %s missing in %s",
                     paste(dQuote(choices), collapse = ", "),
                     sQuote("par.wait")))
    par.wait <- par.wait[i > 0L]        # keep relevant components
    p <- choices[i[i > 0L]]             # keep relevant names
    names(par.wait) <- p                # use full names

    if (wait == "exponential")
    {
        if ("rate" %in% p)
            rate <- par.wait$rate
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("rate"), sQuote("par.wait")))
        m <- length(rate)

        if ("weights" %in% p && m > 1L)
            prob.w <- rep_len(par.wait$weights, m)
        else if (m == 1L)
            prob.w <- 1
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("weights"), sQuote("par.wait")))

        rates.w <- diag(-rate, m)
    }
    else if (wait == "Erlang")
    {
        if ("shape" %in% p)
            shape <- par.wait$shape
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("shape"), sQuote("par.wait")))
        if ("rate" %in% p)
            rate <- par.wait$rate
        else if ("scale" %in% p)
            rate <- 1/par.wait$scale
        else
            stop(sprintf("parameter %s or %s missing in %s",
                         dQuote("rate"), dQuote("scale"), sQuote("par.wait")))
        if (length(shape) < length(rate))
            shape <- rep_len(shape, length(rate))
        else
            rate <-  rep_len(rate, length(shape))
        m <- sum(shape)

        if ("weights" %in% p && length(shape) > 1L)
        {
            prob.w <- numeric(sum(shape))
            prob.w[cumsum(c(1, head(shape, -1L)))] <- par.wait$weights
        }
        else if (length(shape) == 1L)
            prob.w <- c(1, rep.int(0, m - 1))
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("weights"), sQuote("par.wait")))

        rates.w <- diag(rep.int(-rate, shape), m)
        if (m > 1 && any(shape > 1))
        {
            tmp <- -head(diag(rates.w), -1L)
            tmp[cumsum(head(shape, -1L))] <- 0 # insert 0s in "ll corners"
            rates.w[cbind(seq_len(m - 1), seq(2, len = m - 1))] <- tmp
        }
    }
    else                                # wait == "phase-type"
    {
        if ("prob" %in% p)
            prob.w <- par.wait$prob
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("prob"), sQuote("par.wait")))
        if ("rates" %in% p)
            rates.w <- par.wait$rates
        else
            stop(sprintf("parameter %s missing in %s",
                         dQuote("rates"), sQuote("par.wait")))
        m <- length(prob.w)
        if (!(is.matrix(rates.w) && nrow(rates.w) == m))
            stop(sprintf("invalid parameters in %s", sQuote("par.wait")))
    }
    ## =================================================

    ## Empty definition of the output function. The body is set later.
    FUN <- function(u, survival = FALSE, lower.tail = !survival) {}

    ## Cramer-Lundberg model
    if (wait == "exponential" && m == 1L)
    {
        ## Special case with an explicit solution
        if (claims == "exponential" && n == 1L)
        {
            lambda <- -drop(rates.w)/premium.rate
            body(FUN) <- substitute({res <- a * exp(-(b) * u);
                                     if (lower.tail) res else 0.5 - res + 0.5},
                                    list(a = -lambda/drop(rates),
                                         b = -drop(rates) - lambda))
            environment(FUN) <- new.env() # new, empty environment
            class(FUN) <- c("ruin", class(FUN))
            return(FUN)
        }

        ## Use phase-type representation for all other claim severity models.
        pi <- drop(rates.w) * prob %*% solve(rates)/premium.rate
        Q <- rates - rowSums(rates) %*% pi
    }
    ## Sparre Andersen model (interarrival times other than single exponential)
    else
    {
        ## Matrix Q is a "fixed point" of some function (Asmussen &
        ## Rolski, 1992, p. 265-266). Many elements of this function
        ## never change, hence they are computed once and for all
        ## here.
        In <- diag(n)                    # n x n identity matrix
        Im <- diag(m)                    # m x m identity matrix
        t0pi <- -rowSums(rates) %o% prob # "multiple" of A(Q)
        A <- In %x% rbind(prob.w)        # first term of A(Q)
        B <- In %x% rates.w              # rhs of the Kronecker sum
        C <- In %x% -rowSums(rates.w)    # third term of A(Q)

        if (echo)
        {
            cat("Iteration\tMatrix Q (column major order)\n")
            exp <- expression(cat(" ", count, "\t\t ", Q1 <- Q,
                fill = TRUE))
        }
        else
            exp <- expression(Q1 <- Q)

        Q <- rates
        count <- 0L

        repeat
        {
            eval(exp)

            if (maxit < (count <- count + 1L))
            {
                warning("maximum number of iterations reached before obtaining convergence")
                break
            }

            Q1 <- Q
            Q <- rates - t0pi %*% A %*% solve(Q %x% Im + B, C)

            if (max(rowSums(abs(Q - Q1))) < tol)
                break
        }
        pi <- colSums(Q - rates)/(-sum(rates) * premium.rate)
    }

    ## Compute the probability of ruin using the cdf of a phase-type
    ## distribution with parameters pi and Q.
    body(FUN) <- substitute(pphtype(u, a, b, lower.tail = !lower.tail),
                            list(a = pi, b = Q))
    environment(FUN) <- new.env()       # new, empty environment
    class(FUN) <- c("ruin", class(FUN))
    FUN
}

plot.ruin <- function(x, from = NULL, to = NULL, add = FALSE,
                      xlab = "u", ylab = expression(psi(u)),
                      main = "Probability of Ruin", xlim = NULL, ...)
    curve(x, from = from, to = to, add = add, xlab = xlab,
          ylab = ylab, main = main, xlim = xlim, ...)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev,mgf}invgauss functions to compute
### characteristics of the Inverse Gaussian distribution.
###
### Functions [dpq]invgauss rely on C implementations of functions of
### the same name in package statmod. See:
###
### Giner, G. and Smyth, G. K. (2016), "statmod: Probability
###   Calculations for the Inverse Gaussian Distribution", R Journal,
###   vol. 8, no 1, p. 339-351.
###   https://journal.r-project.org/archive/2016-1/giner-smyth.pdf
###
### Chhikara, R. S. and Folk, T. L. (1989), The Inverse Gaussian
###   Distribution: Theory, Methodology and Applications}, Decker.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvgauss <- function(x, mean, shape = 1, dispersion = 1/shape, log = FALSE)
    .External(C_actuar_do_dpq, "dinvgauss", x, mean, dispersion, log)

pinvgauss <- function(q, mean, shape = 1, dispersion = 1/shape,
                      lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvgauss", q, mean, dispersion,
              lower.tail, log.p)

qinvgauss <- function(p, mean, shape = 1, dispersion = 1/shape,
                      lower.tail = TRUE, log.p = FALSE,
                      tol = 1e-14, maxit = 100, echo = FALSE, trace = echo)
    .External(C_actuar_do_dpq, "qinvgauss", p, mean, dispersion,
              lower.tail, log.p, tol, maxit, trace)

rinvgauss <- function(n, mean, shape = 1, dispersion = 1/shape)
    .External(C_actuar_do_random, "rinvgauss", n, mean, dispersion)

minvgauss <- function(order, mean, shape = 1, dispersion = 1/shape)
    .External(C_actuar_do_dpq, "minvgauss", order, mean, dispersion, FALSE)

levinvgauss <- function(limit, mean, shape = 1, dispersion = 1/shape, order = 1)
    .External(C_actuar_do_dpq, "levinvgauss", limit, mean, dispersion, order, FALSE)

mgfinvgauss <- function(t, mean, shape = 1, dispersion = 1/shape, log = FALSE)
    .External(C_actuar_do_dpq, "mgfinvgauss", t, mean, dispersion, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Histogram for grouped data
###
### See Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>, Mathieu Pigeon

hist.grouped.data <-
    function(x, freq = NULL, probability = !freq,
             density = NULL, angle = 45, col = NULL, border = NULL,
             main = paste("Histogram of", xname), xlim = range(x),
             ylim = NULL, xlab = xname, ylab, axes = TRUE,
             plot = TRUE, labels = FALSE, ...)
{
    ## We keep the first frequencies column only; group boundaries are
    ## in the environment of 'x'
    y <- x[, 2L]
    x <- eval(expression(cj), envir = environment(x))

    ## If any frequency is non finite, omit the group
    keep <- which(is.finite(y))
    y <- y[keep]
    x <- x[c(1L, keep + 1L)]

    ## Some useful values
    n <- sum(y)                         # total number of observations
    h <- diff(x)                        # group widths
    dens <- y/(n * h)                   # group "densities"

    ## Cannot plot histogram with infinite group
    if (any(is.infinite(x)))
        stop("infinite group boundaries")

    ## The rest is taken from hist.default()
    xname <- paste(deparse(substitute(x), 500), collapse = "\n")
    equidist <- diff(range(h)) < 1e-07 * mean(h)
    if (is.null(freq))
    {
        freq <- if (!missing(probability))
                    !as.logical(probability)
                else equidist
    }
    else if (!missing(probability) && any(probability == freq))
        stop(sprintf("%s is an alias for %s, however they differ.",
                     sQuote("probability"), sQuote("!freq")))
    mids <- 0.5 * (x[-1L] + x[-length(x)])
    r <- structure(list(breaks = x, counts = y, intensities = dens,
                        density = dens, mids = mids, xname = xname,
                        equidist = equidist),
                   class = "histogram")
    if (plot)
    {
        plot(r, freq = freq, col = col, border = border, angle = angle,
             density = density, main = main, xlim = xlim, ylim = ylim,
             xlab = xlab, ylab = ylab, axes = axes, labels = labels, ...)
        invisible(r)
    }
    else
        r
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Tests for the simulation of compound models with 'rcompound' and
### 'rcomppois'.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

## Load the package
library(actuar)

## Copy of tools::assertError.
assertError <- tools::assertError

## Tests for rcompound
n <- 20
set.seed(123)
x <- numeric(n)
N <- rnbinom(n, 2, 0.8)
y <- rgamma(sum(N), 2, 1)
x[which(N != 0)] <- tapply(y, rep(seq_len(n), N), sum)
stopifnot(exprs = {
    identical(x, {
        set.seed(123)
        rcompound(n, rnbinom(2, 0.8), rgamma(2, 1))
    })
    identical(x, {
        set.seed(123)
        rcompound(n, rnbinom(2, 0.8), expression(rgamma(2, 1)))
    })
    identical(x, {
        set.seed(123)
        rcompound(n, expression(rnbinom(2, 0.8)), rgamma(2, 1))
    })
    identical(x, {
        set.seed(123)
        fmodel <- expression(rnbinom(2, 0.8))
        smodel <- expression(rgamma(2, 1))
        rcompound(n, fmodel, smodel)
    })
})

## Tests for rcomppois
n <- 20
set.seed(123)
lambda <- 2
x <- numeric(n)
N <- rpois(n, lambda)
y <- rgamma(sum(N), 2, 1)
x[which(N != 0)] <- tapply(y, rep(seq_len(n), N), sum)
stopifnot(exprs = {
    identical(x, {
        set.seed(123)
        rcomppois(n, lambda, rgamma(2, 1))
    })
    identical(x, {
        set.seed(123)
        rcomppois(n, lambda, expression(rgamma(2, 1)))
    })
    identical(x, {
        set.seed(123)
        smodel <- expression(rgamma(2, 1))
        rcomppois(n, lambda, smodel)
    })
})

## Test the calling environment, that is that arguments are correctly
## identified when 'rcomppois' is called inside another function.
set.seed(123)
x <- rcomppois(n, lambda, rgamma(2, 1))
f <- function(n, p, model)
{
    ## safe way to pass down all sorts of 'model' objects
    model <- substitute(model)
    if (is.name(model))
        model <- eval.parent(model)
    rcomppois(n, p, model)
}
g1 <- function(n, p, s, r)
    rcomppois(n, p, rgamma(s, r))
g2 <- function(n, p, s, r)
    rcomppois(n, p, expression(rgamma(s, r)))
h <- function(n, p, model)
{
    ## safe way to pass down all sorts of 'model' objects
    model <- substitute(model)
    if (is.name(model))
        model <- eval.parent(model)
    f(n, p, model)
}
stopifnot(exprs = {
    identical(x, {
        set.seed(123)
        f(n, lambda, rgamma(2, 1))
    })
    identical(x, {
        set.seed(123)
        f(n, lambda, expression(rgamma(2, 1)))
    })
    identical(x, {
        set.seed(123)
        f(n, lambda, smodel)
    })
    identical(x, {
        set.seed(123)
        g1(n, lambda, 2, 1)
    })
    identical(x, {
        set.seed(123)
        g2(n, lambda, 2, 1)
    })
    identical(x, {
        set.seed(123)
        h(n, lambda, rgamma(2, 1))
    })
})

## Finally, test invalid arguments.
assertError(rcomppois(-1, lambda, smodel))
assertError(rcomppois(n, -1, smodel))
assertError(rcomppois(n, c(3, -1), smodel))
assertError(rcompound(-1, fmodel, smodel))

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### The "beta integral"
###
###    B(a, b; x) = Gamma(a + b) int_0^x t^(a-1) (1 - t)^(b-1) dt
###
### a > 0, b != -1, -2, ..., 0 < x < 1. This mathematical function is
### only used at the C level in the package. The R function therein
### provides an R interface just in case it could be useful.
###
### The function is *not* exported.
###
### See Appendix A of Klugman, Panjer and Willmot (2012), Loss Models,
### Fourth Edition, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

## see src/betaint.c
betaint <- function(x, a, b)
    .External(C_actuar_do_betaint, x, a, b)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Simulation of discrete mixtures
###
###    f(x) = p_1 f_1(x) + ... + p_n f_n(x).
###
### Uses the syntax of rcomphierarc() for model specfification.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

rmixture <- function(n, probs, models, shuffle = TRUE)
{
    ## Validity checks (similar to other r<dist> functions and to
    ## rmultinom).
    if (any(is.na(n)) || any(n < 0))
        stop(sprintf("invalid first argument %s", sQuote("n")))
    if (all(probs <= 0))
        stop("no positive probabilities")
    if ((!is.expression(models)) || (length(models) == 0L))
        stop(sprintf("invalid third argument %s", sQuote("models")))

    ## Number of models in the mixture.
    m <- max(length(probs), length(models))

    ## Number of variates to generate: 'length(n)' if length of 'n' is
    ## > 1, like other 'r<dist>' functions.
    if (length(n) > 1L)
        n <- length(n)

    ## Number of variates from each model. By definition of the
    ## multinomial distribution, sum(nj) == n.
    ##
    ## Note that 'rmultinom' will normalize probabilities to sum 1.
    nj <- rmultinom(1, size = n, prob = rep_len(probs, m))

    ## Auxiliary function to generate 'n' variates from the model
    ## given in 'expr'. The expressions end up being evaluated three
    ## frames below the current one.
    f <- function(n, expr)
    {
        expr$n <- n
        eval.parent(expr, n = 3)
    }

    ## Simulate from each model the appropriate number of times and
    ## return result as an atomic vector. Variates are ordered by
    ## model: all random variates from model 1, then all random
    ## variates from model 2, and so on.
    x <- unlist(mapply(f, n = nj, expr = rep_len(models, m),
                       SIMPLIFY = FALSE))

    ## Return variates reshuffled or in the order above as per
    ## argument 'shuffle'.
    if (shuffle)
        x[sample.int(n)]
    else
        x
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,r,m,mgf}ph functions to compute
### characteristics of Phase-type distributions with cumulative
### distribution function
###
###       Pr[X <= x] = 1 - pi %*% exp(Tx) %*% e,
###
### where 'pi' is the initial probability vector, 'T' is the
### subintensity matrix and 'e' is 1-vector of R^m.
###
### See Bladt, M. (2005), "A review on phase-type distributions and
### their use in risk theory", Astin Bulletin 35, p. 145-161.
###
### AUTHORS: Christophe Dutang, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dphtype <- function(x, prob, rates, log = FALSE)
    .External(C_actuar_do_dpqphtype, "dphtype", x, prob, rates, log)

pphtype <- function(q, prob, rates, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpqphtype, "pphtype", q, prob, rates, lower.tail, log.p)

rphtype <- function(n, prob, rates)
    .External(C_actuar_do_randomphtype, "rphtype", n, prob, rates)

mphtype <- function(order, prob, rates)
    .External(C_actuar_do_dpqphtype, "mphtype", order, prob, rates, FALSE)

mgfphtype <- function(t, prob, rates, log = FALSE)
    .External(C_actuar_do_dpqphtype, "mgfphtype", t, prob, rates, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Function to discretize a continuous distribution using various
### methods.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

discretize <- function (cdf, from, to, step = 1,
                        method = c("upper", "lower", "rounding", "unbiased"),
                        lev, by = step, xlim = NULL)
{
    method <- match.arg(method)

    ## If 'cdf' is only the name of a function (say f), build a call
    ## 'f(x)'. Otherwise, check that the expression is a function call
    ## containing an 'x'. Taken from 'curve'.
    scdf <- substitute(cdf)
    if (is.name(scdf))
    {
        fcall <- paste(scdf, "(x)")
        cdf <- parse(text = fcall)
    }
    else
    {
        if (!(is.call(scdf) && match("x", all.vars(scdf), nomatch = 0)))
            stop(sprintf("%s must be a function or an expression containing %s",
                         sQuote("cdf"), sQuote("x")))
        cdf <- scdf
    }

    ## If 'from' and/or 'to' are not specified, take their values in 'xlim'.
    if (missing(from))
        from <- xlim[1]
    if (missing(to))
        to <- xlim[2]

    if (method %in% c("upper", "lower"))
    {
        ## The "upper" discretization method assigns to point x =
        ## from, from + step, ..., to - step the probability mass F(x
        ## + step) - F(x).
        ##
        ## The "lower" discretization method assigns to point x = from
        ## the probability mass 0 and to x = from + step, ..., to the
        ## probability mass F(x) - F(x - step).
        ##
        ## Hence, the latter method simply has one more element than the
        ## former.
        x <- seq.int(from, to, by)
        Fx <- eval(cdf, envir = list(x = x), enclos = parent.frame())
        return(c(if(method == "lower") 0, diff(Fx)))
    }

    if (method == "rounding")
    {
        ## Rounding method assigns to point x = from the probability
        ## mass F(from + step/2) - F(from) and to point x = from +
        ## step, ..., to - step the probability mass F(x - step/2) -
        ## F(x + step/2).
        ##
        ## It is possible to make adjustments for the limits of the
        ## intervals (closed or open) for discrete distributions via
        ## 'cdf'.
        x <- c(from, seq.int(from + by/2, to - by/2, by))
        Fx <- eval(cdf, envir = list(x = x), enclos = parent.frame())
        return(diff(Fx))
    }

    if (method == "unbiased")
    {
        ## This is the matching of the first moment method. It
        ## requires a function to compute the first limited moment
        ## which should be provided in argument 'lev'. The latter is
        ## specified just like 'cdf'.
        if (missing(lev))
            stop(sprintf("%s required with method %s",
                         sQuote("lev"), dQuote("unbiased")))

        slev <- substitute(lev)
        if (is.name(slev))
        {
            fcall <- paste(slev, "(x)")
            lev <- parse(text = fcall)
        }
        else
        {
            if (!(is.call(slev) && match("x", all.vars(slev), nomatch = 0)))
            stop(sprintf("%s must be a function or an expression containing %s",
                         sQuote("lev"), sQuote("x")))
            lev <- slev
        }

        ## The first limited moment must be evaluated in x = from,
        ## from + step, ..., to and the cdf in x = from and x = to
        ## only (see below).
        x <- seq.int(from, to, by)
        Ex <- eval(lev, envir = list(x = x), enclos = parent.frame())
        Fx <- eval(cdf, envir = list(x = c(from, to)), enclos = parent.frame())

        ## The probability mass in x = from is
        ##
        ##   (E[X ^ x] - E[X ^ x + step])/step + 1 - F(x).
        ##
        ## The probability mass in x = from + step, ..., to - step is
        ##
        ##   (2 * E[X ^ x] - E[X ^ x - step] - E[X ^ x + step])/step.
        ##
        ## The probability mass in x = to is
        ##
        ##   (E[X ^ x] - E[X ^ x - step])/step - 1 + F(x).
        ##
        ## See exercise 6.36 in Loss Models, 2nd edition.
        return(c(-diff(head(Ex, 2))/by + 1 - Fx[1],
                 (2 * head(Ex[-1], -1) - head(Ex, -2) - tail(Ex, -2))/by,
                 diff(tail(Ex, 2))/by - 1 + Fx[2]))
    }
}

discretise <- discretize

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,r}zmlogarithmic functions to compute
### characteristics of the zero modified logarithmic distribution. See
### ./Logarithmic.R for details on the parametrization.
###
### See p. 93 of Klugman, Panjer & Willmot, Loss Models, Fourth
### Edition, Wiley, 2012.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dzmlogarithmic <- function(x, prob, p0, log = FALSE)
    .External(C_actuar_do_dpq, "dzmlogarithmic", x, prob, p0, log)

pzmlogarithmic <- function(q, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pzmlogarithmic", q, prob, p0, lower.tail, log.p)

qzmlogarithmic <- function(p, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qzmlogarithmic", p, prob, p0, lower.tail, log.p)

rzmlogarithmic <- function(n, prob, p0)
    .External(C_actuar_do_random, "rzmlogarithmic", n, prob, p0)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}burr functions to compute
### characteristics of the Inverse Burr distribution. The version used
### in these functions has cumulative distribution function
###
###   Pr[X <= x] = (u/(1 + u))^shape1, u = (x/scale)^shape2, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvburr <- function (x, shape1, shape2, rate = 1, scale = 1/rate,
                      log = FALSE)
    .External(C_actuar_do_dpq, "dinvburr", x, shape1, shape2, scale, log)

pinvburr <- function(q, shape1, shape2, rate = 1, scale = 1/rate,
                     lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvburr", q, shape1, shape2, scale,
              lower.tail, log.p)

qinvburr <- function(p, shape1, shape2, rate = 1, scale = 1/rate,
                     lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qinvburr", p, shape1, shape2, scale,
              lower.tail, log.p)

rinvburr <- function(n, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rinvburr", n, shape1, shape2, scale)

minvburr <- function(order, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "minvburr", order, shape1, shape2, scale, FALSE)

levinvburr <- function(limit, shape1, shape2, rate = 1, scale = 1/rate,
                       order = 1)
    .External(C_actuar_do_dpq, "levinvburr", limit, shape1, shape2, scale,
              order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev}lnorm functions.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

mlnorm <- function(order, meanlog = 0, sdlog = 1)
    .External(C_actuar_do_dpq, "mlnorm", order, meanlog, sdlog, FALSE)

levlnorm <- function(limit, meanlog = 0, sdlog = 1, order = 1)
    .External(C_actuar_do_dpq, "levlnorm", limit, meanlog, sdlog, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,mgf}norm functions to compute raw and the
### moment generating function for the Normal distribution (as defined
### in R).
###
### See Chapter 13 of Johnson & Kotz, Continuous univariate
### distributions, volume 1, Wiley, 1970
###
### AUTHORS: Christophe Dutang, Vincent Goulet <vincent.goulet@act.ulaval.ca>

mnorm <- function(order, mean = 0, sd = 1)
    .External(C_actuar_do_dpq, "mnorm", order, mean, sd, FALSE)

mgfnorm <- function(t, mean = 0, sd = 1, log = FALSE)
    .External(C_actuar_do_dpq, "mgfnorm", t, mean, sd, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}zmgeom functions to compute
### characteristics of the Zero Modified Geometric distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dzmgeom <- function (x, prob, p0, log = FALSE)
    .External(C_actuar_do_dpq, "dzmgeom", x, prob, p0, log)

pzmgeom <- function(q, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pzmgeom", q, prob, p0, lower.tail, log.p)

qzmgeom <- function(p, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qzmgeom", p, prob, p0, lower.tail, log.p)

rzmgeom <- function(n, prob, p0)
    .External(C_actuar_do_random, "rzmgeom", n, prob, p0)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}pareto functions to compute
### characteristics of the Pareto distribution. The version used in
### these functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - (scale/(x + scale))^shape, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS: Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dpareto <- function (x, shape, scale, log = FALSE)
    .External(C_actuar_do_dpq, "dpareto", x, shape, scale, log)

ppareto <- function (q, shape, scale, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ppareto", q, shape, scale, lower.tail, log.p)

qpareto <- function (p, shape, scale, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qpareto", p, shape, scale, lower.tail, log.p)

rpareto <- function(n, shape, scale)
    .External(C_actuar_do_random, "rpareto", n, shape, scale)

mpareto <- function(order, shape, scale)
     .External(C_actuar_do_dpq, "mpareto", order, shape, scale, FALSE)

levpareto <- function(limit, shape, scale, order = 1)
     .External(C_actuar_do_dpq, "levpareto", limit, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}poisinvgauss functions to compute
### characteristics of the Poisson-Inverse Gaussian discrete
### distribution.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dpoisinvgauss <- function(x, mean, shape = 1, dispersion = 1/shape, log = FALSE)
    .External(C_actuar_do_dpq, "dpoisinvgauss", x, mean, dispersion, log)

ppoisinvgauss <- function(q, mean, shape = 1, dispersion = 1/shape,
                          lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ppoisinvgauss", q, mean, dispersion,
              lower.tail, log.p)

qpoisinvgauss <- function(p, mean, shape = 1, dispersion = 1/shape,
                          lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qpoisinvgauss", p, mean, dispersion,
              lower.tail, log.p)

rpoisinvgauss <- function(n, mean, shape = 1, dispersion = 1/shape)
    .External(C_actuar_do_random, "rpoisinvgauss", n, mean, dispersion)

## Aliases
dpig <- dpoisinvgauss
ppig <- ppoisinvgauss
qpig <- qpoisinvgauss
rpig <- rpoisinvgauss

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Computing summary statistics and accessing components of a
### portfolio.
###
### AUTHORS: Louis-Philippe Pouliot, Tommy Ouellet,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

aggregate.portfolio <- function(x, by = names(x$nodes), FUN = sum,
                                classification = TRUE, prefix = NULL, ...)
{
    level.names <- names(x$nodes)       # level names
    nlevels <- length(level.names)      # number of levels
    years <- level.names[nlevels]       # name of last level

    ## Match level names in 'by' to those in the model
    by <- match.arg(by, level.names, several.ok = TRUE)

    ## Version of FUN able to work on lists
    fun <- function(x, ...) FUN(unlist(x), ...)

    ## The most common case should be to aggregate claim amounts by
    ## node. This case being very simple, it is treated separately.
    if (identical(by, level.names))
        return(cbind(if (classification) x$classification,
                     array(sapply(x$data, FUN, ...), dim(x$data),
                           dimnames = list(NULL, paste(prefix, colnames(x$data), sep = "")))))

    ## Summaries only by last level (years) are also simple to handle.
    if (identical(by, years))
    {
        res <- apply(x$data, 2, fun, ...)
        names(res) <- paste(prefix, colnames(x$data), sep = "")
        return(res)
    }

    ## The other possibilities require to split the data in groups as
    ## specified in argument 'by'. If the last level (years) is in
    ## 'by', then the matrix structure must be retained to make the
    ## summaries. Otherwise, it can just be dropped since summaries
    ## will span the years of observation.
    ##
    ## Convert the sequence of subscripts into factors by pasting the
    ## digits together. It is important *not* to sort the levels in
    ## case the levels in 'by' are not in the same order as in
    ## 'level.names'.
    rows <- setdiff(by, years)          # groups other than years
    s <- x$classification[, rows, drop = FALSE] # subscripts
    f <- apply(s, 1, paste, collapse = "")      # grouping IDs
    f <- factor(f, levels = unique(f))          # factors
    s <- s[match(levels(f), f), , drop = FALSE] # unique subscripts
    xx <- split(x$data, f)                      # split data

    ## Make summaries
    if (years %in% by)
    {
        xx <- lapply(xx, matrix, ncol = ncol(x$data))
        res <- t(sapply(xx, function(x, ...) apply(x, 2, fun, ...), ...))
        cols <- colnames(x$data)
    }
    else
    {
        res <- sapply(xx, fun, ...)
        cols <- deparse(substitute(FUN))
    }

    ## Return results as a matrix
    structure(cbind(if (classification) s, res),
              dimnames = list(NULL, c(if (classification) rows, paste(prefix, cols, sep = ""))))
}

frequency.portfolio <- function(x, by = names(x$nodes),
                                classification = TRUE, prefix = NULL, ...)
{
    chkDots(...)                        # method does not use '...'
    freq <- function(x) if (identical(x, NA)) NA else length(x[!is.na(x)])
    aggregate(x, by, freq, classification, prefix)
}

severity.portfolio <- function(x, by = head(names(x$node), -1),
                               splitcol = NULL, classification = TRUE,
                               prefix = NULL, ...)
{
    chkDots(...)                        # method does not use '...'

    level.names <- names(x$nodes)       # level names
    ci <- seq_len(ncol(x$data))         # column indexes

    ## Match level names in 'by' to those in the model
    by <- match.arg(by, level.names, several.ok = TRUE)

    ## Sanity checks
    if (identical(by, level.names))
    {
        warning("nothing to do")
        return(x)
    }

    ## Convert character 'splitcol' to numeric and then from numeric
    ## or NULL to boolean.
    if (is.character(splitcol))
        splitcol <- pmatch(splitcol, colnames(x$data), duplicates.ok = TRUE)
    if (is.numeric(splitcol) || is.null(splitcol))
        splitcol <- ci %in% splitcol

    ## Unroll claim amounts by column; simplest case
    if (tail(level.names, 1L) %in% by)
    {
        if (length(by) > 1L)
            stop(sprintf("invalid %s specification", sQuote("by")))
        #x <- x$data
        res <- unroll(x$data, bycol = TRUE, drop = FALSE)
        colnames(res) <- paste(prefix, colnames(res), sep = "")
        return(list(main = res[, !splitcol],
                    split = if (all(!splitcol)) NULL else res[, splitcol]))
    }

    ## Unrolling per row (or group of rows) is more work. It requires
    ## to split the columns of the matrix first, and then to apply the
    ## unrolling procedure twice (if 'splitcol' != NULL).
    ##
    ## Utility function
    fun <- function(x) unlist(x[!is.na(x)])

    ## Split rows according to the 'by' argument.
    s <- x$classification[, by, drop = FALSE]   # subscripts
    f <- apply(s, 1, paste, collapse = "")      # grouping IDs
    f <- factor(f, levels = unique(f))          # factors
    s <- s[match(levels(f), f), , drop = FALSE] # unique subscripts

    ## Keep the 'splitcol' columns for later use.
    x.split <- x$data[, splitcol]

    ## If a prefix is not specified, use "claim." as a sensible
    ## choice.
    if (is.null(prefix))
        prefix <- "claim."

    ## Unroll the "main" block of columns.
    if (all(splitcol))
        res.main <- NULL
    else
    {
        x <- cbind(lapply(split(x$data[, !splitcol], f), fun))
        res.main <- unroll(x, bycol = FALSE, drop = FALSE)
        res.main <-
            if (0L < (nc <- ncol(res.main)))
            {
                dimnames(res.main) <-
                    list(NULL, paste(prefix, seq_len(nc), sep = ""))
                cbind(if (classification) s, res.main)
            }
            else
                NULL
    }

    ## Unroll the 'splitcol' block of columns.
    if (all(!splitcol))
        res.split <- NULL
    else
    {
        x <- cbind(lapply(split(x.split, f), fun))     # split data
        res.split <- unroll(x, bycol = FALSE, drop = FALSE)
        res.split <-
            if (0L < (nc <- ncol(res.split)))
            {
                dimnames(res.split) <-
                    list(NULL, paste(prefix, seq_len(nc), sep = ""))
                cbind(if (classification) s, res.split)
            }
            else
                NULL
    }

    ## Return the result as a list.
    list(main = res.main, split = res.split)
}

weights.portfolio <- function(object, classification = TRUE, prefix = NULL, ...)
{
    chkDots(...)                        # method does not use '...'

    if (is.null(object$weights))
        NULL
    else
    {
        w <- object$weights
        colnames(w) <- paste(prefix, colnames(w), sep = "")
        cbind(if (classification) object$classification, w)
    }
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}pareto3 functions to compute
### characteristics of the Pareto (type) II distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = v/(1 + v), x > min,
###
### where v = ((x - min)/scale)^shape.
###
### See Arnold, B. C. (2015), Pareto Distributions, Second Edition,
### CRC Press.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dpareto3 <- function (x, min, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "dpareto3", x, min, shape, scale, log)

ppareto3 <- function (q, min, shape, rate = 1, scale = 1/rate, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ppareto3", q, min, shape, scale, lower.tail, log.p)

qpareto3 <- function (p, min, shape, rate = 1, scale = 1/rate, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qpareto3", p, min, shape, scale, lower.tail, log.p)

rpareto3 <- function(n, min, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rpareto3", n, min, shape, scale)

mpareto3 <- function(order, min, shape, rate = 1, scale = 1/rate)
     .External(C_actuar_do_dpq, "mpareto3", order, min, shape, scale, FALSE)

levpareto3 <- function(limit, min, shape, rate = 1, scale = 1/rate, order = 1)
     .External(C_actuar_do_dpq, "levpareto3", limit, min, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}genpareto functions to compute
### characteristics of the Generalized Pareto distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = Pr[Y <=  x / (x + scale)],
###
### where Y has a Beta distribution with parameters shape2 and shape1.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dgenpareto <- function(x, shape1, shape2, rate = 1, scale = 1/rate,
                       log = FALSE)
     .External(C_actuar_do_dpq, "dgenpareto", x, shape1, shape2, scale, log)

pgenpareto <- function(q, shape1, shape2, rate = 1, scale = 1/rate,
                       lower.tail = TRUE, log.p = FALSE)
     .External(C_actuar_do_dpq, "pgenpareto", q, shape1, shape2, scale,
               lower.tail, log.p)

qgenpareto <- function(p, shape1, shape2, rate = 1, scale = 1/rate,
                       lower.tail = TRUE, log.p = FALSE)
     .External(C_actuar_do_dpq, "qgenpareto", p, shape1, shape2, scale,
               lower.tail, log.p)

rgenpareto <- function(n, shape1, shape2, rate = 1, scale = 1/rate)
     .External(C_actuar_do_random, "rgenpareto", n, shape1, shape2, scale)

mgenpareto <- function(order, shape1, shape2, rate = 1, scale = 1/rate)
     .External(C_actuar_do_dpq, "mgenpareto", order, shape1, shape2, scale, FALSE)

levgenpareto <- function(limit, shape1, shape2, rate = 1, scale = 1/rate,
                         order = 1)
     .External(C_actuar_do_dpq, "levgenpareto", limit, shape1, shape2, scale,
               order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Buhlmann-Straub credibility model calculations.
###
### Computation of the between variance estimators has been moved to
### external functions bvar.unbiased() and bvar.iterative() to share
### with hache().
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Sebastien Auclair, Louis-Philippe Pouliot

bstraub <- function(ratios, weights, method = c("unbiased", "iterative"),
                    tol = sqrt(.Machine$double.eps), maxit = 100, echo = FALSE)
{
    ## If weights are not specified, use equal weights as in
    ## Buhlmann's model.
    if (missing(weights))
    {
        if (any(is.na(ratios)))
            stop("missing ratios not allowed when weights are not supplied")
        weights <- array(1, dim(ratios))
    }

    ## Check other bad arguments.
    if (ncol(ratios) < 2L)
        stop("there must be at least one node with more than one period of experience")
    if (nrow(ratios) < 2L)
        stop("there must be more than one node")
    if (!identical(which(is.na(ratios)), which(is.na(weights))))
        stop(sprintf("missing values are not in the same positions in %s and in %s",
                     sQuote("weights"), sQuote("ratios")))
    if (all(!weights, na.rm = TRUE))
        stop("no available data to fit model")

    ## Individual weighted averages. It could happen that a contract
    ## has no observations, for example when applying the model on
    ## claim amounts. In such a situation, we will put the total
    ## weight of the contract and the weighted average both equal to
    ## zero. That way, the premium will be equal to the credibility
    ## weighted average, as it should, but the contract will have no
    ## contribution in the calculations.
    weights.s <- rowSums(weights, na.rm = TRUE)
    ratios.w <- ifelse(weights.s > 0, rowSums(weights * ratios, na.rm = TRUE)/weights.s, 0)

    ## Size of the portfolio.
    ncontracts <- sum(weights.s > 0)
    ntotal <- sum(!is.na(weights))

    ## Collective weighted average.
    weights.ss <- sum(weights.s)

    ## Estimation of s^2
    s2 <-  sum(weights * (ratios - ratios.w)^2, na.rm = TRUE)/(ntotal - ncontracts)

    ## First estimation of a. Always compute the unbiased estimator.
    a <- bvar.unbiased(ratios.w, weights.s, s2, ncontracts)

    ## Iterative estimation of a. Compute only if
    ## 1. asked to in argument;
    ## 2. weights are not all equal (Buhlmann model).
    ## 3. the unbiased estimator is > 0;
    method <- match.arg(method)

    if (method == "iterative" &&
        diff(range(weights, na.rm = TRUE)) > .Machine$double.eps^0.5)
    {
        a <- if (a > 0)
                 bvar.iterative(ratios.w, weights.s, s2, ncontracts, start = a,
                                tol = tol, maxit = maxit, echo = echo)
             else
                0
    }

    ## Final credibility factors and estimator of the collective mean.
    if (a > 0)
    {
        cred <- 1/(1 + s2/(weights.s * a))
        ratios.zw <- drop(crossprod(cred, ratios.w))/sum(cred)
    }
    else
    {
        cred <- numeric(length(weights.s))
        ratios.zw <- drop(crossprod(weights.s, ratios.w))/sum(weights.s)
    }

    structure(list(means = list(ratios.zw, ratios.w),
                   weights = list(if (a > 0) sum(cred) else weights.ss, weights.s),
                   unbiased = if (method == "unbiased") c(a, s2),
                   iterative = if (method == "iterative") c(a, s2),
                   cred = cred,
                   nodes = list(nrow(weights))),
              class = "bstraub",
              model = "Buhlmann-Straub")
}

predict.bstraub <- function(object, levels = NULL, newdata, ...)
    structure(object$means[[1L]] + object$cred * (object$means[[2L]] - object$means[[1L]]), ...)

## Alias for the linear Bayes case
predict.bayes <- predict.bstraub

bvar.unbiased <- function(x, w, within, n)
{
    w.s <- sum(w)
    x.w <- drop(crossprod(w, x))/w.s

    w.s * (drop(crossprod(w, (x - x.w)^2)) - (n - 1) * within)/(w.s^2 - sum(w^2))
}

### codetools does not like the way 'a1' is defined in function
### 'bvar.iterative' below. Avoid false positive in R CMD check.
if (getRversion() >= "2.15.1") utils::globalVariables(c("a1"))

bvar.iterative <- function(x, w, within, n, start,
                           tol = sqrt(.Machine$double.eps), maxit = 100,
                           echo = FALSE)
{
    if (echo)
    {
        cat("Iteration\tBetween variance estimator\n")
        expr <- expression(cat(" ", count, "\t\t ", a1 <- a, fill = TRUE))
    }
    else
        expr <- expression(a1 <-  a)

    a <- start
    count <- 0L

    repeat
    {
        eval(expr)

        if (maxit < (count <- count + 1L))
        {
            warning("maximum number of iterations reached before obtaining convergence")
            break
        }

        cred <- 1/(1 + within/(w * a))
        x.z <- drop(crossprod(cred, x))/sum(cred)
        a <- drop(crossprod(cred, (x - x.z)^2))/(n - 1)

        if (abs((a - a1)/a1) < tol)
            break
    }
    a
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Credibility in the regression case using the Hachemeister (1975)
### model with possibly an adjustment to put the intercept at the
### barycenter of time (see Buhlmann & Gisler, 2005).
###
### AUTHORS: Xavier Milhaud, Tommy Ouellet, Vincent Goulet
### <vincent.goulet@act.ulaval.ca>

hache <- function(ratios, weights, formula, data, adj.intercept = FALSE,
                  method = c("unbiased", "iterative"),
                  tol = sqrt(.Machine$double.eps),
                  maxit = 100, echo = FALSE)
{
    Call <- match.call()

    ## If weights are not specified, use equal weights as in
    ## Buhlmann's model.
    if (missing(weights))
    {
        if (any(is.na(ratios)))
            stop("missing ratios not allowed when weights are not supplied")
        weights <- array(1, dim(ratios))
    }

    ## Check other bad arguments.
    if (NCOL(ratios) < 2)
        stop("there must be at least one node with more than one period of experience")
    if (NROW(ratios) < 2)
        stop("there must be more than one node")
    if (!identical(which(is.na(ratios)), which(is.na(weights))))
        stop("missing values are not in the same positions in 'weights' and in 'ratios'")
    if (all(!weights, na.rm = TRUE))
        stop("no available data to fit model")

    ## Build the design matrix
    mf <- model.frame(formula, data, drop.unused.levels = TRUE)
    mt <- attr(mf, "terms")
    xreg <- model.matrix(mt, mf)

    ## Do computations in auxiliary functions.
    res <-
        if (adj.intercept)
            hache.barycenter(ratios, weights, xreg,
                             method = match.arg(method),
                             tol = tol, maxit = maxit, echo = echo)
        else
            hache.origin(ratios, weights, xreg,
                         tol = tol, maxit = maxit, echo = echo)

    ## Add the terms object to the result for use in predict.hache()
    ## [and thus predict.lm()].
    res$terms <- mt

    ## Results
    attr(res, "class") <- "hache"
    attr(res, "model") <- "regression"
    res
}

predict.hache <- function(object, levels = NULL, newdata, ...)
{
    ## If model was fitted at the barycenter of time (there is a
    ## transition matrix in the object), then also convert the
    ## regression coefficients in the base of the (original) design
    ## matrix.
    if (!is.null(R <- object$transition))
    {
        for (i in seq_along(object$adj.models))
        {
            b <- coefficients(object$adj.models[[i]])
            object$adj.models[[i]]$coefficients <- solve(R, b)
        }
    }

    ## Prediction (credibility premiums) using predict.lm() on each of
    ## the adjusted individual models. This first requires to add a
    ## 'terms' component to each adjusted model.
    f <- function(z, ...)
    {
        z$terms <- object$terms
        class(z) <- "lm"                # to keep predict.lm() quiet
        unname(predict.lm(z, ...))
    }

    structure(sapply(object$adj.models, f, newdata = newdata), ...)
}

print.hache <- function(x, ...)
    print.default(x)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Auxiliary function to fit regression credibility model using the
### original Hachemeister model.
###
### AUTHORS: Tommy Ouellet, Vincent Goulet <vincent.goulet@act.ulaval.ca>

### codetools does not like the way 'coll1' is defined in
### 'hache.origin' below. Avoid false positive in R CMD check.
if (getRversion() >= "2.15.1")  utils::globalVariables(c("coll1"))

hache.origin <- function(ratios, weights, xreg, tol, maxit, echo)
{
    ## Frequently used values
    weights.s <- rowSums(weights, na.rm = TRUE) # contract total weights
    has.data <- which(weights.s > 0)	# contracts with data
    ncontracts <- nrow(ratios)   	# number of contracts
    eff.ncontracts <- length(has.data)  # effective number of contracts
    p <- ncol(xreg)               	# rank (>= 2) of design matrix
    n <- nrow(xreg)                     # number of observations

    ## Fit linear model to each contract. For contracts without data,
    ## fit some sort of empty model to ease use in predict.hache().
    f <- function(i)
    {
    	z <-
            if (i %in% has.data)            # contract with data
            {
                y <- ratios[i, ]
                not.na <- !is.na(y)
                lm.wfit(xreg[not.na, , drop = FALSE], y[not.na], weights[i, not.na])
            }
            else                            # contract without data
                lm.fit(xreg, rep.int(0, n))
        z[c("coefficients", "residuals", "weights", "rank", "qr")]
    }
    fits <- lapply(seq_len(ncontracts), f)

    ## Individual regression coefficients
    ind <- sapply(fits, coef)
    ind[is.na(ind)] <- 0

    ## Individual variance estimators. The contribution of contracts
    ## without data is 0.
    S <- function(z)                    # from stats::summary.lm
    {
        nQr <- NROW(z$qr$qr)
        r1 <- z$rank
        r <- z$residuals
        w <- z$weights
        sum(w * r^2) / (nQr - r1)
    }
    sigma2 <- sapply(fits[has.data], S)
    sigma2[is.nan(sigma2)] <- 0

    ## Initialization of a few containers: p x p x ncontracts arrays
    ## for the weight and credibility matrices; p x p matrices for the
    ## between variance-covariance matrix and total credibility
    ## matrix.
    cred <- W <- array(0, c(p, p, ncontracts))
    A <- cred.s <- matrix(0, p, p)

    ## Weight matrices: we use directly (X'WX)^{-1}. This is quite
    ## different from hache.barycenter().
    V <- function(z)                # from stats::summary.lm
    {
        r1 <- z$rank
        if (r1 == 1L)
            diag(as.double(chol2inv(z$qr$qr[1L, 1L, drop = FALSE])), p)
        else
            chol2inv(z$qr$qr[1L:r1, 1L:r1, drop = FALSE])
    }
    W[, , has.data] <- sapply(fits[has.data], V)

    ## Starting credibility matrices and collective regression
    ## coefficients.
    cred[, , has.data] <- diag(p)         # identity matrices
    coll <- rowSums(ind) / eff.ncontracts # coherent with above

    ## === ESTIMATION OF WITHIN VARIANCE ===
    s2 <- mean(sigma2)

    ## === ESTIMATION OF THE BETWEEN VARIANCE-COVARIANCE MATRIX ===
    ##
    ## This is an iterative procedure similar to the Bischel-Straub
    ## estimator. Following Goovaerts & Hoogstad, stopping criterion
    ## is based in the collective regression coefficients estimates.
    ##
    ## If printing of iterations was asked for, start by printing a
    ## header and the starting values.
    if (echo)
    {
        cat("Iteration\tCollective regression coefficients\n")
        exp <- expression(cat(" ", count, "\t\t ", coll1 <- coll,
                              fill = TRUE))
    }
    else
        exp <- expression(coll1 <-  coll)

    ## Iterative procedure
    count <- 0
    repeat
    {
        eval(exp)

        ## Stop after 'maxit' iterations
        if (maxit < (count <- count + 1))
        {
            warning("maximum number of iterations reached before obtaining convergence")
            break
        }

        ## Calculation of the between variance-covariance matrix.
        A[] <- rowSums(sapply(has.data, function(i)
                              cred[, , i] %*% tcrossprod(ind[, i] - coll))) /
                                  (eff.ncontracts - 1)

        ## Symmetrize A
        A <- (A + t(A))/2

        ## New credibility matrices
        cred[, , has.data] <- sapply(has.data, function(i)
                                     A %*% solve(A + s2 * W[, , i]))

        ## New collective regression coefficients
        cred.s <- apply(cred[, , has.data], c(1L, 2L), sum)
        coll <- solve(cred.s,
                      rowSums(sapply(has.data, function(i)
                                     cred[, , i] %*% ind[, i])))

        ## Test for convergence
        if (max(abs((coll - coll1)/coll1)) < tol)
            break
    }

    ## Final calculation of the between variance-covariance matrix and
    ## credibility matrices.
    A[] <- rowSums(sapply(has.data, function(i)
                          cred[, , i] %*% tcrossprod(ind[, i] - coll))) /
                              (eff.ncontracts - 1)
    A <- (A + t(A))/2
    cred[, , has.data] <- sapply(has.data, function(i)
                                 A %*% solve(A + s2 * W[, , i]))

    ## Credibility adjusted coefficients. The coefficients of the
    ## models are replaced with these values. That way, prediction
    ## will be trivial using predict.lm().
    for (i in seq_len(ncontracts))
        fits[[i]]$coefficients <- coll + drop(cred[, , i] %*% (ind[, i] - coll))

    ## Add names to the collective coefficients vector.
    names(coll) <- rownames(ind)

    ## Results
    list(means = list(coll, ind),
         weights = list(cred.s, W),
         unbiased = NULL,
         iterative = list(A, s2),
         cred = cred,
         nodes = list(ncontracts),
         adj.models = fits)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Quantiles for objects of class 'aggregateDist'
###
### AUTHORS:  Louis-Philippe Pouliot,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

quantile.aggregateDist <-
    function(x, probs = c(0.25, 0.5, 0.75, 0.9, 0.95, 0.975, 0.99, 0.995),
             smooth = FALSE, names = TRUE, ...)
{
    chkDots(...)                        # method does not use '...'
    label <- comment(x)

    ## The Normal and Normal Power approximations are the only
    ## continuous distributions of class 'aggregateDist'. They are
    ## therefore treated differently, using the 'base' quantile
    ## function qnorm().
    if (label == "Normal approximation")
        res <- qnorm(probs, get("mean", environment(x)),
                     sqrt(get("variance", environment(x))))
    else if (label == "Normal Power approximation")
    {
        m <- get("mean", envir = environment(x))
        sd <- sqrt(get("variance", envir = environment(x)))
        sk <- get("skewness", envir = environment(x))

        ## Calling qnorm() and inverting the Normal Power 'standardization'
        q <- qnorm(probs)
        res <- ifelse(probs <= 0.5, NA, m + sd * (q + sk * (q^2 - 1)/6))
    }
    else
    {
        ## An empirical and discrete approach is used for
        ## 'aggregateDist' objects obtained from methods other than
        ## Normal and Normal Power.
        y <- get("y", environment(x))
        x <- get("x", environment(x))

        ## Create the inverse function of either the cdf or the ogive.
        fun <-
            if (smooth)                     # ogive
                approxfun(y, x, yleft = 0, yright = max(x),
                          method = "linear", ties = "ordered")
            else                            # cdf
                approxfun(y, x, yleft = 0, yright = max(x),
                          method = "constant", f = 1, ties = "ordered")

        ## Quantiles
        res <- fun(probs)
    }

    if (names)
    {
        dig <- max(2, getOption("digits"))
        names(res) <- formatC(paste(100 * probs, "%", sep = ""),
                              format = "fg", width = 1, digits = dig)
    }
    res
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Empirical moments for individual and grouped data.
###
### See Klugman, Panjer & Willmot, Loss Models, Wiley, 1998.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

emm <- function(x, order = 1, ...) UseMethod("emm")

emm.default <- function(x, order = 1, ...)
{
    if (any(order < 0))
      stop(sprintf("%s must be positive", sQuote("order")))

    colMeans(outer(x, order, "^"), ...)
}

emm.grouped.data <- function(x, order = 1, ...)
{
    ## Function does not work for negative moments
    if (any(order < 0))
      stop(sprintf("%s must be positive", sQuote("order")))

    ## Extract group boundaries
    cj <- eval(expression(cj), envir = environment(x))

    ## Compute the factor
    ##
    ## f_j = (c_j^{k + 1} - c_{j-1}^{k+1})/((k+1) * (c_j - c_{j-1}))
    ##
    ## for all values of 'j' and 'k' == 'order'.
    y <- diff(outer(cj, order + 1, "^")) / outer(diff(cj), order + 1)

    ## Drop the group boundaries column
    x <- as.matrix(x[-1L])

    ## Compute sum(n_j * f_j)/sum(nj) for all values of 'order'.
    drop(crossprod(x, y)) / colSums(x, ...)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}trgamma functions to compute
### characteristics of the Transformed Gamma distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = pgamma((x/scale)^shape2, shape1, scale = 1), x > 0
###
### or, equivalently,
###
###   Pr[X <= x] = pgamma(x^shape2, shape1, scale = scale^shape2), x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dtrgamma <- function (x, shape1, shape2, rate = 1, scale = 1/rate,
                      log = FALSE)
    .External(C_actuar_do_dpq, "dtrgamma", x, shape1, shape2, scale, log)

ptrgamma <- function(q, shape1, shape2, rate = 1, scale = 1/rate,
                     lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "ptrgamma", q, shape1, shape2, scale,
              lower.tail, log.p)

qtrgamma <- function(p, shape1, shape2, rate = 1, scale = 1/rate,
                     lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qtrgamma", p, shape1, shape2, scale,
              lower.tail, log.p)

rtrgamma <- function(n, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rtrgamma", n, shape1, shape2, scale)

mtrgamma <- function(order, shape1, shape2, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mtrgamma", order, shape1, shape2, scale, FALSE)

levtrgamma <- function(limit, shape1, shape2, rate = 1, scale = 1/rate,
                       order = 1)
    .External(C_actuar_do_dpq, "levtrgamma", limit, shape1, shape2, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Variance and standard deviation
###
### See Klugman, Panjer & Willmot, Loss Models, Wiley, 1998.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>
### Walter Garcia-Fontes

## New generics for functions of the stats package
var <- function(x, ...) UseMethod("var")
sd <- function(x, ...) UseMethod("sd")

## Default methods are stats::var and stats:sd
var.default <- function(x, y = NULL, na.rm = FALSE, use, ...)
    stats::var(x, y = NULL, na.rm = na.rm, use)
sd.default <- function(x, na.rm = FALSE, ...)
    stats::sd(x, na.rm = na.rm)

## Methods for grouped data
var.grouped.data <- function(x, ...)
{
    ## Get group boundaries
    cj <- eval(expression(cj), envir = environment(x))

    ## Compute group midpoints
    midpoints <- cj[-length(cj)] + diff(cj)/2

    ## Extract frequencies columns by dropping the boundaries column;
    ## convert to matrix for use in crossprod()
    x <- as.matrix(x[-1L])

    ## Compute mean per column (avoiding a call to
    ## 'mean.grouped.data') that would redo most of the computations
    ## above.
    means <- drop(crossprod(x, midpoints))/colSums(x)

    ## Compute midpoints minus mean and square it
    midsquare <- (midpoints - means)^2

    ## Compute mean per column
    drop(crossprod(x, midsquare))/(colSums(x) - 1)
}

sd.grouped.data <- function(x, ...)
{
    ## Square root of variance
    drop(sqrt(var.grouped.data(x)))
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Simulation of hierarchical portfolios of data. Claim number and
### claim amounts in any given node are simulated independently. Both
### frequency and severity models can be mixtures of distributions.
###
### In the code here, numbering of levels starts at 1 at the data
### level, whereas in common discussion of hierarchical models the
### data level is numbered 0.
###
### AUTHORS:  Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Sebastien Auclair and Louis-Philippe Pouliot

rcomphierarc <- function(nodes, model.freq = NULL, model.sev = NULL, weights = NULL)
{
    ## Get level names. Each could be NULL.
    level.names <- names(nodes)
    freq.names <- names(model.freq)
    sev.names <- names(model.sev)

    ## 'nodes' must be a named list. One exception is allowed: there
    ## is only one level. In this case, add a predetermined name if
    ## there isn't one already and make sure 'nodes' is a list.
    if (length(nodes) == 1L)
    {
        if (is.null(level.names))
            names(nodes) <- "X"
        nodes <- as.list(nodes)
    }
    else
    {
        if (!is.list(nodes) || is.null(level.names))
            stop(sprintf("%s must be a named list", sQuote("nodes")))
    }

    ## Determine if frequency and severity models are present. Keep
    ## for future use.
    has.freq <- !all(sapply(model.freq, is.null))
    has.sev  <- !all(sapply(model.sev, is.null))

    ## Check that at least one of 'model.freq' or 'model.sev' is
    ## present and that the level names match with those of 'nodes'.
    ## Perhaps is there a fancier way to do all these tests, but the
    ## version below is at least easy to follow.
    if (has.freq)
    {
        if (has.sev)
        {
            if (! (identical(level.names, freq.names) &&
                   identical(level.names, sev.names)))
                stop(sprintf("level names different in %s, %s and %s",
                             sQuote("nodes"),
                             sQuote("model.freq"),
                             sQuote("model.sev")))
        }
        else
        {
            if (!identical(level.names, freq.names))
                stop(sprintf("level names different in %s, %s and %s",
                             sQuote("nodes"),
                             sQuote("model.freq"),
                             sQuote("model.sev")))
        }
    }
    else
    {
        if (has.sev)
        {
            if (!identical(level.names, sev.names))
                stop(sprintf("level names different in %s, %s and %s",
                             sQuote("nodes"),
                             sQuote("model.freq"),
                             sQuote("model.sev")))
        }
        else
            stop(sprintf("one of %s or %s must be non-NULL",
                         sQuote("model.freq"), sQuote("model.sev")))
    }

    ## The function is written for models with at least two levels
    ## (entity and year). If there is only one, add a dummy level to
    ## avoid scattering the code with conditions.
    if (length(nodes) < 2L)
    {
        nodes <- c(node = 1, nodes)
        model.freq <-
            if (has.freq) c(expression(node = NULL), model.freq) else NULL
        model.sev <-
            if (has.sev) c(expression(node = NULL), model.sev) else NULL
    }

    ## Frequently used quantities
    level.names <- names(nodes)         # need to reset!
    nlevels <- length(nodes)            # number of levels

    ## Recycling of the number of nodes (if needed) must be done
    ## "manually". We do it here once and for all since in any case
    ## below we will need to know the total number of nodes in the
    ## portfolio. Furthermore, the recycled list 'nodes' will be
    ## returned by the function.
    for (i in 2L:nlevels)       # first node doesn't need recycling
        nodes[[i]] <- rep(nodes[[i]], length = sum(nodes[[i - 1L]]))

    ## Simulation of the frequency mixing parameters for each level
    ## (e.g. class, contract) and, at the last level, the actual
    ## frequencies. If 'model.freq' is NULL, this is equivalent to
    ## having one claim per node.
    if (has.freq)
    {
        ## Normally, only the immediately above mixing parameter will
        ## be used in the model for a level, but the code here allows
        ## for more general schemes. For this to work, all mixing
        ## parameters have to be correctly recycled at each level
        ## where they *could* be used. Since the model at any level
        ## could be NULL, 'params' will keep track of the mixing
        ## parameters that were simulated in previous iteration of the
        ## forthcoming loop.
        params <- character(0)

        for (i in seq_len(nlevels))
        {
            ## Number of nodes at the current level
            n.current <- nodes[[i]]

            ## Extract simulation model for the level.
            Call <- model.freq[[i]]

            ## Repeat the mixing parameters of all levels above the
            ## current one that were simulated in the past.
            for (j in seq_along(params))
                eval(substitute(x <- rep.int(x, n.current),
                                list(x = as.name(params[[j]]))))

            ## Simulate data only if there is a model at the current
            ## level.
            if (!is.null(Call))
            {
                ## Add the number of variates to the call.
                Call$n <- sum(n.current)

                ## Simulation of the mixing parameters or the data. In
                ## the latter case, store the results in a fixed
                ## variable name.
                if (i < nlevels)
                {
                    assign(level.names[[i]], eval(Call))
                    params[i] <- level.names[[i]] # remember the parameter
                }
                else
                    frequencies <- eval(Call)
            }
        }
    }
    else
        frequencies <- rep.int(1, sum(nodes[[nlevels]]))

    ## Simulation of the claim amounts. If 'model.sev' is NULL, this
    ## is equivalent to simulating frequencies only.
    if (has.sev)
    {
        ## Repeat the same procedure as for the frequency model, with
        ## one difference: when reaching the last level (claim
        ## amounts), the number of variates to simulate is not given
        ## by the number of nodes but rather by the number of claims
        ## as found in 'frequencies'.
        params <- character(0)

        for (i in seq_len(nlevels))
        {
            n.current <- nodes[[i]]
            Call <- model.sev[[i]]

            for (j in seq_along(params))
                eval(substitute(x <- rep.int(x, n.current),
                                list(x = as.name(params[[j]]))))

            if (!is.null(Call))
            {
                ## The rest of the procedure differs depending if we
                ## are still simulating mixing parameters or claim
                ## amounts.
                if (i < nlevels)
                {
                    ## Simulation of mixing parameters is identical to the
                    ## simulation of frequencies.
                    Call$n <- sum(n.current)
                    assign(level.names[[i]], eval(Call))
                    params[i] <- level.names[[i]]
                }
                else
                {
                    ## For the simulation of claim amounts, the number
                    ## of variates is rather given by the
                    ## 'frequencies' object. Furthermore, the mixing
                    ## parameters must be recycled once more to match
                    ## the vector of frequencies.
                    for (p in intersect(all.vars(Call), params))
                        eval(substitute(x <- rep.int(x, frequencies),
                                        list(x = as.name(p))))
                    Call$n <- sum(frequencies)
                    severities <-eval(Call)
                }
            }
        }
    }
    else
        severities <- rep.int(1, sum(frequencies))

    ## We must now distribute the claim amounts in vector 'severities'
    ## to the appropriate nodes. This is complicated by the
    ## possibility to have different number of nodes (years of
    ## observation) for each entity. The result must be a matrix
    ## with the number of columns equal to the maximum number of last
    ## level nodes.
    ##
    ## The number of nodes (years of observation) per entity is
    ## given by 'n.current' since we reached the last level in (either
    ## one of) the above loops.
    ##
    ## Assign a unique ID to each node, leaving gaps for nodes without
    ## observations.
    ind <- unlist(mapply(seq,
                         from = seq(by = max(n.current), along = n.current),
                         length = n.current))

    ## Repeating the vector of IDs according to the frequencies
    ## effectively assigns a node ID to each claim amount. The vector
    ## of claim amounts is then split by node, yielding a list where
    ## each element corresponds to a node with claims.
    f <- rep.int(ind, frequencies)
    severities <- split(severities, f)

    ## Identify nodes with frequency equal to 0, which is different
    ## from having no observation (NA).
    freq0 <- ind[which(frequencies == 0)]

    ## Rearrange the list of claim amounts in a matrix;
    ##
    ##      number of rows: number of nodes at the penultimate level
    ##                      (number of entities)
    ##   number of columns: maximum number of nodes at the last level
    ##                      (number of years of observation).
    ##
    ## Moreover, assign a value of 'numeric(0)' to nodes with a
    ## frequency of 0.
    nrow <- length(n.current)           # number of entities
    ncol <- max(n.current)              # number of years
    res <- as.list(rep.int(NA, nrow * ncol))
    res[unique(f)] <- severities
    res[freq0] <- lapply(rep.int(0, length(freq0)), numeric)
    res <- matrix(res, nrow, ncol, byrow = TRUE,
                  dimnames = list(NULL,
                                  paste(level.names[nlevels],
                                        seq_len(ncol), sep = ".")))

    ## Reshape weights as a matrix, if necessary.
    weights <- if (is.null(weights))
        NULL
    else
    {
        ## Integrate NAs into the weights matrix as appropriate.
        w <- rep.int(NA, nrow * ncol)
        w[ind] <- weights
        matrix(w, nrow = nrow, byrow = TRUE, dimnames = dimnames(res))
    }

    ## Finally, create a matrix where each row contains the series of
    ## identifiers for an entity in the portfolio, e.g. if the data
    ## is denoted X_{ijkt}, one line of the matrix will contain
    ## subscripts i, j and k. As we move from right to left in the
    ## columns of 'm', the subcripts are increasingly repeated.
    ncol <- nlevels - 1L
    m <- matrix(1, nrow, ncol,
                dimnames = list(NULL, head(level.names, ncol)))
    for (i in seq_len(ncol - 1L))    # all but the last column
    {
        ## Vector 'x' will originally contain all subscripts for one
        ## level. These subscripts are then repeated as needed to give
        ## the desired result. To avoid another explicit loop, I use a
        ## 'lapply' with a direct assignment in the current
        ## frame. Somewhat unusual, but this is the simplest procedure
        ## I managed to come up with.
        x <- unlist(lapply(nodes[[i]], seq))
        lapply(nodes[(i + 1L):(nlevels - 1L)],
               function(v) assign("x", rep.int(x, v), envir = parent.frame(2)))
        m[, i] <- x
    }
    m[, ncol] <- unlist(lapply(nodes[[ncol]], seq)) # last column

    ## Return object of class 'portfolio'
    structure(list(data = res,
                   weights = weights,
                   classification = m,
                   nodes = nodes,
                   model.freq = model.freq,
                   model.sev = model.sev),
              class = "portfolio")
}

### Alias for backward compatibility with actuar < 2.0-0.
simul <- rcomphierarc

### 'print' method for 'portfolio' objects
print.portfolio <- function(x, ...)
{
    cat("\nPortfolio of claim amounts \n\n")
    nn <- names(x$nodes)
    nc <- max(nchar(nn))
    if (!is.null(x$model.freq))
    {
        cat("  Frequency model\n")
        cat(paste("    ",
                  format(nn, width = nc),
                  " ~ ",
                  x$model.freq,
                  "\n", sep = ""), sep = "")
    }
    if (!is.null(x$model.sev))
    {
        cat("  Severity model\n")
        cat(paste("    ",
                  format(nn, width = nc),
                  " ~ ",
                  x$model.sev,
                  "\n", sep = ""), sep = "")
    }
    cat("\n  Number of claims per node: \n\n")
    print(frequency(x), ...)
    invisible(x)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}fpareto functions to compute
### characteristics of the Feller-Pareto distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = Pr[Y <= v/(1 + v)], x > min,
###
### where v = ((x - min)/scale)^shape2 and Y has a Beta distribution
### with parameters shape3 and shape1.
###
### See Arnold, B. C. (2015), Pareto Distributions, Second Edition,
### CRC Press.
###
### AUTHORS: Nicholas Langevin, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dfpareto <-
    function (x, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              log = FALSE)
    .External(C_actuar_do_dpq, "dfpareto", x, min,  shape1, shape2,
              shape3, scale, log)

pfpareto <-
    function (q, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pfpareto", q, min, shape1, shape2, shape3,
              scale, lower.tail, log.p)

qfpareto <-
    function (p, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qfpareto", p, min, shape1, shape2, shape3,
              scale, lower.tail, log.p)

rfpareto <-
    function (n, min, shape1, shape2, shape3, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rfpareto", n, min, shape1, shape2,
              shape3, scale)

mfpareto <-
    function (order, min, shape1, shape2, shape3, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mfpareto", order, min, shape1, shape2,
              shape3, scale, FALSE)

levfpareto <-
    function (limit, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
             order = 1)
    .External(C_actuar_do_dpq, "levfpareto", limit, min, shape1, shape2,
              shape3, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}gumbel functions to compute
### characteristics of the Gumbel distribution. The version used in
### these functions has cumulative distribution function
###
###   Pr[X <= x] = exp(-exp(-(x - alpha)/scale)), -Inf < x < Inf.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dgumbel <- function (x, alpha, scale, log = FALSE)
    .External(C_actuar_do_dpq, "dgumbel", x, alpha, scale, log)

pgumbel <- function(q, alpha, scale, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pgumbel", q, alpha, scale, lower.tail, log.p)

qgumbel <- function(p, alpha, scale, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qgumbel", p, alpha, scale, lower.tail, log.p)

rgumbel <- function(n, alpha, scale)
    .External(C_actuar_do_random, "rgumbel", n, alpha, scale)

mgumbel <- function(order, alpha, scale)
    .External(C_actuar_do_dpq, "mgumbel", order, alpha, scale, FALSE)

mgfgumbel <- function(t, alpha, scale, log = FALSE)
    .External(C_actuar_do_dpq, "mgfgumbel", t, alpha, scale, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}genbeta functions to compute
### characteristics of the Generalized Beta distribution. The version
### used in these functions has cumulative distribution function
###
###   Pr[X <= x] = Pr[Y <= (x/scale)^shape3], 0 < x < scale,
###
### where Y has a Beta distribution with parameters shape1 and shape2.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dgenbeta <-
    function (x, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              log = FALSE)
    .External(C_actuar_do_dpq, "dgenbeta", x, shape1, shape2, shape3, scale, log)

pgenbeta <-
    function (q, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pgenbeta", q, shape1, shape2, shape3, scale,
              lower.tail, log.p)

qgenbeta <-
    function (p, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qgenbeta", p, shape1, shape2, shape3, scale,
              lower.tail, log.p)

rgenbeta <-
    function (n, shape1, shape2, shape3, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rgenbeta", n, shape1, shape2, shape3, scale)

mgenbeta <-
    function (order, shape1, shape2, shape3, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mgenbeta", order, shape1, shape2, shape3, scale, FALSE)

levgenbeta <-
    function (limit, shape1, shape2, shape3, rate = 1, scale = 1/rate,
              order = 1)
    .External(C_actuar_do_dpq, "levgenbeta", limit, shape1, shape2, shape3, scale,
              order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}ztbinom functions to compute
### characteristics of the Zero Truncated Binomial distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Second
### Edition, Wiley, 2004.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dztbinom <- function (x, size, prob, log = FALSE)
    .External(C_actuar_do_dpq, "dztbinom", x, size, prob, log)

pztbinom <- function(q, size, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pztbinom", q, size, prob, lower.tail, log.p)

qztbinom <- function(p, size, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qztbinom", p, size, prob, lower.tail, log.p)

rztbinom <- function(n, size, prob)
    .External(C_actuar_do_random, "rztbinom", n, size, prob)

## not exported; for internal use in panjer()
pgfztbinom <- function(x, size, prob)
{
    qn <- (1 - prob)^size
    (exp(size * log1p(prob * (x - 1))) - qn)/(1 - qn)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}invexp functions to compute
### characteristics of the Inverse Exponential distribution. The
### version used in these functions has cumulative distribution
### function
###
###   Pr[X <= x] = exp(-scale/x), x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dinvexp <- function (x, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "dinvexp", x, scale, log)

pinvexp <- function(q, rate = 1, scale = 1/rate,
                    lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pinvexp", q, scale, lower.tail, log.p)

qinvexp <- function(p, rate = 1, scale = 1/rate,
                    lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qinvexp", p, scale, lower.tail, log.p)

rinvexp <- function(n, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rinvexp", n, scale)

minvexp <- function(order, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "minvexp", order, scale, FALSE)

levinvexp <- function(limit, rate = 1, scale = 1/rate, order)
    .External(C_actuar_do_dpq, "levinvexp", limit, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev,mgf}chisq functions to compute raw and
### limited moments, and the moment generating function for
### the Chi-square distribution (as defined in R)
###
### See Chapter 17 of Johnson & Kotz, Continuous univariate
### distributions, volume 1, Wiley, 1970
###
### AUTHORS: Christophe Dutang, Vincent Goulet <vincent.goulet@act.ulaval.ca>

mchisq <- function(order, df, ncp = 0)
    .External(C_actuar_do_dpq, "mchisq", order, df, ncp, FALSE)

levchisq <- function(limit, df, ncp = 0, order = 1)
    .External(C_actuar_do_dpq, "levchisq", limit, df, ncp, order, FALSE)

mgfchisq <- function(t, df, ncp = 0, log = FALSE)
    .External(C_actuar_do_dpq, "mgfchisq", t, df, ncp, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Mean of grouped data objects
###
### See Klugman, Panjer & Willmot, Loss Models, Wiley, 1998.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

## New method of base::mean generic for grouped data
mean.grouped.data <- function(x, ...)
{
    ## Get group boundaries
    cj <- eval(expression(cj), envir = environment(x))

    ## Compute group midpoints
    midpoints <- cj[-length(cj)] + diff(cj)/2

    ## Extract frequencies columns by dropping the boundaries column;
    ## convert to matrix for use in crossprod()
    x <- as.matrix(x[-1L])

    ## Compute mean per column
    drop(crossprod(x, midpoints))/colSums(x)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Pure bayesian credibility calculations.
###
### AUTHORS: Alexandre Parent <alexandre.parent.12@ulaval.ca>,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

bayes <- function(x, likelihood =
                         c("poisson", "bernoulli", "geometric",
                           "exponential", "normal", "binomial",
                           "negative binomial", "gamma",
                           "pareto"),
                  shape, rate = 1, scale = 1/rate,
                  shape1, shape2, mean = 0, sd = 1,
                  size, shape.lik, sd.lik, min)
{
    likelihood <- match.arg(likelihood)

    ## We need to treat separately the (Single Parameter, or
    ## Translated) Pareto/Gamma case given the different form of the
    ## individual mean and the "credibility factor" (which isn't one,
    ## really).
    if (likelihood == "pareto")
    {
        if (missing(min))
            stop("lower bound of the likelihood missing")
        if (missing(shape) || (missing(rate) && missing(scale)))
            stop(sprintf("one of the Gamma prior parameter %s, %s or %s missing",
                         dQuote("shape"), dQuote("rate"), dQuote("scale")))
        coll <- shape * scale
        vars <- c(NA, NA)                # not pertinent here

        ## Computation of individual means and credibility factors
        ## differs depending on the type of data provided in argument.
        if (is.null(x))                 # no data
            cred <- ind.means <- n <- 0
        else if (is.vector(x, mode = "numeric")) # atomic vector
        {
            n <- length(x)
            sumlog <- sum(log(x)) - n * log(min)
            ind.means <- n/sumlog
            cred <- 1/(1 + 1/(scale * sumlog))
        }
        else                            # matrix or data frame
        {
            n <- ncol(x)
            sumlog <- rowSums(log(x)) - n * log(min)
            ind.means <- n/sumlog
            cred <- 1/(1 + 1/(scale * sumlog))
        }
    }
    ## Now the usual linear Bayes cases.
    else
    {
        if (likelihood == "poisson")
        {
            if (missing(shape) || (missing(rate) && missing(scale)))
                stop(sprintf("one of the Gamma prior parameter %s, %s or %s missing",
                             dQuote("shape"), dQuote("rate"), dQuote("scale")))
            coll <- shape * scale
            vars <- c(coll * scale, coll)
            K <- 1/scale
        }
        else if (likelihood == "bernoulli")
        {
            if (missing(shape1) || missing(shape2))
                stop(sprintf("one of the Beta prior parameter %s or %s missing",
                             dQuote("shape1"), dQuote("shape2")))
            K <- shape1 + shape2
            coll <- shape1/K
            vars <- (shape1 * shape2) * c(1, K)/(K^2 * (K + 1))
        }
        else if (likelihood == "binomial")
        {
            if (missing(shape1) || missing(shape2))
                stop(sprintf("one of the Beta prior parameter %s or %s missing",
                             dQuote("shape1"), dQuote("shape2")))
            if (missing(size))
                stop(sprintf("parameter %s of the likelihood missing",
                             dQuote("size")))
            K <- (shape1 + shape2)/size
            coll <- shape1/K
            vars <- (shape1 * shape2) * c(1, K)/(K^2 * (shape1 + shape2 + 1))
        }
        else if (likelihood == "geometric")
        {
            if (missing(shape1) || missing(shape2))
                stop(sprintf("one of the Beta prior parameter %s or %s missing",
                             dQuote("shape1"), dQuote("shape2")))
            K <- shape1 - 1
            coll <- shape2/K
            vars <- shape2 * (shape1 + shape2 - 1)/(K * (K - 1))
            vars <- c(vars/K, vars)
        }
        else if (likelihood == "negative binomial")
        {
            if (missing(shape1) || missing(shape2))
                stop(sprintf("one of the Beta prior parameter %s or %s missing",
                             dQuote("shape1"), dQuote("shape2")))
            if (missing(size))
                stop(sprintf("parameter %s of the likelihood missing",
                             dQuote("size")))
            K <- (shape1 - 1)/size
            coll <- shape2/K
            vars <- shape2 * (shape1 + shape2 - 1)/(K * (shape1 - 2))
            vars <- c(vars/K, vars)
        }
        else if (likelihood == "exponential")
        {
            if (missing(shape) || (missing(rate) && missing(scale)))
                stop(sprintf("one of the Gamma prior parameter %s, %s or %s missing",
                             dQuote("shape"), dQuote("rate"), dQuote("scale")))
            K <- shape - 1
            coll <- 1/(K * scale)
            vars <- c(coll^2, coll/scale)/(shape - 2)
        }
        else if (likelihood == "gamma")
        {
            if (missing(shape) || (missing(rate) && missing(scale)))
                stop(sprintf("one of the Gamma prior parameter %s, %s or %s missing",
                             dQuote("shape"), dQuote("rate"), dQuote("scale")))
            if (missing(shape.lik))
                stop(sprintf("parameter %s of the likelihood missing",
                             dQuote("shape.lik")))
            K <- (shape - 1)/shape.lik
            coll <- 1/(K * scale)
            vars <- c(coll^2, coll/scale)/(shape - 2)
        }
        else if (likelihood == "normal")
        {
            if (missing(sd.lik))
                stop(sprintf("parameter %s of the likelihood missing",
                             dQuote("sd.lik")))
            coll <- mean
            vars <- c(sd, sd.lik)^2
            K <- vars[2L]/vars[1L]
        }
        else
            stop("unsupported likelihood")

        ## Computation of individual means and credibility factors
        ## differs depending on the type of data provided in argument.
        if (is.null(x))                 # no data
            cred <- ind.means <- n <- 0
        else if (is.vector(x, mode = "numeric")) # atomic vector
        {
            n <- length(x)
            ind.means <- mean(x)
            cred <- n/(n + K)
        }
        else                            # matrix or data frame
        {
            n <- ncol(x)
            ind.means <- rowMeans(x)
            cred <- n/(n + K)
        }
    }

    structure(list(means = list(coll, ind.means),
                   weights = list(NULL, n),
                   unbiased = vars,
                   iterative = NULL,
                   cred = cred,
                   nodes = 1L),
              class = "bayes",
              model = "Linear Bayes")

}

## Premium calculation is identical to the Buhlmann-Straub case; no
## need for another method. See bstraub.R for the definition.
# predict.bayes <- predict.bstraub

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}ztpois functions to compute
### characteristics of the Zero Truncated Poisson distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dztpois <- function (x, lambda, log = FALSE)
    .External(C_actuar_do_dpq, "dztpois", x, lambda, log)

pztpois <- function(q, lambda, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pztpois", q, lambda, lower.tail, log.p)

qztpois <- function(p, lambda, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qztpois", p, lambda, lower.tail, log.p)

rztpois <- function(n, lambda)
    .External(C_actuar_do_random, "rztpois", n, lambda)

## not exported; for internal use in panjer()
pgfztpois <- function(x, lambda)
    expm1(lambda * x)/expm1(lambda)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}zmnbinom functions to compute
### characteristics of the Zero Modified Negative Binomial
### distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dzmnbinom <- function (x, size, prob, p0, log = FALSE)
    .External(C_actuar_do_dpq, "dzmnbinom", x, size, prob, p0, log)

pzmnbinom <- function(q, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pzmnbinom", q, size, prob, p0, lower.tail, log.p)

qzmnbinom <- function(p, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qzmnbinom", p, size, prob, p0, lower.tail, log.p)

rzmnbinom <- function(n, size, prob, p0)
    .External(C_actuar_do_random, "rzmnbinom", n, size, prob, p0)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Create modified density and modified cumulative distribution
### function for data with deductible, limit, coinsurance and
### inflation.
###
### See Chapter 8 of Klugman, Panjer & Willmot, Loss Models, Fourth
### Edition, Wiley, 2012.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

## TODO (2025-03-18): when 'coverage' is called with variables, say
## 'foo' and 'bar', holding the names of the pdf and cdf, the function
## returned contains calls to 'foo' and 'bar' instead of the names
## actually therein. To solve this, replace Call$[pc]df by
## eval(Call$[pc]df).
##
## TODO (2025-03-18): furthermore, the underlying pdf and cdf should
## be put in the environment of the returned function, such that its
## evaluation remains possible even if the pdf and cdf are not in the
## current environment. Perhaps only one of these two changes is
## necessary.
##
## TODO (2025-03-19): in the limit case where there is no deductible
## and no limit, the values at 0 (for densities with the mode at 0)
## and Inf (least likely) may be wrong.

coverage <- function(pdf, cdf, deductible = 0, franchise = FALSE,
                     limit = Inf, coinsurance = 1, inflation = 0,
                     per.loss = FALSE)
{
    Call <- match.call()

    ## First determine if the cdf is needed or not. It is needed when
    ## there is a deductible or a limit and, of course, if the output
    ## function should compute the cdf.
    is.cdf <- missing(pdf) || is.null(pdf) # return cdf?
    has.limit <- limit < Inf               # used often
    needs.cdf <- any(deductible > 0, has.limit, is.cdf) # cdf needed?

    ## Sanity check of arguments
    if (any(deductible < 0, limit < 0, coinsurance < 0, inflation < 0))
        stop("coverage modifications must be positive")
    if (limit <= deductible)
      stop("deductible must be smaller than the limit")
    if (coinsurance > 1)
        stop("coinsurance must be between 0 and 1")
    if (missing(cdf) & needs.cdf)
        stop(sprintf("%s must be supplied", sQuote("cdf")))

    ## Quantites often used. Leave as expressions for the output
    ## function to preserve accuracy.
    r <- 1 + inflation
    d <- if (inflation) substitute(d/r, list(d = deductible, r = r)) else deductible
    u <- if (inflation) substitute(u/r, list(u = limit, r = r)) else limit

    ## The modified cdf or pdf are usually defined in branches. To
    ## avoid using nested ifelse(), we will rather rely on sets of
    ## expressions to make the required calculations for each branch
    ## separately. This is actually much faster.
    ##
    ## The output function will have varying number of said
    ## expressions depending on the case that is dealt with. We will
    ## build the body of the output function piece by piece as we go
    ## along.
    e <- expression(Call <- match.call())

    ## One main discriminating factor is whether the cdf is needed for
    ## the output function of not.
    if (needs.cdf)
    {
        ## Get argument list of 'cdf' to transfert them to the output
        ## function.
        argv <- formals(cdf)            # arguments as list
        argn <- names(argv)             # arguments names as strings

        ## Remember if argument 'lower.tail' is available, so we can
        ## use it later. Then, drop unsupported arguments 'lower.tail'
        ## and 'log.p'.
        has.lower <- "lower.tail" %in% argn
        argn <- setdiff(argn, c("lower.tail", "log.p"))

        ## Calculations in the output function are done by evaluating
        ## function calls built upon the call to the output function
        ## itself. This is convenient as we do not need to fiddle with
        ## the values of the formal arguments.
        if (is.cdf)                     # output function computes F(y)
        {
            ## The output function will have the same formal arguments
            ## as the cdf. Object 'x' holds the symbol of the first
            ## argument.
            argsFUN <- argv[argn]       # arguments of output function
            x <- as.name(argn[1])       # symbol

            ## Calls needed in this case:
            ## 1. one to compute F(y);
            ## 2. one to compute F(d) if there is a deductible;
            ## 3. one to compute 1 - F(d) for the per payment cases.
            ## Never need to compute F(u).
            ##
            ## If 'lower.tail' is available in 'cdf', then set it to
            ## FALSE to compute S(d) = 1 - F(d) more accurately.
            e <- c(e,
                   quote(F <- Call),        # 1.
                   substitute(F[[1L]] <- as.name(fun),
                              list(fun = as.character(Call$cdf))))
            if (deductible)
            {
                e <- c(e,
                       quote(Fd <- F),      # 2.
                       substitute(Fd[[2L]] <- a, list(a = d)))
                if (!per.loss & has.lower)
                    e <- c(e,
                           quote(Sd <- Fd), # 3.
                           quote(Sd$lower.tail <- FALSE))
            }
        }
        else                            # output function computes f(y)
        {
            ## When there is a limit, we will need to compute 1 - F(u)
            ## as is or using 'lower.tail = FALSE' to improve
            ## accuracy. For clarity in the output function, we will
            ## use Fu as object name in the first case and Su in the
            ## second.
            if (has.limit)
            {
                if (has.lower)
                {
                    Fu.name <- as.name("Su")
                    Su.quote <- quote(eval.parent(Su))
                }
                else
                {
                    Fu.name <- as.name("Fu")
                    Su.quote <- quote((1 - eval.parent(Fu)))
                }
            }

            ## Calls needed in this case:
            ## 1. one to compute F(d) if there is a deductible for the
            ##    per loss cases;
            ## 2. one to compute 1 - F(d) if there is a deductible for the
            ##    per payment cases;
            ## 3. one to compute 1 - F(u) when there is a limit.
            ## No function to compute F(y) needed.
            ##
            ## If 'lower.tail' is available in 'cdf', then set it to
            ## FALSE to compute S(d) = 1 - F(d) and S(u) = 1 - F(u)
            ## more accurately.
            if (deductible)             # f(y) with deductible
            {
                Fd.name <- as.name(if (!per.loss & has.lower) "Sd" else "Fd")
                e <- c(e,
                       substitute(G <- Call, list(G = Fd.name)),         # 1. or 2.
                       if (!per.loss & has.lower) quote(Sd$lower.tail <- FALSE),
                       substitute(G[[1L]] <- as.name(fun),
                                  list(G = Fd.name, fun = as.character(Call$cdf))),
                       substitute(names(G)[2L] <- q, list(G = Fd.name, q = argn[1])),
                       substitute(G[[2L]] <- a, list(G = Fd.name, a = d)))
                if (has.limit)
                    e <- c(e,
                           substitute(H <- G, list(H = Fu.name, G = Fd.name)), # 3.
                           if (per.loss & has.lower) quote(Su$lower.tail <- FALSE),
                           substitute(H[[2L]] <- a, list(H = Fu.name, a = u)))
            }
            else                        # f(y) with limit only
            {
                ## Since 'needs.cdf == TRUE', then this case
                ## necessarily has 'limit < Inf'. Only call needed is
                ## one to compute 1 - F(u).
                e <- c(e,
                       substitute(G <- Call, list(G = Fu.name)),
                       if (has.lower) quote(Su$lower.tail <- FALSE),
                       substitute(G[[1L]] <- as.name(fun),
                                  list(G = Fu.name, fun = as.character(Call$cdf))),
                       substitute(names(G)[2L] <- q, list(G = Fu.name, q = argn[1])),
                       substitute(G[[2L]] <- a, list(G = Fu.name, a = u)))
            }
        }
    }

    ## Repeat same steps as above for case needing the pdf. The output
    ## function is a pdf and in this case the arguments of the output
    ## function are those of 'pdf'.
    if (!is.cdf)
    {
        argv <- formals(pdf)                # arguments as list
        argn <- setdiff(names(argv), "log") # drop argument 'log'
        argsFUN <- argv[argn]           # arguments of output function
        x <- as.name(argn[1])           # symbol
        e <- c(e,
               quote(f <- Call),
               substitute(f[[1L]] <- as.name(fun),
                          list(fun = as.character(Call$pdf))))
    }

    ## Build the value at which the underlying pdf/cdf will be called
    ## for non special case values of 'x'. We need to index 'x' to
    ## only compute for the correct values of a given branch.
    x.mod <- as.call(c(as.name("["), x, as.name("w")))
    if (coinsurance < 1)
        x.mod <- substitute(x/alpha, list(x = x.mod, alpha = coinsurance))
    if (deductible & !franchise)
        x.mod <- substitute(x + d, list(x = x.mod, d = deductible))
    if (inflation)
        x.mod <- substitute((x)/r, list(x = x.mod, r = r))

    ## Each pdf/cdf is defined in three branches. Define the
    ## boundaries and conditions for the first two branches. Those for
    ## the third branch are defined further down.
    if (franchise)
    {
        bound1 <- coinsurance * deductible
        bound2 <- coinsurance * limit
        cond1 <- if (is.cdf)
            substitute(0 <= x & x <= b1, list(x = x, b1 = bound1))
        else
            quote(x == 0)
        cond2 <- substitute(b1 < x & x < b2,
                            list(x = x, b1 = bound1, b2 = bound2))
    }
    else
    {
        bound1 <- 0
        bound2 <- coinsurance * (limit - deductible)
        cond1 <- substitute(x == 0, list(x = x))
        cond2 <- substitute(0 < x & x < b, list(x = x, b = bound2))
    }

    ## Initialization of the results vector in the output function
    ## with 0s.
    e <- c(e,
           substitute(res <- numeric(length(x)), list(x = x)))

    ## Definition of the output function for the first branch. There
    ## is a computation to make only if there is a deductible with the
    ## payment per loss random variable. For all other cases, the
    ## value in the first branch is 0 and we rely on the
    ## initialization with numeric() done at the previous step.
    if (per.loss & deductible)
        e <- c(e,
               substitute(res[which(cond1)] <- eval.parent(Fd),
                          list(cond1 = cond1)))

    ## Definition of the output function for the second and third
    ## branches. The 'is.cdf = TRUE' and 'is.cdf = FALSE' cases must
    ## be treated separately.
    if (is.cdf)
    {
        cond3 <- substitute(x >= b, list(x = x, b = bound2))
        f2 <- quote(eval.parent(F))
        if (!per.loss & deductible)
            f2 <- if (has.lower)
                substitute((f - F)/S,
                           list(f = f2,
                                F = quote(eval.parent(Fd)),
                                S = quote(eval.parent(Sd))))
            else
                substitute((f - F)/S,
                           list(f = f2,
                                F = quote((p <- eval.parent(Fd))),
                                S = quote((1 - p))))
        e <- c(e,
               substitute(w <- which(cond), list(cond = cond2)),
               substitute(F[[2L]] <- x, list(x = x.mod)),
               substitute(res[w] <- f, list(f = f2)),
               if (has.limit) substitute(res[cond] <- 1, list(cond = cond3)))
    }
    else
    {
        cond3 <- substitute(x == b, list(x = x, b = bound2))
        f2 <- quote(eval.parent(f))
        if (has.limit) f3 <- Su.quote
        if (!per.loss & deductible)
        {
            if (has.limit)
            {
                f2 <- if (has.lower)
                    substitute(f/(p <- S),
                               list(f = f2, S = quote(eval.parent(Sd))))
                else
                    substitute(f/(p <- S),
                               list(f = f2, S = quote(1 - eval.parent(Fd))))
                f3 <- substitute(f/p, list(f = f3))
            }
            else
                f2 <- if (has.lower)
                    substitute(f/S,
                               list(f = f2, S = quote(eval.parent(Sd))))
                else
                    substitute(f/S,
                               list(f = f2, S = quote((1 - eval.parent(Fd)))))
        }
        if (inflation | coinsurance < 1)
            f2 <- substitute(f/k, list(f = f2, k = coinsurance * r))
        e <- c(e,
               substitute(w <- which(cond), list(cond = cond2)),
               substitute(f[[2L]] <- x, list(x = x.mod)),
               substitute(res[w] <- f, list(f = f2)),
               if (has.limit) substitute(res[cond] <- f, list(cond = cond3, f = f3)))
    }

    ## Last expression of the output function.
    e <- c(e, quote(res))

    ## Wrap up the output function.
    FUN <- function() {}
    body(FUN) <- as.call(c(as.name("{"), e)) # taken from help(body)
    formals(FUN) <- argsFUN             # set arguments
    environment(FUN) <- new.env()       # new, empty environment
    FUN
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r}zmpois functions to compute
### characteristics of the Zero Modified Poisson distribution.
###
### See Appendix B of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dzmpois <- function (x, lambda, p0, log = FALSE)
    .External(C_actuar_do_dpq, "dzmpois", x, lambda, p0, log)

pzmpois <- function(q, lambda, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pzmpois", q, lambda, p0, lower.tail, log.p)

qzmpois <- function(p, lambda, p0, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qzmpois", p, lambda, p0, lower.tail, log.p)

rzmpois <- function(n, lambda, p0)
    .External(C_actuar_do_random, "rzmpois", n, lambda, p0)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Quantiles (inverse of the ogive) for grouped data
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>
### Walter Garcia-Fontes

quantile.grouped.data <- function(x, probs = seq(0, 1, 0.25),
                                  names = TRUE, ...)
{
    ## We keep the first frequencies column only; group boundaries are
    ## in the environment of 'x'
    y <- x[, 2L]
    x <- eval(expression(cj), envir = environment(x))

    ## Inverse of the ogive
    fun <- approxfun(c(0, cumsum(y))/sum(y), x,
                     yleft = min(x), yright = max(x),
                     method = "linear", ties = "ordered")

    ## Quantiles
    res <- fun(probs)

    if (names)
    {
        dig <- max(2, getOption("digits"))
        names(res) <- formatC(paste(100 * probs, "%", sep = ""),
                              format = "fg", width = 1, digits = dig)
    }
    res
}

summary.grouped.data <- function(object, ...)
{
    ## Keep only the first frequencies column
    object <- object[1L:2L]

    res <- quantile(object)
    res <- c(res[1L:3L], mean(object), res[4L:5L])
    names(res) <- c("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.")
    class(res) <- c("summaryDefault", "table")
    res
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Exact calculation of the aggregate claim amount distribution
### function by convolution. Requires a discrete distribution for
### claim amounts.
###
### AUTHORS:  Vincent Goulet <vincent.goulet@act.ulaval.ca>
### and Louis-Philippe Pouliot

exact <- function(fx, pn, x.scale = 1)
{
    ## Some useful lengths
    m <- length(fx)                   # 1 + maximum claim amount
    n <- length(pn) - 1               # maximum number of claims
    r <- n * m - n + 1                # maximum total amount of claims

    ## Initialization of the output vector
    fs <- rep(0, r)
    fs[1] <- pn[1]                    # Pr[N = 0]

    ## Convolutions
    fxc <- 1
    for (i in 1:n)
    {
        pos <- seq_len(i * m - i + 1)
        fxc <- convolve(fx, rev(fxc), type = "open")
        fs[pos] <- fs[pos] + fxc * pn[i + 1]
    }

    FUN <- approxfun((0:(length(fs) - 1)) * x.scale, pmin(cumsum(fs), 1),
                     method = "constant", yleft = 0, yright = 1, f = 0,
                     ties = "ordered")
    class(FUN) <- c("ecdf", "stepfun", class(FUN))
    assign("fs", fs, envir = environment(FUN))
    assign("x.scale", x.scale, envir = environment(FUN))
    FUN
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {m,lev,mgf}exp functions to compute raw and
### limited moments, and the moment generating function for the
### Exponential distribution (as defined in R).
###
### See Chapter 18 of Johnson & Kotz, Continuous univariate
### distributions, volume 1, Wiley, 1970
###
### AUTHORS: Mathieu Pigeon, Christophe Dutang,
### Vincent Goulet <vincent.goulet@act.ulaval.ca>

mexp <- function(order, rate = 1)
    .External(C_actuar_do_dpq, "mexp", order, 1/rate, FALSE)

levexp <- function(limit, rate = 1, order = 1)
    .External(C_actuar_do_dpq, "levexp", limit, 1/rate, order, FALSE)

mgfexp <- function(t, rate = 1, log = FALSE)
    .External(C_actuar_do_dpq, "mgfexp", t, 1/rate, log)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Simulation of standard, non hierarchical, compound models. Uses a
### simplified version of the syntax of 'rcomphierarc' for model
### specfification.
###
### Where 'rcomphierarc' was developed for flexibility, the functions
### therein aim at execution speed. Various algorithms were tested.
### No argument validity checks.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

rcompound <- function(n, model.freq, model.sev, SIMPLIFY = TRUE)
{
    ## Validity checks.
    if (any(is.na(n)) || any(n < 0))
        stop(sprintf("invalid first argument %s", sQuote("n")))

    ## Convert model expressions into language objects.
    cl.freq <- substitute(model.freq)
    cl.sev <- substitute(model.sev)

    ## If a model expression was actually an object containing the
    ## model, we need to evaluate the object to retrieve the model.
    ## If the resulting object is an expression object, its first
    ## element is the language object we are after.
    if (is.name(cl.freq))
    {
        cl.freq <- eval.parent(cl.freq)
        if (is.expression(cl.freq))
            cl.freq <- cl.freq[[1L]]
    }
    if (is.name(cl.sev))
    {
        cl.sev <- eval.parent(cl.sev)
        if (is.expression(cl.sev))
            cl.sev <- cl.sev[[1L]]
    }

    ## If a model expression is wrapped into 'expression' (as in
    ## 'rcomphierarc'), get rid of the call.
    if (cl.freq[[1L]] == "expression")
        cl.freq <- cl.freq[[-1L]]
    if (cl.sev[[1L]] == "expression")
        cl.sev <- cl.sev[[-1L]]

    ## Initialize the output vector. We will use the fact that 'res'
    ## is filled with zeros later.
    res <- numeric(n)

    ## Add the number of variates to the 'model.freq' call.
    cl.freq$n <- n

    ## Generate frequencies.
    N <- eval.parent(cl.freq)

    ## Add the number of variates to the 'model.sev' call.
    cl.sev$n <- sum(N)

    ## Generate all severities.
    x <- eval.parent(cl.sev)

    ## Create a vector that will be used as a factor to regroup
    ## severities for the computation of aggregate values. Idea:
    ## assign one integer to each frequency and repeat that integer a
    ## number of times equal to the frequency. For example, if the
    ## frequencies are (2, 0, 1, 3), then the vector will be (1, 1, 3,
    ## 4, 4, 4).
    f <- rep.int(seq_len(n), N)

    ## Compute aggregate values and put them in the appropriate
    ## positions in the output vector. The positions corresponding to
    ## zero frequencies are already initialized with zeros.
    res[which(N != 0)] <- tapply(x, f, sum)

    if (SIMPLIFY)
        res
    else
        list(aggregate = res,
             frequency = N,
             severity = x)
}

rcomppois <- function(n, lambda, model.sev, SIMPLIFY = TRUE)
{
    ## Validity checks.
    if (any(is.na(n)) || any(n < 0))
        stop(sprintf("invalid first argument %s", sQuote("n")))
    if (any(lambda < 0))
        stop(sprintf("invalid values in %s", sQuote("lambda")))

    ## Convert model expression into language object.
    cl.sev <- substitute(model.sev)

    ## If the model expression was actually an object containing the
    ## model, we need to evaluate the object to retrieve the model.
    ## If the resulting object is an expression object, its first
    ## element is the language object we are after.
    if (is.name(cl.sev))
    {
        cl.sev <- eval.parent(cl.sev)
        if (is.expression(cl.sev))
            cl.sev <- cl.sev[[1L]]
    }

    ## Get rid of the eventual 'expression' call in the language
    ## object.
    if (cl.sev[[1L]] == "expression")
        cl.sev <- cl.sev[[-1L]]

    ## Initialize the output vector.
    res <- numeric(n)

    ## Generate frequencies from Poisson distribution.
    N <- rpois(n, lambda)

    ## Add the number of variates to the 'model.sev' call.
    cl.sev$n <- sum(N)

    ## Generate all severities.
    x <- eval.parent(cl.sev)

    ## Create a vector that will be used as a factor to regroup
    ## severities for the computation of aggregate values. (See
    ## comments in 'rcompound' for details.)
    f <- rep.int(seq_len(n), N)

    ## Compute aggregate values and put them in the appropriate
    ## positions in the output vector.
    res[which(N != 0)] <- tapply(x, f, sum)

    if (SIMPLIFY)
        res
    else
        list(aggregate = res,
             frequency = N,
             severity = x)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Simulation of a aggregate claim amounts
###
### AUTHORS:  Vincent Goulet <vincent.goulet@act.ulaval.ca>
### and Louis-Philippe Pouliot

simS <- function(n, model.freq, model.sev)
{
    ## Prepare the call to simul() by building up 'nodes'
    level.names <- names(if (is.null(model.freq)) model.sev else model.freq)
    nlevels <- length(level.names)
    nodes <- as.list(c(rep(1, nlevels - 1), n))
    names(nodes) <- level.names

    ## Get sample
    x <- aggregate(simul(nodes = nodes,
                         model.freq = model.freq,
                         model.sev = model.sev))[-1]

    ## Compute the empirical cdf of the sample. Done manually instead
    ## of calling stats::ecdf to keep a copy of the empirical pmf in
    ## the environment without computing it twice.
    x <- sort(x)
    vals <- unique(x)
    fs <- tabulate(match(x, vals))/length(x)
    FUN <- approxfun(vals, pmin(cumsum(fs), 1), method = "constant",
                     yleft = 0, yright = 1, f = 0, ties = "ordered")
    class(FUN) <- c("ecdf", "stepfun", class(FUN))
    assign("fs", fs, envir = environment(FUN))
    FUN
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Auxiliary function to fit regression credibility model by
### positioning the intercept at the barycenter of time.
###
### AUTHORS: Xavier Milhaud, Vincent Goulet <vincent.goulet@act.ulaval.ca>

hache.barycenter <- function(ratios, weights, xreg, method,
                             tol, maxit, echo)
{
    ## Frequently used values
    weights.s <- rowSums(weights, na.rm = TRUE) # contract total weights
    has.data <- which(weights.s > 0)	# contracts with data
    ncontracts <- nrow(ratios)   	# number of contracts
    eff.ncontracts <- length(has.data)  # effective number of contracts
    p <- ncol(xreg)               	# rank (>= 2) of design matrix
    n <- nrow(xreg)                     # number of observations

    ## Putting the intercept at the barycenter of time amounts to use
    ## a "weighted orthogonal" design matrix in the regression (that
    ## is, X'WX = I for some diagonal weight matrix W). In theory,
    ## there would be one orthogonal design matrix per contract. In
    ## practice, we orthogonalize with a "collective" barycenter. We
    ## use average weights per period across contracts since these
    ## will be closest to the their individual analogues.
    ##
    ## We orthogonalize the original design matrix using QR
    ## decomposition. We also keep matrix R as a transition matrix
    ## between the original base and the orthogonal base.
    w <- colSums(weights, na.rm = TRUE)/sum(weights.s)
    Xqr <- qr(xreg * sqrt(w))   # QR decomposition
    R <- qr.R(Xqr)              # transition matrix
    x <- qr.Q(Xqr) / sqrt(w)    # weighted orthogonal matrix

    ## Fit linear model to each contract. For contracts without data,
    ## fit some sort of empty model to ease use in predict.hache().
    f <- function(i)
    {
    	z <-
            if (i %in% has.data)            # contract with data
            {
                y <- ratios[i, ]
                not.na <- !is.na(y)
                lm.wfit(x[not.na, , drop = FALSE], y[not.na], weights[i, not.na])
            }
            else                            # contract without data
                lm.fit(x, rep.int(0, n))
        z[c("coefficients", "residuals", "weights", "rank", "qr")]
    }
    fits <- lapply(seq_len(ncontracts), f)

    ## Individual regression coefficients
    ind <- sapply(fits, coef)
    ind[is.na(ind)] <- 0

    ## Individual variance estimators. The contribution of contracts
    ## without data is 0.
    S <- function(z)                    # from stats::summary.lm
    {
        nQr <- NROW(z$qr$qr)
        rank <- z$rank
        r <- z$residuals
        w <- z$weights
        sum(w * r^2) / (nQr - rank)
    }
    sigma2 <- sapply(fits[has.data], S)
    sigma2[is.nan(sigma2)] <- 0

    ## Initialization of a few containers: p x p x ncontracts arrays
    ## for the weight and credibility matrices; p x p matrices for the
    ## between variance-covariance matrix and total weight matrix; a
    ## vector of length p for the collective regression coefficients.
    cred <- W <- array(0, c(p, p, ncontracts))
    A <- W.s <- matrix(0, p, p)
    coll <- numeric(p)

    ## Weight matrices: we need here only the diagonal elements of
    ## X'WX, where W = diag(w_{ij}) (and not w_{ij}/w_{i.} as in the
    ## orthogonalization to keep a w_{i.} lying around). The first
    ## element is w_{i.} and the off-diagonal elements are zero by
    ## construction. Note that array W is quite different from the one
    ## in hache.origin().
    W[1, 1, ] <- weights.s
    for (i in 2:p)
        W[i, i, has.data] <-
            colSums(t(weights[has.data, ]) * x[, i]^2, na.rm = TRUE)

    ## === ESTIMATION OF THE WITHIN VARIANCE ===
    s2 <- mean(sigma2)

    ## === ESTIMATION OF THE BETWEEN VARIANCE-COVARIANCE MATRIX ===
    ##
    ## By construction, we only estimate the diagonal of the matrix.
    ## Variance components are estimated just like in the
    ## Buhlmann-Straub model (see bstraub.R for details).
    ##
    ## Should we compute the iterative estimators?
    do.iter <- method == "iterative" && diff(range(weights, na.rm = TRUE)) > .Machine$double.eps^0.5

    ## Do the computations one regression parameter at a time.
    for (i in seq_len(p))
    {
        ## Unbiased estimator
        a <- A[i, i] <- bvar.unbiased(ind[i, has.data], W[i, i, has.data],
                                      s2, eff.ncontracts)

        ## Iterative estimator
        if (do.iter)
        {
            a <- A[i, i] <-
                if (a > 0)
                    bvar.iterative(ind[i, has.data], W[i, i, has.data],
                                   s2, eff.ncontracts, start = a,
                                   tol = tol, maxit = maxit, echo = echo)
                else
                    0
        }

        ## Credibility factors and estimator of the collective
        ## regression coefficients.
        if (a > 0)
        {
            z <- cred[i, i, has.data] <- 1/(1 + s2/(W[i, i, has.data] * a))
            z. <- W.s[i, i] <- sum(z)
            coll[i] <- drop(crossprod(z, ind[i, has.data])) / z.
        }
        else
        {
            ## (credibility factors were already initialized to 0)
            w <- W[i, i, has.data]
            w. <- W.s[i, i] <- sum(w)
            coll[i] <- drop(crossprod(w, ind[i, ])) / w.
        }
    }

    ## Credibility adjusted coefficients. The coefficients of the
    ## models are replaced with these values. That way, prediction
    ## will be trivial using predict.lm().
    for (i in seq_len(ncontracts))
        fits[[i]]$coefficients <- coll + drop(cred[, , i] %*% (ind[, i] - coll))

    ## Add names to the collective coefficients vector.
    names(coll) <- rownames(ind)

    ## Results
    list(means = list(coll, ind),
         weights = list(W.s, W),
         unbiased = if (method == "unbiased") list(A, s2),
         iterative = if (method == "iterative") list(A, s2),
         cred = cred,
         nodes = list(ncontracts),
         adj.models = fits,
         transition = R)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Panjer recursion formula to compute the approximate aggregate
### claim amount distribution of a portfolio over a period.
###
### AUTHORS:  Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Sebastien Auclair, Louis-Philippe Pouliot and Tommy Ouellet

panjer <- function(fx, dist, p0 = NULL, x.scale = 1, ...,
                   convolve = 0, tol = sqrt(.Machine$double.eps),
                   maxit = 500, echo = FALSE)
{
    ## Express 'tol' as a value close to 1. If needed, modify the
    ## accuracy level so that the user specified level is attained
    ## *after* the additional convolutions (without getting too high).
    tol <- if (convolve > 0)
               min((0.5 - tol + 0.5)^(0.5 ^ convolve),
                   0.5 - sqrt(.Machine$double.eps) + 0.5)
           else
               0.5 - tol + 0.5

    ## Check if p0 is a valid probability.
    if (!is.null(p0))
    {
        if (length(p0) > 1L)
        {
            p0 <- p0[1L]
            warning(sprintf("%s has many elements: only the first used",
                            sQuote("p0")))
        }
        if ((p0 < 0) || (p0 > 1))
            stop(sprintf("%s must be a valid probability (between 0 and 1)",
                         sQuote("p0")))
    }

    ## Treat trivial case where 'p0 == 1' and hence F_S(0) = 1.
    if (identical(p0, 1))
    {
        FUN <- approxfun(0, 1, method = "constant",
                         yleft = 0, yright = 1, f = 0)
        class(FUN) <- c("ecdf", "stepfun", class(FUN))
        assign("fs", 1, envir = environment(FUN))
        assign("x.scale", x.scale, envir = environment(FUN))
        return(FUN)
    }

    ## The call to .External below requires 'p1' to be initialized.
    p1 <- 0

    ## Argument '...' should contain the values of the parameters of
    ## 'dist'.
    par <- list(...)

    ## Distributions are expressed as a member of the (a, b, 0) or (a,
    ## b, 1) families of distributions. Assign parameters 'a' and 'b'
    ## depending of the chosen distribution and compute f_S(0) in
    ## every case, and p1 if p0 is specified in argument.
    ##
    ## At this point, either p0 is NULL or 0 <= p0 < 1.
    if (startsWith(dist, "zero-truncated"))
    {
        if (!(is.null(p0) || identical(p0, 0)))
            warning(sprintf("value of %s ignored with a zero-truncated distribution",
                            sQuote("p0")))
        dist <- sub("zero-truncated ", "", dist) # drop "zero truncated" prefix
        p0 <- 0
    }

    if (startsWith(dist, "zero-modified"))
        dist <- sub("zero-modified ", "", dist) # drop "zero modified" prefix

    if (dist == "geometric")
    {
        dist <- "negative binomial"
        par$size <- 1
    }

    if (dist == "poisson")
    {
        if (!"lambda" %in% names(par))
            stop(sprintf("value of %s missing", sQuote("lambda")))
        lambda <- par$lambda
        a <- 0
        b <- lambda
        if (is.null(p0)) # standard Poisson
            fs0 <- exp(lambda * (fx[1L] - 1))
        else  # 0 <= p0 < 1; zero-truncated/modified Poisson
        {
            fs0 <- p0 + (1 - p0) * pgfztpois(fx[1L], lambda)
            p1 <- (1 - p0) * dztpois(1, lambda)
        }
    }
    else if (dist == "negative binomial")
    {
        if (!all(c("prob", "size") %in% names(par)))
            stop(sprintf("value of %s or %s missing",
                         sQuote("prob"), sQuote("size")))
        r <- par$size
        p <- par$prob
        a <- 1 - p
        b <- (r - 1) * a
        if (is.null(p0)) # standard negative binomial
            fs0 <- exp(-r * log1p(-a/p * (fx[1L] - 1)))
        else  # 0 <= p0 < 1; zero-truncated/modified neg. binomial
        {
            fs0 <- p0 + (1 - p0) * pgfztnbinom(fx[1L], r, p)
            p1 <- (1 - p0) * dztnbinom(1, r, p)
        }
    }
    else if (dist == "binomial")
    {
        if (!all(c("prob", "size") %in% names(par)))
            stop(sprintf("value of %s or %s missing",
                         sQuote("prob"), sQuote("size")))
        n <- par$size
        p <- par$prob
        a <- p/(p - 1)                  # equivalent to -p/(1 - p)
        b <- -(n + 1) * a
        if (is.null(p0)) # standard binomial
            fs0 <- exp(n * log1p(p * (fx[1L] - 1)))
        else  # 0 <= p0 < 1; zero-truncated/modified binomial
        {
            fs0 <- p0 + (1 - p0) * pgfztbinom(fx[1L], n, p)
            p1 <- (1 - p0) * dztbinom(1, n, p)
        }
    }
    else if (dist == "logarithmic")
    {
        if (!"prob" %in% names(par))
            stop(sprintf("value of %s missing", sQuote("prob")))
        a <- par$prob
        b <- -a
        if (is.null(p0) || identical(p0, 0)) # standard logarithmic
            fs0 <- pgflogarithmic(fx[1L], a)
        else # 0 < p0 < 1; zero-modified logarithmic
        {
            fs0 <- p0 + (1 - p0) * pgflogarithmic(fx[1L], a)
            p1 <- (1 - p0) * dlogarithmic(1, a)
        }
    }
    else
        stop("frequency distribution not in the (a, b, 0) or (a, b, 1) families")

    ## If fs0 is equal to zero, the recursion will not start. There is
    ## no provision to automatically cope with this situation in the
    ## current version of this function. Just issue an error message
    ## and let the user do the work by hand.
    if (identical(fs0, 0))
        stop("Pr[S = 0] is numerically equal to 0; impossible to start the recursion")

    ## Recursive calculations in C.
    fs <- .External(C_actuar_do_panjer, p0, p1, fs0, fx, a, b, convolve, tol, maxit, echo)

    FUN <- approxfun((0:(length(fs) - 1)) * x.scale, pmin(cumsum(fs), 1),
                     method = "constant", yleft = 0, yright = 1, f = 0,
                     ties = "ordered")
    class(FUN) <- c("ecdf", "stepfun", class(FUN))
    assign("fs", fs, envir = environment(FUN))
    assign("x.scale", x.scale, envir = environment(FUN))
    FUN
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,r}logarithmic functions to compute
### characteristics of the logarithmic distribution. The version used
### in these functions has probability mass function
###
###   Pr[X = x] = -p^x/(x log(1 - p)), x = 1, 2, ...
###
### This is the standard parametrization in the literature; see for
### example https://en.wikipedia.org/wiki/Logarithmic_distribution.
###
### NOTE: Klugman, Panjer & Willmot (Loss Models) introduce the
### logarithmic distribution as a limiting case of the zero truncated
### negative binomial. In this setting, parameter 'p' above would be
### the probability of *failure* (a.k.a. q) of the zero truncated
### negative binomial.
###
### AUTHOR: Vincent Goulet <vincent.goulet@act.ulaval.ca>

dlogarithmic <- function(x, prob, log = FALSE)
    .External(C_actuar_do_dpq, "dlogarithmic", x, prob, log)

plogarithmic <- function(q, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "plogarithmic", q, prob, lower.tail, log.p)

qlogarithmic <- function(p, prob, lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qlogarithmic", p, prob, lower.tail, log.p)

rlogarithmic <- function(n, prob)
    .External(C_actuar_do_random, "rlogarithmic", n, prob)

## not exported; for internal use in panjer()
pgflogarithmic <- function(x, prob)
    log1p(-prob * x)/log1p(-prob)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Normal and Normal Power Approximation of the total amount of
### claims distribution
###
### See Dayken, Pentikanen and Pesonen, Practical Risk Theory for
### Actuaries, Chapman & Hall, 1994.
###
### AUTHORS:  Vincent Goulet <vincent.goulet@act.ulaval.ca>
### and Louis-Philippe Pouliot

normal <- function(mean, variance)
{
    ## Approximate the total amount of claims distribution using the first
    ## two moments.
    FUN <- function(x) pnorm(x, mean = mean, sd = sqrt(variance))

    environment(FUN) <- new.env()
    assign("mean", mean, envir = environment(FUN))
    assign("variance", variance, envir = environment(FUN))
    attr(FUN, "source") <- "function(x) pnorm(x, mean = mean, sd = sqrt(variance))"
    FUN
}

npower <- function(mean, variance, skewness)
{
    ## Approximate the total amount of claims distribution using the first
    ## three moments.
    FUN <- function(x)
        ifelse(x <= mean, NA,
               pnorm(sqrt(1 + 9/skewness^2 + 6 * (x - mean)/(sqrt(variance) *
                                                             skewness)) -
                     3/skewness))

    environment(FUN) <- new.env()
    assign("mean", mean, envir = environment(FUN))
    assign("variance", variance, envir = environment(FUN))
    assign("skewness", skewness, envir = environment(FUN))
    attr(FUN, "source") <- "function(x) ifelse(x <= mean, NA, pnorm(sqrt(1 + 9/skewness^2 + 6 * (x - mean)/(sqrt(variance) * skewness)) - 3/skewness))"
    FUN
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Definition of the {d,p,q,r,m,lev}paralogis functions to compute
### characteristics of the paralogistic distribution. The version used
### in these functions has cumulative distribution function
###
###   Pr[X <= x] = 1 - (1/(1 + (x/scale)^shape))^shape, x > 0.
###
### See Appendix A of Klugman, Panjer & Willmot, Loss Models, Wiley.
###
### AUTHORS:  Mathieu Pigeon, Vincent Goulet <vincent.goulet@act.ulaval.ca>

dparalogis <- function (x, shape, rate = 1, scale = 1/rate, log = FALSE)
    .External(C_actuar_do_dpq, "dparalogis", x, shape, scale, log)

pparalogis <- function(q, shape, rate = 1, scale = 1/rate,
                       lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "pparalogis", q, shape, scale, lower.tail, log.p)

qparalogis <- function(p, shape, rate = 1, scale = 1/rate,
                       lower.tail = TRUE, log.p = FALSE)
    .External(C_actuar_do_dpq, "qparalogis", p, shape, scale, lower.tail, log.p)

rparalogis <- function(n, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_random, "rparalogis", n, shape, scale)

mparalogis <- function(order, shape, rate = 1, scale = 1/rate)
    .External(C_actuar_do_dpq, "mparalogis", order, shape, scale, FALSE)

levparalogis <- function(limit, shape, rate = 1, scale = 1/rate,
                         order = 1)
    .External(C_actuar_do_dpq, "levparalogis", limit, shape, scale, order, FALSE)

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Sample empirical limited value functions for individual and
### grouped data.
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca> and
###          Mathieu Pigeon

elev <- function(x, ...) UseMethod("elev")

elev.default <- function(x, ...)
{
    chkDots(...)                        # method does not use '...'
    Call <- match.call()
    if (exists(".Generic", inherits = FALSE))
        Call[[1]] <- as.name(.Generic)
    FUN <- function(limit)
        sapply(limit, function(x, y) mean(pmin(x, y)), x = x)
    environment(FUN) <- new.env()
    assign("x", sort(x), envir = environment(FUN))
    assign("n", length(unique(x)), envir = environment(FUN))
    class(FUN) <- c("elev", class(FUN))
    attr(FUN, "call") <- Call
    attr(FUN, "grouped") <- FALSE
    FUN
}

### Function 'elev.grouped.data' below returns a function that uses
### data stored in its environment. Avoid false positive in R CMD
### check.
if (getRversion() >= "2.15.1")  utils::globalVariables(c("cj", "nj"))

### This function assumes right-closed intervals, but the numerical
### values are identical for left-closed intervals.
elev.grouped.data <- function(x, ...)
{
    chkDots(...)                        # method does not use '...'
    Call <- match.call()
    if (exists(".Generic", inherits = FALSE))
        Call[[1]] <- as.name(.Generic)
    FUN <- function(limit)
    {
        ## Explicitely get the data from the function environment.
        ## cj <- eval(expression(cj))
        ## nj <- eval(expression(nj))

        ## Number of classes.
        r <- length(nj)

        ## This is to avoid numerical problems.
        limit <-  pmin(limit, cj[r + 1L])

        ## Class in which the limit is located.
        cl <- findInterval(limit, cj, all.inside = TRUE)

        ## Means for all classes below each limit.
        cjt <- head(cj, max(cl))        # upper bounds
        res1 <- sapply(cl - 1L, function(n, x)
                       drop(crossprod(head(x, n), head(nj, n))),
                       (head(cjt, -1) + tail(cjt, -1))/2)

        ## Means for classes with each limit.
        cjt <- cj[cl]                   # lower bounds
        njt <- nj[cl]                   # frequencies
        p <- (limit - cjt) / (cj[cl + 1L] - cjt) # prop. to take
        res2 <- njt * p * (cjt + limit)/2 + njt * (1 - p) * limit

        ## Means for classes above each limit.
        res3 <- limit * sapply(r - cl, function(n, x) sum(tail(x, n)),
                               tail(nj, -min(cl)))

        ## Total
        (res1 + res2 + res3)/sum(nj)
    }

    environment(FUN) <- new.env()
    assign("cj", eval(expression(cj), envir = environment(x)),
           envir = environment(FUN))
    assign("nj", x[, 2L], envir = environment(FUN))
    assign("n", nrow(x), envir = environment(FUN))
    class(FUN) <- c("elev", class(FUN))
    attr(FUN, "call") <- Call
    attr(FUN, "grouped") <- TRUE
    FUN
}

### Essentially identical to stats::print.ecdf().
print.elev <- function(x, digits = getOption("digits") - 2, ...)
{
    ## Utility function
    numform <- function(x) paste(formatC(x, digits = digits), collapse = ", ")

    ## The rest is adapted from ecdf()
    varname <- if (attr(x, "grouped")) "cj" else "x"
    cat("Empirical LEV \nCall: ")
    print(attr(x, "call"), ...)
    n <- length(xx <- eval(parse(text = varname), envir = environment(x)))
    i1 <- 1L:min(3L, n)
    i2 <- if (n >= 4L) max(4L, n - 1L):n else integer(0)
    cat(" ", varname, "[1:", n, "] = ", numform(xx[i1]), if (n > 3L) ", ",
        if (n > 5L) " ..., ",  numform(xx[i2]), "\n", sep = "")
    invisible(x)
}

### Essentially identical to stats::summary.ecdf().
summary.elev <- function (object, ...)
{
    cat("Empirical LEV:\t ", eval(expression(n), envir = environment(object)),
        "unique values with summary\n")
    summary(knots(object), ...)
}

### Essentially identical to stats::knots.stepfun().
knots.elev <- function(Fn, ...)
{
    if (attr(Fn, "grouped"))
        eval(expression(cj), envir = environment(Fn))
    else
        eval(expression(x), envir = environment(Fn))
}

plot.elev <- function(x, ..., main = NULL, xlab = "x", ylab = "Empirical LEV")
{
    if (missing(main))
        main <- {
            cl <- attr(x, "call")
            deparse(if (!is.null(cl)) cl else sys.call())
        }

    kn <- knots(x)
    Fn <- x(kn)
    plot(kn, Fn,  ..., main = main, xlab = xlab, ylab = ylab)
}

### actuar: Actuarial Functions and Heavy Tailed Distributions
###
### Extraction and replacement methods for grouped data
### objects
###
### AUTHORS: Vincent Goulet <vincent.goulet@act.ulaval.ca>,
### Mathieu Pigeon, Louis-Philippe Pouliot

"[.grouped.data" <- function(x, i, j)
{
    ## Only columns to extract are specified.
    if (nargs() < 3L)
    {
        if (missing(i))
            return(x)
        if (is.matrix(i))
            return(as.matrix(x)[i])
        res <- as.data.frame(NextMethod())
        if (length(i) > 1 && 1 %in% seq(ncol(x))[i])
        {
            environment(res) <- environment(x)
            class(res) <- c("grouped.data", class(res))
        }
        return(res)
    }

    ## Convert row and column indexes to strictly positive integers.
    ii <- if (missing(i)) seq.int(nrow(x)) else seq.int(nrow(x))[i]
    ij <- if (missing(j)) integer(0) else seq.int(ncol(x))[j]

    ## Extraction of at least the group boundaries (the complicated case).
    if (!length(ij) || 1L %in% ij)
    {
        ## Extraction of group boundaries in increasing order only
        ## (untractable otherwise).
        if (is.unsorted(ii))
        {
            warning("rows extracted in increasing order")
            ii <- sort(ii)
        }

        ## Fetch the appropriate group boundaries.
        cj <- eval(expression(cj), envir = environment(x))
        cj <- cj[sort(unique(c(ii, ii + 1L)))]

        ## Extraction of the first column only: return the vector of group
        ## boundaries.
        if (identical(ij, 1L))
            return(cj)

        ## Return a modified 'grouped.data' object.
        res <- NextMethod()
        environment(res) <- new.env()
        assign("cj", cj, environment(res))
        return(res)
    }

    ## All other cases handled like a regular data frame.
    NextMethod()
}

"[<-.grouped.data" <- function(x, i, j, value)
{
    nA <- nargs()
    if (nA == 4L)
    {
        ii <- if (missing(i)) NULL else i
        ij <- if (missing(j)) NULL else j
    }
    else if (nA == 3L)
    {
        ## No arguments inside [ ]: only replacing by NULL is supported.
        if (missing(i) && missing(j))
        {
            if (is.null(value))
                return(x[logical(0)])
            stop("impossible to replace boundaries and frequencies simultaneously")
        }
        ## Indexing by a logical matrix is supported, but only two
        ## types of replacement are allowed: replacing in the
        ## first column only, or replacing in any column but the
        ## first.
        if (is.logical(i) && is.matrix(i) && all(dim(i) == dim(x)))
        {
            ij <- apply(i, 2, any)      # columns with replacements
            if (match(TRUE, ij) == 1)   # boundaries to replace
            {
                if (length(ij) > 1)     # boundaries and frequencies
                    stop("impossible to replace boundaries and frequencies simultaneously")
                ii <- i[, ij]           # boundaries only
            }
            return(NextMethod())        # frequencies only
        }
        ## Indexing by a non logical matrix is not supported.
        if (is.matrix(i))
            stop("only logical matrix subscripts are allowed in replacement")
        ## Indexing by a vector: the argument specifies columns to
        ## replace.
        ij <- i
        ii <- NULL
    }
    else
        stop("need 0, 1, or 2 subscripts")

    ## Convert row and column indexes to integers.
    ii <- if (is.null(ii)) seq.int(nrow(x)) else seq.int(nrow(x))[ii]
    ij <- if (is.null(ij)) integer(0) else seq.int(ncol(x))[ij]

    ## Replacement at least in the group boundaries column.
    if (!length(ij) || 1L %in% ij)
    {
        ## supported: replacement of group boundaries only
        if (identical(ij, 1L))
        {
            cj <- eval(expression(cj), envir = environment(x))
            cj[sort(unique(c(ii, ii + 1L)))] <- value
            res <- grouped.data(cj, x[, -1L])
            names(res) <- names(x)
            return(res)
        }
        ## not supported (untractable): replacement in the column of
        ## boundaries and any other column
        stop("impossible to replace boundaries and frequencies simultaneously")
    }

    ## All other cases handled like a regular data frame.
    NextMethod()
}

\input{share/preamble}

  %\VignetteIndexEntry{Complete formulas used by coverage}
  %\VignettePackage{actuar}
  %\SweaveUTF8

  \title{Complete formulas used by \code{coverage}}
  \author{Vincent Goulet \\ Université Laval}
  \date{}

<<echo=FALSE>>=
library(actuar)
@

\begin{document}

\maketitle

Function \code{coverage} of \pkg{actuar} defines a new function to
compute the probability density function (pdf) of cumulative
distribution function (cdf) of any probability law under the following
insurance coverage modifications: ordinary or franchise deductible,
limit, coinsurance, inflation.

In addition, the function can return the distribution of either the
payment per loss or the payment per payment random variable. This
terminology refers to whether or not the insurer knows that a loss
occurred. For the exact definitions of the terms as used by
\code{coverage}, see Chapter~5 of \cite{LossModels2e}.

In the presence of a deductible, four random variables can be defined:
\begin{enumerate}
\item $Y^P$, the payment per payment with an ordinary deductible;
\item $Y^L$, the payment per loss with an ordinary deductible;
\item $\tilde{Y}^P$, the payment per payment with a franchise
  deductible;
\item $\tilde{Y}^L$, the payment per loss with a franchise deductible.
\end{enumerate}
The most common case in insurance applications is the distribution of
the amount paid per payment with an ordinary deductible, $Y^P$.
Hence, it is the default in \code{coverage}.

When there is no deductible, all four random variables are equivalent.

This document presents the definitions of the above four random
variables and their corresponding cdf and pdf for a deductible $d$, a
limit $u$, a coinsurance level $\alpha$ and an inflation rate $r$. An
illustrative plot of each cdf and pdf is also included. In these
plots, a dot indicates a probability mass at the given point.

In definitions below, $X$ is the nonnegative random variable of the
losses with cdf $F_X(\cdot)$ and pdf $f_X(\cdot)$.

\bibliography{actuar}


<<echo=FALSE>>=
deductible <- 5
limit <- 13
@

\section{Payment per payment, ordinary deductible}

<<echo=FALSE>>=
pgammaL <- coverage(cdf = pgamma, deductible = deductible, limit = limit,
                     per.loss = TRUE)
dgammaL <- coverage(dgamma, pgamma, deductible = deductible, limit = limit,
                     per.loss = TRUE)
pgammaP <- coverage(cdf = pgamma, deductible = deductible, limit = limit)
dgammaP <- coverage(dgamma, pgamma, deductible = deductible, limit = limit)

d <- deductible
u <- limit - d
e <- 0.001
ylim <- c(0, dgammaL(0, 5, 0.6))
@

\begin{align*}
  Y^P
  &=
  \begin{cases}
    \alpha ((1 + r) X - d),
      & \D\frac{d}{1 + r} \leq X < \frac{u}{1 + r} \\
    \alpha (u - d),
      & \D X \geq \frac{u}{1 + r}
  \end{cases} & \\
  F_{Y^P}(y)
  &=
  \begin{cases}
    0,
      & y = 0 \\
    \D\frac{F_X \left( \frac{y + \alpha d}{\alpha (1 + r)} \right) - F_X
      \left( \frac{d}{1 + r} \right)}{%
      1 - F_X \left( \frac{d}{1 + r} \right)},
      & 0 < y < \alpha (u - d) \\
    1,
      & y \geq \alpha(u - d)
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(pgammaP(x, 5, 0.6), from = 0, to = u - e,
      xlim = c(0, limit), ylim = c(0, 1),
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
curve(pgammaP(x, 5, 0.6), from = u, add = TRUE, lwd = 2)
axis(1, at = c(0, u), labels = c("0", "u - d"))
@
  \end{minipage} \\
  f_{Y^P}(y)
  &=
  \begin{cases}
    0,
      & y = 0 \\
    \left( \D\frac{1}{\alpha (1 + r)} \right)
    \D\frac{f_X \left( \frac{y + \alpha d}{\alpha(1 + r)} \right)}{%
      1 - F_X \left( \frac{d}{1 + r} \right)},
      & 0 < y < \alpha (u - d) \\
    \D\frac{1 - F_X \Big( \frac{u}{1 + r} \Big)}{%
      1 - F_X \left( \frac{d}{1 + r} \right)},
      & y = \alpha(u - d)
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(dgammaP(x, 5, 0.6), from = 0 + e, to = u - e,
      xlim = c(0, limit), ylim = ylim,
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
points(u, dgammaP(u, 5, 0.6), pch = 16)
axis(1, at = c(0, u), labels = c("0", "u - d"))
@
  \end{minipage}
\end{align*}


\section{Payment per loss, ordinary deductible}

\begin{align*}
  Y^L
  &=
  \begin{cases}
    0,
      & X < \D \frac{d}{1 + r} \\
    \alpha ((1 + r) X - d),
      & \D\frac{d}{1 + r} \leq X < \frac{u}{1 + r} \\
    \alpha (u - d),
      & \D X \geq \frac{u}{1 + r}
  \end{cases} & \\
  F_{Y^L}(y)
  &=
  \begin{cases}
    F_X \left( \D\frac{d}{1 + r} \right),
      & y = 0 \\
    F_X \left( \D\frac{y + \alpha d}{\alpha(1 + r)} \right),
      & 0 < y < \alpha (u - d) \\
    1,
      & y \geq \alpha(u - d)
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(pgammaL(x, 5, 0.6), from = 0, to = u - e,
      xlim = c(0, limit), ylim = c(0, 1),
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
curve(pgammaL(x, 5, 0.6), from = u, add = TRUE, lwd = 2)
axis(1, at = c(0, u), labels = c("0", "u - d"))
@
  \end{minipage} \\
  f_{Y^L}(y)
  &=
  \begin{cases}
    F_X \left( \D\frac{d}{1 + r} \right),
      & y = 0 \\
    \D\frac{1}{\alpha (1 + r)} f_X \left( \D\frac{y + \alpha
        d}{\alpha(1 + r)} \right),
      & 0 < y < \alpha (u - d) \\
    1 - F_X \left( \D\frac{u}{1 + r} \right),
      & y = \alpha(u - d)
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(dgammaL(x, 5, 0.6), from = 0 + e, to = u - e,
      xlim = c(0, limit), ylim = ylim,
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
points(c(0, u), dgammaL(c(0, u), 5, 0.6), pch = 16)
axis(1, at = c(0, u), labels = c("0", "u - d"))
@
  \end{minipage}
\end{align*}


\section{Payment per payment, franchise deductible}

<<echo=FALSE>>=
pgammaL <- coverage(cdf = pgamma, deductible = deductible, limit = limit,
                     per.loss = TRUE, franchise = TRUE)
dgammaL <- coverage(dgamma, pgamma, deductible = deductible, limit = limit,
                     per.loss = TRUE, franchise = TRUE)
pgammaP <- coverage(cdf = pgamma, deductible = deductible, limit = limit,
                    franchise = TRUE)
dgammaP <- coverage(dgamma, pgamma, deductible = deductible, limit = limit,
                    franchise = TRUE)

d <- deductible
u <- limit
e <- 0.001
ylim <- c(0, dgammaL(0, 5, 0.6))
@

\begin{align*}
  \tilde{Y}^P
  &=
  \begin{cases}
    \alpha (1 + r) X,
      & \D\frac{d}{1 + r} \leq X < \frac{u}{1 + r} \\
    \alpha u,
      & \D X \geq \frac{u}{1 + r}
  \end{cases} & \\
  F_{\tilde{Y}^P}(y)
  &=
  \begin{cases}
    0,
      & 0 \leq y \leq \alpha d \\
    \D\frac{F_X \left( \frac{y}{\alpha (1 + r)} \right) - F_X
      \left( \frac{d}{1 + r} \right)}{%
      1 - F_X \left( \frac{d}{1 + r} \right)},
      & \alpha d < y < \alpha u \\
    1,
      & y \geq \alpha u
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(pgammaP(x, 5, 0.6), from = 0, to = u - e,
      xlim = c(0, limit + d), ylim = c(0, 1),
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
curve(pgammaP(x, 5, 0.6), from = u, add = TRUE, lwd = 2)
axis(1, at = c(0, d, u), labels = c("0", "d", "u"))
@
  \end{minipage} \\
  f_{\tilde{Y}^P}(y)
  &=
  \begin{cases}
    0,
      & 0 \leq y \leq \alpha d \\
    \left( \D\frac{1}{\alpha (1 + r)} \right)
    \D\frac{f_X \left( \frac{y}{\alpha(1 + r)} \right)}{%
      1 - F_X \left( \frac{d}{1 + r} \right)},
      & \alpha d < y < \alpha u \\
    \D\frac{1 - F_X \Big( \frac{u}{1 + r} \Big)}{%
      1 - F_X \left( \frac{d}{1 + r} \right)},
      & y = \alpha u
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(dgammaP(x, 5, 0.6), from = d + e, to = u - e,
      xlim = c(0, limit + d), ylim = ylim,
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
curve(dgammaL(x, 5, 0.6), from = 0 + e, to = d, add = TRUE, lwd = 2)
points(u, dgammaP(u, 5, 0.6), pch = 16)
axis(1, at = c(0, d, u), labels = c("0", "d", "u"))
@
  \end{minipage}
\end{align*}


\section{Payment per loss, franchise deductible}

\begin{align*}
  \tilde{Y}^L
  &=
  \begin{cases}
    0,
      & X < \D \frac{d}{1 + r} \\
    \alpha (1 + r) X,
      & \D\frac{d}{1 + r} \leq X < \frac{u}{1 + r} \\
    \alpha u,
      & \D X \geq \frac{u}{1 + r}
  \end{cases} & \\
  F_{\tilde{Y}^L}(y)
  &=
  \begin{cases}
    F_X \left( \D\frac{d}{1 + r} \right),
      & 0 \leq y \leq \alpha d \\
    F_X \left( \D\frac{y}{\alpha(1 + r)} \right),
      & \alpha d < y < \alpha u \\
    1,
      & y \geq \alpha u
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(pgammaL(x, 5, 0.6), from = 0, to = u - e,
      xlim = c(0, limit + d), ylim = c(0, 1),
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
curve(pgammaL(x, 5, 0.6), from = u, add = TRUE, lwd = 2)
axis(1, at = c(0, d, u), labels = c("0", "d", "u"))
@
  \end{minipage} \\
  f_{\tilde{Y}^L}(y)
  &=
  \begin{cases}
    F_X \left( \D\frac{d}{1 + r} \right),
      & y = 0 \\
    \D\frac{1}{\alpha (1 + r)} f_X \left( \D\frac{y}{\alpha(1 + r)} \right),
      & \alpha d < y < \alpha u \\
    1 - F_X \left( \D\frac{u}{1 + r} \right),
      & y = \alpha u
  \end{cases} &
  \begin{minipage}{0.4\linewidth}
<<echo=FALSE, fig=TRUE, width=4, height=3>>=
par(mar = c(2, 3, 1, 1))
curve(dgammaL(x, 5, 0.6), from = d + e, to = u - e,
      xlim = c(0, limit + d), ylim = ylim,
      xlab = "", ylab = "", xaxt = "n", lwd = 2)
curve(dgammaL(x, 5, 0.6), from = 0 + e, to = d, add = TRUE, lwd = 2)
points(c(0, u), dgammaL(c(0, u), 5, 0.6), pch = 16)
axis(1, at = c(0, d, u), labels = c("0", "d", "u"))
@
  \end{minipage}
\end{align*}

\end{document}

%%% Local Variables:
%%% mode: noweb
%%% TeX-master: t
%%% coding: utf-8
%%% End:

\input{share/preamble}

  %\VignetteIndexEntry{Credibility theory}
  %\VignettePackage{actuar}
  %\SweaveUTF8

  \title{Credibility theory features of \pkg{actuar}}
  \author{Christophe Dutang \\ Université Paris Dauphine \\[3ex]
    Vincent Goulet \\ Université Laval \\[3ex]
    Xavier Milhaud \\ Université Claude Bernard Lyon 1 \\[3ex]
    Mathieu Pigeon \\ Université du Québec à Montréal}
  \date{}

<<echo=FALSE>>=
library(actuar)
options(width = 57, digits = 4, deparse.cutoff = 30L)
@

\begin{document}

\maketitle

\section{Introduction}
\label{sec:introduction}

Credibility models are actuarial tools to distribute premiums fairly
among a heterogeneous group of policyholders (henceforth called
\emph{entities}). More generally, they can be seen as prediction methods
applicable in any setting where repeated measures are made for
subjects with different risk levels.

The credibility theory features of \pkg{actuar} consist of matrix
\code{hachemeister} containing the famous data set of
\cite{Hachemeister_75} and function \code{cm} to fit hierarchical
(including Bühlmann, Bühlmann-Straub), regression and linear Bayes
credibility models. Furthermore, function \code{rcomphierarc} can
simulate portfolios of data satisfying the assumptions of the
aforementioned credibility models; see the \code{"simulation"}
vignette for details.


\section{Hachemeister data set}
\label{sec:hachemeister}

The data set of \cite{Hachemeister_75} consists of private passenger
bodily injury insurance average claim amounts, and the corresponding
number of claims, for five U.S.\ states over 12 quarters between July
1970 and June 1973. The data set is included in the package in the
form of a matrix with 5 rows and 25 columns. The first column contains
a state index, columns 2--13 contain the claim averages and columns
14--25 contain the claim numbers:
<<echo=TRUE>>=
data(hachemeister)
hachemeister
@

\section{Hierarchical credibility model}
\label{sec:hierarchical}

The linear model fitting function of R is \code{lm}. Since credibility
models are very close in many respects to linear models, and since the
credibility model fitting function of \pkg{actuar} borrows much of its
interface from \code{lm}, we named the credibility function \code{cm}.

Function \code{cm} acts as a unified interface for all credibility
models supported by the package. Currently, these are: the
unidimensional models of \cite{Buhlmann_69} and \cite{BS_70}; the
hierarchical model of \cite{Jewell_75} (of which the first two are
special cases); the regression model of \cite{Hachemeister_75},
optionally with the intercept at the barycenter of time
\citep[Section~8.4]{Buhlmann_Gisler}; linear Bayes models. The modular
design of \code{cm} makes it easy to add new models if desired.

This section concentrates on usage of \code{cm} for hierarchical
models.

There are some variations in the formulas of the hierarchical model in
the literature. We compute the credibility premiums as given in
\cite{BJ_87} or \cite{Buhlmann_Gisler}, supporting three types of
estimators of the between variance structure parameters: the unbiased
estimators of \cite{Buhlmann_Gisler} (the default), the slightly
different version of \cite{Ohlsson} and the iterative
pseudo-estimators as found in \cite{LivreVert} or \cite{Goulet_JAP}.

Consider an insurance portfolio where \emph{entities} are classified
into \emph{cohorts}. In our terminology, this is a two-level
hierarchical classification structure. The observations are claim
amounts $S_{ijt}$, where index $i = 1, \dots, I$ identifies the
cohort, index $j = 1, \dots, J_i$ identifies the entity within the
cohort and index $t = 1, \dots, n_{ij}$ identifies the period (usually
a year). To each data point corresponds a weight --- or volume ---
$w_{ijt}$. Then, the best linear prediction for the next period
outcome of a entity based on ratios $X_{ijt} = S_{ijt}/w_{ijt}$ is
\begin{equation}
  \label{eq:hierarchical:premiums}
  \begin{split}
    \hat{\pi}_{ij} &= z_{ij} X_{ijw} + (1 - z_{ij}) \hat{\pi}_i \\
    \hat{\pi}_i    &= z_i X_{izw} + (1 - z_i) m,
  \end{split}
\end{equation}
with the credibility factors
\begin{align*}
  z_{ij}
  &= \frac{w_{ij\pt}}{w_{ij\pt} + s^2/a}, &
  w_{ij\pt}
  &= \sum_{t = 1}^{n_{ij}} w_{ijt} \\
  z_{i}
  &= \frac{z_{i\pt}}{z_{i\pt} + a/b}, &
  z_{i\pt}
  &= \sum_{j = 1}^{J_i} z_{ij}
\end{align*}
and the weighted averages
\begin{align*}
  X_{ijw} &= \sum_{t = 1}^{n_{ij}} \frac{w_{ijt}}{w_{ij\pt}}\, X_{ijt} \\
  X_{izw} &= \sum_{j = 1}^{J_i} \frac{z_{ij}}{z_{i\pt}}\, X_{ijw}.
\end{align*}

The estimator of $s^2$ is
\begin{equation}
  \label{eq:s2}
  \hat{s}^2 = \frac{1}{\sum_{i = 1}^I \sum_{j = 1}^{J_i} (n_{ij} - 1)}
  \sum_{i = 1}^I \sum_{j = 1}^{J_i} \sum_{t = 1}^{n_{ij}}
  w_{ijt} (X_{ijt} - X_{ijw})^2.
\end{equation}

The three types of estimators for the variance components $a$ and $b$
are the following. First, let
\begin{align*}
  A_i
  &= \sum_{j = 1}^{J_i} w_{ij\pt} (X_{ijw} - X_{iww})^2 -
  (J_i - 1) s^2 &
  c_i
  &= w_{i\pt\pt} - \sum_{j = 1}^{J_i} \frac{w_{ij\pt}^2}{w_{i\pt\pt}} \\
  B
  &= \sum_{i = 1}^I z_{i\pt} (X_{izw} - \bar{X}_{zzw})^2 - (I - 1) a &
  d
  &= z_{\pt\pt} - \sum_{i = 1}^I \frac{z_{i\pt}^2}{z_{\pt\pt}},
\end{align*}
with
\begin{equation}
  \label{eq:Xbzzw}
  \bar{X}_{zzw} = \sum_{i = 1}^I \frac{z_{i\pt}}{z_{\pt\pt}}\, X_{izw}.
\end{equation}
(Hence, $\E{A_i} = c_i a$ and $\E{B} = d b$.) Then, the
Bühlmann--Gisler estimators are
\begin{align}
  \label{eq:ac-BG}
  \hat{a}
  &= \frac{1}{I} \sum_{i = 1}^I \max \left( \frac{A_i}{c_i}, 0 \right) \\
  \label{eq:bc-BG}
  \hat{b}
  &= \max \left( \frac{B}{d}, 0 \right),
\end{align}
the Ohlsson estimators are
\begin{align}
  \label{eq:ac-Ohl}
  \hat{a}^\prime
  &= \frac{\sum_{i = 1}^I A_i}{\sum_{i = 1}^I c_i} \\
  \label{eq:bc-Ohl}
  \hat{b}^\prime
  &= \frac{B}{d}
\end{align}
and the iterative (pseudo-)estimators are
\begin{align}
  \label{eq:at}
  \tilde{a}
  &= \frac{1}{\sum_{i = 1}^I (J_i - 1)}
  \sum_{i = 1}^I \sum_{j = 1}^{J_i} z_{ij} (X_{ijw} - X_{izw})^2 \\
  \label{eq:bt}
  \tilde{b}
  &= \frac{1}{I - 1}
  \sum_{i = 1}^I z_i (X_{izw} - X_{zzw})^2,
\end{align}
where
\begin{equation}
  \label{eq:Xzzw}
  X_{zzw} = \sum_{i = 1}^I \frac{z_i}{z_\pt}\, X_{izw}.
\end{equation}
Note the difference between the two weighted averages \eqref{eq:Xbzzw}
and \eqref{eq:Xzzw}. See \cite{cm} for further discussion on this
topic.

Finally, the estimator of the collective mean $m$ is $\hat{m} =
X_{zzw}$.

The credibility modeling function \code{cm} assumes that data is available in
the format most practical applications would use, namely a rectangular
array (matrix or data frame) with entity observations in the rows and
with one or more classification index columns (numeric or character).
One will recognize the output format of \code{rcomphierarc} and its summary
methods.

Then, function \code{cm} works much the same as \code{lm}. It takes in
argument: a formula of the form \code{\~{} terms} describing the
hierarchical interactions in a data set; the data set containing the
variables referenced in the formula; the names of the columns
where the ratios and the weights are to be found in the data set. The
latter should contain at least two nodes in each level and more than
one period of experience for at least one entity. Missing values are
represented by \code{NA}s. There can be entities with no experience
(complete lines of \code{NA}s).

In order to give an easily reproducible example, we group states 1 and
3 of the Hachemeister data set into one cohort and states 2, 4 and 5
into another. This shows that data does not have to be sorted by
level. The fitted model below uses the iterative estimators of the
variance components.
<<echo=TRUE>>=
X <- cbind(cohort = c(1, 2, 1, 2, 2), hachemeister)
fit <- cm(~cohort + cohort:state, data = X,
          ratios = ratio.1:ratio.12,
          weights = weight.1:weight.12,
          method = "iterative")
fit
@

The function returns a fitted model object of class \code{"cm"}
containing the estimators of the structure parameters. To compute the
credibility premiums, one calls a method of \code{predict} for this
class.
<<echo=TRUE>>=
predict(fit)
@

One can also obtain a nicely formatted view of the most important
results with a call to \code{summary}.
<<echo=TRUE>>=
summary(fit)
@

The methods of \code{predict} and \code{summary} can both report for a
subset of the levels by means of an argument \code{levels}.
<<echo=TRUE>>=
summary(fit, levels = "cohort")
predict(fit, levels = "cohort")
@


\section{Bühlmann and Bühlmann--Straub models}
\label{sec:buhlmann}

As mentioned above, the Bühlmann and Bühlmann--Straub models are
simply one-level hierarchical models. In this case, the
Bühlmann--Gisler and Ohlsson estimators of the between variance
parameters are both identical to the usual \cite{BS_70} estimator
\begin{equation}
  \label{eq:a-hat}
  \hat{a} = \frac{w_{\pt\pt}}{w_{\pt\pt}^2 - \sum_{i=1}^I
    w_{i\pt}^2}
  \left(
    \sum_{i=1}^I w_{i\pt} (X_{iw} - X_{ww})^2 - (I - 1) \hat{s}^2
  \right),
\end{equation}
and the iterative estimator
\begin{equation}
  \label{eq:a-tilde}
  \tilde{a} = \frac{1}{I - 1} \sum_{i = 1}^I z_i (X_{iw} - X_{zw})^2
\end{equation}
is better known as the Bichsel--Straub estimator.

To fit the Bühlmann model using \code{cm}, one simply does not
specify any weights.
<<echo=TRUE>>=
cm(~state, hachemeister, ratios = ratio.1:ratio.12)
@

When weights are specified together with a one-level model, \code{cm}
automatically fits the Bühlmann--Straub model to the data. In the
example below, we use the Bichsel--Straub estimator for the between
variance.
<<echo=TRUE>>=
cm(~state, hachemeister, ratios = ratio.1:ratio.12,
   weights = weight.1:weight.12)
@


\section{Regression model of Hachemeister}
\label{sec:regression}

The credibility regression model of \cite{Hachemeister_75} is a generalization of
the Bühlmann--Straub model. If data shows a systematic trend, the
latter model will typically under- or over-estimate the true premium
of an entity. The idea of \citeauthor{Hachemeister_75} was to fit to
the data a regression model where the parameters are a credibility
weighted average of an entity's regression parameters and the group's
parameters.

In order to use \code{cm} to fit a credibility regression model to a
data set, one simply has to supply as additional arguments
\code{regformula} and \code{regdata}. The first one is a formula of
the form \code{\~{} terms} describing the regression model, and the
second is a data frame of regressors. That is, arguments
\code{regformula} and \code{regdata} are in every respect equivalent
to arguments \code{formula} and \code{data} of \code{lm}, with the
minor difference that \code{regformula} does not need to have a left
hand side (and is ignored if present). Below, we fit the model
\begin{displaymath}
  X_{it} = \beta_0 + \beta_1 t + \varepsilon_t, \quad
  t = 1, \dots, 12
\end{displaymath}
to the original data set of \cite{Hachemeister_75}.
<<echo=TRUE>>=
fit <- cm(~state, hachemeister, regformula = ~ time,
          regdata = data.frame(time = 1:12),
          ratios = ratio.1:ratio.12,
          weights = weight.1:weight.12)
fit
@

To compute the credibility premiums, one has to provide the ``future''
values of the regressors as in \code{predict.lm}.
<<echo=TRUE>>=
predict(fit, newdata = data.frame(time = 13))
@

It is well known that the basic regression model has a major drawback:
there is no guarantee that the credibility regression line will lie
between the collective and individual ones. This may lead to grossly
inadequate premiums, as Figure~\ref{fig:state4} shows.

\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
plot(NA, xlim = c(1, 13), ylim = c(1000, 2000), xlab = "", ylab = "")
x <- cbind(1, 1:12)
lines(1:12, x %*% fit$means$portfolio,
      col = "blue", lwd = 2)
lines(1:12, x %*% fit$means$state[, 4],
      col = "red", lwd = 2, lty = 2)
lines(1:12, x %*% coefficients(fit$adj.models[[4]]),
      col = "darkgreen", lwd = 2, lty = 3)
points(13, predict(fit, newdata = data.frame(time = 13))[4],
       pch = 8, col = "darkgreen")
legend("bottomright",
       legend = c("collective", "individual", "credibility"),
       col = c("blue", "red", "darkgreen"), lty = 1:3)
@
  \caption{Collective, individual and credibility regression lines for
    State 4 of the Hachemeister data set. The point indicates the
    credibility premium.}
  \label{fig:state4}
\end{figure}

The solution proposed by \cite{Buhlmann:regression:1997} is simply to
position the intercept not at time origin, but instead at the
barycenter of time \citep[see also][Section~8.4]{Buhlmann_Gisler}. In
mathematical terms, this essentially amounts to using an orthogonal
design matrix. By setting the argument \code{adj.intercept} to
\code{TRUE} in the call, \code{cm} will automatically fit the
credibility regression model with the intercept at the barycenter of
time. The resulting regression coefficients have little meaning, but
the predictions are sensible.
<<echo=TRUE>>=
fit2 <- cm(~state, hachemeister, regformula = ~ time,
           regdata = data.frame(time = 1:12),
           adj.intercept = TRUE,
           ratios = ratio.1:ratio.12,
           weights = weight.1:weight.12)
summary(fit2, newdata = data.frame(time = 13))
@ %
Figure~\ref{fig:state4:2} shows the beneficient effect of the
intercept adjustment on the premium of State~4.

\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
plot(NA, xlim = c(1, 13), ylim = c(1000, 2000), xlab = "", ylab = "")
x <- cbind(1, 1:12)
R <- fit2$transition
lines(1:12, x %*% solve(R, fit2$means$portfolio),
      col = "blue", lwd = 2)
lines(1:12, x %*% solve(R, fit2$means$state[, 4]),
      col = "red", lwd = 2, lty = 2)
lines(1:12, x %*% solve(R, coefficients(fit2$adj.models[[4]])),
      col = "darkgreen", lwd = 2, lty = 3)
points(13, predict(fit2, newdata = data.frame(time = 13))[4],
       pch = 8, col = "darkgreen")
legend("bottomright",
       legend = c("collective", "individual", "credibility"),
       col = c("blue", "red", "darkgreen"), lty = 1:3)
@
  \caption{Collective, individual and credibility regression lines for
    State 4 of the Hachemeister data set when the intercept is
    positioned at the barycenter of time. The point indicates the
    credibility premium.}
  \label{fig:state4:2}
\end{figure}

\section{Linear Bayes model}
\label{sec:bayes}

In the pure bayesian approach to the ratemaking problem, we assume
that the observations $X_t$, $t = 1, \dots, n$, of an entity depend on
its risk level $\theta$, and that this risk level is a realization of
an unobservable random variable $\Theta$. The best (in the mean square
sense) approximation to the unknown risk premium
$\mu(\theta) = \E{X_t|\Theta = \theta}$ based on observations
$X_1, \dots, X_n$ is the Bayesian premium
\begin{equation*}
  B_{n + 1} = \E{\mu(\Theta)|X_1, \dots, X_n}.
\end{equation*}
It is then well known \citep{Buhlmann_Gisler,LossModels4e} that for
some combinaisons of distributions, the Bayesian premium is linear and
can written as a credibility premium
\begin{equation*}
  B_{n + 1} = z \bar{X} + (1 - z) m,
\end{equation*}
where $m = \E{\mu(\Theta)}$ and $z = n/(n + K)$ for some constant
$K$.

The combinations of distributions yielding a linear Bayes premium
involve members of the univariate exponential family for the
distribution of $X|\Theta = \theta$ and their natural conjugate for
the distribution of $\Theta$:
\begin{itemize}
\item $X|\Theta = \theta \sim \text{Poisson}(\theta)$, $\Theta \sim
  \text{Gamma}(\alpha, \lambda)$;
\item $X|\Theta = \theta \sim \text{Exponential}(\theta)$, $\Theta \sim
  \text{Gamma}(\alpha, \lambda)$;
\item $X|\Theta = \theta \sim \text{Normal}(\theta, \sigma^2_2)$, $\Theta \sim
  \text{Normal}(\mu, \sigma^2_1)$;
\item $X|\Theta = \theta \sim \text{Bernoulli}(\theta)$, $\Theta \sim
  \text{Beta}(a, b)$;
\item $X|\Theta = \theta \sim \text{Geometric}(\theta)$, $\Theta \sim
  \text{Beta}(a, b)$;
\end{itemize}
and the convolutions
\begin{itemize}
\item $X|\Theta = \theta \sim \text{Gamma}(\tau, \theta)$, $\Theta \sim
  \text{Gamma}(\alpha, \lambda)$;
\item $X|\Theta = \theta \sim \text{Binomial}(\nu, \theta)$, $\Theta \sim
  \text{Beta}(a, b)$;
\item $X|\Theta = \theta \sim \text{Negative Binomial}(r, \theta)$ and
  $\Theta \sim \text{Beta}(a, b)$.
\end{itemize}
\autoref{sec:formulas} provides the complete formulas for the above
combinations of distributions.

In addition, \citet[section~2.6]{Buhlmann_Gisler} show that if
$X|\Theta = \theta \sim \text{Single Parameter Pareto}(\theta, x_0)$ and
$\Theta \sim \text{Gamma}(\alpha, \lambda)$, then the Bayesian
estimator of parameter $\theta$ --- not of the risk premium! --- is
\begin{equation*}
  \hat{\Theta} = \eta \hat{\theta}^{\text{MLE}} +
  (1 - \eta) \frac{\alpha}{\lambda},
\end{equation*}
where
\begin{equation*}
  \hat{\theta}^{\text{MLE}} =
  \frac{n}{\sum_{i = 1}^n \ln (X_i/x_0)}
\end{equation*}
is the maximum likelihood estimator of $\theta$ and
\begin{equation*}
  \eta = \frac{\sum_{i = 1}^n \ln (X_i/x_0)}{%
    \lambda + \sum_{i = 1}^n \ln (X_i/x_0)}
\end{equation*}
is a weight not restricted to $(0, 1)$. (See the
\code{"distributions"} package vignette for details on the Single
Parameter Pareto distribution.)

When argument \code{formula} is \code{"bayes"}, function \code{cm}
computes pure Bayesian premiums --- or estimator in the Pareto/Gamma
case --- for the combinations of distributions above. We identify
which by means of argument \code{likelihood} that must be one of %
\code{"poisson"}, %
\code{"exponential"}, %
\code{"gamma"}, %
\code{"normal"}, %
\code{"bernoulli"}, %
\code{"binomial"}, %
\code{"geometric"}, %
\code{"negative binomial"} or %
\code{"pareto"}. %
The parameters of the distribution of $X|\Theta = \theta$, if any, and
those of the distribution of $\Theta$ are specified using the argument
names (and default values) of \code{dgamma}, \code{dnorm},
\code{dbeta}, \code{dbinom}, \code{dnbinom} or \code{dpareto1}, as
appropriate.

Consider the case where
\begin{align*}
  X|\Theta = \theta &\sim \text{Poisson}(\theta) \\
  \Theta &\sim \text{Gamma}(\alpha, \lambda).
\end{align*}
The posterior distribution of $\Theta$ is
\begin{equation*}
  \Theta|X_1, \dots, X_n \sim \text{Gamma}
  \left(
    \alpha + \sum_{t = 1}^n X_t, \lambda + n
  \right).
\end{equation*}
Therefore, the Bayesian premium is
\begin{align*}
  B_{n + 1}
  &= \E{\mu(\Theta)|X_1, \dots, X_n} \\
  &= \E{\Theta|X_1, \dots, X_n} \\
  &= \frac{\alpha + \sum_{t = 1}^n X_t}{\lambda + n} \\
  &= \frac{n}{n + \lambda}\, \bar{X} +
    \frac{\lambda}{n + \lambda} \frac{\alpha}{\lambda} \\
  &= z \bar{X} + (1 - z) m,
\end{align*}
with $m = \E{\mu(\Theta)} = \E{\Theta} = \alpha/\lambda$ and
\begin{equation*}
  z = \frac{n}{n + K}, \quad K = \lambda.
\end{equation*}
One may easily check that if $\alpha = \lambda = 3$ and
$X_1 = 5, X_2 = 3, X_3 = 0, X_4 = 1, X_5 = 1$, then $B_6 = 1.625$. We
obtain the same result using \code{cm}.
<<echo=TRUE>>=
x <- c(5, 3, 0, 1, 1)
fit <- cm("bayes", x, likelihood = "poisson",
           shape = 3, rate = 3)
fit
predict(fit)
summary(fit)
@

\appendix

\section{Linear Bayes formulas}
\label{sec:formulas}

This appendix provides the main linear Bayes credibility results for
combinations of a likelihood function member of the univariate
exponential family with its natural conjugate. For each combination,
we provide, other than the names of the distributions of
$X|\Theta = \theta$ and $\Theta$:
\begin{itemize}
\item the posterior distribution $\Theta|X_1 = x_1, \dots, X_n = x_n$,
  always of the same type as the prior, only with updated parameters;
\item the risk premium $\mu(\theta) = \E{X|\Theta = \theta}$;
\item the collective premium $m = \E{\mu(\Theta)}$;
\item the Bayesian premium
  $B_{n+1} = \E{\mu(\Theta)|X_1, \dots, X_n}$, always equal to the
  collective premium evaluated at the parameters of the posterior
  distribution;
\item the credibility factor when the Bayesian premium is expressed as
  a credibility premium.
\end{itemize}

%% Compact Listes à puce compactes et sans puce, justement.
\begingroup
\setlist[itemize]{label={},leftmargin=0pt,align=left,nosep}

\subsection{Bernoulli/beta case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Bernoulli}(\theta)$
\item $\Theta \sim \text{Beta}(a, b)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim \text{Beta}(\tilde{a},
  \tilde{b})$
  \begin{align*}
    \tilde{a} &= a + \sum_{t = 1}^n x_t \\
    \tilde{b} &= b + n - \sum_{t = 1}^n x_t
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \theta
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{a}{a + b}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{a + \sum_{t = 1}^n X_t}{a + b + n}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + a + b}
  \end{equation*}
\end{itemize}

\subsection{Binomial/beta case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Binomial}(\nu, \theta)$
\item $\Theta \sim \text{Beta}(a, b)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim \text{Beta}(\tilde{a},
  \tilde{b})$
  \begin{align*}
    \tilde{a} &= a + \sum_{t = 1}^n x_t \\
    \tilde{b} &= b + n \nu - \sum_{t = 1}^n x_t
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \nu \theta
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{\nu a}{a + b}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{\nu (a + \sum_{t = 1}^n X_t)}{a + b + n \nu}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + (a + b)/\nu}
  \end{equation*}
\end{itemize}

\subsection{Geometric/Beta case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Geometric}(\theta)$
\item $\Theta \sim \text{Beta}(a, b)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim
  \text{Beta}(\tilde{a}, \tilde{b})$
  \begin{align*}
    \tilde{a} &= a + n \\
    \tilde{b} &= b + \sum_{t = 1}^n x_t
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \frac{1 - \theta}{\theta}
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{b}{a - 1}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{b + \sum_{t = 1}^n X_t}{a + n - 1}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + a - 1}
  \end{equation*}
\end{itemize}

\subsection{Negative binomial/Beta case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Negative binomial}(r, \theta)$
\item $\Theta \sim \text{Beta}(a, b)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim
  \text{Beta}(\tilde{a}, \tilde{b})$
  \begin{align*}
    \tilde{a} &= a + n r \\
    \tilde{b} &= b + \sum_{t = 1}^n x_t
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \frac{r (1 - \theta)}{\theta}
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{r b}{a - 1}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{r (b + \sum_{t = 1}^n X_t)}{a + n r - 1}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + (a - 1)/r}
  \end{equation*}
\end{itemize}

\subsection{Poisson/Gamma case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Poisson}(\theta)$
\item $\Theta \sim \text{Gamma}(\alpha, \lambda)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim
  \text{Gamma}(\tilde{\alpha}, \tilde{\lambda})$
  \begin{align*}
    \tilde{\alpha} &= \alpha + \sum_{t = 1}^n x_t \\
    \tilde{\lambda} &= \lambda + n
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \theta
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{\alpha}{\lambda}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{\alpha + \sum_{t = 1}^n X_t}{\lambda + n}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + \lambda}
  \end{equation*}
\end{itemize}

\subsection{Exponential/Gamma case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Exponential}(\theta)$
\item $\Theta \sim \text{Gamma}(\alpha, \lambda)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim
  \text{Gamma}(\tilde{\alpha}, \tilde{\lambda})$
  \begin{align*}
    \tilde{\alpha} &= \alpha + n \\
    \tilde{\lambda} &= \lambda + \sum_{t = 1}^n x_t
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \frac{1}{\theta}
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{\lambda}{\alpha - 1}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{\lambda + \sum_{t = 1}^n X_t}{\alpha + n - 1}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + \alpha - 1}
  \end{equation*}
\end{itemize}

\subsection{Gamma/Gamma case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Gamma}(\tau, \theta)$
\item $\Theta \sim \text{Gamma}(\alpha, \lambda)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim
  \text{Gamma}(\tilde{\alpha}, \tilde{\lambda})$
  \begin{align*}
    \tilde{\alpha} &= \alpha + n \tau \\
    \tilde{\lambda} &= \lambda + \sum_{t = 1}^n x_t
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \frac{\tau}{\theta}
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \frac{\tau \lambda}{\alpha - 1}
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} = \frac{\tau (\lambda + \sum_{t = 1}^n X_t)}{\alpha + n \tau - 1}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + (\alpha - 1)/\tau}
  \end{equation*}
\end{itemize}

\subsection{Normal/Normal case}

\begin{itemize}
\item $X|\Theta = \theta \sim \text{Normal}(\theta, \sigma_2^2)$
\item $\Theta \sim \text{Normal}(\mu, \sigma_1^2)$
\item $\Theta|X_1 = x_1, \dots, X_n = x_n \sim
  \text{Normal}(\tilde{\mu}, \tilde{\sigma}_1^2)$
  \begin{align*}
    \tilde{\mu} &=
      \frac{\sigma_1^2 \sum_{t = 1}^n x_t + \sigma_2^2 \mu}{n \sigma_1^2 +
        \sigma_2^2} \\
    \tilde{\sigma}_1^2 &= \frac{\sigma_1^2 \sigma_2^2}{n
        \sigma_1^2 + \sigma_2^2}
  \end{align*}
\item Risk premium
  \begin{equation*}
    \mu(\theta) = \theta
  \end{equation*}
\item Collective premium
  \begin{equation*}
    m = \mu
  \end{equation*}
\item Bayesian premium
  \begin{equation*}
    B_{n + 1} =
      \frac{\sigma_1^2 \sum_{t = 1}^n X_t + \sigma_2^2 \mu}{n \sigma_1^2 +
        \sigma_2^2}
  \end{equation*}
\item Credibility factor
  \begin{equation*}
    z = \frac{n}{n + \sigma_2^2/\sigma_1^2}
  \end{equation*}
\end{itemize}

\endgroup

\bibliography{actuar}

\end{document}

%%% Local Variables:
%%% mode: noweb
%%% TeX-master: t
%%% coding: utf-8
%%% End:

\input{share/preamble}

  %\VignetteIndexEntry{Introduction to actuar}
  %\VignettePackage{actuar}
  %\SweaveUTF8

  \title{Introduction to \pkg{actuar}}
  \author{Christophe Dutang \\ Université Paris Dauphine \\[3ex]
    Vincent Goulet \\ Université Laval \\[3ex]
    Mathieu Pigeon \\ Université du Québec à Montréal}
  \date{}

\begin{document}

\maketitle

\section{Introduction}
\label{sec:introduction}

\pkg{actuar} \citep{actuar} provides additional actuarial science
functionality and support for heavy tailed distributions to the R
statistical system. The project was officially launched in 2005 and is
under active development.

The current feature set of the package can be split into five main
categories: additional probability distributions; loss distributions
modeling; risk and ruin theory; simulation of compound hierarchical
models; credibility theory.

Furthermore, starting with version 3.0-0, \pkg{actuar} gives easy
access to many of its underlying C workhorses through an API.

As much as possible, the developers have tried to keep the ``user
interface'' of the various functions of the package consistent.
Moreover, the package follows the general R philosophy of working with
model objects. This means that instead of merely returning, say, a
vector of probabilities, many functions will return an object
containing, among other things, the said probabilities. The object can
then be manipulated at one's will using various extraction, summary or
plotting functions.


\section{Documentation}

In addition to the help pages, \pkg{actuar} ships with extensive
vignettes and demonstration scripts; run the following commands at the
R prompt to obtain the list of each.
<<echo=TRUE, eval=FALSE>>=
vignette(package = "actuar")
demo(package = "actuar")
@


\section{Collaboration and citation}

If you use R or \pkg{actuar} for actuarial analysis, please cite the
software in publications. For information on how to cite the software,
use:
<<eval=FALSE>>=
citation()
citation("actuar")
@


\section*{Acknowledgments}

The package would not be at this stage of development without the
stimulating contribution of Sébastien Auclair, Christophe Dutang,
Nicholas Langevin, Xavier Milhaud, Tommy Ouellet and Louis-Philippe
Pouliot.

This research benefited from financial support from the Natural
Sciences and Engineering Research Council of Canada and from the
\emph{Chaire d'actuariat} (Actuarial Science Foundation) of Université
Laval.

\bibliography{actuar}

\end{document}

%%% Local Variables:
%%% mode: noweb
%%% TeX-master: t
%%% coding: utf-8
%%% End:

\input{share/preamble}

  %\VignetteIndexEntry{Risk and ruin theory}
  %\VignettePackage{actuar}
  %\SweaveUTF8

  \title{Risk and ruin theory features of \pkg{actuar}}
  \author{Christophe Dutang \\ Université Paris Dauphine \\[3ex]
    Vincent Goulet \\ Université Laval \\[3ex]
    Mathieu Pigeon \\ Université du Québec à Montréal}
  \date{}

  %% Additional math commands
  \newcommand{\VaR}{\mathrm{VaR}}
  \newcommand{\CTE}{\mathrm{CTE}}

<<echo=FALSE>>=
library(actuar)
options(width = 52, digits = 4)
@

\begin{document}

\maketitle

\section{Introduction}
\label{sec:introduction}

Risk theory refers to a body of techniques to model and measure the
risk associated with a portfolio of insurance contracts. A first
approach consists in modeling the distribution of total claims over a
fixed period of time using the classical collective model of risk
theory. A second input of interest to the actuary is the evolution of
the surplus of the insurance company over many periods of time. In
\emph{ruin theory}, the main quantity of interest is the probability
that the surplus becomes negative, in which case technical ruin of the
insurance company occurs.

The interested reader can read more on these subjects in
\cite{LossModels4e,Gerber_MRT,DenuitCharpentier1,MART:2e}, among others.

The current version of \pkg{actuar} \citep{actuar} contains four
visible functions related to the above problems: two for the
calculation of the aggregate claim amount distribution and two for
ruin probability calculations.



\section{The collective risk model}
\label{sec:collective-risk-model}

Let  random variable $S$ represent the aggregate claim amount (or
total amount of claims) of a portfolio of independent risks over a
fixed period of time, random variable $N$ represent the number of
claims (or frequency) in the portfolio over that period, and random
variable $C_j$ represent the amount of claim $j$ (or severity). Then,
we have the random sum
\begin{equation}
  \label{eq:definition-S}
  S = C_1 + \dots + C_N,
\end{equation}
where we assume that $C_1, C_2, \dots$ are mutually independent and
identically distributed random variables each independent of $N$.
The task at hand consists in evaluating numerically the cdf of $S$,
given by
\begin{align}
  \label{eq:cdf-S}
  F_S(x)
  &= \Pr[S \leq x] \notag \\
  &= \sum_{n = 0}^\infty \Pr[S \leq x|N = n] p_n \notag \\
  &= \sum_{n = 0}^\infty  F_C^{*n}(x) p_n,
\end{align}
where $F_C(x) = \Pr[C \leq x]$ is the common cdf of $C_1, \dots, C_n$,
$p_n = \Pr[N = n]$ and $F_C^{*n}(x) = \Pr[C_1 + \dots + C_n \leq x]$
is the $n$-fold convolution of $F_C(\cdot)$. If $C$ is discrete on $0,
1, 2, \dots$, one has
\begin{equation}
  \label{eq:convolution-formula}
  F_C^{*k}(x) =
  \begin{cases}
    I\{x \geq 0\}, & k = 0 \\
    F_C(x),        & k = 1 \\
    \sum_{y = 0}^x F_C^{*(k - 1)}(x - y) f_C(y), & k = 2, 3, \dots,
  \end{cases}
\end{equation}
where $I\{\mathcal{A}\} = 1$ if $\mathcal{A}$ is true and
$I\{\mathcal{A}\} = 0$ otherwise.



\section{Discretization of claim amount distributions}
\label{sec:discretization}

Some numerical techniques to compute the aggregate claim amount
distribution (see \autoref{sec:aggregate}) require a
discrete arithmetic claim amount distribution; that is, a distribution
defined on $0, h, 2h, \dots$ for some step (or span, or lag) $h$. The
package provides function \code{discretize} to discretize a
continuous distribution. (The function can also be used to modify the
support of an already discrete distribution, but this requires
additional care.)

Let $F(x)$ denote the cdf of the distribution to discretize on some
interval $(a, b)$ and $f_x$ denote the probability mass at $x$ in the
discretized distribution. Currently, \code{discretize} supports the
following four discretization methods.
\begin{enumerate}
\item Upper discretization, or forward difference of $F(x)$:
  \begin{equation}
    \label{eq:discretization:upper}
    f_x = F(x + h) - F(x)
  \end{equation}
  for $x = a, a + h, \dots, b - h$. The discretized cdf is always
  above the true cdf.
\item Lower discretization, or backward difference of $F(x)$:
  \begin{equation}
    \label{eq:discretization:lower}
    f_x =
    \begin{cases}
      F(a),            & x = a \\
      F(x) - F(x - h), & x = a + h, \dots, b.
    \end{cases}
  \end{equation}
  The discretized cdf is always under the true cdf.
\item Rounding of the random variable, or the midpoint method:
  \begin{equation}
    \label{eq:discretization:midpoint}
    f_x =
    \begin{cases}
      F(a + h/2),              & x = a \\
      F(x + h/2) - F(x - h/2), & x = a + h, \dots, b - h.
    \end{cases}
  \end{equation}
  The true cdf passes exactly midway through the steps of the
  discretized cdf.
\item Unbiased, or local matching of the first moment method:
  \begin{equation}
    \label{eq:discretization:unbiased}
    f_x =
    \begin{cases}
      \dfrac{\E{X \wedge a} - \E{X \wedge a + h}}{h} + 1 - F(a),
      & x = a \\
      \dfrac{2 \E{X \wedge x} - \E{X \wedge x - h}
        - \E{X \wedge x + h}}{h},
      & a < x < b \\
      \dfrac{\E{X \wedge b} - \E{X \wedge b - h}}{h} - 1 + F(b), &
      x = b.
    \end{cases}
  \end{equation}
  The discretized and the true distributions have the same total
  probability and expected value on $(a, b)$.
\end{enumerate}

\autoref{fig:discretization-methods} illustrates the four methods.
It should be noted that although very close in this example, the
rounding and unbiased methods are not identical.
\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
fu <- discretize(plnorm(x), method = "upper", from = 0, to = 5)
fl <- discretize(plnorm(x), method = "lower", from = 0, to = 5)
fr <- discretize(plnorm(x), method = "rounding", from = 0, to = 5)
fb <- discretize(plnorm(x), method = "unbiased", from = 0, to = 5,
                 lev = levlnorm(x))
par(mfrow = c(2, 2), mar = c(5, 2, 4, 2))
curve(plnorm(x), from = 0, to = 5, lwd = 2, main = "Upper", ylab = "F(x)")
plot(stepfun(0:4, diffinv(fu)), pch = 20, add = TRUE)
curve(plnorm(x), from = 0, to = 5, lwd = 2, main = "Lower", ylab = "F(x)")
plot(stepfun(0:5, diffinv(fl)), pch = 20, add = TRUE)
curve(plnorm(x), from = 0, to = 5, lwd = 2, main = "Rounding", ylab = "F(x)")
plot(stepfun(0:4, diffinv(fr)), pch = 20, add = TRUE)
curve(plnorm(x), from = 0, to = 5, lwd = 2, main = "Unbiased", ylab = "F(x)")
plot(stepfun(0:5, diffinv(fb)), pch = 20, add = TRUE)
## curve(plnorm(x), from = 0, to = 5, lwd = 2, ylab = "F(x)")
## par(col = "blue")
## plot(stepfun(0:4, diffinv(fu)), pch = 19, add = TRUE)
## par(col = "red")
## plot(stepfun(0:5, diffinv(fl)), pch = 19, add = TRUE)
## par(col = "green")
## plot(stepfun(0:4, diffinv(fr)), pch = 19, add = TRUE)
## par(col = "magenta")
## plot(stepfun(0:5, diffinv(fb)), pch = 19, add = TRUE)
## legend(3, 0.3, legend = c("upper", "lower", "rounding", "unbiased"),
##        col = c("blue", "red", "green", "magenta"), lty = 1, pch = 19,
##        text.col = "black")
@
  \caption{Comparison of four discretization methods}
  \label{fig:discretization-methods}
\end{figure}

Usage of \code{discretize} is similar to R's plotting
function \code{curve}. The cdf to discretize and, for the unbiased
method only, the limited expected value function are passed to
\code{discretize} as expressions in \code{x}. The other arguments
are the upper and lower bounds of the discretization interval, the
step $h$ and the discretization method. For example, upper and
unbiased discretizations of a Gamma$(2, 1)$ distribution on $(0, 17)$
with a step of $0.5$ are achieved with, respectively,
<<echo=TRUE, eval=FALSE>>=
fx <- discretize(pgamma(x, 2, 1), method = "upper",
                 from = 0, to = 17, step = 0.5)
fx <- discretize(pgamma(x, 2, 1), method = "unbiased",
                 lev = levgamma(x, 2, 1),
                 from = 0, to = 17, step = 0.5)
@

Function \code{discretize} is written in a modular fashion making it
simple to add other discretization methods if needed.


\section{Calculation of the aggregate claim amount distribution}
\label{sec:aggregate}

Function \code{aggregateDist} serves as a unique front end for
various methods to compute or approximate the cdf of the aggregate
claim amount random variable $S$. Currently, five methods are
supported.
\begin{enumerate}
\item Recursive calculation using the algorithm of \cite{Panjer_81}.
  This requires the severity distribution to be discrete arithmetic on
  $0, 1, 2, \dots, m$ for some monetary unit and the frequency
  distribution to be a member of either the $(a, b, 0)$ or $(a, b, 1)$
  class of distributions \citep{LossModels4e}. (These classes
  contain the Poisson, binomial, negative binomial and logarithmic
  distributions and their zero-truncated and zero-modified extensions
  allowing for a zero or arbitrary mass at $x = 0$.) The general
  recursive formula is:
  \begin{displaymath}
    f_S(x) = \frac{(p_1 - (a + b)p_0)f_C(x) + \sum_{y=1}^{\min(x,
        m)}(a + by/x)f_C(y)f_S(x - y)}{1 - a f_C(0)},
  \end{displaymath}
  with starting value $f_S(0) = P_N(f_C(0))$, where $P_N(\cdot)$ is
  the probability generating function of $N$. Probabilities are
  computed until their sum is arbitrarily close to 1.

  The recursions are done in C to dramatically increase
  speed. One difficulty the programmer is facing is the unknown length
  of the output. This was solved using a common, simple and fast
  technique: first allocate an arbitrary amount of memory and double
  this amount each time the allocated space gets full.

\item Exact calculation by numerical convolutions using
  \eqref{eq:cdf-S} and \eqref{eq:convolution-formula}. This
  also requires a discrete severity distribution. However, there is no
  restriction on the shape of the frequency distribution. The package
  merely implements the sum \eqref{eq:cdf-S}, the convolutions being
  computed with R's function \code{convolve}, which in turn
  uses the Fast Fourier Transform. This approach
  is practical for small problems only, even on today's fast
  computers.
\item Normal approximation of the cdf, that is
  \begin{equation}
    \label{eq:normal-approximation}
    F_S(x) \approx \Phi
    \left(
      \frac{x - \mu_S}{\sigma_S}
    \right),
  \end{equation}
  where $\mu_S = \E{S}$ and $\sigma_S^2 = \VAR{S}$. For most realistic
  models, this approximation is rather crude in the tails of the
  distribution.
\item Normal Power II approximation of the cdf, that is
  \begin{equation}
    \label{eq:np2-approximation}
    F_S(x) \approx \Phi
    \left(
      -\frac{3}{\gamma_S} +
      \sqrt{\frac{9}{\gamma_S^2} + 1
        + \frac{6}{\gamma_S}
        \frac{x - \mu_S}{\sigma_S}} \right),
  \end{equation}
  where $\gamma_S = \E{(S - \mu_S)^3}/\sigma_S^{3/2}$. The
  approximation is valid for $x > \mu_S$ only and performs reasonably
  well when $\gamma_S < 1$. See \cite{Daykin_et_al} for details.
\item Simulation of a random sample from $S$ and approximation of
  $F_S(x)$ by the empirical cdf
  \begin{equation}
    F_n(x) = \frac{1}{n} \sum_{j = 1}^n I\{x_j \leq x\}.
  \end{equation}
  The simulation itself is done with function \code{simul} (see the
  \code{"simulation"} vignette). This function admits very general
  hierarchical models for both the frequency and the severity
  components.
\end{enumerate}

Here also, adding other methods to \code{aggregateDist} is simple due
to its modular conception.

The arguments of \code{aggregateDist} differ according to the chosen
calculation method; see the help page for details. One interesting
argument to note is \code{x.scale} to specify the monetary unit of
the severity distribution. This way, one does not have to mentally do
the conversion between the support of $0, 1, 2, \dots$ assumed by the
recursive and convolution methods, and the true support of $S$.

The recursive method fails when the expected number of claims is so
large that $f_S(0)$ is numerically equal to zero. One solution
proposed by \citet{LossModels4e} consists in dividing the appropriate
parameter of the frequency distribution by $2^n$, with $n$ such that
$f_S(0) > 0$ and the recursions can start. One then computes the
aggregate claim amount distribution using the recursive method and
then convolves the resulting distribution $n$ times with itself to
obtain the final distribution. Function \code{aggregateDist} supports
this procedure through its argument \code{convolve}.

A common problem with the recursive method is failure to obtain a
cumulative distribution function that reaching (close to) $1$. This is
usually due to too coarse a discretization of the severity
distribution. One should make sure to use a small enough
discretization step and to discretize the severity distribution far in
the right tail.

The function \code{aggregateDist} returns an object of class
\code{"aggregateDist"} inheriting from the \code{"function"} class.
Thus, one can use the object as a function to compute the value of
$F_S(x)$ in any $x$.

For illustration purposes, consider the following model: the
distribution of $S$ is a compound Poisson with parameter $\lambda =
10$ and severity distribution Gamma$(2, 1)$. To obtain an
approximation of the cdf of $S$ we first discretize the gamma
distribution on $(0, 22)$ with the unbiased method and a step of
$0.5$, and then use the recursive method in \code{aggregateDist}:
<<echo=TRUE>>=
fx <- discretize(pgamma(x, 2, 1), method = "unbiased",
                 from = 0, to = 22, step = 0.5,
                 lev = levgamma(x, 2, 1))
Fs <- aggregateDist("recursive", model.freq = "poisson",
                    model.sev = fx, lambda = 10,
                    x.scale = 0.5)
summary(Fs)
@
Although useless here, the following is essentially equivalent, except
in the far right tail for numerical reasons:
<<echo=TRUE>>=
Fsc <- aggregateDist("recursive", model.freq = "poisson",
                     model.sev = fx, lambda = 5,
                     convolve = 1, x.scale = 0.5)
summary(Fsc)
@

We return to object \code{Fs}. It contains an empirical cdf with support
<<echo=TRUE>>=
knots(Fs)
@
A nice graph of this function is obtained with a method of \code{plot} (see
\autoref{fig:Fs}):
<<echo=TRUE, eval=FALSE>>=
plot(Fs, do.points = FALSE, verticals = TRUE,
     xlim = c(0, 60))
@
\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
plot(Fs, do.points = FALSE, verticals = TRUE, xlim = c(0, 60))
@
  \caption{Graphic of the empirical cdf of $S$ obtained with the
    recursive method}
  \label{fig:Fs}
\end{figure}

The package defines a few summary methods to extract information from
\code{"aggregateDist"} objects. First, there are methods of
\code{mean} and \code{quantile} to easily compute the mean and
obtain the quantiles of the approximate distribution:
<<echo=TRUE>>=
mean(Fs)
quantile(Fs)
quantile(Fs, 0.999)
@

Second, a method of \texttt{diff} gives easy access to the underlying
probability mass function:
<<echo=TRUE>>=
diff(Fs)
@
Of course, this is defined (and makes sense) for the recursive, direct
convolution and simulation methods only.

Third, the package introduces the generic functions \code{VaR} and
\code{CTE} (with alias \code{TVaR}) with methods for objects of class
\code{"aggregateDist"}. The former computes the value-at-risk
$\VaR_\alpha$ such that
\begin{equation}
  \label{eq:VaR}
  \Pr[S \leq \VaR_\alpha] = \alpha,
\end{equation}
where $\alpha$ is the confidence level. Thus, the value-at-risk is
nothing else than a quantile. As for the method of \code{CTE}, it
computes the conditional tail expectation (also called Tail
Value-at-Risk)
\begin{equation}
  \label{eq:CTE}
  \CTE_\alpha = \E{S|S > \VaR_\alpha}.
\end{equation}
Here are examples using object \code{Fs} obtained above:
<<echo=TRUE>>=
VaR(Fs)
CTE(Fs)
@

To conclude on the subject, \autoref{fig:Fs-comparison} shows the
cdf of $S$ using five of the many combinations of discretization and
calculation method supported by \pkg{actuar}.
\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
fx.u <- discretize(pgamma(x, 2, 1), from = 0, to = 22, step = 0.5,
                   method = "upper")
Fs.u <- aggregateDist("recursive", model.freq = "poisson",
                      model.sev = fx.u, lambda = 10, x.scale = 0.5)
fx.l <- discretize(pgamma(x, 2, 1), from = 0, to = 22, step = 0.5,
                   method = "lower")
Fs.l <- aggregateDist("recursive", model.freq = "poisson",
                      model.sev = fx.l, lambda = 10, x.scale = 0.5)
Fs.n <- aggregateDist("normal", moments = c(20, 60))
Fs.s <- aggregateDist("simulation",
                      model.freq = expression(y = rpois(10)),
                      model.sev = expression(y = rgamma(2, 1)),
                      nb.simul = 10000)
par(col = "black")
plot(Fs, do.points = FALSE, verticals = TRUE, xlim = c(0, 60), sub = "")
par(col = "blue")
plot(Fs.u, do.points = FALSE, verticals = TRUE, add = TRUE, sub = "")
par(col = "red")
plot(Fs.l, do.points = FALSE, verticals = TRUE, add = TRUE, sub = "")
par(col = "green")
plot(Fs.s, do.points = FALSE, verticals = TRUE, add = TRUE, sub = "")
par(col = "magenta")
plot(Fs.n, add = TRUE, sub = "")
legend(30, 0.4, c("recursive + unbiased", "recursive + upper", "recursive + lower", "simulation", "normal approximation"),
       col = c("black", "blue", "red", "green", "magenta"),
       lty = 1, text.col = "black")
@
  \caption{Comparison between the empirical or approximate cdf of $S$
    obtained with five different methods}
  \label{fig:Fs-comparison}
\end{figure}


\section{The continuous time ruin model}
\label{sec:ruin-model}

We now turn to the multi-period ruin problem. Let $U(t)$ denote the
surplus of an insurance company at time $t$, $c(t)$ denote premiums
collected through time $t$, and $S(t)$ denote aggregate claims paid
through time $t$. If $u$ is the initial surplus at time $t = 0$, then
a mathematically convenient definition of $U(t)$ is
\begin{equation}
  \label{eq:definition-surplus}
  U(t) = u + c(t) - S(t).
\end{equation}
As mentioned previously, technical ruin of the insurance company
occurs when the surplus becomes negative. Therefore, the definition of
the infinite time probability of ruin is
\begin{equation}
  \label{eq:definition-ruin}
  \psi(u) = \Pr[U(t) < 0 \text{ for some } t \geq 0].
\end{equation}

We define some other quantities needed in the sequel. Let $N(t)$
denote the number of claims up to time $t \geq 0$ and $C_j$ denote the
amount of claim $j$. Then the definition of $S(t)$ is analogous to
\eqref{eq:definition-S}:
\begin{equation}
  \label{eq:definition-S(t)}
  S(t) = C_1 + \dots + C_{N(t)},
\end{equation}
assuming $N(0) = 0$ and $S(t) = 0$ as long as $N(t) = 0$. Furthermore,
let $T_j$ denote the time when claim $j$ occurs, such that $T_1 < T_2
< T_3 < \dots$ Then the random variable of the interarrival (or wait)
time between claim $j - 1$ and claim $j$ is defined as $W_1 = T_1$ and
\begin{equation}
  \label{eq:definition-wait}
  W_j = T_j - T_{j - 1}, \quad j \geq 2.
\end{equation}

For the rest of this discussion, we make the following assumptions:
\begin{enumerate}
\item premiums are collected at a constant rate $c$, hence $c(t) =
  ct$;
\item the sequence $\{T_j\}_{j \geq 1}$ forms an ordinary renewal
  process, with the consequence that  random variables $W_1, W_2,
  \dots$ are independent and identically distributed;
\item claim amounts $C_1, C_2, \dots$ are independent and identically
  distributed.
\end{enumerate}



\section{Adjustment coefficient}
\label{sec:adjustment-coefficient}

The quantity known as the adjustment coefficient $\rho$ hardly has any
physical interpretation, but it is useful as an approximation to
the probability of ruin since we have the inequality
\begin{displaymath}
  \psi(u) \leq e^{-\rho u}, \quad u \geq 0.
\end{displaymath}
The adjustment coefficient is defined as the smallest strictly
positive solution (if it exists) of the Lundberg equation
\begin{equation}
  \label{eq:definition-adjcoef}
  h(t) = \E{e^{t C - t c W}} = 1,
\end{equation}
where the premium rate $c$ satisfies the positive safety loading
constraint $\E{C - cW} < 0$. If $C$ and $W$ are independent, as in the
most common models, then the equation can be rewritten as
\begin{equation}
  \label{eq:definition-adjcoef-ind}
  h(t) = M_C(t) M_W(-tc) = 1.
\end{equation}

Function \code{adjCoef} of \pkg{actuar} computes the adjustment
coefficient $\rho$ from the following arguments: either the two moment
generating functions $M_C(t)$ and $M_W(t)$ (thereby assuming
independence) or else function $h(t)$; the premium rate $c$; the upper
bound of the support of $M_C(t)$ or any other upper bound for $\rho$.

For example, if $W$ and $C$ are independent and each follow an
exponential distribution, $W$ with parameter $2$ and
$C$ with parameter $1$, and the premium rate is $c = 2.4$ (for
a safety loading of 20\% using the expected value premium principle),
then the adjustment coefficient is
<<echo=TRUE>>=
adjCoef(mgf.claim = mgfexp(x), mgf.wait = mgfexp(x, 2),
        premium.rate = 2.4, upper = 1)
@

The function also supports models with proportional or excess-of-loss
reinsurance \citep{Centeno_02}. Under the first type of treaty, an
insurer pays a proportion $\alpha$ of every loss and the rest is paid
by the reinsurer. Then, for fixed $\alpha$ the adjustment coefficient
is the solution of
\begin{equation}
  \label{eq:definition-adjcoef-prop}
  h(t) = \E{e^{t \alpha C - t c(\alpha) W}} = 1.
\end{equation}
Under an excess-of-loss treaty, the primary insurer pays each claim up
to a limit $L$. Again, for fixed $L$, the adjustment coefficient is
the solution of
\begin{equation}
  \label{eq:definition-adjcoef-xl}
  h(t) = \E{e^{t \min(C, L) - t c(L) W}} = 1.
\end{equation}

For models with reinsurance, \code{adjCoef} returns an object of
class \code{"adjCoef"} inheriting from the \code{"function"} class.
One can then use the object to compute the adjustment coefficient for
any retention rate $\alpha$ or retention limit $L$. The package also
defines a method of \code{plot} for these objects.

For example, using the same assumptions as above with proportional
reinsurance and a 30\% safety loading for the reinsurer, the
adjustment coefficient as a function of $\alpha \in [0, 1]$ is (see
\autoref{fig:adjcoef} for the graph):
<<echo=TRUE, fig=FALSE>>=
mgfx <- function(x, y) mgfexp(x * y)
p <- function(x) 2.6 * x - 0.2
rho <- adjCoef(mgfx, mgfexp(x, 2), premium = p,
               upper = 1, reins = "prop",
               from = 0, to = 1)
rho(c(0.75, 0.8, 0.9, 1))
plot(rho)
@

\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
plot(rho)
@
  \caption{Adjustment coefficient as a function of the retention rate}
  \label{fig:adjcoef}
\end{figure}


\section{Probability of ruin}
\label{sec:ruin}

In this subsection, we always assume that interarrival times and claim
amounts are independent.

The main difficulty with the calculation of the infinite time
probability of ruin lies in the lack of explicit formulas except for
the most simple models. If interarrival times are
Exponential$(\lambda)$ distributed (Poisson claim number process) and
claim amounts are Exponential$(\beta)$ distributed, then
\begin{equation}
  \label{eq:ruin-cramer-lundberg}
  \psi(u) = \frac{\lambda}{c \beta}\, e^{-(\beta - \lambda/c) u}.
\end{equation}
If the frequency assumption of this model is defensible, the severity
assumption can hardly be used beyond illustration purposes.

Fortunately, phase-type distributions have come to the rescue since
the early 1990s. \cite{AsmussenRolski_91} first show that in the
classical Cramér--Lundberg model where interarrival times are
Exponential$(\lambda)$ distributed, if claim amounts are
Phase-type$(\mat{\pi}, \mat{T})$ distributed, then $\psi(u) = 1
- F(u)$, where $F$ is Phase-type$(\mat{\pi}_+, \mat{Q})$ with
\begin{equation}
  \label{eq:prob-ruin:cramer-lundberg}
  \begin{split}
    \mat{\pi}_+
    &= - \frac{\lambda}{c}\, \mat{\pi} \mat{T}^{-1} \\
    \mat{Q}
    &= \mat{T} + \mat{t} \mat{\pi}_+,
  \end{split}
\end{equation}
and $\mat{t} = -\mat{T} \mat{e}$, $\mat{e}$ is a column vector with
all components equal to 1; see the \code{"lossdist"} vignette for
details.

In the more general Sparre~Andersen model where interarrival times can
have any Phase-type$(\mat{\nu}, \mat{S})$ distribution,
\cite{AsmussenRolski_91} also show that using the same claim severity
assumption as above, one still has $\psi(u) = 1 - F(u)$ where $F$ is
Phase-type$(\mat{\pi}_+, \mat{Q})$, but with parameters
\begin{equation}
  \label{eq:prob-ruin:sparre:pi+}
  \mat{\pi}_+  = \frac{\mat{e}^\prime (\mat{Q} - \mat{T})}{%
    c \mat{e}^\prime \mat{t}}
\end{equation}
and $\mat{Q}$ solution of
\begin{equation}
  \label{eq:eq:prob-ruin:sparre:Q}
  \begin{split}
    \mat{Q}
    &= \Psi(\mat{Q}) \\
    &= \mat{T} - \mat{t} \mat{\pi}
    \left[
      (\mat{I}_n \otimes \mat{\nu})
      (\mat{Q} \oplus \mat{S})^{-1}
      (\mat{I}_n \otimes \mat{s})
    \right].
  \end{split}
\end{equation}
In the above, $\mat{s} = -\mat{S} \mat{e}$, $\mat{I}_n$ is the $n
\times n$ identity matrix, $\otimes$ denotes the usual Kronecker
product between two matrices and $\oplus$ is the Kronecker sum defined
as
\begin{equation}
  \label{eq:kronecker-sum}
  \mat{A}_{m \times m} \oplus \mat{B}_{n \times n} =
  \mat{A} \otimes \mat{I}_n + \mat{B} \otimes \mat{I}_m.
\end{equation}

Function \code{ruin} of \pkg{actuar} returns a function object of
class \code{"ruin"} to compute the probability of ruin for any initial
surplus $u$. In all cases except the exponential/exponential model
where \eqref{eq:ruin-cramer-lundberg} is used, the output object calls
function \code{pphtype} to compute the ruin probabilities.

Some thought went into the interface of \code{ruin}. Obviously, all
models can be specified using phase-type distributions, but the
authors wanted users to have easy access to the most common models
involving exponential and Erlang distributions. Hence, one first
states the claim amount and interarrival times models with any
combination of \code{"exponential"}, \code{"Erlang"} and
\code{"phase-type"}. Then, one passes the parameters of each model
using lists with components named after the corresponding parameters
of \code{dexp}, \code{dgamma} and \code{dphtype}. If a component
\code{"weights"} is found in a list, the model is a mixture of
exponential or Erlang (mixtures of phase-type are not supported).
Every component of the parameter lists is recycled as needed.

The following examples should make the matter clearer. (All examples
use $c = 1$, the default value in \code{ruin}.) First, for the
exponential/exponential model, one has
<<echo=TRUE>>=
psi <- ruin(claims = "e", par.claims = list(rate = 5),
            wait   = "e", par.wait   = list(rate = 3))
psi
psi(0:10)
@

Second, for a mixture of two exponentials claim amount model and
exponential interarrival times, the simplest call to \code{ruin} is
<<echo=FALSE>>=
op <- options(width=50)
@
<<echo=TRUE>>=
ruin(claims = "e",
     par.claims = list(rate = c(3, 7), weights = 0.5),
     wait = "e",
     par.wait = list(rate = 3))
@

Finally, one will obtain a function to compute ruin probabilities in a
model with phase-type claim amounts and mixture of exponentials
interarrival times with
<<echo=TRUE>>=
prob <- c(0.5614, 0.4386)
rates <- matrix(c(-8.64, 0.101, 1.997, -1.095), 2, 2)
ruin(claims = "p",
     par.claims = list(prob = prob, rates = rates),
     wait = "e",
     par.wait = list(rate = c(5, 1), weights = c(0.4, 0.6)))
@

To ease plotting of the probability of ruin function, the package
provides a method of \code{plot} for objects returned by \code{ruin}
that is a simple wrapper for \code{curve} (see \autoref{fig:prob-ruin}):
<<echo=TRUE, fig=FALSE>>=
psi <- ruin(claims = "p",
            par.claims = list(prob = prob, rates = rates),
            wait = "e",
            par.wait = list(rate = c(5, 1),
                            weights = c(0.4, 0.6)))
plot(psi, from = 0, to = 50)
@
<<echo=FALSE>>=
options(op)
@

\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
plot(psi, from = 0, to = 50)
@
  \caption{Graphic of the probability of ruin as a function of the
    initial surplus $u$}
  \label{fig:prob-ruin}
\end{figure}


\section{Approximation to the probability of ruin}
\label{sec:beekman}

When the model for the aggregate claim process
\eqref{eq:definition-S(t)} does not fit nicely into the framework of
the previous section, one can compute ruin probabilities using the
so-called Beekman's convolution formula
\citep{Beekman_68,BeekmanFormula_EAS}.

Let the surplus process and the aggregate claim amount process be
defined as in \eqref{eq:definition-surplus} and
\eqref{eq:definition-S(t)}, respectively, and let $\{N(t)\}$ be a
Poisson process with mean $\lambda$. As before, claim amounts $C_1,
C_2, \dots$ are independent and identically distributed with cdf
$P(\cdot)$ and mean $\mu = \E{C_1}$. Then the infinite time
probability of ruin is given by
\begin{equation}
  \label{eq:beekman:prob-ruin}
  \psi(u) = 1 - F(u),
\end{equation}
where $F(\cdot)$ is Compound~Geometric$(p, H)$ with
\begin{equation}
  \label{eq:beekman:p}
  p = 1 - \frac{\lambda \mu}{c}
\end{equation}
and
\begin{equation}
  \label{eq:beekman:H}
  H(x) = \int_0^x \frac{1 - P(y)}{\mu}\, dy.
\end{equation}
In other words, we have (compare with \eqref{eq:cdf-S}):
\begin{equation}
  \label{eq:beekman:prob-ruin-long}
  \psi(u) = 1 - \sum_{n = 0}^\infty  H^{*n}(u) p (1 - p)^n.
\end{equation}

In most practical situations, numerical evaluation of
\eqref{eq:beekman:prob-ruin-long} is done using Panjer's recursive
formula. This usually requires discretization of $H(\cdot)$. In such
circumstances, Beekman's formula yields approximate ruin
probabilities.

For example, let claim amounts have a Pareto$(5, 4)$ distribution,
that is
\begin{displaymath}
  P(x) = 1 - \left( \frac{4}{4 + x} \right)^5
\end{displaymath}
and $\mu = 1$. Then
\begin{align*}
  H(x)
  &= \int_0^x \left( \frac{4}{4 + y} \right)^5 dy \\
  &= 1 - \left( \frac{4}{4 + x} \right)^4,
\end{align*}
or else $H$ is Pareto$(4, 4)$. Furthermore, we determine the premium
rate $c$ with the expected value premium principle and a safety
loading of 20\%, that is $c = 1.2 \lambda \mu$. Thus, $p = 0.2/1.2 =
1/6$.

One can get functions to compute lower bounds and upper bounds for
$F(u)$ with functions \code{discretize} and \code{aggregateDist}
as follows:
<<echo=TRUE>>=
f.L <- discretize(ppareto(x, 4, 4), from = 0, to = 200,
                  step = 1, method = "lower")
f.U <- discretize(ppareto(x, 4, 4), from = 0, to = 200,
                  step = 1, method = "upper")
F.L <- aggregateDist(method = "recursive",
                     model.freq = "geometric",
                     model.sev = f.L, prob = 1/6)
F.U <- aggregateDist(method = "recursive",
                     model.freq = "geometric",
                     model.sev = f.U, prob = 1/6)
@

Corresponding functions for the probability of ruin $\psi(u)$ lower
and upper bounds are (see \autoref{fig:beekman:prob-ruin} for the
graphic):
<<echo=TRUE>>=
psi.L <- function(u) 1 - F.U(u)
psi.U <- function(u) 1 - F.L(u)
u <- seq(0, 50, by = 5)
cbind(lower = psi.L(u), upper = psi.U(u))
curve(psi.L, from = 0, to = 100, col = "blue")
curve(psi.U, add = TRUE, col = "green")
@

\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
curve(psi.L, from = 0, to = 100, col = "blue")
curve(psi.U, add = TRUE, col = "green")
@
  \caption{Lower and upper bounds for the probability of ruin as
    determined using Beekman's convolution formula.}
  \label{fig:beekman:prob-ruin}
\end{figure}

One can make the bounds as close as one wishes by reducing the
discretization step.

\bibliography{actuar}

\end{document}

%%% Local Variables:
%%% mode: noweb
%%% TeX-master: t
%%% coding: utf-8
%%% End:

\input{share/preamble}

  %\VignetteIndexEntry{Loss distributions modeling}
  %\VignettePackage{actuar}
  %\SweaveUTF8

  \title{Loss modeling features of \pkg{actuar}}
  \author{Christophe Dutang \\ Université Paris Dauphine \\[3ex]
    Vincent Goulet \\ Université Laval \\[3ex]
    Mathieu Pigeon \\ Université du Québec à Montréal}
  \date{}

<<echo=FALSE>>=
library(actuar)
options(width = 52, digits = 4)
@

\begin{document}

\maketitle

\section{Introduction}
\label{sec:introduction}

One important task of actuaries is the modeling of claim amount and
claim count distributions for ratemaking, loss reserving or other risk
evaluation purposes. Package \pkg{actuar} features many support
functions for loss distributions modeling:
\begin{enumerate}
\item support for heavy tail continuous distributions useful in
  loss severity modeling;
\item support for phase-type distributions for ruin theory;
\item functions to compute raw moments, limited moments and the moment
  generating function (when it exists) of continuous distributions;
\item support for zero-truncated and zero-modified extensions of the
  discrete distributions commonly used in loss frequency modeling;
\item extensive support of grouped data;
\item functions to compute empirical raw and limited moments;
\item support for minimum distance estimation using three different
  measures;
\item treatment of coverage modifications (deductibles, limits,
  inflation, coinsurance).
\end{enumerate}

Vignette \code{"distributions"} covers the points 1--4 above in great
detail. This document concentrates on points 5--8.


\section{Grouped data}
\label{sec:grouped-data}

Grouped data is data represented in an interval-frequency manner.
Typically, a grouped data set will report that there were $n_j$ claims
in the interval $(c_{j - 1}, c_j]$, $j = 1, \dots, r$ (with the
possibility that $c_r = \infty$). This representation is much more
compact than an individual data set --- where the value of each claim
is known --- but it also carries far less information. Now that
storage space in computers has essentially become a non issue, grouped
data has somewhat fallen out of fashion. Still, grouped data remains
useful as a means to represent data, if only graphically --- for
example, a histogram is nothing but a density approximation for
grouped data. Moreover, various parameter estimation techniques rely
on grouped data.

For these reasons, \pkg{actuar} provides facilities to store,
manipulate and summarize grouped data. A standard storage method is
needed since there are many ways to represent grouped data in the
computer: using a list or a matrix, aligning $n_j$ with
$c_{j - 1}$ or with $c_j$, omitting $c_0$ or not, etc. With
appropriate extraction, replacement and summary methods, manipulation
of grouped data becomes similar to that of individual data.

Function \code{grouped.data} creates a grouped data object
similar to --- and inheriting from --- a data frame. The function
accepts two types of input:
\begin{enumerate}
\item a vector of group boundaries $c_0, c_1, \dots, c_r$ and one or
  more vectors of group frequencies $n_1, \dots, n_r$ (note that there
  should be one more group boundary than group frequencies);
\item individual data $x_1, \dots, x_n$ and either a vector of
  breakpoints $c_1, \dots, c_r$, a number $r$ of breakpoints or an
  algorithm to determine the latter.
\end{enumerate}
In the second case, \code{grouped.data} will group the individual data
using function \code{hist}. The function always assumes that the intervals
are contiguous.

\begin{example}
  \label{ex:grouped.data-1}
  Consider the following already grouped data set:
  \begin{center}
    \begin{tabular}{lcc}
      \toprule
      Group & Frequency (Line 1) & Frequency (Line 2) \\
      \midrule
      $(0, 25]$    & 30 & 26 \\
      $(25, 50]$   & 31 & 33 \\
      $(50, 100]$  & 57 & 31 \\
      $(100, 150]$ & 42 & 19 \\
      $(150, 250]$ & 65 & 16 \\
      $(250, 500]$ & 84 & 11 \\
      \bottomrule
    \end{tabular}
  \end{center}
  We can conveniently and unambiguously store this data set in
  R as follows:
<<echo=TRUE>>=
x <- grouped.data(Group = c(0, 25, 50, 100,
                            150, 250, 500),
                  Line.1 = c(30, 31, 57, 42, 65, 84),
                  Line.2 = c(26, 33, 31, 19, 16, 11))
@

   Internally, object \code{x} is a list with class
<<echo=TRUE>>=
class(x)
@

  The package provides a suitable \code{print} method to display
  grouped data objects in an intuitive manner:
<<echo=TRUE>>=
x
@
  \qed
\end{example}

\begin{example}
  \label{ex:grouped.data-2}
  Consider Data Set~B of \citet[Table~11.2]{LossModels4e}:
  \begin{center}
    \begin{tabular}{*{10}{r}}
       27 &  82 & 115 & 126 & 155 &  161 &  243 &  294 &  340 &   384 \\
      457 & 680 & 855 & 877 & 974 & \np{1193}
                                  & \np{1340}
                                  & \np{1884}
                                  & \np{2558}
                                  & \np{15743}
    \end{tabular}
  \end{center}
  We can represent this data set as grouped data using either an automatic
  or a suggested number of groups (see \code{?hist} for details):
<<echo=TRUE>>=
y <- c(  27,   82,  115,   126, 155, 161, 243,  294,
        340,  384,  457,   680, 855, 877, 974, 1193,
       1340, 1884, 2558, 15743)
grouped.data(y)
grouped.data(y, breaks = 5)
@

  The above grouping methods use equi-spaced breaks. This is rarely
  appropriate for heavily skewed insurance data. For this reason,
  \code{grouped.data} also supports specified breakpoints (or group
  boundaries):
<<echo=TRUE>>=
grouped.data(y, breaks = c(0, 100, 200, 350, 750,
                           1200, 2500, 5000, 16000))
@
  \qed
\end{example}

The package supports the most common extraction and
replacement methods for \code{"grouped.data"} objects using the
usual \code{[} and \code{[<-} operators. In particular, the
following extraction operations are supported. (In the following,
object \code{x} is the grouped data object of \autoref{ex:grouped.data-1}.)
<<echo=FALSE>>=
x <- grouped.data(Group = c(0, 25, 50, 100, 150, 250, 500),
                  Line.1 = c(30, 31, 57, 42, 65, 84),
                  Line.2 = c(26, 33, 31, 19, 16, 11))
@
\begin{enumerate}[i)]
\item Extraction of the vector of group boundaries (the first column):
<<echo=TRUE>>=
x[, 1]
@
\item Extraction of the vector or matrix of group frequencies (the
  second and third columns):
<<echo=TRUE>>=
x[, -1]
@
\item Extraction of a subset of the whole object (first three lines):
<<echo=TRUE>>=
x[1:3, ]
@
\end{enumerate}
Notice how extraction results in a simple vector or matrix if either
of the group boundaries or the group frequencies are dropped.

As for replacement operations, the package implements the following.
\begin{enumerate}[i)]
\item Replacement of one or more group frequencies:
<<echo=TRUE>>=
x[1, 2] <- 22; x
x[1, c(2, 3)] <- c(22, 19); x
@
\item Replacement of the boundaries of one or more groups:
<<echo=TRUE>>=
x[1, 1] <- c(0, 20); x
x[c(3, 4), 1] <- c(55, 110, 160); x
@
\end{enumerate}
It is not possible to replace the boundaries and the frequencies
simultaneously.

The mean of grouped data is
\begin{equation}
  \hat{\mu} = \frac{1}{n} \sum_{j = 1}^r a_j n_j,
\end{equation}
where $a_j = (c_{j - 1} + c_j)/2$ is the midpoint of the $j$th
interval, and $n = \sum_{j = 1}^r n_j$, whereas the variance is
\begin{equation}
  \frac{1}{n} \sum_{j = 1}^r n_j (a_j - \hat{\mu})^2.
\end{equation}
The standard deviation is the square root of the variance. The package
defines methods to easily compute the above descriptive statistics:
<<echo=TRUE>>=
mean(x)
var(x)
sd(x)
@
Higher empirical moments can be computed with \code{emm}; see
\autoref{sec:empirical-moments}.

The R function \code{hist} splits individual data into
groups and draws an histogram of the frequency distribution. The
package introduces a method for already grouped data. Only the first
frequencies column is considered (see \autoref{fig:histogram} for
the resulting graph):
<<echo=TRUE, fig=FALSE>>=
hist(x[, -3])
@
\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
hist(x[, -3])
@
  \caption{Histogram of a grouped data object}
  \label{fig:histogram}
\end{figure}

\begin{rem}
  One will note that for an individual data set like \code{y} of
  \autoref{ex:grouped.data-2}, the following two expressions yield the
  same result:
<<echo=TRUE, fig=FALSE>>=
hist(y)
hist(grouped.data(y))
@
\end{rem}

R has a function \code{ecdf} to compute the empirical
cdf $F_n(x)$ of an individual data set:
\begin{equation}
  \label{eq:ecdf}
  F_n(x) = \frac{1}{n} \sum_{j = 1}^n I\{x_j \leq x\},
\end{equation}
where $I\{\mathcal{A}\} = 1$ if $\mathcal{A}$ is true and
$I\{\mathcal{A}\} = 0$ otherwise. The function returns a \code{"function"}
object to compute the value of $F_n(x)$
in any $x$.

The approximation of the empirical cdf for grouped data is called an
ogive \citep{LossModels4e,HoggKlugman}. It is obtained by joining the
known values of $F_n(x)$ at group boundaries with straight line
segments:
\begin{equation}
  \tilde{F}_n(x) =
  \begin{cases}
    0, & x \leq c_0 \\
    \dfrac{(c_j - x) F_n(c_{j-1}) + (x - c_{j-1}) F_n(c_j)}{%
      c_j - c_{j - 1}}, & c_{j-1} < x \leq c_j \\
    1, & x > c_r.
  \end{cases}
\end{equation}
The package includes a generic function \code{ogive} with methods for
individual and for grouped data. The function behaves
exactly like \code{ecdf}.

\begin{example}
  \label{ex:ogive}
  Consider first the grouped data set of \autoref{ex:grouped.data-1}.
  Function \code{ogive} returns a function to compute the ogive
  $\tilde{F}_n(x)$ in any point:
<<echo=TRUE>>=
(Fnt <- ogive(x))
@
  Methods for functions \code{knots} and \code{plot} allow,
  respectively, to obtain the knots $c_0, c_1, \dots, c_r$ of the
  ogive and to draw a graph (see \autoref{fig:ogive}):
<<echo=TRUE, fig=FALSE>>=
knots(Fnt)
Fnt(knots(Fnt))
plot(Fnt)
@
\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE>>=
plot(Fnt)
@
  \caption{Ogive of a grouped data object}
  \label{fig:ogive}
\end{figure}

  To add further symmetry between functions \code{hist} and
  \code{ogive}, the latter also accepts in argument a vector individual
  data. It will call \code{grouped.data} and then computes the ogive.
  (Below, \code{y} is the individual data set of
  \autoref{ex:grouped.data-2}.)
<<echo=TRUE>>=
(Fnt <- ogive(y))
knots(Fnt)
@
  \qed
\end{example}

A method of function \code{quantile} for grouped data objects
returns linearly smoothed quantiles, that is, the inverse of the ogive
evaluated at various points:
<<echo=FALSE>>=
Fnt <- ogive(x)
@
<<echo=TRUE>>=
quantile(x)
Fnt(quantile(x))
@

Finally, a \code{summary} method for grouped data objects returns the
quantiles and the mean, as is usual for individual data:
<<echo=TRUE>>=
summary(x)
@

\section{Data sets}
\label{sec:data-sets}

This is certainly not the most spectacular feature of \pkg{actuar},
but it remains useful for illustrations and examples: the package
includes the individual dental claims and grouped dental claims data
of \cite{LossModels4e}:
<<echo=TRUE>>=
data("dental"); dental
data("gdental"); gdental
@

\section{Calculation of empirical moments}
\label{sec:empirical-moments}

The package provides two functions useful for estimation based on
moments. First, function \code{emm} computes the $k$th empirical
moment of a sample, whether in individual or grouped data form. For
example, the following expressions compute the first three moments for
individual and grouped data sets:
<<echo=TRUE>>=
emm(dental, order = 1:3)
emm(gdental, order = 1:3)
@

Second, in the same spirit as \code{ecdf} and \code{ogive},
function \code{elev} returns a function to compute the empirical
limited expected value --- or first limited moment --- of a sample for
any limit. Again, there are methods for individual and grouped data
(see \autoref{fig:elev} for the graphs):
<<echo=TRUE, fig=FALSE>>=
lev <- elev(dental)
lev(knots(lev))
plot(lev, type = "o", pch = 19)

lev <- elev(gdental)
lev(knots(lev))
plot(lev, type = "o", pch = 19)
@
\begin{figure}[t]
  \centering
<<echo=FALSE, fig=TRUE, height=5, width=10>>=
par(mfrow = c(1, 2))
plot(elev(dental), type = "o", pch = 19)
plot(elev(gdental), type = "o", pch = 19)
@
  \caption{Empirical limited expected value function of an individual
    data object (left) and a grouped data object (right)}
  \label{fig:elev}
\end{figure}

\section{Minimum distance estimation}
\label{sec:minimum-distance}

Two methods are widely used by actuaries to fit models to data:
maximum likelihood and minimum distance. The first technique applied
to individual data is well covered by function \code{fitdistr} of the
package \pkg{MASS} \citep{MASS}.

The second technique minimizes a chosen distance function between
theoretical and empirical distributions. Package \pkg{actuar} provides
function \code{mde}, very similar in usage and inner working to
\code{fitdistr}, to fit models according to any of the following three
distance minimization methods.

\begin{enumerate}
\item The Cramér-von~Mises method (\code{CvM}) minimizes the squared
  difference between the theoretical cdf and the empirical cdf or
  ogive at their knots:
  \begin{equation}
    d(\theta) =
    \sum_{j = 1}^n w_j [F(x_j; \theta) - F_n(x_j; \theta)]^2
  \end{equation}
  for individual data and
  \begin{equation}
    d(\theta) =
    \sum_{j = 1}^r w_j [F(c_j; \theta) - \tilde{F}_n(c_j; \theta)]^2
  \end{equation}
  for grouped data. Here, $F(x)$ is the theoretical cdf of a
  parametric family, $F_n(x)$ is the empirical cdf, $\tilde{F}_n(x)$
  is the ogive and $w_1 \geq 0, w_2 \geq 0, \dots$ are arbitrary
  weights (defaulting to $1$).
\item The modified chi-square method (\code{chi-square}) applies to
  grouped data only and minimizes the squared difference between the
  expected and observed frequency within each group:
  \begin{equation}
    d(\theta) =
    \sum_{j = 1}^r w_j [n (F(c_j; \theta) - F(c_{j - 1}; \theta)) - n_j]^2,
  \end{equation}
  where $n = \sum_{j = 1}^r n_j$. By default, $w_j = n_j^{-1}$.
\item The layer average severity method (\code{LAS}) applies to
  grouped data only and minimizes the squared difference between the
  theoretical and empirical limited expected value within each group:
  \begin{equation}
    d(\theta) = \sum_{j = 1}^r w_j
    [\LAS(c_{j - 1}, c_j; \theta)
    - \tilde{\LAS}_n(c_{j - 1}, c_j; \theta)]^2,
  \end{equation}
  where $\LAS(x, y) = \E{X \wedge y} - \E{X \wedge x}$, %
  $\tilde{\LAS}_n(x, y) = \tilde{E}_n[X \wedge y] - \tilde{E}_n[X
  \wedge x]$ and $\tilde{E}_n[X \wedge x]$ is the empirical limited
  expected value for grouped data.
\end{enumerate}

The arguments of \code{mde} are a data set, a function to compute
$F(x)$ or $\E{X \wedge x}$, starting values for the optimization
procedure and the name of the method to use. The empirical functions
are computed with \code{ecdf}, \code{ogive} or \code{elev}.

\begin{example}
  \label{ex:mde}
  The expressions below fit an exponential distribution to the grouped
  dental data set, as per example~2.21 of \cite{LossModels}:
<<echo=FALSE>>=
op <- options(warn = -1)                # hide warnings from mde()
@
<<echo=TRUE>>=
mde(gdental, pexp, start = list(rate = 1/200),
    measure = "CvM")
mde(gdental, pexp, start = list(rate = 1/200),
    measure = "chi-square")
mde(gdental, levexp, start = list(rate = 1/200),
    measure = "LAS")
@
<<echo=FALSE>>=
options(op)                             # restore warnings
@
  \qed
\end{example}

It should be noted that optimization is not always as simple to
achieve as in \autoref{ex:mde}. For example, consider the problem of
fitting a Pareto distribution to the same data set using the
Cramér--von~Mises method:
<<echo=TRUE, eval=FALSE>>=
mde(gdental, ppareto,
    start = list(shape = 3, scale = 600),
    measure = "CvM")
@
<<echo=FALSE, eval=TRUE>>=
out <- try(mde(gdental, ppareto, start = list(shape = 3, scale = 600),
        measure = "CvM"), silent = TRUE)
cat(sub(", scale", ",\n             scale", out))
@

Working in the log of the parameters often solves the problem since
the optimization routine can then flawlessly work with negative
parameter values:
<<echo=TRUE>>=
pparetolog <- function(x, logshape, logscale)
    ppareto(x, exp(logshape), exp(logscale))
(p <- mde(gdental, pparetolog,
          start = list(logshape = log(3),
                       logscale = log(600)),
          measure = "CvM"))
@
The actual estimators of the parameters are obtained with
<<echo=TRUE>>=
exp(p$estimate)
@ %$
This procedure may introduce additional bias in the estimators, though.



\section{Coverage modifications}
\label{sec:coverage}

Let $X$ be the random variable of the actual claim amount for an
insurance policy, $Y^L$ be the random variable of the amount paid per
loss and $Y^P$ be the random variable of the amount paid per payment.
The terminology for the last two random variables refers to whether or
not the insurer knows that a loss occurred. Now, the random variables
$X$, $Y^L$ and $Y^P$ will differ if any of the following coverage
modifications are present for the policy: an ordinary or a franchise
deductible, a limit, coinsurance or inflation adjustment
\cite[see][chapter~8 for precise definitions of these
terms]{LossModels4e}. \autoref{tab:coverage} summarizes the
definitions of $Y^L$ and $Y^P$.

\begin{table}
  \centering
  \begin{tabular}{lll}
    \toprule
    Coverage modification & Per-loss variable ($Y^L$) &
    Per-payment variable ($Y^P$)\\
    \midrule
    Ordinary deductible ($d$)  &
    $\begin{cases}
      0, & X \leq d \\
      X - d, & X > d
    \end{cases}$ &
    $\begin{cases}
      X - d, & X > d
    \end{cases}$ \medskip \\
    Franchise deductible ($d$) &
    $\begin{cases}
      0, & X \leq d \\
      X, & X > d
    \end{cases}$ &
    $\begin{cases}
      X, & X > d
    \end{cases} $ \medskip \\
    Limit ($u$) &
    $\begin{cases}
      X, & X \leq u \\
      u, & X > u
    \end{cases}$ &
    $\begin{cases} X, & X \leq u \\
      u, & X > u
    \end{cases}$ \bigskip \\
    Coinsurance ($\alpha$) & $\alpha X$ & $\alpha X$ \medskip \\
    Inflation ($r$) & $(1 + r)X$ & $(1 + r)X$ \\
    \bottomrule
  \end{tabular}
  \caption{Coverage modifications for per-loss variable ($Y^L$) and
    per-payment variable ($Y^P$) as defined in \cite{LossModels4e}.}
  \label{tab:coverage}
\end{table}

Often, one will want to use data $Y^P_1, \dots, Y^P_n$ (or $Y^L_1,
\dots, Y^L_n$) from the random variable $Y^P$ ($Y^L$) to fit a model
on the unobservable random variable $X$. This requires expressing the
pdf or cdf of $Y^P$ ($Y^L$) in terms of the pdf or cdf of $X$.
Function \code{coverage} of \pkg{actuar} does just that: given a pdf
or cdf and any combination of the coverage modifications mentioned
above, \code{coverage} returns a function object to compute the pdf or
cdf of the modified random variable. The function can then be used in
modeling like any other \code{dfoo} or \code{pfoo} function.

\begin{example}
  \label{ex:coverage}
  Let $Y^P$ represent the amount paid by an insurer for a policy with
  an ordinary deductible $d$ and a limit $u - d$ (or maximum covered
  loss of $u$). Then the definition of $Y^P$ is
  \begin{equation}
    Y^P =
    \begin{cases}
      X - d,            & d \leq X \leq u \\
      u - d,            & X \geq u
    \end{cases}
  \end{equation}
  and its pdf is
  \begin{equation}
    \label{eq:pdf-YP}
    f_{Y^P}(y)
    =
    \begin{cases}
      0,                               & y = 0 \\
      \dfrac{f_X(y + d)}{1 - F_X(d)},  & 0 < y < u - d \\
      \dfrac{1 - F_X(u)}{1 - F_X(d)},  & y = u - d \\
      0,                               & y > u - d.
    \end{cases}
  \end{equation}

  Assume $X$ has a gamma distribution. Then an R function
  to compute the pdf \eqref{eq:pdf-YP} in any $y$ for a deductible
  $d = 1$ and a limit $u = 10$ is obtained with \code{coverage} as
  follows:
<<echo=TRUE>>=
f <- coverage(pdf = dgamma, cdf = pgamma,
              deductible = 1, limit = 10)
f
f(0, shape = 5, rate = 1)
f(5, shape = 5, rate = 1)
f(9, shape = 5, rate = 1)
f(12, shape = 5, rate = 1)
@
  \qed
\end{example}

Note how function \code{f} in the previous example is built
specifically for the coverage modifications submitted and contains as
little useless code as possible.

The function returned by \code{coverage} may be used for various
purposes, most notably parameter estimation, as the following example
illustrates.

\begin{example}
  Let object \code{y} contain a sample of claims amounts from policies
  with the deductible and limit of \autoref{ex:coverage}. One can
  fit a gamma distribution by maximum likelihood to the claim severity
  distribution as follows:
<<echo=FALSE>>=
x <- rgamma(100, 2, 0.5)
y <- pmin(x[x > 1], 9)
op <- options(warn = -1)                # hide warnings from fitdistr()
@
<<echo=TRUE>>=
library(MASS)
fitdistr(y, f, start = list(shape = 2, rate = 0.5))
@
<<echo=FALSE>>=
options(op)                             # restore warnings
@
  \qed
\end{example}

Vignette \code{"coverage"} contains more detailed formulas for the pdf
and the cdf under various combinations of coverage modifications.

\bibliography{actuar}

\end{document}

%%% Local Variables:
%%% mode: noweb
%%% TeX-master: t
%%% coding: utf-8
%%% End:

\name{Pareto4}
\alias{Pareto4}
\alias{dpareto4}
\alias{ppareto4}
\alias{qpareto4}
\alias{rpareto4}
\alias{mpareto4}
\alias{levpareto4}
\title{The Pareto IV Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Pareto IV distribution with
  parameters \code{min}, \code{shape1}, \code{shape2} and \code{scale}.
}
\usage{
dpareto4(x, min, shape1, shape2, rate = 1, scale = 1/rate,
         log = FALSE)
ppareto4(q, min, shape1, shape2, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
qpareto4(p, min, shape1, shape2, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
rpareto4(n, min, shape1, shape2, rate = 1, scale = 1/rate)
mpareto4(order, min, shape1, shape2, rate = 1, scale = 1/rate)
levpareto4(limit, min, shape1, shape2, rate = 1, scale = 1/rate,
           order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{min}{lower bound of the support of the distribution.}
  \item{shape1, shape2, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Pareto IV (or \dQuote{type IV}) distribution with parameters
  \code{min} \eqn{= \mu}{= m},
  \code{shape1} \eqn{= \alpha}{= a},
  \code{shape2} \eqn{= \gamma}{= b} and
  \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\alpha \gamma ((x - \mu)/\theta)^{\gamma - 1}}{%
      \theta [1 + ((x - \mu)/\theta)^\gamma]^{\alpha + 1}}}{%
    f(x) = (a b ((x - m)/s)^(b - 1))/(s [1 + ((x - m)/s)^b]^(a + 1))}
  for \eqn{x > \mu}{x > m}, \eqn{-\infty < \mu < \infty}{-Inf < m < Inf},
  \eqn{\alpha > 0}{a > 0}, \eqn{\gamma > 0}{b > 0}
  and \eqn{\theta > 0}{s > 0}.

  The Pareto IV is the distribution of the random variable
  \deqn{\mu + \theta \left(\frac{X}{1 - X}\right)^{1/\gamma},}{%
    m + s (X/(1 - X))^(1/b),}
  where \eqn{X} has a beta distribution with parameters \eqn{1} and
  \eqn{\alpha}{a}. It derives from the \link[=dfpareto]{Feller-Pareto}
  distribution with \eqn{\tau = 1}{shape3 = 1}. Setting \eqn{\mu = 0}{min
  = 0} yields the \link[=dburr]{Burr} distribution.

  The Pareto IV distribution also has the following direct special
  cases:
  \itemize{
    \item A \link[=dpareto3]{Pareto III} distribution when \code{shape1
      == 1};
    \item A \link[=dpareto2]{Pareto II} distribution when \code{shape1
      == 1}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} for nonnegative integer values of \eqn{k <
  \alpha\gamma}{k < shape1 * shape2}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]} for nonnegative integer values of \eqn{k}
  and \eqn{\alpha - j/\gamma}{shape1 - j/shape2}, \eqn{j = 1, \dots, k}
  not a negative integer.
}
\value{
  \code{dpareto4} gives the density,
  \code{ppareto4} gives the distribution function,
  \code{qpareto4} gives the quantile function,
  \code{rpareto4} generates random deviates,
  \code{mpareto4} gives the \eqn{k}th raw moment, and
  \code{levpareto4} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levpareto4} computes the limited expected value using
  \code{\link{betaint}}.

  For Pareto distributions, we use the classification of Arnold (2015)
  with the parametrization of Klugman et al. (2012).

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Arnold, B.C. (2015), \emph{Pareto Distributions}, Second Edition, CRC
  Press.

  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dburr}} for the Burr distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
exp(dpareto4(1, min = 10, 2, 3, log = TRUE))
p <- (1:10)/10
ppareto4(qpareto4(p, min = 10, 2, 3, 2), min = 10, 2, 3, 2)

## variance
mpareto4(2, min = 10, 2, 3, 1) - mpareto4(1, min = 10, 2, 3, 1) ^ 2

## case with shape1 - order/shape2 > 0
levpareto4(10, min = 10, 2, 3, 1, order = 2)

## case with shape1 - order/shape2 < 0
levpareto4(10, min = 10, 1.5, 0.5, 1, order = 2)
}
\keyword{distribution}

\name{Extract.grouped.data}
\alias{Extract.grouped.data}
\alias{[.grouped.data}
\alias{[<-.grouped.data}
\title{Extract or Replace Parts of a Grouped Data Object}
\description{
  Extract or replace subsets of grouped data objects.
}
\usage{
\method{[}{grouped.data}(x, i, j)
\method{[}{grouped.data}(x, i, j) <- value
}
\arguments{
  \item{x}{an object of class \code{grouped.data}.}

  \item{i, j}{elements to extract or replace. \code{i, j} are
    \code{numeric} or \code{character} or, for \code{[} only, empty.
    Numeric values are coerced to integer as if by
    \code{\link[base]{as.integer}}. For replacement by \code{[}, a logical
    matrix is allowed, but not replacement in the group boundaries and
    group frequencies simultaneously.}

  \item{value}{a suitable replacement value.}
}
\details{
  Objects of class \code{"grouped.data"} can mostly be indexed like data
  frames, with the following restrictions:
  \enumerate{
    \item For \code{[}, the extracted object must keep a group
      boundaries column and at least one group frequencies column to
      remain of class \code{"grouped.data"};
    \item For \code{[<-}, it is not possible to replace group boundaries
      and group frequencies simultaneously;
    \item When replacing group boundaries, \code{length(value) ==
      length(i) + 1}.
  }

  \code{x[, 1]} will return the plain vector of group boundaries.
  
  Replacement of non adjacent group boundaries is not possible for
  obvious reasons.
  
  Otherwise, extraction and replacement should work just like for data
  frames.
}
\value{
  For \code{[} an object of class \code{"grouped.data"}, a data frame or a
  vector.

  For \code{[<-} an object of class \code{"grouped.data"}.
}
\note{
  Currently \code{[[}, \code{[[<-}, \code{$} and \code{$<-} are not 
  specifically supported, but should work as usual on group frequency
  columns.
}
\seealso{
  \code{\link[base]{[.data.frame}} for extraction and replacement
  methods of data frames, \code{\link{grouped.data}} to create grouped
  data objects.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
data(gdental)

(x <- gdental[1])         # select column 1
class(x)                  # no longer a grouped.data object
class(gdental[2])         # same
gdental[, 1]              # group boundaries
gdental[, 2]              # group frequencies

gdental[1:4,]             # a subset
gdental[c(1, 3, 5),]      # avoid this

gdental[1:2, 1] <- c(0, 30, 60) # modified boundaries
gdental[, 2] <- 10              # modified frequencies
\dontrun{gdental[1, ] <- 2}   # not allowed
}
\keyword{manip}
\keyword{array}

\name{Pareto2}
\alias{Pareto2}
\alias{dpareto2}
\alias{ppareto2}
\alias{qpareto2}
\alias{rpareto2}
\alias{mpareto2}
\alias{levpareto2}
\title{The Pareto II Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Pareto II distribution with
  parameters \code{min}, \code{shape} and \code{scale}.
}
\usage{
dpareto2(x, min, shape, rate = 1, scale = 1/rate,
         log = FALSE)
ppareto2(q, min, shape, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
qpareto2(p, min, shape, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
rpareto2(n, min, shape, rate = 1, scale = 1/rate)
mpareto2(order, min, shape, rate = 1, scale = 1/rate)
levpareto2(limit, min, shape, rate = 1, scale = 1/rate,
           order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{min}{lower bound of the support of the distribution.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Pareto II (or \dQuote{type II}) distribution with parameters
  \code{min} \eqn{= \mu}{= m},
  \code{shape} \eqn{= \alpha}{= a} and
  \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\alpha}{%
      \theta [1 + (x - \mu)/\theta]^{\alpha + 1}}}{%
    f(x) = a/(s [1 + (x - m)/s]^(a + 1))}
  for \eqn{x > \mu}{x > m}, \eqn{-\infty < \mu < \infty}{-Inf < m < Inf},
  \eqn{\alpha > 0}{a > 0} and \eqn{\theta > 0}{s > 0}.

  The Pareto II is the distribution of the random variable
  \deqn{\mu + \theta \left(\frac{X}{1 - X}\right),}{%
    m + s X/(1 - X),}
  where \eqn{X} has a beta distribution with parameters \eqn{1} and
  \eqn{\alpha}{a}. It derives from the \link[=dfpareto]{Feller-Pareto}
  distribution with \eqn{\tau = \gamma = 1}{shape2 = shape3 = 1}.
  Setting \eqn{\mu = 0}{min = 0} yields the familiar
  \link[=dpareto]{Pareto} distribution.

  The \link[=dpareto1]{Pareto I} (or Single parameter Pareto)
  distribution is a special case of the Pareto II with \code{min ==
  scale}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} for nonnegative integer values of \eqn{k <
  \alpha}{k < shape}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]} for nonnegative integer values of \eqn{k}
  and \eqn{\alpha - j}{shape1 - j}, \eqn{j = 1, \dots, k}
  not a negative integer.
}
\value{
  \code{dpareto2} gives the density,
  \code{ppareto2} gives the distribution function,
  \code{qpareto2} gives the quantile function,
  \code{rpareto2} generates random deviates,
  \code{mpareto2} gives the \eqn{k}th raw moment, and
  \code{levpareto2} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levpareto2} computes the limited expected value using
  \code{\link{betaint}}.

  For Pareto distributions, we use the classification of Arnold (2015)
  with the parametrization of Klugman et al. (2012).

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Arnold, B.C. (2015), \emph{Pareto Distributions}, Second Edition, CRC
  Press.

  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dpareto}} for the Pareto distribution without a location
  parameter.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
exp(dpareto2(1, min = 10, 3, 4, log = TRUE))
p <- (1:10)/10
ppareto2(qpareto2(p, min = 10, 2, 3), min = 10, 2, 3)

## variance
mpareto2(2, min = 10, 4, 1) - mpareto2(1, min = 10, 4, 1)^2

## case with shape - order > 0
levpareto2(10, min = 10, 3, scale = 1, order = 2)

## case with shape - order < 0
levpareto2(10, min = 10, 1.5, scale = 1, order = 2)
}
\keyword{distribution}

\name{rcomphierarc}
\alias{rcomphierarc}
\alias{simul}
\alias{print.portfolio}
\title{Simulation from Compound Hierarchical Models}
\description{
  Simulate data for insurance applications allowing hierarchical
  structures and separate models for the frequency and severity of
  claims distributions.

  \code{rcomphierarc} is an alias for \code{simul}.
}
\usage{
rcomphierarc(nodes, model.freq = NULL, model.sev = NULL, weights = NULL)

\method{print}{portfolio}(x, \dots)
}
\arguments{
  \item{nodes}{a vector or a named list giving the number of "nodes" at each level
    in the hierarchy of the portfolio. The nodes are listed from top
    (portfolio) to bottom (usually the years of experience).}
  \item{model.freq}{a named vector of expressions specifying the
    frequency of claims model (see Details); if \code{NULL}, only claim
    amounts are simulated.}
  \item{model.sev}{a named vector of expressions specifying the severity
    of claims model (see Details); if \code{NULL}, only claim numbers
    are simulated.}
  \item{weights}{a vector of weights.}
  \item{x}{a \code{portfolio} object.}
  \item{\dots}{potential further arguments required by generic.}
}
\details{
  The order and the names of the elements in \code{nodes},
  \code{model.freq} and \code{model.sev} must match. At least one of
  \code{model.freq} and \code{model.sev} must be non \code{NULL}.

  \code{nodes} may be a basic vector, named or not, for non hierarchical
  models. The rule above still applies, so \code{model.freq} and
  \code{model.sev} should not be named if \code{nodes} is not. However,
  for non hierarchical models, \code{\link{rcompound}} is faster and has
  a simpler interface.

  \code{nodes} specifies the hierarchical layout of the portfolio. Each
  element of the list is a vector of the number of nodes at a given
  level. Vectors are recycled as necessary.

  \code{model.freq} and \code{model.sev} specify the simulation models
  for claim numbers and claim amounts, respectively. A model is
  expressed in a semi-symbolic fashion using an object of mode
  \code{\link[base]{expression}}. Each element of the object
  must be named and should be a complete call to a random number
  generation function, with the number of variates omitted. Hierarchical
  (or mixtures of) models are achieved by replacing one or more
  parameters of a distribution at a given level by any combination of
  the names of the levels above. If no mixing is to take place at a
  level, the model for this level can be \code{NULL}.

  The argument of the random number generation functions for the number
  of variates to simulate \strong{must} be named \code{n}.

  Weights will be used wherever the name \code{"weights"} appears in a
  model. It is the user's responsibility to ensure that the length of
  \code{weights} will match the number of nodes when weights are to be
  used. Normally, there should be one weight per node at the lowest
  level of the model.

  Data is generated in lexicographic order, that is by row in the output
  matrix.
}
\value{
  An object of \code{\link[base]{class}} \code{"portfolio"}. A
  \code{print} method for this class displays the models used in the
  simulation as well as the frequency of claims for each year and entity
  in the portfolio.

  An object of class \code{"portfolio"} is a list containing the
  following components:
  \item{data}{a two dimension list where each element is a vector of
    claim amounts;}
  \item{weights}{the vector of weights given in argument reshaped as a
    matrix matching element \code{data}, or \code{NULL};}
  \item{classification}{a matrix of integers where each row is a unique
    set of subscripts identifying an entity in the portfolio
    (e.g. integers \eqn{i}, \eqn{j} and \eqn{k} for data
    \eqn{X_{ijkt}}{X[ijkt]});}
  \item{nodes}{the \code{nodes} argument, appropriately recycled;}
  \item{model.freq}{the frequency model as given in argument;}
  \item{model.sev}{the severity model as given in argument.}

  It is recommended to manipulate objects of class \code{"portfolio"} by
  means of the corresponding methods of functions \code{aggregate},
  \code{frequency} and \code{severity}.
}
\references{
  Goulet, V. and Pouliot, L.-P. (2008), Simulation of compound
  hierarchical models in R, \emph{North American Actuarial Journal}
  \bold{12}, 401--412.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Sébastien Auclair and Louis-Philippe Pouliot
}
\seealso{
  \code{\link{rcomphierarc.summaries}} for the functions to create the
  matrices of aggregate claim amounts, frequencies and individual claim
  amounts.

  \code{\link{rcompound}} for a simpler and much faster way to generate
  variates from standard, non hierarchical, compound models.
}
\examples{
## Two level (contracts and years) portfolio with frequency model
## Nit|Theta_i ~ Poisson(Theta_i), Theta_i ~ Gamma(2, 3) and severity
## model X ~ Lognormal(5, 1)
rcomphierarc(nodes = list(contract = 10, year = 5),
             model.freq = expression(contract = rgamma(2, 3),
                                     year = rpois(contract)),
             model.sev = expression(contract = NULL,
                                    year = rlnorm(5, 1)))

## Model with weights and mixtures for both frequency and severity
## models
nodes <- list(entity = 8, year = c(5, 4, 4, 5, 3, 5, 4, 5))
mf <- expression(entity = rgamma(2, 3),
                 year = rpois(weights * entity))
ms <- expression(entity = rnorm(5, 1),
                 year = rlnorm(entity, 1))
wit <- sample(2:10, 35, replace = TRUE)
pf <- rcomphierarc(nodes, mf, ms, wit)
pf 				# print method
weights(pf)			# extraction of weights
aggregate(pf)[, -1]/weights(pf)[, -1] # ratios

## Four level hierarchical model for frequency only
nodes <- list(sector = 3, unit = c(3, 4),
              employer = c(3, 4, 3, 4, 2, 3, 4), year = 5)
mf <- expression(sector = rexp(1),
                 unit = rexp(sector),
                 employer = rgamma(unit, 1),
                 year = rpois(employer))
pf <- rcomphierarc(nodes, mf, NULL)
pf 				# print method
aggregate(pf) 			# aggregate claim amounts
frequency(pf)  			# frequencies
severity(pf)			# individual claim amounts

## Standard, non hierarchical, compound model with simplified
## syntax (function rcompound() is much faster for such cases)
rcomphierarc(10,
             model.freq = expression(rpois(2)),
             model.sev = expression(rgamma(2, 3)))
}
\keyword{datagen}

\name{GeneralizedBeta}
\alias{GeneralizedBeta}
\alias{dgenbeta}
\alias{pgenbeta}
\alias{qgenbeta}
\alias{rgenbeta}
\alias{mgenbeta}
\alias{levgenbeta}
\title{The Generalized Beta Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Generalized Beta distribution
  with parameters \code{shape1}, \code{shape2}, \code{shape3} and
  \code{scale}.
}
\usage{
dgenbeta(x, shape1, shape2, shape3, rate = 1, scale = 1/rate,
         log = FALSE)
pgenbeta(q, shape1, shape2, shape3, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
qgenbeta(p, shape1, shape2, shape3, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
rgenbeta(n, shape1, shape2, shape3, rate = 1, scale = 1/rate)
mgenbeta(order, shape1, shape2, shape3, rate = 1, scale = 1/rate)
levgenbeta(limit, shape1, shape2, shape3, rate = 1, scale = 1/rate,
           order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, shape3, scale}{parameters. Must be strictly
    positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The generalized beta distribution with parameters \code{shape1} \eqn{=
    \alpha}{= a}, \code{shape2} \eqn{= \beta}{= b}, \code{shape3}
  \eqn{= \tau}{= c} and \code{scale} \eqn{= \theta}{= s}, has
  density:
  \deqn{f(x) = \frac{\Gamma(\alpha + \beta)}{\Gamma(\alpha)\Gamma(\beta)}
    (x/\theta)^{\alpha \tau} (1 - (x/\theta)^\tau)^{\beta - 1}
    \frac{\tau}{x}}{%
      f(x) = Gamma(a + b)/(Gamma(a) * Gamma(b))
             (c (x/s)^(ac) [1 - (x/s)^c]^(b - 1))/x}
  for \eqn{0 < x < \theta}{0 < x < s}, \eqn{\alpha > 0}{a > 0},
  \eqn{\beta > 0}{b > 0}, \eqn{\tau > 0}{c > 0} and \eqn{\theta > 0}{s >
    0}. (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The generalized beta is the distribution of the random variable
  \deqn{\theta X^{1/\tau},}{s X^(1/c),}
  where \eqn{X} has a beta distribution with parameters \eqn{\alpha}{a}
  and \eqn{\beta}{b}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)]}{E[min(X, d)]}, \eqn{k > -\alpha\tau}{k
  > -shape1 * shape3}.
}
\value{
  \code{dgenbeta} gives the density,
  \code{pgenbeta} gives the distribution function,
  \code{qgenbeta} gives the quantile function,
  \code{rgenbeta} generates random deviates,
  \code{mgenbeta} gives the \eqn{k}th raw moment, and
  \code{levgenbeta} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  This is \emph{not} the generalized three-parameter beta distribution
  defined on page 251 of Johnson et al, 1995.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Johnson, N. L., Kotz, S. and Balakrishnan, N. (1995)
  \emph{Continuous Univariate Distributions, Volume 2}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
exp(dgenbeta(2, 2, 3, 4, 0.2, log = TRUE))
p <- (1:10)/10
pgenbeta(qgenbeta(p, 2, 3, 4, 0.2), 2, 3, 4, 0.2)
mgenbeta(2, 1, 2, 3, 0.25) - mgenbeta(1, 1, 2, 3, 0.25) ^ 2
levgenbeta(10, 1, 2, 3, 0.25, order = 2)
}
\keyword{distribution}

\name{rmixture}
\alias{rmixture}
\title{Simulation from Discrete Mixtures}
\description{
  Generate random variates from a discrete mixture of distributions.
}
\usage{
rmixture(n, probs, models, shuffle = TRUE)
}
\arguments{
  \item{n}{number of random variates to generate. If \code{length(n) >
    1}, the length is taken to be the number required.}
  \item{probs}{numeric non-negative vector specifying the probability
    for each model; is internally normalized to sum 1. Infinite
    and missing values are not allowed. Values are recycled as necessary
    to match the length of \code{models}.}
  \item{models}{vector of expressions specifying the simulation models
    with the number of variates omitted; see Details. Models are
    recycled as necessary to match the length of \code{probs}.}
  \item{shuffle}{logical; should the random variates from the
    distributions be shuffled?}
}
\details{
  \code{rmixture} generates variates from a discrete mixture, that is
  the random variable with a probability density function of the form
  \deqn{f(x) = p_1 f_1(x) + ... + p_n f_n(x),}
  where \eqn{f_1, \dots, f_n} are densities and \eqn{\sum_{i = 1}^n p_i
    = 1}{p_1 + \dots + p_n = 1}.

  The values in \code{probs} will be internally normalized to be
  used as probabilities \eqn{p_1 + \dots + p_n}.

  The specification of simulation models uses the syntax of
  \code{\link{rcomphierarc}}. Models \eqn{f_1, \dots, f_n} are expressed in a
  semi-symbolic fashion using an object of mode
  \code{\link[base]{expression}} where each element is a complete call
  to a random number generation function, with the number of variates
  omitted.

  The argument of the random number generation functions for the number
  of variates to simulate \strong{must} be named \code{n}.

  If \code{shuffle} is \code{FALSE}, the output vector contains all the
  random variates from the first model, then all the random variates
  from the second model, and so on. If the order of the variates is
  irrelevant, this cuts the time to generate the variates roughly in
  half.
}
\note{
  Building the expressions in \code{models} from the arguments of
  another function is delicate. The expressions must be such that
  evaluation is possible in the frame of \code{rmixture} or its parent.
  See the examples.  
}
\value{
  A vector of random variates from the mixture with density \eqn{f(x)}.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\seealso{
  \code{\link{rcompound}} to simulate from compound models.

  \code{\link{rcomphierarc}} to simulate from compound hierarchical models.
}
\examples{
## Mixture of two exponentials (with means 1/3 and 1/7) with equal
## probabilities.
rmixture(10, 0.5, expression(rexp(3), rexp(7)))
rmixture(10, 42, expression(rexp(3), rexp(7))) # same

## Mixture of two lognormals with different probabilities.
rmixture(10, probs = c(0.55, 0.45),
         models = expression(rlnorm(3.6, 0.6),
                             rlnorm(4.6, 0.3)))

## Building the model expressions in the following example
## works as 'rate' is defined in the parent frame of
## 'rmixture'.
probs <- c(2, 5)
g <- function(n, p, rate)
    rmixture(n, p, expression(rexp(rate[1]), rexp(rate[2])))
g(10, probs, c(3, 7))

## The following example does not work: 'rate' does not exist
## in the evaluation frame of 'rmixture'.
f <- function(n, p, model) rmixture(n, p, model)
h <- function(n, p, rate)
    f(n, p, expression(rexp(rate[1]), rexp(rate[2])))
\dontrun{h(10, probs, c(3, 7))}

## Fix: substitute the values in the model expressions.
h <- function(n, p, rate)
{
    models <- eval(substitute(expression(rexp(a[1]), rexp(a[2])),
                              list(a = rate)))
    f(n, p, models)
}
h(10, probs, c(3, 7))
}
\keyword{datagen}

\name{Loglogistic}
\alias{Loglogistic}
\alias{dllogis}
\alias{pllogis}
\alias{qllogis}
\alias{rllogis}
\alias{mllogis}
\alias{levllogis}
\title{The Loglogistic Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Loglogistic distribution with
  parameters \code{shape} and \code{scale}.
}
\usage{
dllogis(x, shape, rate = 1, scale = 1/rate, log = FALSE)
pllogis(q, shape, rate = 1, scale = 1/rate,
        lower.tail = TRUE, log.p = FALSE)
qllogis(p, shape, rate = 1, scale = 1/rate,
        lower.tail = TRUE, log.p = FALSE)
rllogis(n, shape, rate = 1, scale = 1/rate)
mllogis(order, shape, rate = 1, scale = 1/rate)
levllogis(limit, shape, rate = 1, scale = 1/rate,
          order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The loglogistic distribution with parameters \code{shape} \eqn{=
    \gamma}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\gamma (x/\theta)^\gamma}{%
      x [1 + (x/\theta)^\gamma]^2}}{%
    f(x) = a (x/s)^a / (x [1 + (x/s)^a]^2)}
  for \eqn{x > 0}, \eqn{\gamma > 0}{a > 0} and \eqn{\theta > 0}{b > 0}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}, \eqn{-\gamma < k < \gamma}{-shape < k < shape}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]}, \eqn{k > -\gamma}{k > -shape}
  and \eqn{1 - k/\gamma}{1 - k/shape} not a negative integer.
}
\value{
  \code{dllogis} gives the density,
  \code{pllogis} gives the distribution function,
  \code{qllogis} gives the quantile function,
  \code{rllogis} generates random deviates,
  \code{mllogis} gives the \eqn{k}th raw moment, and
  \code{levllogis} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levllogis} computes the limited expected value using
  \code{\link{betaint}}.

  Also known as the Fisk distribution. See also Kleiber and Kotz (2003)
  for alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dpareto3}} for an equivalent distribution with a location
  parameter.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dllogis(2, 3, 4, log = TRUE))
p <- (1:10)/10
pllogis(qllogis(p, 2, 3), 2, 3)

## mean
mllogis(1, 2, 3)

## case with 1 - order/shape > 0
levllogis(10, 2, 3, order = 1)

## case with 1 - order/shape < 0
levllogis(10, 2/3, 3, order = 1)
}
\keyword{distribution}

\name{dental}
\docType{data}
\alias{dental}
\title{Individual Dental Claims Data Set}
\description{
  Basic dental claims on a policy with a deductible of 50.
}
\usage{dental}
\format{A vector containing 10 observations}
\source{Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\keyword{datasets}

\name{ZeroModifiedPoisson}
\alias{ZeroModifiedPoisson}
\alias{ZMpoisson}
\alias{dzmpois}
\alias{pzmpois}
\alias{qzmpois}
\alias{rzmpois}
\title{The Zero-Modified Poisson Distribution}
\description{
  Density function, distribution function, quantile function, random
  generation for the Zero-Modified Poisson distribution with parameter
  \code{lambda} and arbitrary probability at zero \code{p0}.
}
\usage{
dzmpois(x, lambda, p0, log = FALSE)
pzmpois(q, lambda, p0, lower.tail = TRUE, log.p = FALSE)
qzmpois(p, lambda, p0, lower.tail = TRUE, log.p = FALSE)
rzmpois(n, lambda, p0)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of values to return.}
  \item{lambda}{vector of (non negative) means.}
  \item{p0}{probability mass at zero. \code{0 <= p0 <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-modified Poisson distribution is a discrete mixture between a
  degenerate distribution at zero and a (standard) Poisson. The
  probability mass function is \eqn{p(0) = p_0}{p(0) = p0} and
  \deqn{%
    p(x) = \frac{(1-p_0)}{(1-e^{-\lambda})} f(x)}{%
    p(x) = (1-p0)/(1-exp(-lambda)) f(x)}
  for \eqn{x = 1, 2, ...}, \eqn{\lambda > 0} and \eqn{0 \le
  p_0 \le 1}{0 \le p0 \le 1}, where \eqn{f(x)} is the probability mass
  function of the Poisson.
  The cumulative distribution function is
  \deqn{P(x) = p_0 + (1 - p_0) \left(\frac{F(x) - F(0)}{1 - F(0)}\right).}{%
    P(x) = p0 + (1 - p0) [F(x) - F(0)]/[1 - F(0)].}

  The mean is \eqn{(1-p_0) \mu}{(1-p0)m} and the variance is
  \eqn{(1-p_0) \sigma^2 + p_0(1-p_0) \mu^2}{(1-p0)v + p0(1-p0)m^2},
  where \eqn{\mu}{m} and \eqn{\sigma^2}{v} are the mean and variance of
  the zero-truncated Poisson.

  In the terminology of Klugman et al. (2012), the zero-modified
  Poisson is a member of the \eqn{(a, b, 1)} class of distributions
  with \eqn{a = 0} and \eqn{b = \lambda}.

  The special case \code{p0 == 0} is the zero-truncated Poisson.

  If an element of \code{x} is not integer, the result of
  \code{dzmpois} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dzmpois} gives the (log) probability mass function,
  \code{pzmpois} gives the (log) distribution function,
  \code{qzmpois} gives the quantile function, and
  \code{rzmpois} generates random deviates.

  Invalid \code{lambda} or \code{p0} will result in return value
  \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rzmpois}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}zmpois} use \code{\{d,p,q\}pois} for all
  but the trivial input values and \eqn{p(0)}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dpois}} for the standard Poisson distribution.

  \code{\link{dztpois}} for the zero-truncated Poisson distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
dzmpois(0:5, lambda = 1, p0 = 0.2)
(1-0.2) * dpois(0:5, lambda = 1)/ppois(0, 1, lower = FALSE) # same

## simple relation between survival functions
pzmpois(0:5, 1, p0 = 0.2, lower = FALSE)
(1-0.2) * ppois(0:5, 1, lower = FALSE) /
    ppois(0, 1, lower = FALSE) # same

qzmpois(pzmpois(0:10, 1, p0 = 0.7), 1, p0 = 0.7)
}
\keyword{distribution}

\name{Paralogistic}
\alias{Paralogistic}
\alias{dparalogis}
\alias{pparalogis}
\alias{qparalogis}
\alias{rparalogis}
\alias{mparalogis}
\alias{levparalogis}
\title{The Paralogistic Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Paralogistic distribution with
  parameters \code{shape} and \code{scale}.
}
\usage{
dparalogis(x, shape, rate = 1, scale = 1/rate, log = FALSE)
pparalogis(q, shape, rate = 1, scale = 1/rate,
           lower.tail = TRUE, log.p = FALSE)
qparalogis(p, shape, rate = 1, scale = 1/rate,
           lower.tail = TRUE, log.p = FALSE)
rparalogis(n, shape, rate = 1, scale = 1/rate)
mparalogis(order, shape, rate = 1, scale = 1/rate)
levparalogis(limit, shape, rate = 1, scale = 1/rate,
             order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The paralogistic distribution with parameters \code{shape} \eqn{=
    \alpha}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\alpha^2 (x/\theta)^\alpha}{%
      x [1 + (x/\theta)^\alpha)^{\alpha + 1}}}{%
    f(x) = a^2 (x/s)^a / (x [1 + (x/s)^a]^(a + 1))}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0} and \eqn{\theta > 0}{b > 0}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{-\alpha < k < \alpha^2}{-shape < k <
  shape^2}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]}, \eqn{k > -\alpha}{k > -shape}
  and \eqn{\alpha - k/\alpha}{shape - k/shape} not a negative integer.
}
\value{
  \code{dparalogis} gives the density,
  \code{pparalogis} gives the distribution function,
  \code{qparalogis} gives the quantile function,
  \code{rparalogis} generates random deviates,
  \code{mparalogis} gives the \eqn{k}th raw moment, and
  \code{levparalogis} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levparalogis} computes the limited expected value using
  \code{\link{betaint}}.

  See Kleiber and Kotz (2003) for alternative names and
  parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dparalogis(2, 3, 4, log = TRUE))
p <- (1:10)/10
pparalogis(qparalogis(p, 2, 3), 2, 3)

## variance
mparalogis(2, 2, 3) - mparalogis(1, 2, 3)^2

## case with shape - order/shape > 0
levparalogis(10, 2, 3, order = 2)

## case with shape - order/shape < 0
levparalogis(10, 1.25, 3, order = 2)
}
\keyword{distribution}

\name{PhaseType}
\alias{PhaseType}
\alias{dphtype}
\alias{pphtype}
\alias{rphtype}
\alias{mphtype}
\alias{mgfphtype}
\title{The Phase-type Distribution}
\description{
  Density, distribution function, random generation, raw moments and
  moment generating function for the (continuous) Phase-type
  distribution with parameters \code{prob} and \code{rates}.
}
\usage{
dphtype(x, prob, rates, log = FALSE)
pphtype(q, prob, rates, lower.tail = TRUE, log.p = FALSE)
rphtype(n, prob, rates)
mphtype(order, prob, rates)
mgfphtype(t, prob, rates, log = FALSE)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{prob}{vector of initial probabilities for each of the transient
    states of the underlying Markov chain. The initial probability of
    the absorbing state is \code{1 - sum(prob)}.}
  \item{rates}{square matrix of the rates of transition among the states
    of the underlying Markov chain.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{t}{numeric vector.}
}
\details{
  The phase-type distribution with parameters \code{prob} \eqn{= \pi}{=
    pi} and \code{rates} \eqn{= \boldsymbol{T}}{= T} has density:%
  \deqn{f(x) = \pi e^{\boldsymbol{T} x} \boldsymbol{t}}{%
    f(x) = pi \%*\% exp(T * x) \%*\% t}%
  for \eqn{x \ge 0} and \eqn{f(0) = 1 - \pi \boldsymbol{e}}{f(0) = 1 - pi
    \%*\% e}, where %
  \eqn{\boldsymbol{e}}{e} %
  is a column vector with all components equal to one, %
  \eqn{\boldsymbol{t} = -\boldsymbol{T} \boldsymbol{e}}{%
    t = -T \%*\% e} %
  is the exit rates vector and %
  \eqn{e^{\boldsymbol{T}x}}{exp(T * x)} %
  denotes the matrix exponential of \eqn{\boldsymbol{T}x}{T * x}. The
  matrix exponential of a matrix \eqn{\boldsymbol{M}}{M} is defined as
  the Taylor series%
  \deqn{e^{\boldsymbol{M}} = \sum_{n = 0}^{\infty}
    \frac{\boldsymbol{M}^n}{n!}.}{%
    exp(M) = sum(n = 0:Inf; (M^n)/(n!)).}

  The parameters of the distribution must satisfy
  \eqn{\pi \boldsymbol{e} \leq 1}{pi \%*\% e <= 1},
  \eqn{\boldsymbol{T}_{ii} < 0}{T[i, i] < 0},
  \eqn{\boldsymbol{T}_{ij} \geq 0}{T[i, j] >= 0} and
  \eqn{\boldsymbol{T} \boldsymbol{e} \leq 0}{T \%*\% e <= 0}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the moment generating function is
  \eqn{E[e^{tX}]}.
}
\value{
  \code{dphasetype} gives the density,
  \code{pphasetype} gives the distribution function,
  \code{rphasetype} generates random deviates,
  \code{mphasetype} gives the \eqn{k}th raw moment, and
  \code{mgfphasetype} gives the moment generating function in \code{x}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}  
\references{
  \url{https://en.wikipedia.org/wiki/Phase-type_distribution}

  Neuts, M. F. (1981), \emph{Generating random variates from a
    distribution of phase type}, WSC '81: Proceedings of the 13th
  conference on Winter simulation, IEEE Press.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Christophe Dutang
}
\examples{
## Erlang(3, 2) distribution
T <- cbind(c(-2, 0, 0), c(2, -2, 0), c(0, 2, -2))
pi <- c(1,0,0)
x <- 0:10

dphtype(x, pi, T)		# density
dgamma(x, 3, 2)			# same
pphtype(x, pi, T)		# cdf
pgamma(x, 3, 2)			# same

rphtype(10, pi, T)		# random values

mphtype(1, pi, T)		# expected value

curve(mgfphtype(x, pi, T), from = -10, to = 1)
}
\keyword{distribution}

\name{InverseGaussian}
\alias{InverseGaussian}
\alias{dinvgauss}
\alias{pinvgauss}
\alias{qinvgauss}
\alias{rinvgauss}
\alias{minvgauss}
\alias{levinvgauss}
\alias{mgfinvgauss}
\title{The Inverse Gaussian Distribution}
\description{
  Density function, distribution function, quantile function, random
  generation, raw moments, limited moments and moment generating
  function for the Inverse Gaussian distribution with parameters
  \code{mean} and \code{shape}.
}
\usage{
dinvgauss(x, mean, shape = 1, dispersion = 1/shape,
          log = FALSE)
pinvgauss(q, mean, shape = 1, dispersion = 1/shape,
          lower.tail = TRUE, log.p = FALSE)
qinvgauss(p, mean, shape = 1, dispersion = 1/shape,
          lower.tail = TRUE, log.p = FALSE,
          tol = 1e-14, maxit = 100, echo = FALSE, trace = echo)
rinvgauss(n, mean, shape = 1, dispersion = 1/shape)
minvgauss(order, mean, shape = 1, dispersion = 1/shape)
levinvgauss(limit, mean, shape = 1, dispersion = 1/shape, order = 1)
mgfinvgauss(t, mean, shape = 1, dispersion = 1/shape, log = FALSE)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{mean, shape}{parameters. Must be strictly positive. Infinite
    values are supported.}
  \item{dispersion}{an alternative way to specify the shape.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment. Only \code{order = 1} is
    supported by \code{levinvgauss}.}
  \item{limit}{limit of the loss variable.}
  \item{tol}{small positive value. Tolerance to assess convergence in
    the Newton computation of quantiles.}
  \item{maxit}{positive integer; maximum number of recursions in the
    Newton computation of quantiles.}
  \item{echo, trace}{logical; echo the recursions to screen in the
    Newton computation of quantiles.}
  \item{t}{numeric vector.}
}
\details{
  The inverse Gaussian distribution with parameters \code{mean} \eqn{=
    \mu} and \code{dispersion} \eqn{= \phi} has density:
  \deqn{f(x) = \left( \frac{1}{2 \pi \phi x^3} \right)^{1/2}
    \exp\left( -\frac{(x - \mu)^2}{2 \mu^2 \phi x} \right),}{%
    f(x) = sqrt(1/(2 \pi \phi x^3)) *
           exp(-((x - \mu)^2)/(2 \mu^2 \phi x)),}
  for \eqn{x \ge 0}, \eqn{\mu > 0} and \eqn{\phi > 0}.

  The limiting case \eqn{\mu = \infty}{\mu = Inf} is an inverse
  chi-squared distribution (or inverse gamma with \code{shape} \eqn{=
  1/2} and \code{rate} \eqn{= 2}\code{phi}). This distribution has no
  finite strictly positive, integer moments.

  The limiting case \eqn{\phi = 0} is an infinite spike in \eqn{x = 0}.

  If the random variable \eqn{X} is IG\eqn{(\mu, \phi)}, then
  \eqn{X/\mu} is IG\eqn{(1, \phi \mu)}{(1, \phi * \mu)}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{k = 1, 2, \dots}, the limited expected
  value at some limit \eqn{d} is \eqn{E[\min(X, d)]}{E[min(X, d)]} and
  the moment generating function is \eqn{E[e^{tX}]}.

  The moment generating function of the inverse guassian is defined for
    \code{t <= 1/(2 * mean^2 * phi)}.
}
\value{
  \code{dinvgauss} gives the density,
  \code{pinvgauss} gives the distribution function,
  \code{qinvgauss} gives the quantile function,
  \code{rinvgauss} generates random deviates,
  \code{minvgauss} gives the \eqn{k}th raw moment,
  \code{levinvgauss} gives the limited expected value, and
  \code{mgfinvgauss} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  Functions \code{dinvgauss}, \code{pinvgauss} and \code{qinvgauss} are
  C implementations of functions of the same name in package
  \pkg{statmod}; see Giner and Smyth (2016).

  Devroye (1986, chapter 4) provides a nice presentation of the
  algorithm to generate random variates from an inverse Gaussian
  distribution.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Giner, G. and Smyth, G. K. (2016), \dQuote{\pkg{statmod}: Probability
    Calculations for the Inverse Gaussian Distribution}, \emph{The R
    Journal}, vol. 8, no 1, p. 339-351. \doi{10.32614/RJ-2016-024}

  Chhikara, R. S. and Folk, T. L. (1989), \emph{The Inverse Gaussian
    Distribution: Theory, Methodology and Applications}, Decker.

  Devroye, L. (1986), \emph{Non-Uniform Random Variate Generation},
  Springer-Verlag. \url{https://luc.devroye.org/rnbookindex.html}
}
\seealso{
  \code{\link{dinvgamma}} for the inverse gamma distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
dinvgauss(c(-1, 0, 1, 2, Inf), mean = 1.5, dis = 0.7)
dinvgauss(c(-1, 0, 1, 2, Inf), mean = Inf, dis = 0.7)
dinvgauss(c(-1, 0, 1, 2, Inf), mean = 1.5, dis = Inf) # spike at zero

## Typical graphical representations of the inverse Gaussian
## distribution. First fixed mean and varying shape; second
## varying mean and fixed shape.
col = c("red", "blue", "green", "cyan", "yellow", "black")
par = c(0.125, 0.5, 1, 2, 8, 32)
curve(dinvgauss(x, 1, par[1]), from = 0, to = 2, col = col[1])
for (i in 2:6)
    curve(dinvgauss(x, 1, par[i]), add = TRUE, col = col[i])

curve(dinvgauss(x, par[1], 1), from = 0, to = 2, col = col[1])
for (i in 2:6)
    curve(dinvgauss(x, par[i], 1), add = TRUE, col = col[i])

pinvgauss(qinvgauss((1:10)/10, 1.5, shape = 2), 1.5, 2)

minvgauss(1:4, 1.5, 2)

levinvgauss(c(0, 0.5, 1, 1.2, 10, Inf), 1.5, 2)
}
\keyword{distribution}

\name{coverage}
\alias{coverage}
\alias{Coverage}
\title{Density and Cumulative Distribution Function for Modified Data}
\description{
  Compute probability density function or cumulative distribution
  function of the payment per payment or payment per loss random
  variable under any combination of the following coverage
  modifications: deductible, limit, coinsurance, inflation.
}
\usage{
coverage(pdf, cdf, deductible = 0, franchise = FALSE,
         limit = Inf, coinsurance = 1, inflation = 0,
         per.loss = FALSE)
}
\arguments{
  \item{pdf, cdf}{function object or character string naming a function
    to compute, respectively, the probability density function and cumulative
    distribution function of a probability law.}
  \item{deductible}{a unique positive numeric value.}
  \item{franchise}{logical; \code{TRUE} for a franchise deductible,
    \code{FALSE} (default) for an ordinary deductible.}
  \item{limit}{a unique positive numeric value larger than
    \code{deductible}.}
  \item{coinsurance}{a unique value between 0 and 1; the proportion of
    coinsurance.}
  \item{inflation}{a unique value between 0 and 1; the rate of
    inflation.}
  \item{per.loss}{logical; \code{TRUE} for the per loss distribution,
    \code{FALSE} (default) for the per payment distribution.}
}
\details{
  \code{coverage} returns a function to compute the probability
  density function (pdf) or the cumulative distribution function (cdf)
  of the distribution of losses under coverage modifications. The pdf
  and cdf of unmodified losses are \code{pdf} and \code{cdf},
  respectively.

  If \code{pdf} is specified, the pdf is returned; if \code{pdf} is
  missing or \code{NULL}, the cdf is returned. Note that \code{cdf} is
  needed if there is a deductible or a limit.
}
\value{
  An object of mode \code{"function"} with the same arguments as
  \code{pdf} or \code{cdf}, except \code{"lower.tail"},
  \code{"log.p"} and \code{"log"}, which are not supported.
}
\note{
  Setting arguments of the function returned by \code{coverage} using
  \code{\link{formals}} may very well not work as expected.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\seealso{
  \code{vignette("coverage")} for the exact definitions of the
  per payment and per loss random variables under an ordinary or
  franchise deductible.
}
\examples{
## Default case: pdf of the per payment random variable with
## an ordinary deductible
coverage(dgamma, pgamma, deductible = 1)

## Add a limit
f <- coverage(dgamma, pgamma, deductible = 1, limit = 7)
f <- coverage("dgamma", "pgamma", deductible = 1, limit = 7) # same
f(0, shape = 3, rate = 1)
f(2, shape = 3, rate = 1)
f(6, shape = 3, rate = 1)
f(8, shape = 3, rate = 1)
curve(dgamma(x, 3, 1), xlim = c(0, 10), ylim = c(0, 0.3))    # original
curve(f(x, 3, 1), xlim = c(0.01, 5.99), col = 4, add = TRUE) # modified
points(6, f(6, 3, 1), pch = 21, bg = 4)

## Cumulative distribution function
F <- coverage(cdf = pgamma, deductible = 1, limit = 7)
F(0, shape = 3, rate = 1)
F(2, shape = 3, rate = 1)
F(6, shape = 3, rate = 1)
F(8, shape = 3, rate = 1)
curve(pgamma(x, 3, 1), xlim = c(0, 10), ylim = c(0, 1))    # original
curve(F(x, 3, 1), xlim = c(0, 5.99), col = 4, add = TRUE)  # modified
curve(F(x, 3, 1), xlim = c(6, 10), col = 4, add = TRUE)    # modified

## With no deductible, all distributions below are identical
coverage(dweibull, pweibull, limit = 5)
coverage(dweibull, pweibull, per.loss = TRUE, limit = 5)
coverage(dweibull, pweibull, franchise = TRUE, limit = 5)
coverage(dweibull, pweibull, per.loss = TRUE, franchise = TRUE,
         limit = 5)

## Coinsurance alone; only case that does not require the cdf
coverage(dgamma, coinsurance = 0.8)
}
\keyword{models}

\name{TransformedBeta}
\alias{TransformedBeta}
\alias{dtrbeta}
\alias{ptrbeta}
\alias{qtrbeta}
\alias{rtrbeta}
\alias{mtrbeta}
\alias{levtrbeta}
\alias{Pearson6}
\alias{dpearson6}
\alias{ppearson6}
\alias{qpearson6}
\alias{rpearson6}
\alias{mpearson6}
\alias{levpearson6}
\title{The Transformed Beta Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Transformed Beta distribution
  with parameters \code{shape1}, \code{shape2}, \code{shape3} and
  \code{scale}.
}
\usage{
dtrbeta(x, shape1, shape2, shape3, rate = 1, scale = 1/rate,
        log = FALSE)
ptrbeta(q, shape1, shape2, shape3, rate = 1, scale = 1/rate,
        lower.tail = TRUE, log.p = FALSE)
qtrbeta(p, shape1, shape2, shape3, rate = 1, scale = 1/rate,
        lower.tail = TRUE, log.p = FALSE)
rtrbeta(n, shape1, shape2, shape3, rate = 1, scale = 1/rate)
mtrbeta(order, shape1, shape2, shape3, rate = 1, scale = 1/rate)
levtrbeta(limit, shape1, shape2, shape3, rate = 1, scale = 1/rate,
          order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, shape3, scale}{parameters. Must be strictly
    positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The transformed beta distribution with parameters \code{shape1} \eqn{=
    \alpha}{= a}, \code{shape2} \eqn{= \gamma}{= b}, \code{shape3}
  \eqn{= \tau}{= c} and \code{scale} \eqn{= \theta}{= s}, has
  density:
  \deqn{f(x) = \frac{\Gamma(\alpha + \tau)}{\Gamma(\alpha)\Gamma(\tau)}
    \frac{\gamma (x/\theta)^{\gamma \tau}}{%
      x [1 + (x/\theta)^\gamma]^{\alpha + \tau}}}{%
    f(x) = Gamma(a + c)/(Gamma(a) * Gamma(c)) (b (x/s)^(bc))/%
    (x [1 + (x/s)^b]^(a + c))}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0}, \eqn{\gamma > 0}{b > 0},
  \eqn{\tau > 0}{c > 0} and \eqn{\theta > 0}{s > 0}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The transformed beta is the distribution of the random variable
  \deqn{\theta \left(\frac{X}{1 - X}\right)^{1/\gamma},}{%
    s (X/(1 - X))^(1/b),}
  where \eqn{X} has a beta distribution with parameters \eqn{\tau}{c}
  and \eqn{\alpha}{a}.

  The transformed beta distribution defines a family of distributions
  with the following special cases:
  \itemize{
    \item A \link[=dburr]{Burr} distribution when \code{shape3 == 1};
    \item A \link[=dllogis]{loglogistic} distribution when \code{shape1
      == shape3 == 1};
    \item A \link[=dparalogis]{paralogistic} distribution when
    \code{shape3 == 1} and \code{shape2 == shape1};
    \item A \link[=dgenpareto]{generalized Pareto} distribution when
    \code{shape2 == 1};
    \item A \link[=dpareto]{Pareto} distribution when \code{shape2 ==
      shape3 == 1};
    \item An \link[=dinvburr]{inverse Burr} distribution when
    \code{shape1 == 1};
    \item An \link[=dinvpareto]{inverse Pareto} distribution when
    \code{shape2 == shape1 == 1};
    \item An \link[=dinvparalogis]{inverse paralogistic} distribution
    when \code{shape1 == 1} and \code{shape3 == shape2}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}, \eqn{-\tau\gamma < k < \alpha\gamma}{-shape3 * shape2 <
  k < shape1 * shape2}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]}, \eqn{k > -\tau\gamma}{k > -shape3 * shape2}
  and \eqn{\alpha - k/\gamma}{shape1 - k/shape2} not a negative integer.
}
\value{
  \code{dtrbeta} gives the density,
  \code{ptrbeta} gives the distribution function,
  \code{qtrbeta} gives the quantile function,
  \code{rtrbeta} generates random deviates,
  \code{mtrbeta} gives the \eqn{k}th raw moment, and
  \code{levtrbeta} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levtrbeta} computes the limited expected value using
  \code{\link{betaint}}.

  Distribution also known as the Generalized Beta of the Second
  Kind and Pearson Type VI. See also Kleiber and Kotz (2003) for
  alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dfpareto}} for an equivalent distribution with a location
  parameter.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dtrbeta(2, 2, 3, 4, 5, log = TRUE))
p <- (1:10)/10
ptrbeta(qtrbeta(p, 2, 3, 4, 5), 2, 3, 4, 5)
qpearson6(0.3, 2, 3, 4, 5, lower.tail = FALSE)

## variance
mtrbeta(2, 2, 3, 4, 5) - mtrbeta(1, 2, 3, 4, 5)^2

## case with shape1 - order/shape2 > 0
levtrbeta(10, 2, 3, 4, scale = 1, order = 2)

## case with shape1 - order/shape2 < 0
levtrbeta(10, 1/3, 0.75, 4, scale = 0.5, order = 2)
}
\keyword{distribution}

\name{GammaSupp}
\alias{GammaSupp}
\alias{mgamma}
\alias{levgamma}
\alias{mgfgamma}
\title{Moments and Moment Generating Function of the Gamma Distribution}
\description{
  Raw moments, limited moments and moment generating function for the
  Gamma distribution with parameters \code{shape} and \code{scale}.
}
\usage{
mgamma(order, shape, rate = 1, scale = 1/rate)
levgamma(limit, shape, rate = 1, scale = 1/rate, order = 1)
mgfgamma(t, shape, rate = 1, scale = 1/rate, log = FALSE)
}
\arguments{
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{rate}{an alternative way to specify the scale.}
  \item{shape, scale}{shape and scale parameters. Must be strictly
    positive.}
  \item{t}{numeric vector.}
  \item{log}{logical; if \code{TRUE}, the cumulant generating function
    is returned.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]} and the moment
  generating function is \eqn{E[e^{tX}]}, \eqn{k >
  -\alpha}{k > -shape}.
}
\value{
  \code{mgamma} gives the \eqn{k}th raw moment,
  \code{levgamma} gives the \eqn{k}th moment of the limited loss
  variable, and
  \code{mgfgamma} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link[stats]{GammaDist}}
}
\references{
  Johnson, N. L. and Kotz, S. (1970), \emph{Continuous Univariate
    Distributions, Volume 1}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Christophe Dutang and Mathieu Pigeon
}
\examples{
mgamma(2, 3, 4) - mgamma(1, 3, 4)^2
levgamma(10, 3, 4, order = 2)
mgfgamma(1,3,2)
}
\keyword{distribution}

\name{CTE}
\alias{CTE}
\alias{TVaR}
\alias{CTE.aggregateDist}
\title{Conditional Tail Expectation}
\description{
  Conditional Tail Expectation, also called Tail Value-at-Risk.

  \code{TVaR} is an alias for \code{CTE}.
}
\usage{
CTE(x, \dots)

\method{CTE}{aggregateDist}(x, conf.level = c(0.9, 0.95, 0.99),
         names = TRUE, \dots)

TVaR(x, \dots)
}
\arguments{
  \item{x}{an \R object.}
  \item{conf.level}{numeric vector of probabilities with values
    in \eqn{[0, 1)}.}
  \item{names}{logical; if true, the result has a \code{names}
    attribute.  Set to \code{FALSE} for speedup with many \code{probs}.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  The Conditional Tail Expectation (or Tail Value-at-Risk) measures the
  average of losses above the Value at Risk for some given confidence
  level, that is \eqn{E[X|X > \mathrm{VaR}(X)]} where \eqn{X} is the loss random
  variable.

  \code{CTE} is a generic function with, currently, only a method for
  objects of class \code{"aggregateDist"}.

  For the recursive, convolution and simulation methods of
  \code{\link{aggregateDist}}, the CTE is computed from the definition
  using the empirical cdf.

  For the normal approximation method, an explicit formula exists:
  \deqn{\mu + \frac{\sigma}{(1 - \alpha) \sqrt{2 \pi}}
    e^{-\mathrm{VaR}(X)^2/2},}{%
    m + s exp(-VaR(X)^2/2)/((1 - a) * sqrt(2 pi)),}
  where \eqn{\mu}{m} is the mean, \eqn{\sigma}{s} the standard
  deviation and \eqn{\alpha}{a} the confidence level.

  For the Normal Power approximation, the explicit formula given in
  Castañer et al. (2013) is
  \deqn{\mu + \frac{\sigma}{(1 - \alpha) \sqrt{2 \pi}}
    e^{-\mathrm{VaR}(X)^2/2}
    \left( 1 + \frac{\gamma}{6} \mathrm{VaR}(X) \right),}{%
    m + s exp(-VaR(X)^2/2)/((1 - a) * sqrt(2 pi)) (1 + g * VaR(X)/6),}
  where, as above, \eqn{\mu}{m} is the mean, \eqn{\sigma}{s} the standard
  deviation, \eqn{\alpha}{a} the confidence level and \eqn{\gamma}{g} is
  the skewness.
}
\value{
  A numeric vector, named if \code{names} is \code{TRUE}.
}
\seealso{
  \code{\link{aggregateDist}}; \code{\link{VaR}}
}
\references{
  Castañer, A. and Claramunt, M.M. and Mármol, M. (2013), Tail value at
  risk. An analysis with the Normal-Power approximation. In \emph{Statistical
  and Soft Computing Approaches in Insurance Problems}, pp. 87-112. Nova
  Science Publishers, 2013. ISBN 978-1-62618-506-7.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Tommy Ouellet
}
\examples{
model.freq <- expression(data = rpois(7))
model.sev <- expression(data = rnorm(9, 2))
Fs <- aggregateDist("simulation", model.freq, model.sev, nb.simul = 1000)
CTE(Fs)
}
\keyword{univar}

\name{SingleParameterPareto}
\alias{SingleParameterPareto}
\alias{dpareto1}
\alias{ppareto1}
\alias{qpareto1}
\alias{rpareto1}
\alias{mpareto1}
\alias{levpareto1}
\title{The Single-parameter Pareto Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments, and limited moments for the Single-parameter Pareto
  distribution with parameter \code{shape}.
}
\usage{
dpareto1(x, shape, min, log = FALSE)
ppareto1(q, shape, min, lower.tail = TRUE, log.p = FALSE)
qpareto1(p, shape, min, lower.tail = TRUE, log.p = FALSE)
rpareto1(n, shape, min)
mpareto1(order, shape, min)
levpareto1(limit, shape, min, order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length
    is taken to be the number required.}
  \item{shape}{parameter. Must be strictly positive.}
  \item{min}{lower bound of the support of the distribution.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The single-parameter Pareto, or Pareto I, distribution with parameter
  \code{shape} \eqn{= \alpha}{= a} has density:
  \deqn{f(x) = \frac{\alpha \theta^\alpha}{x^{\alpha + 1}}}{%
    f(x) = a b^a/x^(a + 1)}
  for \eqn{x > \theta}{x > b}, \eqn{\alpha > 0}{a > 0} and \eqn{\theta >
    0}{b > 0}.

  Although there appears to be two parameters, only \code{shape} is a true
  parameter. The value of \code{min} \eqn{= \theta}{= b} must be set in
  advance.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{k < \alpha}{k < shape} and the \eqn{k}th
  limited moment at some limit \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X,
  d)^k]}, \eqn{x \ge \theta}{x \ge min}.
}
\value{
  \code{dpareto1} gives the density,
  \code{ppareto1} gives the distribution function,
  \code{qpareto1} gives the quantile function,
  \code{rpareto1} generates random deviates,
  \code{mpareto1} gives the \eqn{k}th raw moment, and
  \code{levpareto1} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  For Pareto distributions, we use the classification of Arnold (2015)
  with the parametrization of Klugman et al. (2012).

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Arnold, B.C. (2015), \emph{Pareto Distributions}, Second Edition, CRC
  Press.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dpareto}} for the two-parameter Pareto distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dpareto1(5, 3, 4, log = TRUE))
p <- (1:10)/10
ppareto1(qpareto1(p, 2, 3), 2, 3)
mpareto1(2, 3, 4) - mpareto(1, 3, 4) ^ 2
levpareto(10, 3, 4, order = 2)
}
\keyword{distribution}

\name{unroll}
\alias{unroll}
\title{Display a Two-Dimension Version of a Matrix of Vectors}
\description{
  Displays all values of a matrix of vectors by \dQuote{unrolling} the
  object vertically or horizontally.
}
\usage{
unroll(x, bycol = FALSE, drop = TRUE)
}
\arguments{
  \item{x}{a list of vectors with a \code{\link[base]{dim}} attribute of
    length 0, 1 or 2.}
  \item{bycol}{logical; whether to unroll horizontally
    (\code{FALSE}) or vertically (\code{TRUE}).}
  \item{drop}{logical; if \code{TRUE}, the result is coerced to the
    lowest possible dimension.}
}
\details{
  \code{unroll} returns a matrix where elements of \code{x}
  are concatenated (\dQuote{unrolled}) by row (\code{bycol = FALSE}) or
  by column (\code{bycol = TRUE}). \code{NA} is used to make
  rows/columns of equal length.

  Vectors and one dimensional arrays are coerced to \strong{row}
  matrices.
}
\value{
  A vector or matrix.
}
\seealso{
  This function was originally written for use in
  \code{\link{severity.portfolio}}.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Louis-Philippe Pouliot
}
\examples{
x <- list(c(1:3), c(1:8), c(1:4), c(1:3))
(mat <- matrix(x, 2, 2))

unroll(mat)
unroll(mat, bycol = TRUE)

unroll(mat[1, ])
unroll(mat[1, ], drop = FALSE)
}
\keyword{manip}

\name{ZeroModifiedBinomial}
\alias{ZeroModifiedBinomial}
\alias{ZMBinomial}
\alias{dzmbinom}
\alias{pzmbinom}
\alias{qzmbinom}
\alias{rzmbinom}
\title{The Zero-Modified Binomial Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Modified Binomial distribution with
  parameters \code{size} and \code{prob}, and probability at zero
  \code{p0}.
}
\usage{
dzmbinom(x, size, prob, p0, log = FALSE)
pzmbinom(q, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
qzmbinom(p, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
rzmbinom(n, size, prob, p0)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{size}{number of trials (strictly positive integer).}
  \item{prob}{probability of success on each trial. \code{0 <= prob <= 1}.}
  \item{p0}{probability mass at zero. \code{0 <= p0 <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-modified binomial distribution with \code{size} \eqn{= n},
  \code{prob} \eqn{= p} and \code{p0} \eqn{= p_0}{= p0} is a discrete
  mixture between a degenerate distribution at zero and a (standard)
  binomial. The probability mass function is \eqn{p(0) = p_0}{p(0) = p0}
  and
  \deqn{%
    p(x) = \frac{(1-p_0)}{(1 - (1-p)^n)} f(x)}{%
    p(x) = (1-p0)/[1 - (1-p)^n] f(x)}
  for \eqn{x = 1, \ldots, n}, \eqn{0 < p \le 1} and \eqn{0 \le
  p_0 \le 1}{0 \le p0 \le 1}, where \eqn{f(x)} is the probability mass
  function of the binomial.
  The cumulative distribution function is
  \deqn{P(x) = p_0 + (1 - p_0) \left(\frac{F(x) - F(0)}{1 - F(0)}\right)}{%
    P(x) = p0 + (1 - p0) [F(x) - F(0)]/[1 - F(0)].}

  The mean is \eqn{(1-p_0) \mu}{(1-p0)m} and the variance is
  \eqn{(1-p_0) \sigma^2 + p_0(1-p_0) \mu^2}{(1-p0)v + p0(1-p0)m^2},
  where \eqn{\mu}{m} and \eqn{\sigma^2}{v} are the mean and variance of
  the zero-truncated binomial.

  In the terminology of Klugman et al. (2012), the zero-modified
  binomial is a member of the \eqn{(a, b, 1)} class of
  distributions with \eqn{a = -p/(1-p)} and \eqn{b = (n+1)p/(1-p)}.

  The special case \code{p0 == 0} is the zero-truncated binomial.

  If an element of \code{x} is not integer, the result of
  \code{dzmbinom} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dzmbinom} gives the probability mass function,
  \code{pzmbinom} gives the distribution function,
  \code{qzmbinom} gives the quantile function, and
  \code{rzmbinom} generates random deviates.

  Invalid \code{size}, \code{prob} or \code{p0} will result in return
  value \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rzmbinom}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}zmbinom} use \code{\{d,p,q\}binom} for all
  but the trivial input values and \eqn{p(0)}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dbinom}} for the binomial distribution.

  \code{\link{dztbinom}} for the zero-truncated binomial distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
dzmbinom(1:5, size = 5, prob = 0.4, p0 = 0.2)
(1-0.2) * dbinom(1:5, 5, 0.4)/pbinom(0, 5, 0.4, lower = FALSE) # same

## simple relation between survival functions
pzmbinom(0:5, 5, 0.4, p0 = 0.2, lower = FALSE)
(1-0.2) * pbinom(0:5, 5, 0.4, lower = FALSE) /
    pbinom(0, 5, 0.4, lower = FALSE) # same

qzmbinom(pzmbinom(1:10, 10, 0.6, p0 = 0.1), 10, 0.6, p0 = 0.1)

n <- 8; p <- 0.3; p0 <- 0.025
x <- 0:n
title <- paste("ZM Binomial(", n, ", ", p, ", p0 = ", p0,
               ") and Binomial(", n, ", ", p,") PDF",
               sep = "")
plot(x, dzmbinom(x, n, p, p0), type = "h", lwd = 2, ylab = "p(x)",
     main = title)
points(x, dbinom(x, n, p), pch = 19, col = "red")
legend("topright", c("ZT binomial probabilities", "Binomial probabilities"),
       col = c("black", "red"), lty = c(1, 0), lwd = 2, pch = c(NA, 19))
}
\keyword{distribution}

\name{InverseParalogistic}
\alias{InverseParalogistic}
\alias{dinvparalogis}
\alias{pinvparalogis}
\alias{qinvparalogis}
\alias{rinvparalogis}
\alias{minvparalogis}
\alias{levinvparalogis}
\title{The Inverse Paralogistic Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Inverse Paralogistic
  distribution with parameters \code{shape} and \code{scale}.
}
\usage{
dinvparalogis(x, shape, rate = 1, scale = 1/rate, log = FALSE)
pinvparalogis(q, shape, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
qinvparalogis(p, shape, rate = 1, scale = 1/rate,
              lower.tail = TRUE, log.p = FALSE)
rinvparalogis(n, shape, rate = 1, scale = 1/rate)
minvparalogis(order, shape, rate = 1, scale = 1/rate)
levinvparalogis(limit, shape, rate = 1, scale = 1/rate,
                order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The inverse paralogistic distribution with parameters \code{shape}
  \eqn{= \tau}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\tau^2 (x/\theta)^{\tau^2}}{%
      x [1 + (x/\theta)^\tau]^{\tau + 1}}}{%
    f(x) = a^2 (x/s)^(a^2)/(x [1 + (x/s)^a]^(a + 1))}
  for \eqn{x > 0}, \eqn{\tau > 0}{a > 0} and \eqn{\theta > 0}{b > 0}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{-\tau^2 < k < \tau}{-shape^2 < k <
  shape}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]}, \eqn{k > -\tau^2}{k > -shape^2}
  and \eqn{1 - k/\tau}{1 - k/shape} not a negative integer.
}
\value{
  \code{dinvparalogis} gives the density,
  \code{pinvparalogis} gives the distribution function,
  \code{qinvparalogis} gives the quantile function,
  \code{rinvparalogis} generates random deviates,
  \code{minvparalogis} gives the \eqn{k}th raw moment, and
  \code{levinvparalogis} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levinvparalogis} computes computes the limited expected value using
  \code{\link{betaint}}.

  See Kleiber and Kotz (2003) for alternative names and
  parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvparalogis(2, 3, 4, log = TRUE))
p <- (1:10)/10
pinvparalogis(qinvparalogis(p, 2, 3), 2, 3)

## first negative moment
minvparalogis(-1, 2, 2)

## case with 1 - order/shape > 0
levinvparalogis(10, 2, 2, order = 1)

## case with 1 - order/shape < 0
levinvparalogis(10, 2/3, 2, order = 1)
}
\keyword{distribution}

\name{FellerPareto}
\alias{FellerPareto}
\alias{dfpareto}
\alias{pfpareto}
\alias{qfpareto}
\alias{rfpareto}
\alias{mfpareto}
\alias{levfpareto}
\title{The Feller Pareto Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Feller Pareto distribution
  with parameters \code{min}, \code{shape1}, \code{shape2}, \code{shape3} and
  \code{scale}.
}
\usage{
dfpareto(x, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
        log = FALSE)
pfpareto(q, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
        lower.tail = TRUE, log.p = FALSE)
qfpareto(p, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
        lower.tail = TRUE, log.p = FALSE)
rfpareto(n, min, shape1, shape2, shape3, rate = 1, scale = 1/rate)
mfpareto(order, min, shape1, shape2, shape3, rate = 1, scale = 1/rate)
levfpareto(limit, min, shape1, shape2, shape3, rate = 1, scale = 1/rate,
          order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{min}{lower bound of the support of the distribution.}
  \item{shape1, shape2, shape3, scale}{parameters. Must be strictly
    positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Feller-Pareto distribution with parameters \code{min} \eqn{= \mu}{= m},
  \code{shape1} \eqn{= \alpha}{= a}, \code{shape2} \eqn{= \gamma}{= b},
  \code{shape3} \eqn{= \tau}{= c} and \code{scale} \eqn{= \theta}{= s}, has
  density:
  \deqn{f(x) = \frac{\Gamma(\alpha + \tau)}{\Gamma(\alpha)\Gamma(\tau)}
    \frac{\gamma ((x - \mu)/\theta)^{\gamma \tau - 1}}{%
      \theta [1 + ((x - \mu)/\theta)^\gamma]^{\alpha + \tau}}}{%
    f(x) = Gamma(a + c)/(Gamma(a) * Gamma(c)) (b ((x - m)/s)^(bc - 1))/%
    (s [1 + ((x - m)/s)^b]^(a + c))}
  for \eqn{x > \mu}{x > m}, \eqn{-\infty < \mu < \infty}{-Inf < m < Inf},
  \eqn{\alpha > 0}{a > 0}, \eqn{\gamma > 0}{b > 0},
  \eqn{\tau > 0}{c > 0} and \eqn{\theta > 0}{s > 0}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The Feller-Pareto is the distribution of the random variable
  \deqn{\mu + \theta \left(\frac{1 - X}{X}\right)^{1/\gamma},}{%
   m + s ((1 - X)/X)^(1/b),}
  where \eqn{X} has a beta distribution with parameters \eqn{\alpha}{a}
  and \eqn{\tau}{c}.

  The Feller-Pareto defines a large family of distributions encompassing
  the transformed beta family and many variants of the Pareto
  distribution. Setting \eqn{\mu = 0}{min = 0} yields the
  \link[=dtrbeta]{transformed beta} distribution.

  The Feller-Pareto distribution also has the following direct special
  cases:
  \itemize{
    \item A \link[=dpareto4]{Pareto IV} distribution when \code{shape3
      == 1};
    \item A \link[=dpareto3]{Pareto III} distribution when \code{shape1
      shape3 == 1};
    \item A \link[=dpareto2]{Pareto II} distribution when \code{shape1
      shape2 == 1};
    \item A \link[=dpareto1]{Pareto I} distribution when \code{shape1
      shape2 == 1} and \code{min = scale}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} for nonnegative integer values of \eqn{k <
  \alpha\gamma}{k < shape1 * shape2}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]} for nonnegative integer values of \eqn{k}
  and \eqn{\alpha - j/\gamma}{shape1 - j/shape2}, \eqn{j = 1, \dots, k}
  not a negative integer.

  Note that the range of admissible values for \eqn{k} in raw and
  limited moments is larger when \eqn{\mu = 0}{min == 0}.
}
\value{
  \code{dfpareto} gives the density,
  \code{pfpareto} gives the distribution function,
  \code{qfpareto} gives the quantile function,
  \code{rfpareto} generates random deviates,
  \code{mfpareto} gives the \eqn{k}th raw moment, and
  \code{levfpareto} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levfpareto} computes the limited expected value using
  \code{\link{betaint}}.

  For the Feller-Pareto and other Pareto distributions, we use the
  classification of Arnold (2015) with the parametrization of Klugman et
  al. (2012).  

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Dutang, C., Goulet, V., Langevin, N. (2022).
  Feller-Pareto and Related Distributions: Numerical Implementation and Actuarial
  Applications.
  \emph{Journal of Statistical Software}, \bold{103}(6), 1--22.
  \doi{10.18637/jss.v103.i06}.

  Abramowitz, M. and Stegun, I. A. (1972), \emph{Handbook of 
  Mathematical Functions}, Dover.
  
  Arnold, B. C. (2015), \emph{Pareto Distributions}, Second Edition, CRC
  Press.

  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dtrbeta}} for the transformed beta distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Nicholas Langevin
}
\examples{
exp(dfpareto(2, 1, 2, 3, 4, 5, log = TRUE))
p <- (1:10)/10
pfpareto(qfpareto(p, 1, 2, 3, 4, 5), 1, 2, 3, 4, 5)

## variance
mfpareto(2, 1, 2, 3, 4, 5) - mfpareto(1, 1, 2, 3, 4, 5)^2

## case with shape1 - order/shape2 > 0
levfpareto(10, 1, 2, 3, 4, scale = 1, order = 2)

## case with shape1 - order/shape2 < 0
levfpareto(20, 10, 0.1, 14, 2, scale = 1.5, order = 2)
}
\keyword{distribution}

\name{VaR}
\alias{VaR}
\title{Value at Risk}
\description{
  Value at Risk.
}
\usage{
VaR(x, \dots)
}  
\arguments{
  \item{x}{an \R object.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  This is a generic function with, currently, only a method for objects
  of class \code{"aggregateDist"}.
}
\value{
  An object of class \code{numeric}.
}
\seealso{
  \code{\link{VaR.aggregateDist}}, \code{\link{aggregateDist}}
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and 
  Tommy Ouellet
}
\keyword{univar}

\name{ZeroModifiedGeometric}
\alias{ZeroModifiedGeometric}
\alias{Zmgeometric}
\alias{dzmgeom}
\alias{pzmgeom}
\alias{qzmgeom}
\alias{rzmgeom}
\title{The Zero-Modified Geometric Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Modified Geometric distribution with
  parameter \code{prob} and arbitrary probability at zero \code{p0}.
}
\usage{
dzmgeom(x, prob, p0, log = FALSE)
pzmgeom(q, prob, p0, lower.tail = TRUE, log.p = FALSE)
qzmgeom(p, prob, p0, lower.tail = TRUE, log.p = FALSE)
rzmgeom(n, prob, p0)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{prob}{parameter. \code{0 < prob <= 1}.}
  \item{p0}{probability mass at zero. \code{0 <= p0 <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-modified geometric distribution with \code{prob} \eqn{= p}
  and \code{p0} \eqn{= p_0}{= p0} is a discrete mixture between a
  degenerate distribution at zero and a (standard) geometric. The
  probability mass function is \eqn{p(0) = p_0}{p(0) = p0} and
  \deqn{%
    p(x) = \frac{(1-p_0)}{(1-p)} f(x)}{%
    p(x) = (1-p0)/(1-p) f(x)}
  for \eqn{x = 1, 2, \ldots}, \eqn{0 < p < 1} and \eqn{0 \le
  p_0 \le 1}{0 \le p0 \le 1}, where \eqn{f(x)} is the probability mass
  function of the geometric.
  The cumulative distribution function is
  \deqn{P(x) = p_0 + (1 - p_0) \left(\frac{F(x) - F(0)}{1 - F(0)}\right)}{%
    P(x) = p0 + (1 - p0) [F(x) - F(0)]/[1 - F(0)].}

  The mean is \eqn{(1-p_0) \mu}{(1-p0)m} and the variance is
  \eqn{(1-p_0) \sigma^2 + p_0(1-p_0) \mu^2}{(1-p0)v + p0(1-p0)m^2},
  where \eqn{\mu}{m} and \eqn{\sigma^2}{v} are the mean and variance of
  the zero-truncated geometric.

  In the terminology of Klugman et al. (2012), the zero-modified
  geometric is a member of the \eqn{(a, b, 1)} class of
  distributions with \eqn{a = 1-p} and \eqn{b = 0}.

  The special case \code{p0 == 0} is the zero-truncated geometric.

  If an element of \code{x} is not integer, the result of
  \code{dzmgeom} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dzmgeom} gives the (log) probability mass function,
  \code{pzmgeom} gives the (log) distribution function,
  \code{qzmgeom} gives the quantile function, and
  \code{rzmgeom} generates random deviates.

  Invalid \code{prob} or \code{p0} will result in return value
  \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rzmgeom}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}zmgeom} use \code{\{d,p,q\}geom} for all but
  the trivial input values and \eqn{p(0)}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dgeom}} for the geometric distribution.

  \code{\link{dztgeom}} for the zero-truncated geometric distribution.

  \code{\link{dzmnbinom}} for the zero-modified negative binomial, of
  which the zero-modified geometric is a special case.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
p <- 1/(1 + 0.5)
dzmgeom(1:5, prob = p, p0 = 0.6)
(1-0.6) * dgeom(1:5, p)/pgeom(0, p, lower = FALSE) # same

## simple relation between survival functions
pzmgeom(0:5, p, p0 = 0.2, lower = FALSE)
(1-0.2) * pgeom(0:5, p, lower = FALSE)/pgeom(0, p, lower = FALSE) # same

qzmgeom(pzmgeom(0:10, 0.3, p0 = 0.6), 0.3, p0 = 0.6)
}
\keyword{distribution}

\name{UniformSupp}
\alias{UniformSupp}
\alias{munif}
\alias{levunif}
\alias{mgfunif}
\title{Moments and Moment Generating Function of the Uniform Distribution}
\description{
  Raw moments, limited moments and moment generating function for the
  Uniform distribution from \code{min} to \code{max}.
}
\usage{
munif(order, min = 0, max = 1)
levunif(limit, min = 0, max =1, order = 1)
mgfunif(t, min = 0, max = 1, log = FALSE)
}
\arguments{
  \item{order}{order of the moment.}
  \item{min, max}{lower and upper limits of the distribution. Must be finite.}
  \item{limit}{limit of the random variable.}
  \item{t}{numeric vector.}
  \item{log}{logical; if \code{TRUE}, the cumulant generating function
    is returned.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]} and the moment
  generating function is \eqn{E[e^{tX}]}.
}
\value{
  \code{munif} gives the \eqn{k}th raw moment,
  \code{levunif} gives the \eqn{k}th moment of the limited random
  variable, and
  \code{mgfunif} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link{Uniform}}.
}
\references{
  \url{https://en.wikipedia.org/wiki/Uniform_distribution_\%28continuous\%29}
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Christophe Dutang
}
\examples{
munif(-1)
munif(1:5)
levunif(3, order=1:5)
levunif(3, 2, 4)
mgfunif(1, 1, 2)
}
\keyword{distribution}

\name{Pareto}
\alias{Pareto}
\alias{dpareto}
\alias{ppareto}
\alias{qpareto}
\alias{rpareto}
\alias{mpareto}
\alias{levpareto}
\title{The Pareto Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Pareto distribution with
  parameters \code{shape} and \code{scale}.
}
\usage{
dpareto(x, shape, scale, log = FALSE)
ppareto(q, shape, scale, lower.tail = TRUE, log.p = FALSE)
qpareto(p, shape, scale, lower.tail = TRUE, log.p = FALSE)
rpareto(n, shape, scale)
mpareto(order, shape, scale)
levpareto(limit, shape, scale, order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Pareto distribution with parameters \code{shape} \eqn{=
    \alpha}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\alpha \theta^\alpha}{(x + \theta)^{\alpha + 1}}}{%
    f(x) = a s^a / (x + s)^(a + 1)}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0} and \eqn{\theta}{s > 0}.

  There are many different definitions of the Pareto distribution in the
  literature; see Arnold (2015) or Kleiber and Kotz (2003). In the
  nomenclature of \pkg{actuar}, The \dQuote{Pareto distribution} does
  not have a location parameter. The version with a location parameter
  is the \link[=dpareto2]{Pareto II}.
  
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}, \eqn{-1 < k < \alpha}{-1 < k < shape}.

  The \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]},
  \eqn{k > -1} and \eqn{\alpha - k}{shape - k} not a
  negative integer.
}
\value{
  \code{dpareto} gives the density,
  \code{ppareto} gives the distribution function,
  \code{qpareto} gives the quantile function,
  \code{rpareto} generates random deviates,
  \code{mpareto} gives the \eqn{k}th raw moment, and
  \code{levpareto} gives the \eqn{k}th moment of the limited loss variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levpareto} computes the limited expected value using
  \code{\link{betaint}}.

  The version of the Pareto defined for \eqn{x > \theta}{x > s} is named
  Single Parameter Pareto, or Pareto I, in \pkg{actuar}.  
}
\seealso{
  \code{\link{dpareto2}} for an equivalent distribution with location
  parameter.

  \code{\link{dpareto1}} for the Single Parameter Pareto distribution.

  \code{"distributions"} package vignette for details on the
  interrelations between the continuous size distributions in
  \pkg{actuar} and complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dpareto(2, 3, 4, log = TRUE))
p <- (1:10)/10
ppareto(qpareto(p, 2, 3), 2, 3)

## variance
mpareto(2, 4, 1) - mpareto(1, 4, 1)^2

## case with shape - order > 0
levpareto(10, 3, scale = 1, order = 2)

## case with shape - order < 0
levpareto(10, 1.5, scale = 1, order = 2)
}
\keyword{distribution}

\name{Logarithmic}
\alias{Logarithmic}
\alias{dlogarithmic}
\alias{plogarithmic}
\alias{qlogarithmic}
\alias{rlogarithmic}
\alias{log-series}
\title{The Logarithmic Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Logarithmic (or log-series) distribution with parameter
  \code{prob}.
}
\usage{
dlogarithmic(x, prob, log = FALSE)
plogarithmic(q, prob, lower.tail = TRUE, log.p = FALSE)
qlogarithmic(p, prob, lower.tail = TRUE, log.p = FALSE)
rlogarithmic(n, prob)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{prob}{parameter. \code{0 <= prob < 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The logarithmic (or log-series) distribution with parameter
  \code{prob} \eqn{= \theta}{= p} has probability mass function
  \deqn{%
    p(x) = \frac{a \theta^x}{x},}{%
    p(x) = a p^x / x,}
  with \eqn{a = -1/\log(1 - \theta)}{a = -1/log(1-p)} and for \eqn{x =
    1, 2, \ldots}, \eqn{0 \le \theta < 1}{0 \le p < 1}.

  The logarithmic distribution is the limiting case of the
  zero-truncated negative binomial distribution with \code{size}
  parameter equal to \eqn{0}. Note that in this context, parameter
  \code{prob} generally corresponds to the probability of \emph{failure}
  of the zero-truncated negative binomial.

  If an element of \code{x} is not integer, the result of
  \code{dlogarithmic} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{F(x) \ge p}, where \eqn{F} is the distribution function.
}
\value{
  \code{dlogarithmic} gives the probability mass function,
  \code{plogarithmic} gives the distribution function,
  \code{qlogarithmic} gives the quantile function, and
  \code{rlogarithmic} generates random deviates.

  Invalid \code{prob} will result in return value \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rlogarithmic}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  \code{qlogarithmic} is based on \code{qbinom} et al.; it uses the
  Cornish--Fisher Expansion to include a skewness correction to a normal
  approximation, followed by a search.

  \code{rlogarithmic} is an implementation of the LS and LK algorithms
  of Kemp (1981) with automatic selection. As suggested by Devroye
  (1986), the LS algorithm is used when \code{prob < 0.95}, and the LK
  algorithm otherwise.
}
\references{
  Johnson, N. L., Kemp, A. W. and Kotz, S. (2005), \emph{Univariate
  Discrete Distributions, Third Edition}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.

  Kemp, A. W. (1981), \dQuote{Efficient Generation of Logarithmically
  Distributed Pseudo-Random Variables}, \emph{Journal of the Royal
  Statistical Society, Series C}, vol. 30, p. 249-253.

  Devroye, L. (1986), \emph{Non-Uniform Random Variate Generation},
  Springer-Verlag. \url{https://luc.devroye.org/rnbookindex.html}
}
\seealso{
  \code{\link{dztnbinom}} for the zero-truncated negative binomial
  distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
## Table 1 of Kemp (1981) [also found in Johnson et al. (2005), chapter 7]
p <- c(0.1, 0.3, 0.5, 0.7, 0.8, 0.85, 0.9, 0.95, 0.99, 0.995, 0.999, 0.9999)
round(rbind(dlogarithmic(1, p),
            dlogarithmic(2, p),
            plogarithmic(9, p, lower.tail = FALSE),
            -p/((1 - p) * log(1 - p))), 2)

qlogarithmic(plogarithmic(1:10, 0.9), 0.9)

x <- rlogarithmic(1000, 0.8)
y <- sort(unique(x))
plot(y, table(x)/length(x), type = "h", lwd = 2,
     pch = 19, col = "black", xlab = "x", ylab = "p(x)",
     main = "Empirical vs theoretical probabilities")
points(y, dlogarithmic(y, prob = 0.8),
       pch = 19, col = "red")
legend("topright", c("empirical", "theoretical"),
       lty = c(1, NA), pch = c(NA, 19), col = c("black", "red"))
}
\keyword{distribution}

\name{severity}
\alias{severity}
\alias{severity.default}
\title{Manipulation of Individual Claim Amounts}
\description{
  \code{severity} is a generic function created to manipulate individual
  claim amounts. The function invokes particular \emph{methods} which
  depend on the \code{\link[base]{class}} of the first argument.
}
\usage{
severity(x, ...)

\method{severity}{default}(x, bycol = FALSE, drop = TRUE, \dots)
}
\arguments{
  \item{x}{an \R object.}
  \item{bycol}{logical; whether to \dQuote{unroll} horizontally
    (\code{FALSE}) or vertically (\code{TRUE})}
  \item{\dots}{further arguments to be passed to or from other methods.}
  \item{drop}{logical; if \code{TRUE}, the result is coerced to the
    lowest possible dimension.}
}
\details{
  Currently, the default method is equivalent to
  \code{\link{unroll}}. This is liable to change since the link between
  the name and the use of the function is rather weak.
}
\value{
  A vector or matrix.
}
\seealso{
  \code{\link{severity.portfolio}} for the original motivation of these
  functions.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Louis-Philippe Pouliot
}
\examples{
x <- list(c(1:3), c(1:8), c(1:4), c(1:3))
(mat <- matrix(x, 2, 2))
severity(mat)
severity(mat, bycol = TRUE)
}
\keyword{datagen}
\keyword{manip}

\name{InverseWeibull}
\alias{InverseWeibull}
\alias{dinvweibull}
\alias{pinvweibull}
\alias{qinvweibull}
\alias{rinvweibull}
\alias{minvweibull}
\alias{levinvweibull}
\alias{dlgompertz}
\alias{plgompertz}
\alias{qlgompertz}
\alias{rlgompertz}
\alias{mlgompertz}
\alias{levlgompertz}
\title{The Inverse Weibull Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Inverse Weibull distribution
  with parameters \code{shape} and \code{scale}.
}
\usage{
dinvweibull(x, shape, rate = 1, scale = 1/rate, log = FALSE)
pinvweibull(q, shape, rate = 1, scale = 1/rate,
            lower.tail = TRUE, log.p = FALSE)
qinvweibull(p, shape, rate = 1, scale = 1/rate,
            lower.tail = TRUE, log.p = FALSE)
rinvweibull(n, shape, rate = 1, scale = 1/rate)
minvweibull(order, shape, rate = 1, scale = 1/rate)
levinvweibull(limit, shape, rate = 1, scale = 1/rate,
              order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The inverse Weibull distribution with parameters \code{shape} \eqn{=
    \tau}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\tau (\theta/x)^\tau e^{-(\theta/x)^\tau}}{x}}{%
    f(x) = a (s/x)^a exp(-(s/x)^a)/x}
  for \eqn{x > 0}, \eqn{\tau > 0}{a > 0} and \eqn{\theta > 0}{s > 0}.

  The special case \code{shape == 1} is an
  \link[=dinvexp]{Inverse Exponential} distribution.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{k < \tau}{k < shape}, and the \eqn{k}th
  limited moment at some limit \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X,
  d)^k]}, all \eqn{k}.
}
\value{
  \code{dinvweibull} gives the density,
  \code{pinvweibull} gives the distribution function,
  \code{qinvweibull} gives the quantile function,
  \code{rinvweibull} generates random deviates,
  \code{minvweibull} gives the \eqn{k}th raw moment, and
  \code{levinvweibull} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levinvweibull} computes the limited expected value using
  \code{gammainc} from package \pkg{expint}.

  Distribution also knonw as the log-Gompertz. See also Kleiber and
  Kotz (2003) for alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvweibull(2, 3, 4, log = TRUE))
p <- (1:10)/10
pinvweibull(qinvweibull(p, 2, 3), 2, 3)
mlgompertz(-1, 3, 3)
levinvweibull(10, 2, 3, order = 1)
}
\keyword{distribution}

\name{NormalSupp}
\alias{NormalSupp}
\alias{mnorm}
\alias{mgfnorm}
\title{Moments and Moment generating function of the Normal Distribution}
\description{
  Raw moments and moment generating function for the normal distribution with
  mean equal to \code{mean} and standard deviation equal to \code{sd}.
}
\usage{
mnorm(order, mean = 0, sd = 1)
mgfnorm(t, mean = 0, sd = 1, log = FALSE)
}
\arguments{
  \item{order}{vector of integers; order of the moment.}
  \item{mean}{vector of means.}
  \item{sd}{vector of standard deviations.}
  \item{t}{numeric vector.}
  \item{log}{logical; if \code{TRUE}, the cumulant generating function
    is returned.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the moment generating function is
  \eqn{E[e^{tX}]}.

  Only integer moments are supported.
}
\value{
  \code{mnorm} gives the \eqn{k}th raw moment and
  \code{mgfnorm} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link{Normal}}
}
\references{
  Johnson, N. L. and Kotz, S. (1970), \emph{Continuous Univariate
    Distributions, Volume 1}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Christophe Dutang
}
\examples{
mgfnorm(0:4,1,2)
mnorm(3)
}
\keyword{distribution}

\name{ogive}
\alias{ogive}
\alias{ogive.default}
\alias{ogive.grouped.data}
\alias{print.ogive}
\alias{summary.ogive}
\alias{knots.ogive}
\alias{plot.ogive}
\title{Ogive for Grouped Data}
\description{
  Compute a smoothed empirical distribution function for grouped data.
}
\usage{
ogive(x, \dots)

\method{ogive}{default}(x, y = NULL, breaks = "Sturges", nclass = NULL, \dots)

\method{ogive}{grouped.data}(x, \dots)

\method{print}{ogive}(x, digits = getOption("digits") - 2, \dots)

\method{summary}{ogive}(object, \dots)

\method{knots}{ogive}(Fn, \dots)

\method{plot}{ogive}(x, main = NULL, xlab = "x", ylab = "F(x)", \dots)
}
\arguments{
  \item{x}{for the generic and all but the default method, an object of
    class \code{"grouped.data"}; for the default method, a vector of
    individual data if \code{y} is \code{NULL}, a vector of group
    boundaries otherwise.}
  \item{y}{a vector of group frequencies.}
  \item{breaks, nclass}{arguments passed to \code{\link{grouped.data}};
    used only for individual data (when \code{y} is \code{NULL}).}
  \item{digits}{number of significant digits to use, see
    \code{\link{print}}.}
  \item{Fn, object}{an \R object inheriting from \code{"ogive"}.}
  \item{main}{main title.}
  \item{xlab, ylab}{labels of x and y axis.}
  \item{\dots}{arguments to be passed to subsequent methods.}
}
\details{
  The ogive is a linear interpolation of the empirical cumulative
  distribution function.

  The equation of the ogive is
  \deqn{G_n(x) = \frac{(c_j - x) F_n(c_{j - 1}) +
      (x - c_{j - 1}) F_n(c_j)}{c_j - c_{j - 1}}}{%
    Gn(x) = 1/(c[j] - c[j-1]) * [(c[j] - x) Fn(c[j-1])
            + (x - c[j-1]) Fn(c[j])]}
  for \eqn{c_{j-1} < x \leq c_j}{c[j-1] < x <= c[j]} and where
  \eqn{c_0, \dots, c_r}{c[0], \dots, c[r]} are the \eqn{r + 1} group
  boundaries and \eqn{F_n}{Fn} is the empirical distribution function of
  the sample.
}
\value{
  For \code{ogive}, a function of class \code{"ogive"}, inheriting from the
  \code{"\link{function}"} class.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\seealso{
  \code{\link{grouped.data}} to create grouped data objects;
  \code{\link{quantile.grouped.data}} for the inverse function;
  \code{\link{approxfun}}, which is used to compute the ogive;
  \code{\link{stepfun}} for related documentation (even though the ogive
  is not a step function).
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
## Most common usage: create ogive from grouped data object.
Fn <- ogive(gdental)
Fn
summary(Fn)
knots(Fn)                      # the group boundaries

Fn(knots(Fn))                  # true values of the empirical cdf
Fn(c(80, 200, 2000))           # linear interpolations

plot(Fn)                       # graphical representation

## Alternative 1: create ogive directly from individual data
## without first creating a grouped data object.
ogive(dental)                  # automatic class boundaries
ogive(dental, breaks = c(0, 50, 200, 500, 1500, 2000))

## Alternative 2: create ogive from set of group boundaries and
## group frequencies.
cj <- c(0, 25, 50, 100, 250, 500, 1000)
nj <- c(30, 31, 57, 42, 45, 10)
ogive(cj, nj)
}
\keyword{dplot}
\keyword{hplot}

\name{ZeroTruncatedNegativeBinomial}
\alias{ZeroTruncatedNegativeBinomial}
\alias{ZTNegativeBinomial}
\alias{ZTNegBinomial}
\alias{dztnbinom}
\alias{pztnbinom}
\alias{qztnbinom}
\alias{rztnbinom}
\title{The Zero-Truncated Negative Binomial Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Truncated Negative Binomial distribution with
  parameters \code{size} and \code{prob}.
}
\usage{
dztnbinom(x, size, prob, log = FALSE)
pztnbinom(q, size, prob, lower.tail = TRUE, log.p = FALSE)
qztnbinom(p, size, prob, lower.tail = TRUE, log.p = FALSE)
rztnbinom(n, size, prob)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{size}{target for number of successful trials, or dispersion
    parameter. Must be positive, need not be integer.}
  \item{prob}{parameter. \code{0 < prob <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-truncated negative binomial distribution with \code{size}
  \eqn{= r} and \code{prob} \eqn{= p} has probability mass function
  \deqn{%
    p(x) = \frac{\Gamma(x + r) p^r (1 - p)^x}{\Gamma(r) x! (1 - p^r)}}{%
    p(x) = [\Gamma(x+r) p^r (1-p)^x]/[\Gamma(n) x! (1-p^r)]}
  for \eqn{x = 1, 2, \ldots}, \eqn{r \ge 0} and \eqn{0 < p < 1}, and
  \eqn{p(1) = 1} when \eqn{p = 1}.
  The cumulative distribution function is
  \deqn{P(x) = \frac{F(x) - F(0)}{1 - F(0)},}{%
    P(x) = [F(x) - F(0)]/[1 - F(0)],}
  where \eqn{F(x)} is the distribution function of the standard negative
  binomial.

  The mean is \eqn{r(1-p)/(p(1-p^r))} and the variance is
  \eqn{[r(1-p)(1 - (1 + r(1-p))p^r)]/[p(1-p^r)]^2}.

  In the terminology of Klugman et al. (2012), the zero-truncated
  negative binomial is a member of the \eqn{(a, b, 1)} class of
  distributions with \eqn{a = 1-p} and \eqn{b = (r-1)(1-p)}.

  The limiting case \code{size == 0} is the
  \link[=Logarithmic]{logarithmic} distribution with parameter \code{1 -
  prob}.

  Unlike the standard negative binomial functions, parametrization
  through the mean \code{mu} is not supported to avoid ambiguity as
  to whether \code{mu} is the mean of the underlying negative binomial
  or the mean of the zero-truncated distribution.

  If an element of \code{x} is not integer, the result of
  \code{dztnbinom} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dztnbinom} gives the (log) probability mass function,
  \code{pztnbinom} gives the (log) distribution function,
  \code{qztnbinom} gives the quantile function, and
  \code{rztnbinom} generates random deviates.

  Invalid \code{size} or \code{prob} will result in return value
  \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rztnbinom}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}ztnbinom} use \code{\{d,p,q\}nbinom} for all
  but the trivial input values and \eqn{p(0)}.

  \code{rztnbinom} uses the simple inversion algorithm suggested by
  Peter Dalgaard on the r-help mailing list on 1 May 2005 %
  (\url{https://stat.ethz.ch/pipermail/r-help/2005-May/070680.html}).
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dnbinom}} for the negative binomial distribution.

  \code{\link{dztgeom}} for the zero-truncated geometric and
  \code{\link{dlogarithmic}} for the logarithmic, which are special
  cases of the zero-truncated negative binomial.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
## Example 6.3 of Klugman et al. (2012)
p <- 1/(1 + 0.5)
dztnbinom(c(1, 2, 3), size = 2.5, prob = p)
dnbinom(c(1, 2, 3), 2.5, p)/pnbinom(0, 2.5, p, lower = FALSE) # same

pztnbinom(1, 2, prob = 1)        # point mass at 1
dztnbinom(2, size = 1, 0.25)     # == dztgeom(2, 0.25)
dztnbinom(2, size = 0, 0.25)     # == dlogarithmic(2, 0.75)

qztnbinom(pztnbinom(1:10, 2.5, 0.3), 2.5, 0.3)

x <- rztnbinom(1000, size = 2.5, prob = 0.4)
y <- sort(unique(x))
plot(y, table(x)/length(x), type = "h", lwd = 2,
     pch = 19, col = "black", xlab = "x", ylab = "p(x)",
     main = "Empirical vs theoretical probabilities")
points(y, dztnbinom(y, size = 2.5, prob = 0.4),
       pch = 19, col = "red")
legend("topright", c("empirical", "theoretical"),
       lty = c(1, NA), lwd = 2, pch = c(NA, 19), col = c("black", "red"))
}
\keyword{distribution}

\name{discretize}
\alias{discretize}
\alias{discretise}
\title{Discretization of a Continuous Distribution}
\description{
  Compute a discrete probability mass function from a continuous
  cumulative distribution function (cdf) with various methods.

  \code{discretise} is an alias for \code{discretize}.
}
\usage{
discretize(cdf, from, to, step = 1,
           method = c("upper", "lower", "rounding", "unbiased"),
           lev, by = step, xlim = NULL)

discretise(cdf, from, to, step = 1,
           method = c("upper", "lower", "rounding", "unbiased"),
           lev, by = step, xlim = NULL)
}
\arguments{
  \item{cdf}{an expression written as a function of \code{x}, or
    alternatively the name of a function, giving the cdf to discretize.}
  \item{from, to}{the range over which the function will be discretized.}
  \item{step}{numeric; the discretization step (or span, or lag).}
  \item{method}{discretization method to use.}
  \item{lev}{an expression written as a function of \code{x}, or
    alternatively the name of a function, to compute the limited
    expected value of the distribution corresponding to
    \code{cdf}. Used only with the \code{"unbiased"} method.}
  \item{by}{an alias for \code{step}.}
  \item{xlim}{numeric of length 2; if specified, it serves as default
    for \code{c(from, to)}.}
}
\details{
  Usage is similar to \code{\link{curve}}.

  \code{discretize} returns the probability mass function (pmf) of the
  random variable obtained by discretization of the cdf specified in
  \code{cdf}.

  Let \eqn{F(x)} denote the cdf, \eqn{E[\min(X, x)]}{E[min(X, x)]]} the
  limited expected value at \eqn{x}, \eqn{h} the step, \eqn{p_x}{p[x]}
  the probability mass at \eqn{x} in the discretized distribution and
  set \eqn{a =} \code{from} and \eqn{b =} \code{to}.

  Method \code{"upper"} is the forward difference of the cdf \eqn{F}:
  \deqn{p_x = F(x + h) - F(x)}{p[x] = F(x + h) - F(x)}
  for \eqn{x = a, a + h, \dots, b - step}.

  Method \code{"lower"} is the backward difference of the cdf \eqn{F}:
  \deqn{p_x = F(x) - F(x - h)}{p[x] = F(x) - F(x - h)} for \eqn{x = a +
    h, \dots, b} and \eqn{p_a = F(a)}{p[a] = F(a)}.

  Method \code{"rounding"} has the true cdf pass through the
  midpoints of the intervals \eqn{[x - h/2, x + h/2)}:
  \deqn{p_x = F(x + h/2) - F(x - h/2)}{p[x] = F(x + h/2) - F(x - h/2)}
  for \eqn{x = a + h, \dots, b - step} and \eqn{p_a = F(a + h/2)}{p[a] =
  F(a + h/2)}. The function assumes the cdf is continuous. Any
  adjusment necessary for discrete distributions can be done via
  \code{cdf}.

  Method \code{"unbiased"} matches the first moment of the discretized
  and the true distributions. The probabilities are as follows:
  \deqn{p_a = \frac{E[\min(X, a)] - E[\min(X, a + h)]}{h} + 1 - F(a)}{%
    p[a] = (E[min(X, a)] - E[min(X, a + h)])/h + 1 - F(a)}
  \deqn{p_x = \frac{2 E[\min(X, x)] - E[\min(X, x - h)] - E[\min(X, x +
      h)]}{h}, \quad a < x < b}{%
    p[x] = (2 E[min(X, x)] - E[min(X, x - h)] - E[min(X, x + h)])/h, a < x < b}
  \deqn{p_b = \frac{E[\min(X, b)] - E[\min(X, b - h)]}{h} - 1 + F(b),}{%
    p[b] = (E[min(X, b)] - E[min(X, b - h)])/h - 1 + F(b).}
}
\value{
  A numeric vector of probabilities suitable for use in
  \code{\link{aggregateDist}}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\seealso{
  \code{\link{aggregateDist}}
}
\examples{
x <- seq(0, 5, 0.5)

op <- par(mfrow = c(1, 1), col = "black")

## Upper and lower discretization
fu <- discretize(pgamma(x, 1), method = "upper",
                 from = 0, to = 5, step = 0.5)
fl <- discretize(pgamma(x, 1), method = "lower",
                 from = 0, to = 5, step = 0.5)
curve(pgamma(x, 1), xlim = c(0, 5))
par(col = "blue")
plot(stepfun(head(x, -1), diffinv(fu)), pch = 19, add = TRUE)
par(col = "green")
plot(stepfun(x, diffinv(fl)), pch = 19, add = TRUE)
par(col = "black")

## Rounding (or midpoint) discretization
fr <- discretize(pgamma(x, 1), method = "rounding",
                 from = 0, to = 5, step = 0.5)
curve(pgamma(x, 1), xlim = c(0, 5))
par(col = "blue")
plot(stepfun(head(x, -1), diffinv(fr)), pch = 19, add = TRUE)
par(col = "black")

## First moment matching
fb <- discretize(pgamma(x, 1), method = "unbiased",
                 lev = levgamma(x, 1), from = 0, to = 5, step = 0.5)
curve(pgamma(x, 1), xlim = c(0, 5))
par(col = "blue")
plot(stepfun(x, diffinv(fb)), pch = 19, add = TRUE)

par(op)
}
\keyword{distribution}
\keyword{models}

\name{rcomphierarc.summaries}
\alias{rcomphierarc.summaries}
\alias{aggregate.portfolio}
\alias{frequency.portfolio}
\alias{severity.portfolio}
\alias{weights.portfolio}
\title{Summary Statistics of a Portfolio}
\description{
  Methods for \link[base]{class} \code{"portfolio"} objects.

  \code{aggregate} splits portfolio data into subsets and computes
  summary statistics for each.

  \code{frequency} computes the frequency of claims for subsets of
  portfolio data.

  \code{severity} extracts the individual claim amounts.

  \code{weights} extracts the matrix of weights.
}
\usage{
\method{aggregate}{portfolio}(x, by = names(x$nodes), FUN = sum,
        classification = TRUE, prefix = NULL, \dots)

\method{frequency}{portfolio}(x, by = names(x$nodes),
        classification = TRUE, prefix = NULL, \dots)

\method{severity}{portfolio}(x, by = head(names(x$node), -1), splitcol = NULL,
        classification = TRUE, prefix = NULL, \dots)

\method{weights}{portfolio}(object, classification = TRUE, prefix = NULL, \dots)
}
\arguments{
  \item{x, object}{an object of class \code{"portfolio"}, typically
    created with \code{\link{simul}}.}
  \item{by}{character vector of grouping elements using the level names
    of the portfolio in \code{x}. The names can be abbreviated.}
  \item{FUN}{the function to be applied to data subsets.}
  \item{classification}{boolean; if \code{TRUE}, the node identifier
    columns are included in the output.}
  \item{prefix}{characters to prefix column names with; if \code{NULL},
        sensible defaults are used when appropriate.}
  \item{splitcol}{columns of the data matrix to extract separately;
    usual matrix indexing methods are supported.}
  \item{\dots}{optional arguments to \code{FUN}, or passed to or from
    other methods.}
}
\details{
  By default, \code{aggregate.portfolio} computes the aggregate claim amounts
  for the grouping specified in \code{by}. Any other statistic based on
  the individual claim amounts can be used through argument \code{FUN}.

  \code{frequency.portfolio} is equivalent to using \code{aggregate.portfolio}
  with argument \code{FUN} equal to \code{if (identical(x, NA)) NA else
    length(x)}.

  \code{severity.portfolio} extracts individual claim amounts of a portfolio
  by groupings using the default method of \code{\link{severity}}.
  Argument \code{splitcol} allows to get the individual claim amounts of
  specific columns separately.

  \code{weights.portfolio} extracts the weight matrix of a portfolio.
}
\value{
  A matrix or vector depending on the groupings specified in \code{by}.

  For the \code{aggregate} and \code{frequency} methods: if at least one
  level other than the last one is used for grouping, the result is a
  matrix obtained by binding the appropriate node identifiers extracted
  from \code{x$classification} if \code{classification = TRUE}, and the
  summaries per grouping. If the last level is used for grouping, the
  column names of \code{x$data} are retained; if the last level is not
  used for grouping, the column name is replaced by the deparsed name of
  \code{FUN}. If only the last level is used (column summaries), a named
  vector is returned.

  For the \code{severity} method: a list of two elements:
  \item{main}{\code{NULL} or a matrix of claim amounts for the
    columns not specified in \code{splitcol}, with the appropriate node
    identifiers extracted from \code{x$classification} if
    \code{classification = TRUE};}
  \item{split}{same as above, but for the columns specified in
    \code{splitcol}.}

  For the \code{weights} method: the weight matrix of the portfolio with
  node identifiers if \code{classification = TRUE}.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Louis-Philippe Pouliot.
}
\seealso{
  \code{\link{rcomphierarc}}
}
\examples{
nodes <- list(sector = 3, unit = c(3, 4),
              employer = c(3, 4, 3, 4, 2, 3, 4), year = 5)
model.freq <- expression(sector = rexp(1),
                         unit = rexp(sector),
                         employer = rgamma(unit, 1),
                         year = rpois(employer))
model.sev <- expression(sector = rnorm(6, 0.1),
                        unit = rnorm(sector, 1),
                        employer = rnorm(unit, 1),
                        year = rlnorm(employer, 1))
pf <- rcomphierarc(nodes, model.freq, model.sev)

aggregate(pf)            # aggregate claim amount by employer and year
aggregate(pf, classification = FALSE) # same, without node identifiers
aggregate(pf, by = "sector")	      # by sector
aggregate(pf, by = "y")		      # by year
aggregate(pf, by = c("s", "u"), mean) # average claim amount

frequency(pf)			      # number of claims
frequency(pf, prefix = "freq.")       # more explicit column names

severity(pf)			      # claim amounts by row
severity(pf, by = "year")	      # by column
severity(pf, by = c("s", "u"))        # by unit
severity(pf, splitcol = "year.5")     # last year separate
severity(pf, splitcol = 5)            # same
severity(pf, splitcol = c(FALSE, FALSE, FALSE, FALSE, TRUE)) # same

weights(pf)

## For portfolios with weights, the following computes loss ratios.
\dontrun{aggregate(pf, classif = FALSE) / weights(pf, classif = FALSE)}
}
\keyword{models}
\keyword{methods}

\name{ChisqSupp}
\alias{ChisqSupp}
\alias{mchisq}
\alias{levchisq}
\alias{mgfchisq}
\title{Moments and Moment Generating Function of the (non-central) Chi-Squared Distribution}
\description{
  Raw moments, limited moments and moment generating function for the
  chi-squared (\eqn{\chi^2}{chi^2}) distribution with \code{df} degrees
  of freedom and optional non-centrality parameter \code{ncp}.
}
\usage{
mchisq(order, df, ncp = 0)
levchisq(limit, df, ncp = 0, order = 1)
mgfchisq(t, df, ncp = 0, log= FALSE)
}
\arguments{
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{df}{degrees of freedom (non-negative, but can be non-integer).}
  \item{ncp}{non-centrality parameter (non-negative).}
  \item{t}{numeric vector.}
  \item{log}{logical; if \code{TRUE}, the cumulant generating function
    is returned.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)]}{E[min(X, d)]} and the moment generating
  function is \eqn{E[e^{tX}]}.

  Only integer moments are supported for the non central Chi-square
  distribution (\code{ncp > 0}).

  The limited expected value is supported for the centered Chi-square
  distribution (\code{ncp = 0}).
}
\value{
  \code{mchisq} gives the \eqn{k}th raw moment,
  \code{levchisq} gives the \eqn{k}th moment of the limited loss
  variable, and
  \code{mgfchisq} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link[stats]{Chisquare}}
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.

  Johnson, N. L. and Kotz, S. (1970), \emph{Continuous Univariate
    Distributions, Volume 1}, Wiley.
}
\author{
  Christophe Dutang, Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
mchisq(2, 3, 4)
levchisq(10, 3, order = 2)
mgfchisq(0.25, 3, 2)
}
\keyword{distribution}

\name{Pareto3}
\alias{Pareto3}
\alias{dpareto3}
\alias{ppareto3}
\alias{qpareto3}
\alias{rpareto3}
\alias{mpareto3}
\alias{levpareto3}
\title{The Pareto III Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Pareto III distribution with
  parameters \code{min}, \code{shape} and \code{scale}.
}
\usage{
dpareto3(x, min, shape, rate = 1, scale = 1/rate,
         log = FALSE)
ppareto3(q, min, shape, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
qpareto3(p, min, shape, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
rpareto3(n, min, shape, rate = 1, scale = 1/rate)
mpareto3(order, min, shape, rate = 1, scale = 1/rate)
levpareto3(limit, min, shape, rate = 1, scale = 1/rate,
           order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{min}{lower bound of the support of the distribution.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Pareto III (or \dQuote{type III}) distribution with parameters
  \code{min} \eqn{= \mu}{= m},
  \code{shape} \eqn{= \gamma}{= b} and
  \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\gamma ((x - \mu)/\theta)^{\gamma - 1}}{%
      \theta [1 + ((x - \mu)/\theta)^\gamma]^2}}{%
    f(x) = (b ((x - m)/s)^(b - 1))/(s [1 + ((x - m)/s)^b]^2)}
  for \eqn{x > \mu}{x > m}, \eqn{-\infty < \mu < \infty}{-Inf < m < Inf},
  \eqn{\gamma > 0}{b > 0} and \eqn{\theta > 0}{s > 0}.

  The Pareto III is the distribution of the random variable
  \deqn{\mu + \theta \left(\frac{X}{1 - X}\right)^{1/\gamma},}{%
    m + s (X/(1 - X))^(1/b),}
  where \eqn{X} has a uniform distribution on \eqn{(0, 1)}. It derives
  from the \link[=dfpareto]{Feller-Pareto} 
  distribution with \eqn{\alpha = \tau = 1}{shape1 = shape3 = 1}.
  Setting \eqn{\mu = 0}{min = 0} yields the \link[=dllogis]{loglogistic}
  distribution.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} for nonnegative integer values of \eqn{k <
    \gamma}{k < shape}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]} for nonnegative integer values of \eqn{k}
  and \eqn{1 - j/\gamma}{1 - j/shape}, \eqn{j = 1, \dots, k}
  not a negative integer.
}
\value{
  \code{dpareto3} gives the density,
  \code{ppareto3} gives the distribution function,
  \code{qpareto3} gives the quantile function,
  \code{rpareto3} generates random deviates,
  \code{mpareto3} gives the \eqn{k}th raw moment, and
  \code{levpareto3} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levpareto3} computes the limited expected value using
  \code{\link{betaint}}.

  For Pareto distributions, we use the classification of Arnold (2015)
  with the parametrization of Klugman et al. (2012).

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Arnold, B.C. (2015), \emph{Pareto Distributions}, Second Edition, CRC
  Press.

  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dllogis}} for the loglogistic distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
exp(dpareto3(1, min = 10, 3, 4, log = TRUE))
p <- (1:10)/10
ppareto3(qpareto3(p, min = 10, 2, 3), min = 10, 2, 3)

## mean
mpareto3(1, min = 10, 2, 3)

## case with 1 - order/shape > 0
levpareto3(20, min = 10, 2, 3, order = 1)

## case with 1 - order/shape < 0
levpareto3(20, min = 10, 2/3, 3, order = 1)
}
\keyword{distribution}

\name{InverseTransformedGamma}
\alias{InverseTransformedGamma}
\alias{dinvtrgamma}
\alias{pinvtrgamma}
\alias{qinvtrgamma}
\alias{rinvtrgamma}
\alias{minvtrgamma}
\alias{levinvtrgamma}
\title{The Inverse Transformed Gamma Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments, and limited moments for the Inverse Transformed Gamma
  distribution with parameters \code{shape1}, \code{shape2} and
  \code{scale}.
}
\usage{
dinvtrgamma(x, shape1, shape2, rate = 1, scale = 1/rate,
            log = FALSE)
pinvtrgamma(q, shape1, shape2, rate = 1, scale = 1/rate,
            lower.tail = TRUE, log.p = FALSE)
qinvtrgamma(p, shape1, shape2, rate = 1, scale = 1/rate,
            lower.tail = TRUE, log.p = FALSE)
rinvtrgamma(n, shape1, shape2, rate = 1, scale = 1/rate)
minvtrgamma(order, shape1, shape2, rate = 1, scale = 1/rate)
levinvtrgamma(limit, shape1, shape2, rate = 1, scale = 1/rate,
              order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The inverse transformed gamma distribution with parameters
  \code{shape1} \eqn{= \alpha}{= a}, \code{shape2} \eqn{= \tau}{= b} and
  \code{scale} \eqn{= \theta}{= s}, has density:
  \deqn{f(x) = \frac{\tau u^\alpha e^{-u}}{x \Gamma(\alpha)}, %
      \quad u = (\theta/x)^\tau}{%
    f(x) = b u^a exp(-u) / (x Gamma(a)), u = (s/x)^b}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0}, \eqn{\tau > 0}{b > 0}
  and \eqn{\theta > 0}{s > 0}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The inverse transformed gamma is the distribution of the random
  variable
  \eqn{\theta X^{-1/\tau},}{s X^(-1/b),}
  where \eqn{X} has a gamma distribution with shape parameter
  \eqn{\alpha}{a} and scale parameter \eqn{1} or, equivalently, of the
  random variable
  \eqn{Y^{-1/\tau}}{Y^(-1/b)}
  with \eqn{Y} a gamma distribution with shape parameter \eqn{\alpha}{a}
  and scale parameter \eqn{\theta^{-\tau}}{s^(-b)}.

  The inverse transformed gamma distribution defines a family of
  distributions with the following special cases:
  \itemize{
    \item An \link[=dinvgamma]{Inverse Gamma} distribution when
    \code{shape2 == 1};
    \item An \link[=dinvweibull]{Inverse Weibull} distribution when
    \code{shape1 == 1};
    \item An \link[=dinvexp]{Inverse Exponential} distribution when
    \code{shape1 == shape2 == 1};
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{k < \alpha\tau}{k < shape1 * shape2}, and
  the \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]} for all \eqn{k}.
}
\value{
  \code{dinvtrgamma} gives the density,
  \code{pinvtrgamma} gives the distribution function,
  \code{qinvtrgamma} gives the quantile function,
  \code{rinvtrgamma} generates random deviates,
  \code{minvtrgamma} gives the \eqn{k}th raw moment, and
  \code{levinvtrgamma} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levinvtrgamma} computes the limited expected value using
  \code{gammainc} from package \pkg{expint}.

  Distribution also known as the Inverse Generalized Gamma. See also Kleiber and
  Kotz (2003) for alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvtrgamma(2, 3, 4, 5, log = TRUE))
p <- (1:10)/10
pinvtrgamma(qinvtrgamma(p, 2, 3, 4), 2, 3, 4)
minvtrgamma(2, 3, 4, 5)
levinvtrgamma(200, 3, 4, 5, order = 2)
}
\keyword{distribution}

\name{gdental}
\docType{data}
\alias{gdental}
\title{Grouped Dental Claims Data Set}
\description{
  Grouped dental claims, that is presented in a number of claims per
  claim amount group form.
}
\usage{gdental}
\format{
  An object of class \code{"grouped.data"} (inheriting from class
  \code{"data.frame"}) consisting of 10 rows and 2 columns. The
  environment of the object contains the plain vector of \code{cj} of
  group boundaries
}
\source{Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\seealso{
  \code{\link{grouped.data}} for a description of grouped data objects.
}
\keyword{datasets}

\name{ZeroTruncatedPoisson}
\alias{ZeroTruncatedPoisson}
\alias{ZTPoisson}
\alias{dztpois}
\alias{pztpois}
\alias{qztpois}
\alias{rztpois}
\title{The Zero-Truncated Poisson Distribution}
\description{
  Density function, distribution function, quantile function, random
  generation for the Zero-Truncated Poisson distribution with parameter
  \code{lambda}.
}
\usage{
dztpois(x, lambda, log = FALSE)
pztpois(q, lambda, lower.tail = TRUE, log.p = FALSE)
qztpois(p, lambda, lower.tail = TRUE, log.p = FALSE)
rztpois(n, lambda)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of values to return.}
  \item{lambda}{vector of (non negative) means.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-truncated Poisson distribution has probability mass function
  \deqn{%
    p(x) = \frac{e^{-/lambda} \lambda^x}{x! (1 - e^{-\lambda})}
    = \frac{\lambda^x}{x! (e^{\lambda} - 1)}}{%
    p(x) = lambda^x exp(-lambda)/[x! (1 - exp(-lambda))]
         = lambda^x/[x! (e^lambda - 1)]}
  for \eqn{x = 1, 2, ...}, and \eqn{p(1) = 1} when \eqn{\lambda = 0}.
  The cumulative distribution function is
  \deqn{P(x) = \frac{F(x) - F(0)}{1 - F(0)},}{%
    P(x) = [F(x) - F(0)]/[1 - F(0)],}
  where \eqn{F(x)} is the distribution function of the standard Poisson.

  The mean is \eqn{\lambda/(1 - e^{-\lambda})}{\lambda/(1 -
    exp(-\lambda))} and the variance is
  \eqn{\lambda[1 - (\lambda+1)e^{-\lambda}]/(1 - e^{-\lambda})^2}{%
    \lambda[1 - (\lambda+1)exp(-\lambda)]/(1 - exp(-\lambda))^2}.

  In the terminology of Klugman et al. (2012), the zero-truncated
  Poisson is a member of the \eqn{(a, b, 1)} class of distributions
  with \eqn{a = 0} and \eqn{b = \lambda}.

  If an element of \code{x} is not integer, the result of
  \code{dztpois} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dztpois} gives the (log) probability mass function,
  \code{pztpois} gives the (log) distribution function,
  \code{qztpois} gives the quantile function, and
  \code{rztpois} generates random deviates.

  Invalid \code{lambda} will result in return value \code{NaN}, with a
  warning.

  The length of the result is determined by \code{n} for
  \code{rztpois}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}ztpois} use \code{\{d,p,q\}pois} for all
  but the trivial input values and \eqn{p(0)}.

  \code{rztpois} uses the simple inversion algorithm suggested by
  Peter Dalgaard on the r-help mailing list on 1 May 2005 %
  (\url{https://stat.ethz.ch/pipermail/r-help/2005-May/070680.html}).
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dpois}} for the standard Poisson distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
dztpois(1:5, lambda = 1)
dpois(1:5, lambda = 1)/ppois(0, 1, lower = FALSE) # same

pztpois(1, lambda = 0)          # point mass at 1

qztpois(pztpois(1:10, 1), 1)

x <- seq(0, 8)
plot(x, dztpois(x, 2), type = "h", lwd = 2, ylab = "p(x)",
     main = "Zero-Truncated Poisson(2) and Poisson(2) PDF")
points(x, dpois(x, 2), pch = 19, col = "red")
legend("topright", c("ZT Poisson probabilities", "Poisson probabilities"),
       col = c("black", "red"), lty = c(1, 0), lwd = 2, pch = c(NA, 19))
}
\keyword{distribution}

\name{BetaMoments}
\alias{BetaMoments}
\alias{mbeta}
\alias{levbeta}
\title{Raw and Limited Moments of the Beta Distribution}
\description{
  Raw moments and limited moments for the (central) Beta distribution
  with parameters \code{shape1} and \code{shape2}.
}
\usage{
mbeta(order, shape1, shape2)
levbeta(limit, shape1, shape2, order = 1)
}
\arguments{
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{shape1, shape2}{positive parameters of the Beta distribution.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]}, \eqn{k > -\alpha}{k
    > -shape1}.

  The noncentral beta distribution is not supported.
}
\value{
  \code{mbeta} gives the \eqn{k}th raw moment and
  \code{levbeta} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a
  warning.
}
\seealso{
  \code{\link[stats]{Beta}} for details on the beta distribution and
  functions \code{[dpqr]beta}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
mbeta(2, 3, 4) - mbeta(1, 3, 4)^2
levbeta(10, 3, 4, order = 2)
}
\keyword{distribution}

\name{rcompound}
\alias{rcompound}
\alias{rcomppois}
\title{Simulation from Compound Models}
\description{
  \code{rcompound} generates random variates from a compound model.

  \code{rcomppois} is a simplified version for a common case.
}
\usage{
rcompound(n, model.freq, model.sev, SIMPLIFY = TRUE)

rcomppois(n, lambda, model.sev, SIMPLIFY = TRUE)}
\arguments{
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{model.freq, model.sev}{expressions specifying the frequency and
    severity simulation models with the number of variates omitted; see
    Details.}
  \item{lambda}{Poisson parameter.}
  \item{SIMPLIFY}{boolean; if \code{FALSE} the frequency and severity
    variates are returned along with the aggregate variates.}
}
\details{
  \code{rcompound} generates variates from a random variable of the form
  \deqn{S = X_1 + ... + X_N,}
  where \eqn{N} is the frequency random variable and \eqn{X_1, X_2,
    \dots} are the severity random variables. The latter are mutually
  independent, identically distributed and independent from \eqn{N}.

  \code{model.freq} and \code{model.sev} specify the simulation models
  for the frequency and the severity random variables, respectively. A
  model is a complete call to a random number generation function, with
  the number of variates omitted. This is similar to
  \code{\link{rcomphierarc}}, but the calls need not be wrapped into
  \code{\link{expression}}. Either argument may also be the name of an
  object containing an expression, in which case the object will be
  evaluated in the parent frame to retrieve the expression.

  The argument of the random number generation functions for the number
  of variates to simulate \strong{must} be named \code{n}.

  \code{rcomppois} generates variates from the common Compound Poisson
  model, that is when random variable \eqn{N} is Poisson distributed
  with mean \code{lambda}.
}
\value{
  When \code{SIMPLIFY = TRUE}, a vector of aggregate amounts \eqn{S_1,
  \dots, S_n}. 

  When \code{SIMPLIFY = FALSE}, a list of three elements:
  \item{\code{aggregate}}{vector of aggregate amounts \eqn{S_1, \dots,
      S_n};}
  \item{\code{frequency}}{vector of frequencies \eqn{N_1, \dots,
      N_n};}
  \item{\code{severity}}{vector of severities \eqn{X_1, X_2, \dots}.}
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\seealso{
  \code{\link{rcomphierarc}} to simulate from compound hierarchical models.
}
\examples{
## Compound Poisson model with gamma severity.
rcompound(10, rpois(2), rgamma(2, 3))
rcomppois(10, 2, rgamma(2, 3))          # same

## Frequencies and individual claim amounts along with aggregate
## values.
rcomppois(10, 2, rgamma(2, 3), SIMPLIFY = FALSE)

## Wrapping the simulation models into expression() is allowed, but
## not needed.
rcompound(10, expression(rpois(2)), expression(rgamma(2, 3)))

\dontrun{## Speed comparison between rcompound() and rcomphierarc().
## [Also note the simpler syntax for rcompound().]
system.time(rcompound(1e6, rpois(2), rgamma(2, 3)))
system.time(rcomphierarc(1e6, expression(rpois(2)), expression(rgamma(2, 3))))}
## The severity can itself be a compound model. It makes sense
## in such a case to use a zero-truncated frequency distribution
## for the second level model.
rcomppois(10, 2,
          rcompound(rztnbinom(1.5, 0.7), rlnorm(1.2, 1)))
}
\keyword{datagen}

\name{ExponentialSupp}
\alias{ExponentialSupp}
\alias{mexp}
\alias{levexp}
\alias{mgfexp}
\title{Moments and Moment Generating Function of the Exponential Distribution}
\description{
  Raw moments, limited moments and moment generating function for the
  exponential distribution with rate \code{rate} (i.e., mean
  \code{1/rate}).
}
\usage{
mexp(order, rate = 1)
levexp(limit, rate = 1, order = 1)
mgfexp(t, rate = 1, log = FALSE)
}
\arguments{
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{rate}{vector of rates.}
  \item{t}{numeric vector.}
  \item{log}{logical; if \code{TRUE}, the cumulant generating function
    is returned.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]} and the moment
  generating function is \eqn{E[e^{tX}]}, \eqn{k > -1}.
}
\value{
  \code{mexp} gives the \eqn{k}th raw moment,
  \code{levexp} gives the \eqn{k}th moment of the limited loss
  variable, and
  \code{mgfexp} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link[stats]{Exponential}}
}
\references{
  Johnson, N. L. and Kotz, S. (1970), \emph{Continuous Univariate
    Distributions, Volume 1}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Christophe Dutang and Mathieu Pigeon.
}
\examples{
mexp(2, 3) - mexp(1, 3)^2
levexp(10, 3, order = 2)
mgfexp(1,2)
}
\keyword{distribution}

\name{adjCoef}
\alias{adjCoef}
\alias{plot.adjCoef}
\title{Adjustment Coefficient}
\description{
  Compute the adjustment coefficient in ruin theory, or return a
  function to compute the adjustment coefficient for various reinsurance
  retentions.
}
\usage{
adjCoef(mgf.claim, mgf.wait = mgfexp, premium.rate, upper.bound,
        h, reinsurance = c("none", "proportional", "excess-of-loss"),
        from, to, n = 101)

\method{plot}{adjCoef}(x, xlab = "x", ylab = "R(x)",
     main = "Adjustment Coefficient", sub = comment(x),
     type = "l", add = FALSE, \dots)
}
\arguments{
  \item{mgf.claim}{an expression written as a function of \code{x} or of
    \code{x} and \code{y}, or alternatively the name of a function,
    giving the moment generating function (mgf) of the claim severity
    distribution.}
  \item{mgf.wait}{an expression written as a function of \code{x}, or
    alternatively the name of a function, giving the mgf of the
    claims interarrival time distribution. Defaults to an exponential
    distribution with mean 1.}
  \item{premium.rate}{if \code{reinsurance = "none"}, a numeric value of
    the premium rate; otherwise, an expression written as a function of
    \code{y}, or alternatively the name of a function, giving the
    premium rate function.}
  \item{upper.bound}{numeric; an upper bound for the coefficient,
    usually the upper bound of the support of the claim severity mgf.}
  \item{h}{an expression written as a function of \code{x} or of
    \code{x} and \code{y}, or alternatively the name of a function,
    giving function \eqn{h} in the Lundberg equation (see below);
    ignored if \code{mgf.claim} is provided.}
  \item{reinsurance}{the type of reinsurance for the portfolio; can
     be abbreviated.}
  \item{from, to}{the range over which the adjustment coefficient will
    be calculated.}
  \item{n}{integer; the number of values at which to evaluate the
    adjustment coefficient.}
  \item{x}{an object of class \code{"adjCoef"}.}
  \item{xlab, ylab}{label of the x and y axes, respectively.}
  \item{main}{main title.}
  \item{sub}{subtitle, defaulting to the type of reinsurance.}
  \item{type}{1-character string giving the type of plot desired; see
    \code{\link[graphics]{plot}} for details.}
  \item{add}{logical; if \code{TRUE} add to already existing plot.}
  \item{\dots}{further graphical parameters accepted by
    \code{\link[graphics]{plot}} or \code{\link[graphics]{lines}}.}
}
\details{
  In the typical case \code{reinsurance = "none"}, the coefficient of
  determination is the smallest (strictly) positive root of the Lundberg
  equation%
  \deqn{h(x) = E[e^{x B - x c W}] = 1}{h(x) = E[exp(x B - x c W)] = 1}%
  on \eqn{[0, m)}, where \eqn{m =} \code{upper.bound}, \eqn{B} is the
  claim severity random variable, \eqn{W} is the claim interarrival
  (or wait) time random variable and \eqn{c =} \code{premium.rate}. The
  premium rate must satisfy the positive safety loading constraint
  \eqn{E[B - c W] < 0}.

  With \code{reinsurance = "proportional"}, the equation becomes
  \deqn{h(x, y) = E[e^{x y B - x c(y) W}] = 1,}{%
    h(x, y) = E[exp(x y B - x c(y) W)] = 1,}
  where \eqn{y} is the retention rate and \eqn{c(y)} is the premium rate
  function.

  With \code{reinsurance = "excess-of-loss"}, the equation becomes
  \deqn{h(x, y) = E[e^{x \min(B, y) - x c(y) W}] = 1,}{%
    h(x, y) = E[exp(x min(B, y) - x c(y) W)] = 1,}
  where \eqn{y} is the retention limit and \eqn{c(y)} is the premium rate
  function.

  One can use argument \code{h} as an alternative way to provide
  function \eqn{h(x)} or \eqn{h(x, y)}. This is necessary in cases where
  random variables \eqn{B} and \eqn{W} are not independent.

  The root of \eqn{h(x) = 1} is found by minimizing \eqn{(h(x) - 1)^2}.
}
\value{
  If \code{reinsurance = "none"}, a numeric vector of length one.
  Otherwise, a function of class \code{"adjCoef"} inheriting from the
  \code{"function"} class.
}
\references{
  Bowers, N. J. J., Gerber, H. U., Hickman, J., Jones, D. and Nesbitt, C.
  (1986), \emph{Actuarial Mathematics}, Society of Actuaries.

  Centeno, M. d. L. (2002), Measuring the effects of reinsurance by the
  adjustment coefficient in the Sparre-Anderson model, \emph{Insurance:
    Mathematics and Economics} \bold{30}, 37--49.

  Gerber, H. U. (1979), \emph{An Introduction to Mathematical Risk
    Theory}, Huebner Foundation.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2008),
  \emph{Loss Models, From Data to Decisions, Third Edition}, Wiley.
}
\author{
  Christophe Dutang, Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
## Basic example: no reinsurance, exponential claim severity and wait
## times, premium rate computed with expected value principle and
## safety loading of 20\%.
adjCoef(mgfexp, premium = 1.2, upper = 1)

## Same thing, giving function h.
h <- function(x) 1/((1 - x) * (1 + 1.2 * x))
adjCoef(h = h, upper = 1)

## Example 11.4 of Klugman et al. (2008)
mgfx <- function(x) 0.6 * exp(x) + 0.4 * exp(2 * x)
adjCoef(mgfx(x), mgfexp(x, 4), prem = 7, upper = 0.3182)

## Proportional reinsurance, same assumptions as above, reinsurer's
## safety loading of 30\%.
mgfx <- function(x, y) mgfexp(x * y)
p <- function(x) 1.3 * x - 0.1
h <- function(x, a) 1/((1 - a * x) * (1 + x * p(a)))
R1 <- adjCoef(mgfx, premium = p, upper = 1, reins = "proportional",
              from = 0, to = 1, n = 11)
R2 <- adjCoef(h = h, upper = 1, reins = "p",
             from = 0, to = 1, n = 101)
R1(seq(0, 1, length = 10))	# evaluation for various retention rates
R2(seq(0, 1, length = 10))	# same
plot(R1)		        # graphical representation
plot(R2, col = "green", add = TRUE) # smoother function

## Excess-of-loss reinsurance
p <- function(x) 1.3 * levgamma(x, 2, 2) - 0.1
mgfx <- function(x, l)
    mgfgamma(x, 2, 2) * pgamma(l, 2, 2 - x) +
    exp(x * l) * pgamma(l, 2, 2, lower = FALSE)
h <- function(x, l) mgfx(x, l) * mgfexp(-x * p(l))
R1 <- adjCoef(mgfx, upper = 1, premium = p, reins = "excess-of-loss",
             from = 0, to = 10, n = 11)
R2 <- adjCoef(h = h, upper = 1, reins = "e",
             from = 0, to = 10, n = 101)
plot(R1)
plot(R2, col = "green", add = TRUE)
}
\keyword{optimize}
\keyword{univar}

\name{mean.grouped.data}
\alias{mean.grouped.data}
\title{Arithmetic Mean}
\description{
  Mean of grouped data objects.
}
\usage{
\method{mean}{grouped.data}(x, \dots)
}
\arguments{
  \item{x}{an object of class \code{"grouped.data"}.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  The mean of grouped data with group boundaries \eqn{c_0, c_1, \dots,
    c_r}{c[0], c[1], \dots, c[r]} and group frequencies \eqn{n_1, \dots,
    n_r}{n[1], \dots, n[r]} is
  \deqn{\frac{1}{n} \sum_{j = 1}^r a_j n_j,}{%
    (1/n) * sum(j; a[j] * n[j]),}
  where
  \eqn{a_j = (c_{j - 1} + c_j)/2}{a[j] = (c[j - 1] + c[j])/2}
  is the midpoint of the \eqn{j}th interval, and
  \eqn{n = \sum_{j = 1}^r n_j}{n = sum(j; n[j])}.
}
\value{
  A named vector of means.
}
\seealso{
  \code{\link{grouped.data}} to create grouped data objects;
  \code{\link{emm}} to compute higher moments.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
data(gdental)
mean(gdental)
}
\keyword{univar}

\name{InverseBurr}
\alias{InverseBurr}
\alias{dinvburr}
\alias{pinvburr}
\alias{qinvburr}
\alias{rinvburr}
\alias{minvburr}
\alias{levinvburr}
\title{The Inverse Burr Distribution}
\description{
  Density function, distribution function, quantile function, random
  generation, raw moments and limited moments for the Inverse Burr
  distribution with parameters \code{shape1}, \code{shape2} and
  \code{scale}.
}
\usage{
dinvburr(x, shape1, shape2, rate = 1, scale = 1/rate,
         log = FALSE)
pinvburr(q, shape1, shape2, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
qinvburr(p, shape1, shape2, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
rinvburr(n, shape1, shape2, rate = 1, scale = 1/rate)
minvburr(order, shape1, shape2, rate = 1, scale = 1/rate)
levinvburr(limit, shape1, shape2, rate = 1, scale = 1/rate,
           order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The inverse Burr distribution with parameters \code{shape1} \eqn{=
    \tau}{= a}, \code{shape2} \eqn{= \gamma}{= b} and \code{scale}
  \eqn{= \theta}{= s}, has density:
  \deqn{f(x) = \frac{\tau \gamma (x/\theta)^{\gamma \tau}}{%
      x [1 + (x/\theta)^\gamma]^{\tau + 1}}}{%
    f(x) = a b (x/s)^(ba)/(x [1 + (x/s)^b]^(a + 1))}
  for \eqn{x > 0}, \eqn{\tau > 0}{a > 0}, \eqn{\gamma > 0}{b > 0} and
  \eqn{\theta > 0}{s > 0}.

  The inverse Burr is the distribution of the random variable
  \deqn{\theta \left(\frac{X}{1 - X}\right)^{1/\gamma},}{%
    s (X/(1 - X))^(1/b),}
  where \eqn{X} has a beta distribution with parameters \eqn{\tau}{a}
  and \eqn{1}.

  The inverse Burr distribution has the following special cases:
  \itemize{
    \item A \link[=dllogis]{Loglogistic} distribution when \code{shape1
      == 1};
    \item An \link[=dinvpareto]{Inverse Pareto} distribution when
    \code{shape2 == 1};
    \item An \link[=dinvparalogis]{Inverse Paralogistic} distribution
    when \code{shape1 == shape2}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{-\tau\gamma < k < \gamma}{-shape1 * shape2
  < k < shape2}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]}, \eqn{k > -\tau\gamma}{k > -shape1 * shape2}
  and \eqn{1 - k/\gamma}{1 - k/shape2} not a negative integer.
}
\value{
  \code{dinvburr} gives the density,
  \code{invburr} gives the distribution function,
  \code{qinvburr} gives the quantile function,
  \code{rinvburr} generates random deviates,
  \code{minvburr} gives the \eqn{k}th raw moment, and
  \code{levinvburr} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levinvburr} computes the limited expected value using
  \code{\link{betaint}}.

  Also known as the Dagum distribution. See also Kleiber and Kotz (2003)
  for alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvburr(2, 2, 3, 1, log = TRUE))
p <- (1:10)/10
pinvburr(qinvburr(p, 2, 3, 1), 2, 3, 1)

## variance
minvburr(2, 2, 3, 1) - minvburr(1, 2, 3, 1) ^ 2

## case with 1 - order/shape2 > 0
levinvburr(10, 2, 3, 1, order = 2)

## case with 1 - order/shape2 < 0
levinvburr(10, 2, 1.5, 1, order = 2)
}
\keyword{distribution}

\name{hist.grouped.data}
\alias{hist.grouped.data}
\title{Histogram for Grouped Data}
\description{
  This method for the generic function \code{\link{hist}} is mainly
  useful to plot the histogram of grouped data. If \code{plot = FALSE},
  the resulting object of class \code{"histogram"} is returned for
  compatibility with \code{\link{hist.default}}, but does not contain
  much information not already in \code{x}.
}
\usage{
\method{hist}{grouped.data}(x, freq = NULL, probability = !freq,
     density = NULL, angle = 45, col = NULL, border = NULL,
     main = paste("Histogram of" , xname),
     xlim = range(x), ylim = NULL, xlab = xname, ylab,
     axes = TRUE, plot = TRUE, labels = FALSE, \dots)
}
\arguments{
  \item{x}{an object of class \code{"grouped.data"}; only the first
    column of frequencies is used.}
  \item{freq}{logical; if \code{TRUE}, the histogram graphic is a
    representation of frequencies, the \code{counts} component of
    the result; if \code{FALSE}, probability densities, component
    \code{density}, are plotted (so that the histogram has a total area
    of one). Defaults to \code{TRUE} \emph{iff} group boundaries are
    equidistant (and \code{probability} is not specified).}
  \item{probability}{an \emph{alias} for \code{!freq}, for S compatibility.}
  \item{density}{the density of shading lines, in lines per inch.
    The default value of \code{NULL} means that no shading lines
    are drawn. Non-positive values of \code{density} also inhibit the
    drawing of shading lines.}
  \item{angle}{the slope of shading lines, given as an angle in
    degrees (counter-clockwise).}
  \item{col}{a colour to be used to fill the bars.
    The default of \code{NULL} yields unfilled bars.}
  \item{border}{the color of the border around the bars. The default
    is to use the standard foreground color.}
  \item{main, xlab, ylab}{these arguments to \code{title} have useful
    defaults here.}
  \item{xlim, ylim}{the range of x and y values with sensible defaults.
    Note that \code{xlim} is \emph{not} used to define the histogram
    (breaks), but only for plotting (when \code{plot = TRUE}).}
  \item{axes}{logical. If \code{TRUE} (default), axes are draw if the
    plot is drawn.}
  \item{plot}{logical. If \code{TRUE} (default), a histogram is
    plotted, otherwise a list of breaks and counts is returned.}
  \item{labels}{logical or character. Additionally draw labels on top
    of bars, if not \code{FALSE}; see \code{\link{plot.histogram}}.}
  \item{\dots}{further graphical parameters passed to
    \code{\link{plot.histogram}} and their to \code{\link{title}} and
    \code{\link{axis}} (if \code{plot=TRUE}).}
}
\value{
  An object of class \code{"histogram"} which is a list with components:
  \item{breaks}{the \eqn{r + 1} group boundaries.}
  \item{counts}{\eqn{r} integers; the frequency within each group.}
  \item{density}{the relative frequencies within each group
    \eqn{n_j/n}{n[j]/n}, where \eqn{n_j}{n[j]} = \code{counts[j]}.}
  \item{intensities}{same as \code{density}. Deprecated, but retained
    for compatibility.}
  \item{mids}{the \eqn{r} group midpoints.}
  \item{xname}{a character string with the actual \code{x} argument name.}
  \item{equidist}{logical, indicating if the distances between
    \code{breaks} are all the same.}
}
\note{
  The resulting value does \emph{not} depend on the values of
  the arguments \code{freq} (or \code{probability})
  or \code{plot}. This is intentionally different from S.
}
\seealso{
  \code{\link{hist}} and \code{\link{hist.default}} for histograms of
  individual data and fancy examples.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\examples{
data(gdental)
hist(gdental)
}
\keyword{dplot}
\keyword{hplot}
\keyword{distribution}

\name{elev}
\alias{elev}
\alias{elev.default}
\alias{elev.grouped.data}
\alias{print.elev}
\alias{summary.elev}
\alias{knots.elev}
\alias{plot.elev}
\title{Empirical Limited Expected Value}
\description{
  Compute the empirical limited expected value for individual or grouped
  data.
}
\usage{
elev(x, ...)

\method{elev}{default}(x, \dots)

\method{elev}{grouped.data}(x, \dots)

\method{print}{elev}(x, digits = getOption("digits") - 2, \dots)

\method{summary}{elev}(object, \dots)

\method{knots}{elev}(Fn, \dots)

\method{plot}{elev}(x, \dots, main = NULL, xlab = "x", ylab = "Empirical LEV")
}
\arguments{
  \item{x}{a vector or an object of class \code{"grouped.data"} (in
    which case only the first column of frequencies is used); for the
    methods, an object of class \code{"elev"}, typically.}
  \item{digits}{number of significant digits to use, see
    \code{\link{print}}.}
  \item{Fn, object}{an \R object inheriting from \code{"ogive"}.}
  \item{main}{main title.}
  \item{xlab, ylab}{labels of x and y axis.}
  \item{\dots}{arguments to be passed to subsequent methods.}
}
\details{
  The limited expected value (LEV) at \eqn{u} of a random variable
  \eqn{X} is
  \eqn{E[X \wedge u] = E[\min(X, u)]}{E[X ^ u] = E[min(X, u)]}. For
  individual data \eqn{x_1, \dots, x_n}{x[1], \dots, x[n]}, the
  empirical LEV \eqn{E_n[X \wedge u]}{En[X ^ u]} is thus
  \deqn{E_n[X \wedge u] = \frac{1}{n}
    \left( \sum_{x_j < u} x_j + \sum_{x_j \geq u} u \right).}{%
    En[X ^ u] = (sum(x[j] < u; 1) + sum(x[j] >= u; u))/n.}

  Methods of \code{elev} exist for individual data or for grouped data
  created with \code{\link{grouped.data}}. The formula in this case is
  too long to show here. See the reference for details.
}
\value{
  For \code{elev}, a function of class \code{"elev"}, inheriting from the
  \code{"\link{function}"} class.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\seealso{
  \code{\link{grouped.data}} to create grouped data objects;
  \code{\link{stepfun}} for related documentation (even though the
  empirical LEV is not a step function).
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
data(gdental)
lev <- elev(gdental)
lev
summary(lev)
knots(lev)            # the group boundaries

lev(knots(lev))       # empirical lev at boundaries
lev(c(80, 200, 2000)) # and at other limits

plot(lev, type = "o", pch = 16)
}
\keyword{dplot}
\keyword{hplot}

\name{PoissonInverseGaussian}
\alias{PoissonInverseGaussian}
\alias{PIG}
\alias{dpoisinvgauss}
\alias{ppoisinvgauss}
\alias{qpoisinvgauss}
\alias{rpoisinvgauss}
\alias{dpig}
\alias{ppig}
\alias{qpig}
\alias{rpig}
\title{The Poisson-Inverse Gaussian Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Poisson-inverse Gaussian discrete distribution with
  parameters \code{mean} and \code{shape}.
}
\usage{
dpoisinvgauss(x, mean, shape = 1, dispersion = 1/shape,
              log = FALSE)
ppoisinvgauss(q, mean, shape = 1, dispersion = 1/shape,
              lower.tail = TRUE, log.p = FALSE)
qpoisinvgauss(p, mean, shape = 1, dispersion = 1/shape,
              lower.tail = TRUE, log.p = FALSE)
rpoisinvgauss(n, mean, shape = 1, dispersion = 1/shape)
}
\arguments{
  \item{x}{vector of (positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{mean, shape}{parameters. Must be strictly positive. Infinite
    values are supported.}
  \item{dispersion}{an alternative way to specify the shape.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The Poisson-inverse Gaussian distribution is the result of the
  continuous mixture between a Poisson distribution and an inverse
  Gaussian, that is, the distribution with probability mass function
  \deqn{%
    p(x) = \int_0^\infty \frac{\lambda^x e^{-\lambda}}{x!}\,
           g(\lambda; \mu, \phi)\, d\lambda,}{%
    p(x) = int_0^Inf (y^x exp(-y))/x! g(y; \mu, \phi) dy,}
  where \eqn{g(\lambda; \mu, \phi)}{g(y; \mu, \phi)} is the density
  function of the inverse Gaussian distribution with parameters
  \code{mean} \eqn{= \mu} and \code{dispersion} \eqn{= \phi} (see
  \code{\link{dinvgauss}}).

  The resulting probability mass function is
  \deqn{%
    p(x) = \sqrt{\frac{2}{\pi \phi}}
    \frac{e^{(\phi\mu)^{-1}}}{x!}
    \left(
      \sqrt{2\phi\left(1 + \frac{1}{2\phi\mu^2}\right)}
    \right)^{-(x - \frac{1}{2})}
    K_{x - \frac{1}{2}}
    \left(
      \sqrt{\frac{2}{\phi}\left(1 + \frac{1}{2\phi\mu^2}\right)}
    \right),}{%
    p(x) = sqrt(2/(\pi \phi)) exp(1/(\phi \mu))/x!
           * [\sqrt(2 \phi (1 + 1/(2 \phi \mu^2)))]^(-(x-1/2))
           * K(\sqrt((2/\phi) (1 + 1/(2 \phi \mu^2))); x-1/2),}
  for \eqn{x = 0, 1, \dots}, \eqn{\mu > 0}, \eqn{\phi > 0} and where
  \eqn{K_\nu(x)}{K(x; \nu)} is the modified Bessel function of the third
  kind implemented by \R's \code{\link{besselK}()} and defined in its
  help.

  The limiting case \eqn{\mu = \infty}{\mu = Inf} has well defined
  probability mass and distribution functions, but has no finite
  strictly positive, integer moments. The pmf in this case reduces to
  \deqn{%
    p(x) = \sqrt{\frac{2}{\pi \phi}}
    \frac{1}{x!}
    (\sqrt{2\phi})^{-(x - \frac{1}{2})}
    K_{x - \frac{1}{2}}(\sqrt{2/\phi}).}{%
    p(x) = sqrt(2/(\pi \phi)) 1/x! [\sqrt(2 \phi)]^(-(x-1/2))
           * K(\sqrt(2/\phi); x-1/2).}

  The limiting case \eqn{\phi = 0} is a degenerate distribution in
  \eqn{x = 0}.

  If an element of \code{x} is not integer, the result of
  \code{dpoisinvgauss} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{F(x) \ge p}, where \eqn{F} is the distribution function.
}
\value{
  \code{dpoisinvgauss} gives the probability mass function,
  \code{ppoisinvgauss} gives the distribution function,
  \code{qpoisinvgauss} gives the quantile function, and
  \code{rpoisinvgauss} generates random deviates.

  Invalid arguments will result in return value \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rpoisinvgauss}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  \code{[dpqr]pig} are aliases for \code{[dpqr]poisinvgauss}.

  \code{qpoisinvgauss} is based on \code{qbinom} et al.; it uses the
  Cornish--Fisher Expansion to include a skewness correction to a normal
  approximation, followed by a search.
}
\references{
  Holla, M. S. (1966), \dQuote{On a Poisson-Inverse Gaussian
  Distribution}, \emph{Metrika}, vol. 15, p. 377-384.

  Johnson, N. L., Kemp, A. W. and Kotz, S. (2005), \emph{Univariate
  Discrete Distributions, Third Edition}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.

  Shaban, S. A., (1981) \dQuote{Computation of the poisson-inverse
    gaussian distribution}, \emph{Communications in Statistics - Theory
    and Methods}, vol. 10, no. 14, p. 1389-1399.
}
\seealso{
  \code{\link{dpois}} for the Poisson distribution,
  \code{\link{dinvgauss}} for the inverse Gaussian distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
## Tables I and II of Shaban (1981)
x <- 0:2
sapply(c(0.4, 0.8, 1), dpoisinvgauss, x = x, mean = 0.1)
sapply(c(40, 80, 100, 130), dpoisinvgauss, x = x, mean = 1)

qpoisinvgauss(ppoisinvgauss(0:10, 1, dis = 2.5), 1, dis = 2.5)

x <- rpoisinvgauss(1000, 1, dis = 2.5)
y <- sort(unique(x))
plot(y, table(x)/length(x), type = "h", lwd = 2,
     pch = 19, col = "black", xlab = "x", ylab = "p(x)",
     main = "Empirical vs theoretical probabilities")
points(y, dpoisinvgauss(y, 1, dis = 2.5),
       pch = 19, col = "red")
legend("topright", c("empirical", "theoretical"),
       lty = c(1, NA), pch = c(NA, 19), col = c("black", "red"))
}
\keyword{distribution}

\name{emm}
\alias{emm}
\alias{emm.default}
\alias{emm.grouped.data}
\title{Empirical Moments}
\description{
  Raw empirical moments for individual and grouped data.
}
\usage{
emm(x, order = 1, \dots)

\method{emm}{default}(x, order = 1, \dots)

\method{emm}{grouped.data}(x, order = 1, \dots)
}
\arguments{
  \item{x}{a vector or matrix of individual data, or an object of class
    \code{"grouped data"}.}
  \item{order}{order of the moment. Must be positive.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  Arguments \code{\dots} are passed to \code{\link{colMeans}};
  \code{na.rm = TRUE} may be useful for individual data with missing
  values. 

  For individual data, the \eqn{k}th empirical moment is
  \eqn{\sum_{j = 1}^n x_j^k}{sum(j; x[j]^k)}.

  For grouped data with group boundaries \eqn{c_0, c_1, \dots,
    c_r}{c[0], c[1], \dots, c[r]} and group frequencies \eqn{n_1, \dots,
    n_r}{n[1], \dots, n[r]}, the \eqn{k}th empirical moment is
  \deqn{\frac{1}{n} \sum_{j = 1}^r \frac{n_j (c_j^{k + 1} - c_{j - 1}^{k + 1})}{%
      (k + 1) (c_j - c_{j - 1})},}{%
    (1/n) * sum(j; (n[j] * {c[j]^(k+1) - c[j-1]^(k+1)})/%
    ((k+1) * {c[j] - c[j-1]})),}
  where \eqn{n = \sum_{j = 1}^r n_j}{n = sum(j; n[j])}.
}
\value{
  A named vector or matrix of moments.
}
\seealso{
  \code{\link{mean}} and \code{\link{mean.grouped.data}} for simpler
  access to the first moment.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
## Individual data
data(dental)
emm(dental, order = 1:3)

## Grouped data
data(gdental)
emm(gdental)
x <- grouped.data(cj = gdental[, 1],
                  nj1 = sample(1:100, nrow(gdental)),
                  nj2 = sample(1:100, nrow(gdental)))
emm(x) # same as mean(x)
}
\keyword{univar}

\name{WeibullMoments}
\alias{WeibullMoments}
\alias{mweibull}
\alias{levweibull}
\title{Raw and Limited Moments of the Weibull Distribution}
\description{
  Raw moments and limited moments for the Weibull distribution with
  parameters \code{shape} and \code{scale}.
}
\usage{
mweibull(order, shape, scale = 1)
levweibull(limit, shape, scale = 1, order = 1)
}
\arguments{
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{shape, scale}{shape and scale parameters, the latter defaulting
    to 1.}
}
\details{
  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]}, \eqn{k >
  -\tau}{k > -shape}.
}
\value{
  \code{mweibull} gives the \eqn{k}th raw moment and
  \code{levweibull} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link{Weibull}} for details on the Weibull distribution and
  functions \code{[dpqr]weibull}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
mweibull(2, 3, 4) - mweibull(1, 3, 4)^2
levweibull(10, 3, 4, order = 2)
}
\keyword{distribution}

\name{Loggamma}
\alias{Loggamma}
\alias{dlgamma}
\alias{plgamma}
\alias{qlgamma}
\alias{rlgamma}
\alias{mlgamma}
\alias{levlgamma}
\title{The Loggamma Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Loggamma distribution with
  parameters \code{shapelog} and \code{ratelog}.
}
\usage{
dlgamma(x, shapelog, ratelog, log = FALSE)
plgamma(q, shapelog, ratelog, lower.tail = TRUE, log.p = FALSE)
qlgamma(p, shapelog, ratelog, lower.tail = TRUE, log.p = FALSE)
rlgamma(n, shapelog, ratelog)
mlgamma(order, shapelog, ratelog)
levlgamma(limit, shapelog, ratelog, order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shapelog, ratelog}{parameters. Must be strictly positive.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The loggamma distribution with parameters \code{shapelog} \eqn{=
    \alpha}{= a} and \code{ratelog} \eqn{= \lambda}{= b} has density:
  \deqn{f(x) = \frac{\lambda^\alpha}{\Gamma(\alpha)}%
    \frac{(\log x)^{\alpha - 1}}{x^{\lambda + 1}}}{%
    f(x) = (b^a (log(x))^(a - 1))/(Gamma(a) * x^(b + 1))}
  for \eqn{x > 1}, \eqn{\alpha > 0}{a > 0} and \eqn{\lambda > 0}{b >
    0}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The loggamma is the distribution of the random variable
  \eqn{e^X}{exp(X)},  where \eqn{X} has a gamma distribution with
  shape parameter \eqn{alpha}{a} and scale parameter
  \eqn{1/\lambda}{1/b}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]}, \eqn{k < \lambda}{k
  < ratelog}.
}
\value{
  \code{dlgamma} gives the density,
  \code{plgamma} gives the distribution function,
  \code{qlgamma} gives the quantile function,
  \code{rlgamma} generates random deviates,
  \code{mlgamma} gives the \eqn{k}th raw moment, and
  \code{levlgamma} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Hogg, R. V. and Klugman, S. A. (1984), \emph{Loss Distributions},
  Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dlgamma(2, 3, 4, log = TRUE))
p <- (1:10)/10
plgamma(qlgamma(p, 2, 3), 2, 3)
mlgamma(2, 3, 4) - mlgamma(1, 3, 4)^2
levlgamma(10, 3, 4, order = 2)
}
\keyword{distribution}

\name{ZeroTruncatedBinomial}
\alias{ZeroTruncatedBinomial}
\alias{ZTBinomial}
\alias{dztbinom}
\alias{pztbinom}
\alias{qztbinom}
\alias{rztbinom}
\title{The Zero-Truncated Binomial Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Truncated Binomial distribution with
  parameters \code{size} and \code{prob}.
}
\usage{
dztbinom(x, size, prob, log = FALSE)
pztbinom(q, size, prob, lower.tail = TRUE, log.p = FALSE)
qztbinom(p, size, prob, lower.tail = TRUE, log.p = FALSE)
rztbinom(n, size, prob)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{size}{number of trials (strictly positive integer).}
  \item{prob}{probability of success on each trial. \code{0 <= prob <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-truncated binomial distribution with \code{size}
  \eqn{= n} and \code{prob} \eqn{= p} has probability mass function
  \deqn{%
    p(x) = {n \choose x} \frac{p^x (1 - p)^{n-x}}{1 - (1 - p)^n}}{%
    p(x) = choose(n, x) [p^x (1-p)^(n-x)]/[1 - (1-p)^n]}
  for \eqn{x = 1, \ldots, n} and \eqn{0 < p \le 1}, and \eqn{p(1) = 1}
  when \eqn{p = 0}.
  The cumulative distribution function is
  \deqn{P(x) = \frac{F(x) - F(0)}{1 - F(0)},}{%
    P(x) = [F(x) - F(0)]/[1 - F(0)],}
  where \eqn{F(x)} is the distribution function of the standard binomial.

  The mean is \eqn{np/(1 - (1-p)^n)} and the variance is
  \eqn{np[(1-p) - (1-p+np)(1-p)^n]/[1 - (1-p)^n]^2}.

  In the terminology of Klugman et al. (2012), the zero-truncated
  binomial is a member of the \eqn{(a, b, 1)} class of
  distributions with \eqn{a = -p/(1-p)} and \eqn{b = (n+1)p/(1-p)}.

  If an element of \code{x} is not integer, the result of
  \code{dztbinom} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dztbinom} gives the probability mass function,
  \code{pztbinom} gives the distribution function,
  \code{qztbinom} gives the quantile function, and
  \code{rztbinom} generates random deviates.

  Invalid \code{size} or \code{prob} will result in return value
  \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rztbinom}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}ztbinom} use \code{\{d,p,q\}binom} for all
  but the trivial input values and \eqn{p(0)}.

  \code{rztbinom} uses the simple inversion algorithm suggested by
  Peter Dalgaard on the r-help mailing list on 1 May 2005 %
  (\url{https://stat.ethz.ch/pipermail/r-help/2005-May/070680.html}).
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dbinom}} for the binomial distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
dztbinom(1:5, size = 5, prob = 0.4)
dbinom(1:5, 5, 0.4)/pbinom(0, 5, 0.4, lower = FALSE) # same

pztbinom(1, 2, prob = 0)        # point mass at 1

qztbinom(pztbinom(1:10, 10, 0.6), 10, 0.6)

n <- 8; p <- 0.3
x <- 0:n
title <- paste("ZT Binomial(", n, ", ", p,
               ") and Binomial(", n, ", ", p,") PDF",
               sep = "")
plot(x, dztbinom(x, n, p), type = "h", lwd = 2, ylab = "p(x)",
     main = title)
points(x, dbinom(x, n, p), pch = 19, col = "red")
legend("topright", c("ZT binomial probabilities", "Binomial probabilities"),
       col = c("black", "red"), lty = c(1, 0), lwd = 2, pch = c(NA, 19))
}
\keyword{distribution}

\name{GeneralizedPareto}
\alias{GeneralizedPareto}
\alias{dgenpareto}
\alias{pgenpareto}
\alias{qgenpareto}
\alias{rgenpareto}
\alias{mgenpareto}
\alias{levgenpareto}
\title{The Generalized Pareto Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Generalized Pareto
  distribution with parameters \code{shape1}, \code{shape2} and
  \code{scale}.
}
\usage{
dgenpareto(x, shape1, shape2, rate = 1, scale = 1/rate,
           log = FALSE)
pgenpareto(q, shape1, shape2, rate = 1, scale = 1/rate,
           lower.tail = TRUE, log.p = FALSE)
qgenpareto(p, shape1, shape2, rate = 1, scale = 1/rate,
           lower.tail = TRUE, log.p = FALSE)
rgenpareto(n, shape1, shape2, rate = 1, scale = 1/rate)
mgenpareto(order, shape1, shape2, rate = 1, scale = 1/rate)
levgenpareto(limit, shape1, shape2, rate = 1, scale = 1/rate,
             order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Generalized Pareto distribution with parameters \code{shape1}
  \eqn{= \alpha}{= a}, \code{shape2} \eqn{= \tau}{= b} and \code{scale}
  \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\Gamma(\alpha + \tau)}{\Gamma(\alpha)\Gamma(\tau)}
    \frac{\theta^\alpha x^{\tau - 1}}{%
      (x + \theta)^{\alpha + \tau}}}{%
    f(x) = Gamma(a + b)/(Gamma(a) * Gamma(b)) *
           (s^a x^(b - 1))/(x + s)^(a + b)}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0}, \eqn{\tau > 0}{b > 0} and
  \eqn{\theta > 0}{s > 0}.
  (Here \eqn{\Gamma(\alpha)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The Generalized Pareto is the distribution of the random variable
  \deqn{\theta \left(\frac{X}{1 - X}\right),}{\theta (X/(1 - X)),}
  where \eqn{X} has a beta distribution with parameters \eqn{\alpha}
  and \eqn{\tau}.

  The Generalized Pareto distribution has the following special cases:
  \itemize{
    \item A \link[=dpareto]{Pareto} distribution when \code{shape2 ==
      1};
    \item An \link[=dinvpareto]{Inverse Pareto} distribution when
    \code{shape1 == 1}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}, \eqn{-\tau < k < \alpha}{-shape2 < k < shape1}.

  The \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]},
  \eqn{k > -\tau}{k > -shape2} and \eqn{\alpha - k}{shape1 - k} not a
  negative integer.
}
\value{
  \code{dgenpareto} gives the density,
  \code{pgenpareto} gives the distribution function,
  \code{qgenpareto} gives the quantile function,
  \code{rgenpareto} generates random deviates,
  \code{mgenpareto} gives the \eqn{k}th raw moment, and
  \code{levgenpareto} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levgenpareto} computes the limited expected value using
  \code{\link{betaint}}.

  Distribution also known as the Beta of the Second Kind. See also
  Kleiber and Kotz (2003) for alternative names and parametrizations.

  The Generalized Pareto distribution defined here is different from the
  one in Embrechts et al. (1997) and in
  \href{https://en.wikipedia.org/wiki/Generalized_Pareto_distribution}{Wikipedia};
  see also Kleiber and Kotz (2003, section 3.12). One may most likely
  compute quantities for the latter using functions for the
  \link[=dpareto]{Pareto} distribution with the appropriate change of
  parametrization.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Embrechts, P., Klüppelberg, C. and Mikisch, T. (1997), \emph{Modelling
    Extremal Events for Insurance and Finance}, Springer.
  
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dgenpareto(3, 3, 4, 4, log = TRUE))
p <- (1:10)/10
pgenpareto(qgenpareto(p, 3, 3, 1), 3, 3, 1)
qgenpareto(.3, 3, 4, 4, lower.tail = FALSE)

## variance
mgenpareto(2, 3, 2, 1) - mgenpareto(1, 3, 2, 1)^2

## case with shape1 - order > 0
levgenpareto(10, 3, 3, scale = 1, order = 2)

## case with shape1 - order < 0
levgenpareto(10, 1.5, 3, scale = 1, order = 2)
}
\keyword{distribution}

\name{quantile.aggregateDist}
\alias{quantile.aggregateDist}
\alias{VaR.aggregateDist}
\title{Quantiles of Aggregate Claim Amount Distribution}
\description{
  Quantile and Value-at-Risk methods for objects of class
  \code{"aggregateDist"}.
}
\usage{
\method{quantile}{aggregateDist}(x, 
         probs = c(0.25, 0.5, 0.75, 0.9, 0.95, 0.975, 0.99, 0.995),
         smooth = FALSE, names = TRUE, \dots)

\method{VaR}{aggregateDist}(x, conf.level = c(0.9, 0.95, 0.99),
         smooth = FALSE, names = TRUE, \dots)
}  
\arguments{
  \item{x}{an object of class \code{"aggregateDist"}.}
  \item{probs, conf.level}{numeric vector of probabilities with values
    in \eqn{[0, 1)}.}
  \item{smooth}{logical; when \code{TRUE} and \code{x} is a step
    function, quantiles are linearly interpolated between knots.}
  \item{names}{logical; if true, the result has a \code{names}
    attribute.  Set to \code{FALSE} for speedup with many \code{probs}.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  The quantiles are taken directly from the cumulative distribution
  function defined in \code{x}. Linear interpolation is available for
  step functions.
}
\value{
  A numeric vector, named if \code{names} is \code{TRUE}.
}
\seealso{
  \code{\link{aggregateDist}}
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and 
  Louis-Philippe Pouliot
}
\examples{
model.freq <- expression(data = rpois(3))
model.sev <- expression(data = rlnorm(10, 1.5))
Fs <- aggregateDist("simulation", model.freq, model.sev, nb.simul = 1000)
quantile(Fs, probs = c(0.25, 0.5, 0.75))
VaR(Fs)
}
\keyword{univar}

\name{var}
\alias{var}
\alias{var.default}
\alias{var.grouped.data}
\alias{sd}
\alias{sd.default}
\alias{sd.grouped.data}
\title{Variance and Standard Deviation}
\description{
  Generic functions for the variance and standard deviation, and methods
  for individual and grouped data.

  The default methods for individual data are the functions from the
  \pkg{stats} package.
}
\usage{
var(x, \dots)

\method{var}{default}(x, y = NULL, na.rm = FALSE, use, \dots)

\method{var}{grouped.data}(x, \dots)

sd(x, \dots)

\method{sd}{default}(x, na.rm = FALSE, \dots)

\method{sd}{grouped.data}(x, \dots)
}
\arguments{
  \item{x}{a vector or matrix of individual data, or an object of class
    \code{"grouped data"}.}
  \item{y}{see \code{\link[stats:var]{stats::var}}.}
  \item{na.rm}{see \code{\link[stats:var]{stats::var}}.}
  \item{use}{see \code{\link[stats:var]{stats::var}}.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  This page documents variance and standard deviation computations for
  grouped data. For individual data, see \code{\link[stats]{var}} and
  \code{\link[stats]{sd}} from the \pkg{stats} package.

  For grouped data with group boundaries \eqn{c_0, c_1, \dots,
  c_r}{c[0], c[1], \dots, c[r]} and group frequencies \eqn{n_1, \dots,
  n_r}{n[1], \dots, n[r]}, \code{var} computes the sample variance
  \deqn{\frac{1}{n - 1} \sum_{j = 1}^r n_j (a_j - m_1)^2,}{%
    (1/(n - 1)) * sum(j; n[j] * (a[j] - m)^2,}
  where
  \eqn{a_j = (c_{j - 1} + c_j)/2}{a[j] = (c[j - 1] + c[j])/2}
  is the midpoint of the \eqn{j}th interval,
  \eqn{m_1}{m} is the sample mean (or sample first moment) of the data,
  and
  \eqn{n = \sum_{j = 1}^r n_j}{n = sum(j; n[j])}.
  The sample sample standard deviation is the square root of the sample
  variance.

  The sample variance for grouped data differs from the variance
  computed from the empirical raw moments with \code{\link{emm}} in two
  aspects. First, it takes into account the degrees of freedom. Second,
  it applies Sheppard's correction factor to compensate for the
  overestimation of the true variation in the data. For groups of equal
  width \eqn{k}, Sheppard's correction factor is equal to \eqn{-k^2/12}.
}
\value{
  A named vector of variances or standard deviations.
}
\seealso{
  \code{\link{grouped.data}} to create grouped data objects;
  \code{\link{mean.grouped.data}} for the mean and \code{\link{emm}} for
  higher moments.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.

  Heumann, C., Schomaker, M., Shalabh (2016), \emph{Introduction to
  Statistics and Data Analysis}, Springer.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}. Variance and
  standard deviation methods for grouped data contributed by Walter
  Garcia-Fontes \email{walter.garcia@upf.edu}.
}
\examples{
data(gdental)
var(gdental)
sd(gdental)

## Illustration of Sheppard's correction factor
cj <- c(0, 2, 4, 6, 8)
nj <- c(1, 5,  3,  2)
gd <- grouped.data(Group = cj, Frequency = nj)
(sum(nj) - 1)/sum(nj) * var(gd)
(emm(gd, 2) - emm(gd)^2) - 4/12
}
\keyword{univar}

\name{Gumbel}
\alias{Gumbel}
\alias{dgumbel}
\alias{pgumbel}
\alias{qgumbel}
\alias{rgumbel}
\alias{mgumbel}
\alias{mgfgumbel}
\title{The Gumbel Distribution}
\description{
  Density function, distribution function, quantile function, random
  generation and raw moments for the Gumbel extreme value distribution
  with parameters \code{alpha} and \code{scale}.
}
\usage{
dgumbel(x, alpha, scale, log = FALSE)
pgumbel(q, alpha, scale, lower.tail = TRUE, log.p = FALSE)
qgumbel(p, alpha, scale, lower.tail = TRUE, log.p = FALSE)
rgumbel(n, alpha, scale)
mgumbel(order, alpha, scale)
mgfgumbel(t, alpha, scale, log = FALSE)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{alpha}{location parameter.}
  \item{scale}{parameter. Must be strictly positive.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment. Only values \eqn{1} and \eqn{2} are
    supported.}
  \item{t}{numeric vector.}
}
\details{
  The Gumbel distribution with parameters \code{alpha} \eqn{=
    \alpha}{= a} and \code{scale} \eqn{= \theta}{= s} has distribution
  function:
  \deqn{F(x) = \exp[-\exp(-(x - \alpha)/\theta)]}{%
    F(x) = exp[-exp(-(x - a)/s)],}
  for \eqn{-\infty < x < \infty}{-Inf < x < Inf}, \eqn{-\infty < a <
    \infty}{-Inf < a < Inf} and \eqn{\theta > 0}{s > 0}.

  The mode of the distribution is in \eqn{\alpha}{a}, the mean is
  \eqn{\alpha + \gamma\theta}{a + g * s}, where \eqn{\gamma}{g} \eqn{=
  0.57721566} is the Euler-Mascheroni constant, and the variance is
  \eqn{\pi^2 \theta^2/6}{(pi * s)^2/6}.
}
\value{
  \code{dgumbel} gives the density,
  \code{pgumbel} gives the distribution function,
  \code{qgumbel} gives the quantile function,
  \code{rgumbel} generates random deviates,
  \code{mgumbel} gives the \eqn{k}th raw moment, \eqn{k = 1, 2}, and
  \code{mgfgamma} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  Distribution also knonw as the generalized extreme value distribution
  Type-I.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
dgumbel(c(-5, 0, 10, 20), 0.5, 2)

p <- (1:10)/10
pgumbel(qgumbel(p, 2, 3), 2, 3)

curve(pgumbel(x, 0.5, 2), from = -5, to = 20, col = "red")
curve(pgumbel(x, 1.0, 2), add = TRUE, col = "green")
curve(pgumbel(x, 1.5, 3), add = TRUE, col = "blue")
curve(pgumbel(x, 3.0, 4), add = TRUE, col = "cyan")

a <- 3; s <- 4
mgumbel(1, a, s)                        # mean
a - s * digamma(1)                      # same

mgumbel(2, a, s) - mgumbel(1, a, s)^2   # variance
(pi * s)^2/6                            # same
}
\keyword{distribution}

\name{quantile.grouped.data}
\alias{quantile.grouped.data}
\alias{summary.grouped.data}
\title{Quantiles of Grouped Data}
\description{
  Sample quantiles corresponding to the given probabilities for objects
  of class \code{"grouped.data"}.
}
\usage{
\method{quantile}{grouped.data}(x, probs = seq(0, 1, 0.25),
         names = TRUE, \dots)

\method{summary}{grouped.data}(object, \dots)
}
\arguments{
  \item{x, object}{an object of class \code{"grouped.data"}.}
  \item{probs}{numeric vector of probabilities with values
    in \eqn{[0, 1]}.}
  \item{names}{logical; if true, the result has a \code{names}
    attribute. Set to \code{FALSE} for speedup with many \code{probs}.}
  \item{\dots}{further arguments passed to or from other methods.}
}
\details{
  The quantile function is the inverse of the ogive, that is a linear
  interpolation of the empirical quantile function.

  The equation of the quantile function is
  \deqn{x = \frac{c_j (F_n(c_{j - 1}) - q) +
      c_{j - 1} (q - F_n(c_j)}{F_n(c_j) - F_n(c_{j - 1})}}{%
    x = (c[j] (Fn(c[j-1]) - q) +
    c[j-1] (q - Fn(c[j])))/(Fn(c[j]) - Fn(c[j-1]))}
  for \eqn{0 \leq q \leq c_j}{0 <= q <= 1} and where \eqn{c_0, \dots,
  c_r}{c[0], \dots, c[r]} are the \eqn{r + 1} group
  boundaries and \eqn{F_n}{Fn} is the empirical distribution function of
  the sample.
}
\value{
  For \code{quantile}, a numeric vector, named if \code{names} is
  \code{TRUE}.

  For the \code{summary} method, an object of class
  \code{c("summaryDefault", "\link{table}")} which has specialized
  \code{\link{format}} and \code{\link{print}} methods.
}
\seealso{
  \code{\link{ogive}} for the smoothed empirical distribution of which
  \code{quantile.grouped.data} is an inverse;
  \code{\link{mean.grouped.data}} and \code{\link{var.grouped.data}} to
  compute the mean and variance of grouped data.
  
  \code{\link{grouped.data}} to create grouped data objects.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
data(gdental)
quantile(gdental)
summary(gdental)
Fn <- ogive(gdental)
Fn(quantile(gdental))		# inverse function
}
\keyword{univar}

\name{LognormalMoments}
\alias{LognormalMoments}
\alias{mlnorm}
\alias{levlnorm}
\title{Raw and Limited Moments of the Lognormal Distribution}
\description{
  Raw moments and limited moments for the Lognormal distribution whose
  logarithm has mean equal to \code{meanlog} and standard deviation
  equal to \code{sdlog}.
}
\usage{
mlnorm(order, meanlog = 0, sdlog = 1)
levlnorm(limit, meanlog = 0, sdlog = 1, order = 1)
}
\arguments{
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{meanlog, sdlog}{mean and standard deviation of the distribution
    on the log scale with default values of \code{0} and \code{1}
    respectively.}
}
\value{
  \code{mlnorm} gives the \eqn{k}th raw moment and
  \code{levlnorm} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\seealso{
  \code{\link{Lognormal}} for details on the lognormal distribution and
  functions \code{[dpqr]lnorm}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
mlnorm(2, 3, 4) - mlnorm(1, 3, 4)^2
levlnorm(10, 3, 4, order = 2)
}
\keyword{distribution}

\name{ZeroModifiedNegativeBinomial}
\alias{ZeroModifiedNegativeBinomial}
\alias{ZMNegativeBinomial}
\alias{ZMNegBinomial}
\alias{dzmnbinom}
\alias{pzmnbinom}
\alias{qzmnbinom}
\alias{rzmnbinom}
\title{The Zero-Modified Negative Binomial Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Modified Negative Binomial distribution with
  parameters \code{size} and \code{prob}, and arbitrary probability at
  zero \code{p0}.
}
\usage{
dzmnbinom(x, size, prob, p0, log = FALSE)
pzmnbinom(q, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
qzmnbinom(p, size, prob, p0, lower.tail = TRUE, log.p = FALSE)
rzmnbinom(n, size, prob, p0)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{size}{target for number of successful trials, or dispersion
    parameter. Must be positive, need not be integer.}
  \item{prob}{parameter. \code{0 < prob <= 1}.}
  \item{p0}{probability mass at zero. \code{0 <= p0 <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-modified negative binomial distribution with \code{size}
  \eqn{= r}, \code{prob} \eqn{= p} and \code{p0} \eqn{= p_0}{= p0} is a
  discrete mixture between a degenerate distribution at zero and a
  (standard) negative binomial. The probability mass function is
  \eqn{p(0) = p_0}{p(0) = p0} and
  \deqn{%
    p(x) = \frac{(1-p_0)}{(1-p^r)} f(x)}{%
    p(x) = (1-p0)/(1-p^r) f(x)}
  for \eqn{x = 1, 2, \ldots}, \eqn{r \ge 0}, \eqn{0 < p < 1} and \eqn{0 \le
  p_0 \le 1}{0 \le p0 \le 1}, where \eqn{f(x)} is the probability mass
  function of the negative binomial.
  The cumulative distribution function is
  \deqn{P(x) = p_0 + (1 - p_0) \left(\frac{F(x) - F(0)}{1 - F(0)}\right)}{%
    P(x) = p0 + (1 - p0) [F(x) - F(0)]/[1 - F(0)].}

  The mean is \eqn{(1-p_0) \mu}{(1-p0)m} and the variance is
  \eqn{(1-p_0) \sigma^2 + p_0(1-p_0) \mu^2}{(1-p0)v + p0(1-p0)m^2},
  where \eqn{\mu}{m} and \eqn{\sigma^2}{v} are the mean and variance of
  the zero-truncated negative binomial.

  In the terminology of Klugman et al. (2012), the zero-modified
  negative binomial is a member of the \eqn{(a, b, 1)} class of
  distributions with \eqn{a = 1-p} and \eqn{b = (r-1)(1-p)}.

  The special case \code{p0 == 0} is the zero-truncated negative
  binomial.

  The limiting case \code{size == 0} is the zero-modified logarithmic
  distribution with parameters \code{1 - prob} and \code{p0}.

  Unlike the standard negative binomial functions, parametrization
  through the mean \code{mu} is not supported to avoid ambiguity as
  to whether \code{mu} is the mean of the underlying negative binomial
  or the mean of the zero-modified distribution.

  If an element of \code{x} is not integer, the result of
  \code{dzmnbinom} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dzmnbinom} gives the (log) probability mass function,
  \code{pzmnbinom} gives the (log) distribution function,
  \code{qzmnbinom} gives the quantile function, and
  \code{rzmnbinom} generates random deviates.

  Invalid \code{size}, \code{prob} or \code{p0} will result in return
  value \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rzmnbinom}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}zmnbinom} use \code{\{d,p,q\}nbinom} for all
  but the trivial input values and \eqn{p(0)}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dnbinom}} for the negative binomial distribution.

  \code{\link{dztnbinom}} for the zero-truncated negative binomial
  distribution.

  \code{\link{dzmgeom}} for the zero-modified geometric and
  \code{\link{dzmlogarithmic}} for the zero-modified logarithmic, which
  are special cases of the zero-modified negative binomial.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
## Example 6.3 of Klugman et al. (2012)
p <- 1/(1 + 0.5)
dzmnbinom(1:5, size = 2.5, prob = p, p0 = 0.6)
(1-0.6) * dnbinom(1:5, 2.5, p)/pnbinom(0, 2.5, p, lower = FALSE) # same

## simple relation between survival functions
pzmnbinom(0:5, 2.5, p, p0 = 0.2, lower = FALSE)
(1-0.2) * pnbinom(0:5, 2.5, p, lower = FALSE) /
    pnbinom(0, 2.5, p, lower = FALSE) # same

qzmnbinom(pzmnbinom(0:10, 2.5, 0.3, p0 = 0.1), 2.5, 0.3, p0 = 0.1)
}
\keyword{distribution}

\name{InverseGamma}
\alias{InverseGamma}
\alias{dinvgamma}
\alias{pinvgamma}
\alias{qinvgamma}
\alias{rinvgamma}
\alias{minvgamma}
\alias{levinvgamma}
\alias{mgfinvgamma}
\title{The Inverse Gamma Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments, and limited moments for the Inverse Gamma distribution
  with parameters \code{shape} and \code{scale}.
}
\usage{
dinvgamma(x, shape, rate = 1, scale = 1/rate, log = FALSE)
pinvgamma(q, shape, rate = 1, scale = 1/rate,
          lower.tail = TRUE, log.p = FALSE)
qinvgamma(p, shape, rate = 1, scale = 1/rate,
          lower.tail = TRUE, log.p = FALSE)
rinvgamma(n, shape, rate = 1, scale = 1/rate)
minvgamma(order, shape, rate = 1, scale = 1/rate)
levinvgamma(limit, shape, rate = 1, scale = 1/rate,
            order = 1)
mgfinvgamma(t, shape, rate =1, scale = 1/rate, log =FALSE)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
  \item{t}{numeric vector.}
}
\details{
  The inverse gamma distribution with parameters \code{shape} \eqn{=
    \alpha}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{u^\alpha e^{-u}}{x \Gamma(\alpha)}, %
    \quad u = \theta/x}{%
    f(x) = u^a exp(-u)/(x Gamma(a)), u = s/x}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0} and \eqn{\theta > 0}{s > 0}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The special case \code{shape == 1} is an
  \link[=dinvexp]{Inverse Exponential} distribution.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{k < \alpha}{k < shape}, and the \eqn{k}th
  limited moment at some limit \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X,
  d)^k]}, all \eqn{k}.

  The moment generating function is given by \eqn{E[e^{tX}]}.
}
\value{
  \code{dinvgamma} gives the density,
  \code{pinvgamma} gives the distribution function,
  \code{qinvgamma} gives the quantile function,
  \code{rinvgamma} generates random deviates,
  \code{minvgamma} gives the \eqn{k}th raw moment,
  \code{levinvgamma} gives the \eqn{k}th moment of the limited loss
  variable, and
  \code{mgfinvgamma} gives the moment generating function in \code{t}.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levinvgamma} computes the limited expected value using
  \code{gammainc} from package \pkg{expint}.

  Also known as the Vinci distribution. See also Kleiber and
  Kotz (2003) for alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvgamma(2, 3, 4, log = TRUE))
p <- (1:10)/10
pinvgamma(qinvgamma(p, 2, 3), 2, 3)
minvgamma(-1, 2, 2) ^ 2
levinvgamma(10, 2, 2, order = 1)
mgfinvgamma(-1, 3, 2)
}
\keyword{distribution}

\name{cm}
\alias{cm}
\alias{print.cm}
\alias{predict.cm}
\alias{summary.cm}
\alias{print.summary.cm}
\title{Credibility Models}
\description{
  Fit the following credibility models: \enc{Bühlmann}{Buhlmann},
  \enc{Bühlmann}{Buhlmann}-Straub, hierarchical, regression
  (Hachemeister) or linear Bayes.
}
\usage{
cm(formula, data, ratios, weights, subset,
   regformula = NULL, regdata, adj.intercept = FALSE,
   method = c("Buhlmann-Gisler", "Ohlsson", "iterative"),
   likelihood, ...,
   tol = sqrt(.Machine$double.eps), maxit = 100, echo = FALSE)

\method{print}{cm}(x, \dots)

\method{predict}{cm}(object, levels = NULL, newdata, \dots)

\method{summary}{cm}(object, levels = NULL, newdata, \dots)

\method{print}{summary.cm}(x, \dots)
}
\arguments{
  \item{formula}{character string \code{"bayes"} or an object of
    class \code{"\link[stats]{formula}"}: a symbolic description of the
    model to be fit. The details of model specification are given
    below.}
  \item{data}{a matrix or a data frame containing the portfolio
    structure, the ratios or claim amounts and their associated weights,
    if any.}
  \item{ratios}{expression indicating the columns of \code{data}
    containing the ratios or claim amounts.}
  \item{weights}{expression indicating the columns of \code{data}
    containing the weights associated with \code{ratios}.}
  \item{subset}{an optional logical expression indicating a subset of
    observations to be used in the modeling process. All observations
    are included by default.}
  \item{regformula}{an object of class \code{"\link[stats]{formula}"}:
    symbolic description of the regression component (see
    \code{\link[stats]{lm}} for details). No left hand side is needed
    in the formula; if present it is ignored. If \code{NULL}, no
    regression is done on the data.}
  \item{regdata}{an optional data frame, list or environment (or object
    coercible by \code{\link[base]{as.data.frame}} to a data frame)
    containing the variables in the regression model.}
  \item{adj.intercept}{if \code{TRUE}, the intercept of the regression
    model is located at the barycenter of the regressor instead of the
    origin.}
  \item{method}{estimation method for the variance components of the
    model; see Details.}
  \item{likelihood}{a character string giving the name of the likelihood
    function in one of the supported linear Bayes cases; see Details.}
  \item{tol}{tolerance level for the stopping criteria for iterative
    estimation method.}
  \item{maxit}{maximum number of iterations in iterative estimation
    method.}
  \item{echo}{logical; whether to echo the iterative procedure or not.}
  \item{x, object}{an object of class \code{"cm"}.}
  \item{levels}{character vector indicating the levels to predict or to
    include in the summary; if \code{NULL} all levels are included.}
  \item{newdata}{data frame containing the variables used to predict
    credibility regression models.}
  \item{\dots}{parameters of the prior distribution for \code{cm};
    additional attributes to attach to the result for the
    \code{predict} and \code{summary} methods; further arguments to
    \code{\link[base]{format}} for the \code{print.summary} method;
    unused for the \code{print} method.}
}
\details{
  \code{cm} is the unified front end for credibility models fitting. The
  function supports hierarchical models with any number of levels (with
  \enc{Bühlmann}{Buhlmann} and \enc{Bühlmann}{Buhlmann}-Straub models as
  special cases) and the regression model of Hachemeister. Usage of
  \code{cm} is similar to \code{\link[stats]{lm}} for these cases.
  \code{cm} can also fit linear Bayes models, in which case usage is
  much simplified; see the section on linear Bayes below.

  When not \code{"bayes"}, the \code{formula} argument symbolically
  describes the structure of the portfolio in the form \eqn{~ terms}.
  Each term is an interaction between risk factors contributing to the
  total variance of the portfolio data. Terms are separated by \code{+}
  operators and interactions within each term by \code{:}. For a
  portfolio divided first into sectors, then units and finally
  contracts, \code{formula} would be \code{~ sector + sector:unit +
  sector:unit:contract}, where \code{sector}, \code{unit} and
  \code{contract} are column names in \code{data}. In general, the
  formula should be of the form \code{~ a + a:b + a:b:c + a:b:c:d +
  ...}.

  If argument \code{regformula} is not \code{NULL}, the regression model
  of Hachemeister is fit to the data. The response is usually time. By
  default, the intercept of the model is located at time origin. If
  argument \code{adj.intercept} is \code{TRUE}, the intercept is moved
  to the (collective) barycenter of time, by orthogonalization of the
  design matrix. Note that the regression coefficients may be difficult
  to interpret in this case.

  Arguments \code{ratios}, \code{weights} and \code{subset} are used
  like arguments \code{select}, \code{select} and \code{subset},
  respectively, of function \code{\link[base]{subset}}.

  Data does not have to be sorted by level. Nodes with no data (complete
  lines of \code{NA} except for the portfolio structure) are allowed,
  with the restriction mentioned above.

  The \code{print} methods use the option \code{deparse.cutoff} to
  control the printing of the call to \code{cm}.
}
\section{Hierarchical models}{
  The credibility premium at one level is a convex combination between
  the linearly sufficient statistic of a node and the credibility
  premium of the level above. (For the first level, the complement of
  credibility is given to the collective premium.) The linearly
  sufficient statistic of a node is the credibility weighted average of
  the data of the node, except at the last level, where natural weights
  are used. The credibility factor of node \eqn{i} is equal to
  \deqn{\frac{w_i}{w_i + a/b},}{w[i]/(w[i] + a/b),}
  where \eqn{w_i}{w[i]} is the weight of the node used in the linearly
  sufficient statistic, \eqn{a} is the average within node variance and
  \eqn{b} is the average between node variance.
}
\section{Regression models}{
  The credibility premium of node \eqn{i} is equal to
  \deqn{y^\prime b_i^a,}{y' ba[i],}
  where \eqn{y} is a matrix created from \code{newdata} and
  \eqn{b_i^a}{ba[i]} is the vector of credibility adjusted regression
  coefficients of node \eqn{i}. The latter is given by
  \deqn{b_i^a = Z_i b_i + (I - Z_I) m,}{
    ba[i] = Z[i] b[i] + (I - Z[i]) m,}
  where \eqn{b_i}{b[i]} is the vector of regression coefficients based
  on data of node \eqn{i} only, \eqn{m} is the vector of collective
  regression coefficients, \eqn{Z_i}{Z[i]} is the credibility matrix and
  \eqn{I} is the identity matrix. The credibility matrix of node \eqn{i}
  is equal to
  \deqn{A^{-1} (A + s^2 S_i),}{A^(-1) (A + s2 S[i]),}
  where \eqn{S_i}{S[i]} is the unscaled regression covariance matrix of
  the node, \eqn{s^2}{s2} is the average within node variance and
  \eqn{A} is the within node covariance matrix.

  If the intercept is positioned at the barycenter of time, matrices
  \eqn{S_i}{S[i]} and \eqn{A} (and hence \eqn{Z_i}{Z[i]}) are diagonal.
  This amounts to use \enc{Bühlmann}{Buhlmann}-Straub models for each
  regression coefficient.

  Argument \code{newdata} provides the \dQuote{future} value of the
  regressors for prediction purposes. It should be given as specified in
  \code{\link[stats]{predict.lm}}.
}
\section{Variance components estimation}{
  For hierarchical models, two sets of estimators of the variance
  components (other than the within node variance) are available:
  unbiased estimators and iterative estimators.

  Unbiased estimators are based on sums of squares of the form
  \deqn{B_i = \sum_j w_{ij} (X_{ij} - \bar{X}_i)^2 - (J - 1) a}{%
    B[i] =  sum(j; w[ij] (X[ij] - Xb[i])^2 - (J - 1) a)}%
  and constants of the form
  \deqn{c_i = w_i - \sum_j \frac{w_{ij}^2}{w_i},}{%
    c[i] = w[i] - sum(j; w[ij]^2)/w[i],}%
  where \eqn{X_{ij}}{X[ij]} is the linearly sufficient statistic of
  level \eqn{(ij)}; \eqn{\bar{X_{i}}}{Xb[i]} is the weighted average of
  the latter using weights \eqn{w_{ij}}{w[ij]}; \eqn{w_i = \sum_j
  w_{ij}}{w[i] = sum(j; w[ij])}; \eqn{J} is the effective number of
  nodes at level \eqn{(ij)}; \eqn{a} is the within variance of this
  level. Weights \eqn{w_{ij}}{w[ij]} are the natural weights at the
  lowest level, the sum of the natural weights the next level and the
  sum of the credibility factors for all upper levels.

  The \enc{Bühlmann}{Buhlmann}-Gisler estimators (\code{method =
    "Buhlmann-Gisler"}) are given by%
  \deqn{b = \frac{1}{I} \sum_i \max \left( \frac{B_i}{c_i}, 0
    \right),}{%
    b = mean(max(B[i]/c[i], 0)),}%
  that is the average of the per node variance estimators truncated at
  0.

  The Ohlsson estimators (\code{method = "Ohlsson"}) are given by
  \deqn{b = \frac{\sum_i B_i}{\sum_i c_i},}{%
    b = sum(i; B[i]) / sum(i; c[i]),}%
  that is the weighted average of the per node variance estimators
  without any truncation. Note that negative estimates will be truncated
  to zero for credibility factor calculations.

  In the \enc{Bühlmann}{Buhlmann}-Straub model, these estimators are
  equivalent.

  Iterative estimators \code{method = "iterative"} are pseudo-estimators
  of the form
  \deqn{b = \frac{1}{d} \sum_i w_i (X_i - \bar{X})^2,}{%
    b = sum(i; w[i] * (X[i] - Xb)^2)/d,}
  where \eqn{X_i}{X[i]} is the linearly sufficient statistic of one
  level, \eqn{\bar{X}}{Xb} is the linearly sufficient statistic of
  the level above and \eqn{d} is the effective number of nodes at one
  level minus the effective number of nodes of the level above. The
  Ohlsson estimators are used as starting values.

  For regression models, with the intercept at time origin, only
  iterative estimators are available. If \code{method} is different from
  \code{"iterative"}, a warning is issued. With the intercept at the
  barycenter of time, the choice of estimators is the same as in the
  \enc{Bühlmann}{Buhlmann}-Straub model.
}
\section{Linear Bayes}{
  When \code{formula} is \code{"bayes"}, the function computes pure
  Bayesian premiums for the following combinations of distributions
  where they are linear credibility premiums:
  \itemize{
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Poisson}(\theta)}{X|\Theta ~ Poisson(\Theta)} and
      \eqn{\Theta \sim \mathrm{Gamma}(\alpha, \lambda)}{\Theta ~ Gamma(\alpha, \lambda)};}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Exponential}(\theta)}{X|\Theta ~ Exponential(\Theta)} and
      \eqn{\Theta \sim  \mathrm{Gamma}(\alpha, \lambda)}{\Theta ~ Gamma(\alpha, \lambda)};}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Gamma}(\tau, \theta)}{X|\Theta ~ Gamma(\tau, \Theta)} and
      \eqn{\Theta \sim  \mathrm{Gamma}(\alpha, \lambda)}{\Theta ~ Gamma(\alpha, \lambda)};}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Normal}(\theta, \sigma_2^2)}{X|\Theta ~ Normal(\Theta, \sigma_2^2)} and
      \eqn{\Theta \sim \mathrm{Normal}(\mu, \sigma_1^2)}{\Theta ~ Normal(\mu, \sigma_1^2)};}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Bernoulli}(\theta)}{X|\Theta ~ Bernoulli(\Theta)} and
      \eqn{\Theta \sim \mathrm{Beta}(a, b)}{\Theta ~ Beta(a, b)};}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Binomial}(\nu, \theta)}{X|\Theta ~ Binomial(\nu, \Theta)} and
      \eqn{\Theta \sim \mathrm{Beta}(a, b)}{\Theta ~ Beta(a, b)};}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Geometric}(\theta)}{X|\Theta = \theta ~ Geometric(\theta)} and
      \eqn{\Theta \sim \mathrm{Beta}(a, b)}{\Theta ~ Beta(a, b)}.}
    \item{\eqn{X|\Theta = \theta \sim \mathrm{Negative~Binomial}(r, \theta)}{X|\Theta ~ Negative Binomial(r, \Theta)} and
      \eqn{\Theta \sim \mathrm{Beta}(a, b)}{\Theta ~ Beta(a, b)}.}}

  The following combination is also supported:
  \eqn{X|\Theta = \theta \sim \mathrm{Single~Parameter~Pareto}(\theta)}{X|\Theta ~ Single Parameter Pareto(\Theta)}
  and \eqn{\Theta \sim \mathrm{Gamma}(\alpha, \lambda)}{\Theta ~
  Gamma(\alpha, \lambda)}. In this case, the Bayesian estimator not of
  the risk premium, but rather of parameter \eqn{\theta} is linear with
  a \dQuote{credibility} factor that is not restricted to \eqn{(0, 1)}.

  Argument \code{likelihood} identifies the distribution of \eqn{X|\Theta
    = \theta} as one of
  \code{"poisson"},
  \code{"exponential"},
  \code{"gamma"},
  \code{"normal"},
  \code{"bernoulli"},
  \code{"binomial"},
  \code{"geometric"},
  \code{"negative binomial"} or
  \code{"pareto"}.

  The parameters of the distributions of \eqn{X|\Theta = \theta} (when
  needed) and \eqn{\Theta} are set in \code{\dots} using the argument
  names (and default values) of \code{\link[stats]{dgamma}},
  \code{\link[stats]{dnorm}}, \code{\link[stats]{dbeta}},
  \code{\link[stats]{dbinom}}, \code{\link[stats]{dnbinom}} or
  \code{dpareto1}, as appropriate. For the Gamma/Gamma case, use
  \code{shape.lik} for the shape parameter \eqn{\tau} of the Gamma
  likelihood. For the Normal/Normal case, use \code{sd.lik} for the
  standard error \eqn{\sigma_2} of the Normal likelihood.

  Data for the linear Bayes case may be a matrix or data frame as usual;
  an atomic vector to fit the model to a single contract; missing or
  \code{NULL} to fit the prior model. Arguments \code{ratios},
  \code{weights} and \code{subset} are ignored.
}
\value{
  Function \code{cm} computes the structure parameters estimators of the
  model specified in \code{formula}. The value returned is an object of
  class \code{cm}.

  An object of class \code{"cm"} is a list with at least the following
  components:
  \item{means}{a list containing, for each level, the vector of linearly
    sufficient statistics.}
  \item{weights}{a list containing, for each level, the vector of total
    weights.}
  \item{unbiased}{a vector containing the unbiased variance components
    estimators, or \code{NULL}.}
  \item{iterative}{a vector containing the iterative variance components
    estimators, or \code{NULL}.}
  \item{cred}{for multi-level hierarchical models: a list containing,
    the vector of credibility factors for each level. For one-level
    models: an array or vector of credibility factors.}
  \item{nodes}{a list containing, for each level, the vector of the
    number of nodes in the level.}
  \item{classification}{the columns of \code{data} containing the
    portfolio classification structure.}
  \item{ordering}{a list containing, for each level, the affiliation of
    a node to the node of the level above.}

  Regression fits have in addition the following components:
  \item{adj.models}{a list containing, for each node, the credibility
    adjusted regression model as obtained with
    \code{\link[stats]{lm.fit}} or \code{\link[stats]{lm.wfit}}.}
  \item{transition}{if \code{adj.intercept} is \code{TRUE}, a transition
    matrix from the basis of the orthogonal design matrix to the basis
    of the original design matrix.}
  \item{terms}{the \code{\link[stats]{terms}} object used.}

  The method of \code{predict} for objects of class \code{"cm"} computes
  the credibility premiums for the nodes of every level included in
  argument \code{levels} (all by default). Result is a list the same
  length as \code{levels} or the number of levels in \code{formula}, or
  an atomic vector for one-level models.
}
\references{
  \enc{Bühlmann}{Buhlmann}, H. and Gisler, A. (2005), \emph{A Course in
    Credibility Theory and its Applications}, Springer.

  Belhadj, H., Goulet, V. and Ouellet, T. (2009), On parameter
  estimation in hierarchical credibility, \emph{Astin Bulletin} \bold{39}.

  Goulet, V. (1998), Principles and application of credibility
  theory, \emph{Journal of Actuarial Practice} \bold{6}, ISSN
  1064-6647.

  Goovaerts, M. J. and Hoogstad, W. J. (1987), \emph{Credibility Theory},
  Surveys of Actuarial Studies, No. 4, Nationale-Nederlanden N.V.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Xavier Milhaud, Tommy Ouellet, Louis-Philippe Pouliot
}
\seealso{
  \code{\link[base]{subset}}, \code{\link[stats]{formula}},
  \code{\link[stats]{lm}}, \code{\link[stats]{predict.lm}}.
}
\examples{
data(hachemeister)

## Buhlmann-Straub model
fit <- cm(~state, hachemeister,
          ratios = ratio.1:ratio.12, weights = weight.1:weight.12)
fit				# print method
predict(fit)			# credibility premiums
summary(fit)			# more details

## Two-level hierarchical model. Notice that data does not have
## to be sorted by level
X <- data.frame(unit = c("A", "B", "A", "B", "B"), hachemeister)
fit <- cm(~unit + unit:state, X, ratio.1:ratio.12, weight.1:weight.12)
predict(fit)
predict(fit, levels = "unit")	# unit credibility premiums only
summary(fit)
summary(fit, levels = "unit")	# unit summaries only

## Regression model with intercept at time origin
fit <- cm(~state, hachemeister,
          regformula = ~time, regdata = data.frame(time = 12:1),
          ratios = ratio.1:ratio.12, weights = weight.1:weight.12)
fit
predict(fit, newdata = data.frame(time = 0))
summary(fit, newdata = data.frame(time = 0))

## Same regression model, with intercept at barycenter of time
fit <- cm(~state, hachemeister, adj.intercept = TRUE,
          regformula = ~time, regdata = data.frame(time = 12:1),
          ratios = ratio.1:ratio.12, weights = weight.1:weight.12)
fit
predict(fit, newdata = data.frame(time = 0))
summary(fit, newdata = data.frame(time = 0))

## Poisson/Gamma pure Bayesian model
fit <- cm("bayes", data = c(5, 3, 0, 1, 1),
          likelihood = "poisson", shape = 3, rate = 3)
fit
predict(fit)
summary(fit)

## Normal/Normal pure Bayesian model
cm("bayes", data = c(5, 3, 0, 1, 1),
   likelihood = "normal", sd.lik = 2,
   mean = 2, sd = 1)
}
\keyword{models}

\name{InversePareto}
\alias{InversePareto}
\alias{dinvpareto}
\alias{pinvpareto}
\alias{qinvpareto}
\alias{rinvpareto}
\alias{minvpareto}
\alias{levinvpareto}
\title{The Inverse Pareto Distribution}
\description{
  Density function, distribution function, quantile function, random generation
  raw moments and limited moments for the Inverse Pareto distribution
  with parameters \code{shape} and \code{scale}.
}
\usage{
dinvpareto(x, shape, scale, log = FALSE)
pinvpareto(q, shape, scale, lower.tail = TRUE, log.p = FALSE)
qinvpareto(p, shape, scale, lower.tail = TRUE, log.p = FALSE)
rinvpareto(n, shape, scale)
minvpareto(order, shape, scale)
levinvpareto(limit, shape, scale, order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape, scale}{parameters. Must be strictly positive.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The inverse Pareto distribution with parameters \code{shape} \eqn{=
    \tau}{= a} and \code{scale} \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\tau \theta x^{\tau - 1}}{%
      (x + \theta)^{\tau + 1}}}{%
    f(x) = a s x^(a - 1)/(x + s)^(a + 1)}
  for \eqn{x > 0}, \eqn{\tau > 0}{a > 0} and \eqn{\theta > 0}{s > 0}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}, \eqn{-\tau < k < 1}{-shape < k < 1}.

  The \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]},
  \eqn{k > -\tau}{k > -shape}.
}
\value{
  \code{dinvpareto} gives the density,
  \code{pinvpareto} gives the distribution function,
  \code{qinvpareto} gives the quantile function,
  \code{rinvpareto} generates random deviates,
  \code{minvpareto} gives the \eqn{k}th raw moment, and
  \code{levinvpareto} calculates the \eqn{k}th limited moment.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  Evaluation of \code{levinvpareto} is done using numerical integration.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvpareto(2, 3, 4, log = TRUE))
p <- (1:10)/10
pinvpareto(qinvpareto(p, 2, 3), 2, 3)
minvpareto(0.5, 1, 2)
}
\keyword{distribution}

\name{Burr}
\alias{Burr}
\alias{dburr}
\alias{pburr}
\alias{qburr}
\alias{rburr}
\alias{mburr}
\alias{levburr}
\title{The Burr Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Burr distribution with
  parameters \code{shape1}, \code{shape2} and \code{scale}.
}
\usage{
dburr(x, shape1, shape2, rate = 1, scale = 1/rate,
      log = FALSE)
pburr(q, shape1, shape2, rate = 1, scale = 1/rate,
      lower.tail = TRUE, log.p = FALSE)
qburr(p, shape1, shape2, rate = 1, scale = 1/rate,
      lower.tail = TRUE, log.p = FALSE)
rburr(n, shape1, shape2, rate = 1, scale = 1/rate)
mburr(order, shape1, shape2, rate = 1, scale = 1/rate)
levburr(limit, shape1, shape2, rate = 1, scale = 1/rate,
        order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The Burr distribution with parameters \code{shape1} \eqn{=
    \alpha}{= a}, \code{shape2} \eqn{= \gamma}{= b} and \code{scale}
  \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\alpha \gamma (x/\theta)^\gamma}{%
      x [1 + (x/\theta)^\gamma]^{\alpha + 1}}}{%
    f(x) = (a b (x/s)^b)/(x [1 + (x/s)^b]^(a + 1))}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0}, \eqn{\gamma > 0}{b > 0}
  and \eqn{\theta > 0}{s > 0}.

  The Burr is the distribution of the random variable
  \deqn{\theta \left(\frac{X}{1 - X}\right)^{1/\gamma},}{%
    s (X/(1 - X))^(1/b),}
  where \eqn{X} has a beta distribution with parameters \eqn{1}
  and \eqn{\alpha}{a}.

  The Burr distribution has the following special cases:
  \itemize{
    \item A \link[=dllogis]{Loglogistic} distribution when \code{shape1
      == 1};
    \item A \link[=dparalogis]{Paralogistic} distribution when
    \code{shape2 == shape1};
    \item A \link[=dpareto]{Pareto} distribution when \code{shape2 ==
      1}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{-\gamma < k < \alpha\gamma}{-shape2 < k <
  shape1 * shape2}.

  The \eqn{k}th limited moment at some limit \eqn{d} is \eqn{E[\min(X,
  d)^k]}{E[min(X, d)^k]}, \eqn{k > -\gamma}{k > -shape2}
  and \eqn{\alpha - k/\gamma}{shape1 - k/shape2} not a negative integer.
}
\value{
  \code{dburr} gives the density,
  \code{pburr} gives the distribution function,
  \code{qburr} gives the quantile function,
  \code{rburr} generates random deviates,
  \code{mburr} gives the \eqn{k}th raw moment, and
  \code{levburr} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levburr} computes the limited expected value using
  \code{\link{betaint}}.

  Distribution also known as the Burr Type XII or Singh-Maddala
  distribution. See also Kleiber and Kotz (2003) for alternative names
  and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dpareto4}} for an equivalent distribution with a location
  parameter.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dburr(1, 2, 3, log = TRUE))
p <- (1:10)/10
pburr(qburr(p, 2, 3, 2), 2, 3, 2)

## variance
mburr(2, 2, 3, 1) - mburr(1, 2, 3, 1) ^ 2

## case with shape1 - order/shape2 > 0
levburr(10, 2, 3, 1, order = 2)

## case with shape1 - order/shape2 < 0
levburr(10, 1.5, 0.5, 1, order = 2)
}
\keyword{distribution}

\name{ruin}
\alias{ruin}
\alias{plot.ruin}
\title{Probability of Ruin}
\description{
  Calulation of infinite time probability of ruin in the models of
  \enc{Cramér}{Cramer}-Lundberg and Sparre Andersen, that is with
  exponential or phase-type (including mixtures of exponentials, Erlang
  and mixture of Erlang) claims interarrival time.
}
\usage{
ruin(claims = c("exponential", "Erlang", "phase-type"), par.claims,
     wait = c("exponential", "Erlang", "phase-type"), par.wait,
     premium.rate = 1, tol = sqrt(.Machine$double.eps),
     maxit = 200L, echo = FALSE)

\method{plot}{ruin}(x, from = NULL, to = NULL, add = FALSE,
     xlab = "u", ylab = expression(psi(u)),
     main = "Probability of Ruin", xlim = NULL, \dots)
}
\arguments{
  \item{claims}{character; the type of claim severity distribution.}
  \item{wait}{character; the type of claim interarrival (wait) time
    distribution.}
  \item{par.claims, par.wait}{named list containing the parameters of
    the distribution; see Details.}
  \item{premium.rate}{numeric vector of length 1; the premium rate.}
  \item{tol, maxit, echo}{respectively the tolerance level of the
    stopping criteria, the maximum number of iterations and whether or
    not to echo the procedure when the transition rates matrix is
    determined iteratively. Ignored if \code{wait = "exponential"}.}
  \item{x}{an object of class \code{"ruin"}.}
  \item{from, to}{the range over which the function will be plotted.}
  \item{add}{logical; if \code{TRUE} add to already existing plot.}
  \item{xlim}{numeric of length 2; if specified, it serves as default
    for \code{c(from, to)}.}
  \item{xlab, ylab}{label of the x and y axes, respectively.}
  \item{main}{main title.}
  \item{\dots}{further graphical parameters accepted by
    \code{\link[graphics]{curve}}.}
}
\details{
  The names of the parameters in \code{par.claims} and \code{par.wait}
  must be the same as in \code{\link[stats]{dexp}},
  \code{\link[stats]{dgamma}} or \code{\link{dphtype}}, as appropriate.
  A model will be a mixture of exponential or Erlang distributions (but
  not phase-type) when the parameters are vectors of length \eqn{> 1}
  and the parameter list contains a vector \code{weights} of the
  coefficients of the mixture.

  Parameters are recycled when needed. Their names can be abbreviated.

  Combinations of exponentials as defined in Dufresne and Gerber (1988)
  are \emph{not} supported.

  Ruin probabilities are evaluated using \code{\link{pphtype}} except
  when both distributions are exponential, in which case an explicit
  formula is used.

  When \code{wait != "exponential"} (Sparre Andersen model), the
  transition rate matrix \eqn{\boldsymbol{Q}}{Q} of the distribution of
  the probability of ruin is determined iteratively using a fixed
  point-like algorithm. The stopping criteria used is%
  \deqn{\max \left\{
    \sum_{j = 1}^n |\boldsymbol{Q}_{ij} - \boldsymbol{Q}_{ij}^\prime|
    \right\} < \code{tol},}{%
    max(rowSum(|Q - Q'|)) < tol,}%
  where \eqn{\boldsymbol{Q}}{Q} and \eqn{\boldsymbol{Q}^\prime}{Q'} are
  two successive values of the matrix.
}
\value{
  A function of class \code{"ruin"} inheriting from the
  \code{"function"} class to compute the probability of ruin given
  initial surplus levels. The function has arguments:
  \item{u}{numeric vector of initial surplus levels;}
  \item{survival}{logical; if \code{FALSE} (default), probabilities are
    \eqn{\psi(u)}{psi(u)}, otherwise, \eqn{\phi(u) = 1 - \psi(u)}{phi(u)
      = 1 - psi(u)};}
  \item{lower.tail}{an alias for \code{!survival}.}
}
\references{
  Asmussen, S. and Rolski, T. (1991), Computational methods in risk theory:
  A matrix algorithmic approach, \emph{Insurance: Mathematics and
    Economics} \bold{10}, 259--274.

  Dufresne, F. and Gerber, H. U. (1988), Three methods to calculate
  the probability of ruin, \emph{Astin Bulletin} \bold{19}, 71--90.

  Gerber, H. U. (1979), \emph{An Introduction to Mathematical Risk
    Theory}, Huebner Foundation.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  and Christophe Dutang
}
\examples{
## Case with an explicit formula: exponential claims and exponential
## interarrival times.
psi <- ruin(claims = "exponential", par.claims = list(rate = 5),
            wait   = "exponential", par.wait   = list(rate = 3))
psi
psi(0:10)
plot(psi, from = 0, to = 10)

## Mixture of two exponentials for claims, exponential interarrival
## times (Gerber 1979)
psi <- ruin(claims = "exponential",
            par.claims = list(rate = c(3, 7), w = 0.5),
            wait   = "exponential",
            par.wait   = list(rate = 3), pre = 1)
u <- 0:10
psi(u)
(24 * exp(-u) + exp(-6 * u))/35	# same

## Mixtures of Erlang distributions for claims and interarrival times.
psi <- ruin(claims = 'Erlang',
            par.claims = list(shape = c(1, 2, 3),
                              scale = c(3, 2, 1),
                              weights = c(3, 4, 1)/8),
            wait = 'Erlang',
            par.wait = list(shape = c(6, 4, 4),
                            scale= c(3, 2, 2),
                            weights = c(3, 4, 1)/8),
            premium.rate = 0.6)
psi(0:10)

## Phase-type claims, exponential interarrival times (Asmussen and
## Rolski 1991)
p <- c(0.5614, 0.4386)
r <- matrix(c(-8.64, 0.101, 1.997, -1.095), 2, 2)
lambda <- 1/(1.1 * mphtype(1, p, r))
psi <- ruin(claims = "p", par.claims = list(prob = p, rates = r),
            wait   = "e", par.wait   = list(rate = lambda))
psi
plot(psi, xlim = c(0, 50))

## Phase-type claims, mixture of two exponentials for interarrival times
## (Asmussen and Rolski 1991)
a <- (0.4/5 + 0.6) * lambda
ruin(claims = "p", par.claims = list(prob = p, rates = r),
     wait   = "e", par.wait   = list(rate = c(5 * a, a), weights =
                                     c(0.4, 0.6)),
     maxit = 225L)
}
\keyword{models}

\name{ZeroTruncatedGeometric}
\alias{ZeroTruncatedGeometric}
\alias{ZTGeometric}
\alias{dztgeom}
\alias{pztgeom}
\alias{qztgeom}
\alias{rztgeom}
\title{The Zero-Truncated Geometric Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Truncated Geometric distribution with
  parameter \code{prob}.
}
\usage{
dztgeom(x, prob, log = FALSE)
pztgeom(q, prob, lower.tail = TRUE, log.p = FALSE)
qztgeom(p, prob, lower.tail = TRUE, log.p = FALSE)
rztgeom(n, prob)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{prob}{parameter. \code{0 < prob <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-truncated geometric distribution with \code{prob} \eqn{= p}
  has probability mass function
  \deqn{%
    p(x) = p (1-p)^{x - 1}}{%
    p(x) = p (1-p)^(x-1)}
  for \eqn{x = 1, 2, \ldots} and \eqn{0 < p < 1}, and
  \eqn{p(1) = 1} when \eqn{p = 1}.
  The cumulative distribution function is
  \deqn{P(x) = \frac{F(x) - F(0)}{1 - F(0)},}{%
    P(x) = [F(x) - F(0)]/[1 - F(0)],}
  where \eqn{F(x)} is the distribution function of the standard geometric.

  The mean is \eqn{1/p} and the variance is \eqn{(1-p)/p^2}.

  In the terminology of Klugman et al. (2012), the zero-truncated
  geometric is a member of the \eqn{(a, b, 1)} class of
  distributions with \eqn{a = 1-p} and \eqn{b = 0}.

  If an element of \code{x} is not integer, the result of
  \code{dztgeom} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{P(x) \ge p}, where \eqn{P} is the distribution function.
}
\value{
  \code{dztgeom} gives the (log) probability mass function,
  \code{pztgeom} gives the (log) distribution function,
  \code{qztgeom} gives the quantile function, and
  \code{rztgeom} generates random deviates.

  Invalid \code{prob} will result in return value \code{NaN}, with a
  warning.

  The length of the result is determined by \code{n} for
  \code{rztgeom}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}ztgeom} use \code{\{d,p,q\}geom} for all but
  the trivial input values and \eqn{p(0)}.

  \code{rztgeom} uses the simple inversion algorithm suggested by
  Peter Dalgaard on the r-help mailing list on 1 May 2005 %
  (\url{https://stat.ethz.ch/pipermail/r-help/2005-May/070680.html}).
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dgeom}} for the geometric distribution.

  \code{\link{dztnbinom}} for the zero-truncated negative binomial, of
  which the zero-truncated geometric is a special case.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
p <- 1/(1 + 0.5)
dztgeom(c(1, 2, 3), prob = p)
dgeom(c(1, 2, 3), p)/pgeom(0, p, lower = FALSE) # same
dgeom(c(1, 2, 3) - 1, p)                        # same

pztgeom(1, prob = 1)        # point mass at 1

qztgeom(pztgeom(1:10, 0.3), 0.3)
}
\keyword{distribution}

\name{actuar-package}
\alias{actuar-package}
\alias{actuar}
\docType{package}
\title{\packageTitle{actuar}}
\description{\packageDescription{actuar}}
\details{
  \pkg{actuar} provides additional actuarial science
  functionality and support for heavy tailed distributions to the \R
  statistical system.

  The current feature set of the package can be split into five main
  categories.
  \enumerate{
    \item{Additional probability distributions: 23 continuous heavy
      tailed distributions from the Feller-Pareto and Transformed
      Gamma families, the loggamma, the Gumbel, the inverse Gaussian
      and the generalized beta; phase-type distributions; the
      Poisson-inverse Gaussian discrete distribution; zero-truncated and
      zero-modified extensions of the standard discrete distributions;
      computation of raw moments, limited moments and the moment
      generating function (when it exists) of continuous distributions.
      See the \dQuote{distributions} package vignette for details.}
    \item{Loss distributions modeling: extensive support of grouped
      data; functions to compute empirical raw and limited moments;
      support for minimum distance estimation using three different
      measures; treatment of coverage modifications (deductibles,
      limits, inflation, coinsurance). See the \dQuote{modeling} and
      \dQuote{coverage} package vignettes for details.}
    \item{Risk and ruin theory: discretization of the claim amount
      distribution; calculation of the aggregate claim amount
      distribution; calculation of the adjustment coefficient;
      calculation of the probability of ruin, including using phase-type
      distributions. See the \dQuote{risk} package vignette for details.}
    \item{Simulation of discrete mixtures, compound models (including
      the compound Poisson), and compound hierarchical models. See the
      \dQuote{simulation} package vignette for details.}
    \item{Credibility theory: function \code{cm} fits hierarchical
      (including Bühlmann, Bühlmann-Straub), regression and linear Bayes
      credibility models. See the \dQuote{credibility} package vignette
      for details.}
  }
}
\author{
  Christophe Dutang, Vincent Goulet, Mathieu Pigeon and many other
  contributors; use \code{packageDescription("actuar")} for the complete
  list.

  Maintainer: Vincent Goulet.
}
\references{
  Dutang, C., Goulet, V. and Pigeon, M. (2008).
  actuar: An R Package for Actuarial Science.
  \emph{Journal of Statistical Software}, \bold{25}(7), 1--37.
  \doi{10.18637/jss.v025.i07}.
  
  Dutang, C., Goulet, V., Langevin, N. (2022).
  Feller-Pareto and Related Distributions: Numerical Implementation and Actuarial
  Applications.
  \emph{Journal of Statistical Software}, \bold{103}(6), 1--22.
  \doi{10.18637/jss.v103.i06}.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  For probability distributions support functions, use as starting
  points:
  \code{\link{FellerPareto}}, 
  \code{\link{TransformedGamma}}, 
  \code{\link{Loggamma}}, 
  \code{\link{Gumbel}}, 
  \code{\link{InverseGaussian}}, 
  \code{\link{PhaseType}}, 
  \code{\link{PoissonInverseGaussian}} and, e.g.,
  \code{\link{ZeroTruncatedPoisson}}, 
  \code{\link{ZeroModifiedPoisson}}.

  For loss modeling support functions:
  \code{\link{grouped.data}},
  \code{\link{ogive}},
  \code{\link{emm}},
  \code{\link{elev}},
  \code{\link{mde}},
  \code{\link{coverage}}.

  For risk and ruin theory functions:
  \code{\link{discretize}},
  \code{\link{aggregateDist}},
  \code{\link{adjCoef}},
  \code{\link{ruin}}.

  For credibility theory functions and datasets:
  \code{\link{cm}},
  \code{\link{hachemeister}}.
}
\examples{
## The package comes with extensive demonstration scripts;
## use the following command to obtain the list.
\dontrun{demo(package = "actuar")}
}
\keyword{package}
\keyword{distribution}
\keyword{models}
\keyword{univar}

\name{grouped.data}
\alias{grouped.data}
\title{Grouped data}
\description{
  Creation of grouped data objects, from either a provided set of group
  boundaries and group frequencies, or from individual data using
  automatic or specified breakpoints.
}
\usage{
grouped.data(\dots, breaks = "Sturges", include.lowest = TRUE,
             right = TRUE, nclass = NULL, group = FALSE,
             row.names = NULL, check.rows = FALSE,
             check.names = TRUE)
}
\arguments{
  \item{\dots}{arguments of the form \code{value} or \code{tag = value};
    see Details.}
  \item{breaks}{same as for \code{\link{hist}}, namely one of:
    \itemize{
      \item{a vector giving the breakpoints between groups;}
      \item{a function to compute the vector of breakpoints;}
      \item{a single number giving the number of groups;}
      \item{a character string naming an algorithm to compute the
      number of groups (see \code{\link{hist}});}
      \item{a function to compute the number of groups.}
    }
    In the last three cases the number is a suggestion only; the
    breakpoints will be set to \code{\link{pretty}} values.  If
    \code{breaks} is a function, the first element in \code{\dots}
    is supplied to it as the only argument.
  }
  \item{include.lowest}{logical; if \code{TRUE}, a data point equal to
    the \code{breaks} value will be included in the first (or last, for
    \code{right = FALSE}) group. Used only for individual data; see
    Details.}
  \item{right}{logical; indicating if the intervals should be closed on
    the right (and open on the left) or vice versa.}
  \item{nclass}{numeric (integer); equivalent to \code{breaks} for a
    scalar or character argument.}
  \item{group}{logical; an alternative way to force grouping of
    individual data.}
  \item{row.names, check.rows, check.names}{arguments identical to those
    of \code{\link{data.frame}}.}
}
\details{
  A grouped data object is a special form of data frame consisting of
  one column of contiguous group boundaries and one or more columns of
  frequencies within each group.

  The function can create a grouped data object from two types of
  arguments.
  \enumerate{
    \item{Group boundaries and frequencies. This is the default mode of
      operation if the call has at least two elements in \code{\dots}.

      The first argument will then be taken as the vector of group
      boundaries. This vector must be exactly one element longer than
      the other arguments, which will be taken as vectors of group
      frequencies. All arguments are coerced to data frames.}
    \item{Individual data. This mode of operation is active if there
      is a single argument in \code{\dots}, or if either \code{breaks}
      or \code{nclass} is specified or \code{group} is \code{TRUE}.

      Arguments of \code{\dots} are first grouped using
      \code{\link{hist}}. If needed, breakpoints are set using the first
      argument.}
  }

  Missing (\code{NA}) frequencies are replaced by zeros, with a
  warning.

  Extraction and replacement methods exist for \code{grouped.data}
  objects, but working on non adjacent groups will most likely yield
  useless results.
}
\value{
  An object of \code{class} \code{c("grouped.data", "data.frame")} with
  an environment containing the vector \code{cj} of group boundaries.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\seealso{
  \code{\link{[.grouped.data}} for extraction and replacement methods.

  \code{\link{data.frame}} for usual data frame creation and
  manipulation.

  \code{\link{hist}} for details on the calculation of breakpoints.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca},
  Mathieu Pigeon and Louis-Philippe Pouliot
}
\examples{
## Most common usage using a predetermined set of group
## boundaries and group frequencies.
cj <- c(0, 25, 50, 100, 250, 500, 1000)
nj <- c(30, 31, 57, 42, 45, 10)
(x <- grouped.data(Group = cj, Frequency = nj))
class(x)

x[, 1] # group boundaries
x[, 2] # group frequencies

## Multiple frequency columns are supported
x <- sample(1:100, 9)
y <- sample(1:100, 9)
grouped.data(cj = 1:10, nj.1 = x, nj.2 = y)

## Alternative usage with grouping of individual data.
grouped.data(x)                         # automatic breakpoints
grouped.data(x, breaks = 7)             # forced number of groups
grouped.data(x, breaks = c(0,25,75,100))    # specified groups
grouped.data(x, y, breaks = c(0,25,75,100)) # multiple data sets

\dontrun{## Providing two or more data sets and automatic breakpoints is
## very error-prone since the range of the first data set has to
## include the ranges of all the other data sets.
range(x)
range(y)
grouped.data(x, y, group = TRUE)}
}
\keyword{classes}
\keyword{methods}

\name{InverseExponential}
\alias{InverseExponential}
\alias{dinvexp}
\alias{pinvexp}
\alias{qinvexp}
\alias{rinvexp}
\alias{minvexp}
\alias{levinvexp}
\title{The Inverse Exponential Distribution}
\description{
  Density function, distribution function, quantile function, random generation
  raw moments and limited moments for the Inverse Exponential
  distribution with parameter \code{scale}.
}
\usage{
dinvexp(x, rate = 1, scale = 1/rate, log = FALSE)
pinvexp(q, rate = 1, scale = 1/rate, lower.tail = TRUE, log.p = FALSE)
qinvexp(p, rate = 1, scale = 1/rate, lower.tail = TRUE, log.p = FALSE)
rinvexp(n, rate = 1, scale = 1/rate)
minvexp(order, rate = 1, scale = 1/rate)
levinvexp(limit, rate = 1, scale = 1/rate, order)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{scale}{parameter. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The inverse exponential distribution with parameter \code{scale}
  \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\theta e^{-\theta/x}}{x^2}}{f(x) = s exp(-s/x)/x^2}
  for \eqn{x > 0} and \eqn{\theta > 0}{s > 0}.

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]}, \eqn{k < 1}, and the \eqn{k}th limited moment at
  some limit \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]}, all
  \eqn{k}.
}
\value{
  \code{dinvexp} gives the density,
  \code{pinvexp} gives the distribution function,
  \code{qinvexp} gives the quantile function,
  \code{rinvexp} generates random deviates,
  \code{minvexp} gives the \eqn{k}th raw moment, and
  \code{levinvexp} calculates the \eqn{k}th limited moment.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  \code{levinvexp} computes the limited expected value using
  \code{gammainc} from package \pkg{expint}.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dinvexp(2, 2, log = TRUE))
p <- (1:10)/10
pinvexp(qinvexp(p, 2), 2)
minvexp(0.5, 2)
}
\keyword{distribution}

\name{betaint}
\alias{betaint}
\title{The \dQuote{Beta Integral}}
\description{
  The \dQuote{beta integral} is just a multiple of the non regularized
  incomplete beta function. This function provides an R interface to the
  C level routine. It is not exported by the package.
}
\usage{
betaint(x, a, b)
}
\arguments{
  \item{x}{vector of quantiles.}
  \item{a, b}{parameters. See Details for admissible values.}
}
\details{
  Function \code{betaint} computes the \dQuote{beta integral}
  \deqn{
    B(a, b; x) = \Gamma(a + b) \int_0^x t^{a-1} (1-t)^{b-1} dt}{%
    B(a, b; x) = Gamma(a + b) int_0^x t^(a-1) (1-t)^(b-1) dt}
  for \eqn{a > 0}, \eqn{b \neq -1, -2, \ldots}{b != -1, -2, \ldots} and
  \eqn{0 < x < 1}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)
  When \eqn{b > 0},
  \deqn{
    B(a, b; x) = \Gamma(a) \Gamma(b) I_x(a, b),}
  where \eqn{I_x(a, b)} is \code{pbeta(x, a, b)}. When \eqn{b < 0},
  \eqn{b \neq -1, -2, \ldots}{b != -1, -2, \ldots}, and \eqn{a > 1 +
  [-b]}{a > 1 + floor(-b)},
  \deqn{%
    \begin{array}{rcl}
      B(a, b; x)
      &=& \displaystyle
          -\Gamma(a + b) \left[ \frac{x^{a-1} (1-x)^b}{b}
          + \frac{(a-1) x^{a-2} (1-x)^{b+1}}{b (b+1)} \right. \\
      & & \displaystyle\left.
	  + \cdots + \frac{(a-1) \cdots (a-r) x^{a-r-1}
	  (1-x)^{b+r}}{b (b+1) \cdots (b+r)} \right] \\
      & &  \displaystyle
  	  + \frac{(a-1) \cdots (a-r-1)}{b (b+1) \cdots (b+r)}
	  \Gamma(a-r-1) \\
      & & \times \Gamma(b+r+1) I_x(a-r-1, b+r+1),
      \end{array}}{%
    B(a, b; x)
    = -Gamma(a+b) \{(x^(a-1) (1-x)^b)/b
        + [(a-1) x^(a-2) (1-x)^(b+1)]/[b(b+1)]
        + \dots
	+ [(a-1)\dots(a-r) x^(a-r-1) (1-x)^(b+r)]/[b(b+1)\dots(b+r)]\}
      + [(a-1)\dots(a-r-1)]/[b(b+1)\dots(b+r)] Gamma(a-r-1)
        * Gamma(b+r+1) I_x(a-r-1, b+r+1),}
  where \eqn{r = [-b]}{r = floor(-b)}.

  This function is used (at the C level) to compute the
  limited expected value for distributions of the transformed beta
  family; see, for example, \code{\link{levtrbeta}}.
}
\value{
  The value of the integral.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  The need for this function in the package is well explained in the
  introduction of Appendix A of Klugman et al. (2012). See also chapter
  6 and 15 of Abramowitz and Stegun (1972) for definitions and relations
  to the hypergeometric series.
}
\references{
  Abramowitz, M. and Stegun, I. A. (1972), \emph{Handbook of Mathematical
  Functions}, Dover.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
x <- 0.3
a <- 7

## case with b > 0
b <- 2
actuar:::betaint(x, a, b)
gamma(a) * gamma(b) * pbeta(x, a, b)    # same

## case with b < 0
b <- -2.2
r <- floor(-b)        # r = 2
actuar:::betaint(x, a, b)

## "manual" calculation
s <- (x^(a-1) * (1-x)^b)/b +
    ((a-1) * x^(a-2) * (1-x)^(b+1))/(b * (b+1)) +
    ((a-1) * (a-2) * x^(a-3) * (1-x)^(b+2))/(b * (b+1) * (b+2))
-gamma(a+b) * s +
    (a-1)*(a-2)*(a-3) * gamma(a-r-1)/(b*(b+1)*(b+2)) *
    gamma(b+r+1)*pbeta(x, a-r-1, b+r+1)
}
\keyword{math}
\keyword{distribution}

\name{hachemeister}
\docType{data}
\alias{hachemeister}
\title{Hachemeister Data Set}
\description{
  Hachemeister (1975) data set giving average claim amounts in private
  passenger bodily injury insurance in five U.S. states over 12 quarters
  between July 1970 and June 1973 and the corresponding number of
  claims.
}
\usage{hachemeister}
\format{
  A matrix with 5 rows and the following 25 columns:
  \describe{
    \item{\code{state}}{the state number;}
    \item{\code{ratio.1}, \dots, \code{ratio.12}}{the average claim
      amounts;}
    \item{\code{weight.1}, \dots, \code{weight.12}}{the
      corresponding number of claims.}
  }
}
\source{
  Hachemeister, C. A. (1975), \emph{Credibility for regression
    models with application to trend}, Proceedings of the Berkeley
  Actuarial Research Conference on Credibility, Academic Press.
}
\keyword{datasets}

\name{TransformedGamma}
\alias{TransformedGamma}
\alias{dtrgamma}
\alias{ptrgamma}
\alias{qtrgamma}
\alias{rtrgamma}
\alias{mtrgamma}
\alias{levtrgamma}
\title{The Transformed Gamma Distribution}
\description{
  Density function, distribution function, quantile function, random generation,
  raw moments and limited moments for the Transformed Gamma distribution
  with parameters \code{shape1}, \code{shape2} and \code{scale}.
}
\usage{
dtrgamma(x, shape1, shape2, rate = 1, scale = 1/rate,
         log = FALSE)
ptrgamma(q, shape1, shape2, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
qtrgamma(p, shape1, shape2, rate = 1, scale = 1/rate,
         lower.tail = TRUE, log.p = FALSE)
rtrgamma(n, shape1, shape2, rate = 1, scale = 1/rate)
mtrgamma(order, shape1, shape2, rate = 1, scale = 1/rate)
levtrgamma(limit, shape1, shape2, rate = 1, scale = 1/rate,
           order = 1)
}
\arguments{
  \item{x, q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{shape1, shape2, scale}{parameters. Must be strictly positive.}
  \item{rate}{an alternative way to specify the scale.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities/densities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}{P[X <= x]}, otherwise, \eqn{P[X > x]}.}
  \item{order}{order of the moment.}
  \item{limit}{limit of the loss variable.}
}
\details{
  The transformed gamma distribution with parameters \code{shape1} \eqn{=
    \alpha}{= a}, \code{shape2} \eqn{= \tau}{= b} and \code{scale}
  \eqn{= \theta}{= s} has density:
  \deqn{f(x) = \frac{\tau u^\alpha e^{-u}}{x \Gamma(\alpha)}, %
      \quad u = (x/\theta)^\tau}{%
    f(x) = b u^a exp(-u) / (x Gamma(a)), u = (x/s)^b}
  for \eqn{x > 0}, \eqn{\alpha > 0}{a > 0}, \eqn{\tau > 0}{b > 0}
  and \eqn{\theta > 0}{s > 0}.
  (Here \eqn{\Gamma(\alpha)}{Gamma(a)} is the function implemented
  by \R's \code{\link{gamma}()} and defined in its help.)

  The transformed gamma is the distribution of the random variable
  \eqn{\theta X^{1/\tau},}{s X^(1/b),}
  where \eqn{X} has a gamma distribution with shape parameter
  \eqn{\alpha}{a} and scale parameter \eqn{1} or, equivalently, of the
  random variable
  \eqn{Y^{1/\tau}}{Y^(1/b)}
  with \eqn{Y} a gamma distribution with shape parameter \eqn{\alpha}{a}
  and scale parameter \eqn{\theta^\tau}{s^b}.

  The transformed gamma probability distribution defines a family of
  distributions with the following special cases:
  \itemize{
    \item A \link[=dgamma]{Gamma} distribution when \code{shape2 == 1};
    \item A \link[=dweibull]{Weibull} distribution when \code{shape1 ==
      1};
    \item An \link[=dexp]{Exponential} distribution when \code{shape2 ==
      shape1 == 1}.
  }

  The \eqn{k}th raw moment of the random variable \eqn{X} is
  \eqn{E[X^k]}{E[X^k]} and the \eqn{k}th limited moment at some limit
  \eqn{d} is \eqn{E[\min(X, d)^k]}{E[min(X, d)^k]}, \eqn{k >
  -\alpha\tau}{k > -shape1 * shape2}.
}
\value{
  \code{dtrgamma} gives the density,
  \code{ptrgamma} gives the distribution function,
  \code{qtrgamma} gives the quantile function,
  \code{rtrgamma} generates random deviates,
  \code{mtrgamma} gives the \eqn{k}th raw moment, and
  \code{levtrgamma} gives the \eqn{k}th moment of the limited loss
  variable.

  Invalid arguments will result in return value \code{NaN}, with a warning.
}
\note{
  Distribution also known as the Generalized Gamma. See also Kleiber and
  Kotz (2003) for alternative names and parametrizations.

  The \code{"distributions"} package vignette provides the
  interrelations between the continuous size distributions in
  \pkg{actuar} and the complete formulas underlying the above functions.
}
\references{
  Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions
  in Economics and Actuarial Sciences}, Wiley.

  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
exp(dtrgamma(2, 3, 4, 5, log = TRUE))
p <- (1:10)/10
ptrgamma(qtrgamma(p, 2, 3, 4), 2, 3, 4)
mtrgamma(2, 3, 4, 5) - mtrgamma(1, 3, 4, 5) ^ 2
levtrgamma(10, 3, 4, 5, order = 2)
}
\keyword{distribution}

\name{ZeroModifiedLogarithmic}
\alias{ZeroModifiedLogarithmic}
\alias{ZMLogarithmic}
\alias{dzmlogarithmic}
\alias{pzmlogarithmic}
\alias{qzmlogarithmic}
\alias{rzmlogarithmic}
\title{The Zero-Modified Logarithmic Distribution}
\description{
  Density function, distribution function, quantile function and random
  generation for the Zero-Modified Logarithmic (or log-series)
  distribution with parameter \code{prob} and arbitrary probability at
  zero \code{p0}.
}
\usage{
dzmlogarithmic(x, prob, p0, log = FALSE)
pzmlogarithmic(q, prob, p0, lower.tail = TRUE, log.p = FALSE)
qzmlogarithmic(p, prob, p0, lower.tail = TRUE, log.p = FALSE)
rzmlogarithmic(n, prob, p0)
}
\arguments{
  \item{x}{vector of (strictly positive integer) quantiles.}
  \item{q}{vector of quantiles.}
  \item{p}{vector of probabilities.}
  \item{n}{number of observations. If \code{length(n) > 1}, the length is
    taken to be the number required.}
  \item{prob}{parameter. \code{0 <= prob < 1}.}
  \item{p0}{probability mass at zero. \code{0 <= p0 <= 1}.}
  \item{log, log.p}{logical; if \code{TRUE}, probabilities
    \eqn{p} are returned as \eqn{\log(p)}{log(p)}.}
  \item{lower.tail}{logical; if \code{TRUE} (default), probabilities are
    \eqn{P[X \le x]}, otherwise, \eqn{P[X > x]}.}
}
\details{
  The zero-modified logarithmic distribution with \code{prob} \eqn{= p}
  and \code{p0} \eqn{= p_0}{= p0} is a discrete mixture between a
  degenerate distribution at zero and a (standard) logarithmic. The
  probability mass function is \eqn{p(0) = p_0}{p(0) = p0} and
  \deqn{%
    p(x) = (1-p_0) f(x)}{p(x) = (1-p0) f(x)}
  for \eqn{x = 1, 2, \ldots}, \eqn{0 < p < 1} and \eqn{0 \le
  p_0 \le 1}{0 \le p0 \le 1}, where \eqn{f(x)} is the probability mass
  function of the logarithmic.
  The cumulative distribution function is
  \deqn{P(x) = p_0 + (1 - p_0) F(x)}{P(x) = p0 + (1 - p0) F(x).}

  The special case \code{p0 == 0} is the standard logarithmic.

  The zero-modified logarithmic distribution is the limiting case of the
  zero-modified negative binomial distribution with \code{size}
  parameter equal to \eqn{0}. Note that in this context, parameter
  \code{prob} generally corresponds to the probability of \emph{failure}
  of the zero-truncated negative binomial.

  If an element of \code{x} is not integer, the result of
  \code{dzmlogarithmic} is zero, with a warning.

  The quantile is defined as the smallest value \eqn{x} such that
  \eqn{F(x) \ge p}, where \eqn{F} is the distribution function.
}
\value{
  \code{dzmlogarithmic} gives the probability mass function,
  \code{pzmlogarithmic} gives the distribution function,
  \code{qzmlogarithmic} gives the quantile function, and
  \code{rzmlogarithmic} generates random deviates.

  Invalid \code{prob} or \code{p0} will result in return value
  \code{NaN}, with a warning.

  The length of the result is determined by \code{n} for
  \code{rzmlogarithmic}, and is the maximum of the lengths of the
  numerical arguments for the other functions.
}
\note{
  Functions \code{\{d,p,q\}zmlogarithmic} use
  \code{\{d,p,q\}logarithmic} for all but the trivial input values and
  \eqn{p(0)}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
  \code{\link{dlogarithmic}} for the logarithmic distribution.

  \code{\link{dztnbinom}} for the zero modified negative binomial
  distribution.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca}
}
\examples{
p <- 1/(1 + 0.5)
dzmlogarithmic(1:5, prob = p, p0 = 0.6)
(1-0.6) * dlogarithmic(1:5, p)/plogarithmic(0, p, lower = FALSE) # same

## simple relation between survival functions
pzmlogarithmic(0:5, p, p0 = 0.2, lower = FALSE)
(1-0.2) * plogarithmic(0:5, p, lower = FALSE)/plogarithmic(0, p, lower = FALSE) # same

qzmlogarithmic(pzmlogarithmic(0:10, 0.3, p0 = 0.6), 0.3, p0 = 0.6)
}
\keyword{distribution}

\name{mde}
\alias{Mde}
\alias{mde}
\title{Minimum Distance Estimation}
\description{
  Minimum distance fitting of univariate distributions, allowing
  parameters to be held fixed if desired.
}
\usage{
mde(x, fun, start, measure = c("CvM", "chi-square", "LAS"),
    weights = NULL, ...)
}
\arguments{
  \item{x}{a vector or an object of class \code{"grouped data"} (in
    which case only the first column of frequencies is used).}
  \item{fun}{function returning a cumulative distribution (for
    \code{measure = "CvM"} and \code{measure = "chi-square"}) or a
    limited expected value (for \code{measure = "LAS"}) evaluated at
    its first argument.}
  \item{start}{a named list giving the parameters to be optimized with
    initial values}
  \item{measure}{either \code{"CvM"} for the Cramer-von Mises method,
    \code{"chi-square"} for the modified chi-square method, or \code{"LAS"}
    for the layer average severity method.}
  \item{weights}{weights; see Details.}
  \item{\dots}{Additional parameters, either for \code{fun} or for
    \code{optim}. In particular, it can be used to specify bounds via
    \code{lower} or \code{upper} or both.  If arguments of \code{fun}
    are included they will be held fixed.}
}
\details{
  The Cramer-von Mises method (\code{"CvM"}) minimizes the squared
  difference between the theoretical cdf and the empirical cdf at the
  data points (for individual data) or the ogive at the knots (for
  grouped data).

  The modified chi-square method (\code{"chi-square"}) minimizes the
  modified chi-square statistic for grouped data, that is the squared
  difference between the expected and observed frequency within each
  group.

  The layer average severity method (\code{"LAS"}) minimizes the
  squared difference between the theoretical and empirical limited
  expected value within each group for grouped data.

  All sum of squares can be weighted. If arguments \code{weights} is
  missing, weights default to 1 for \code{measure = "CvM"} and
  \code{measure = "LAS"}; for \code{measure = "chi-square"}, weights
  default to \eqn{1/n_j}{1/n[j]}, where \eqn{n_j}{n[j]} is the frequency
  in group \eqn{j = 1, \dots, r}.
  
  Optimization is performed using \code{\link{optim}}. For
  one-dimensional problems the Nelder-Mead method is used and for
  multi-dimensional problems the BFGS method, unless arguments named
  \code{lower} or \code{upper} are supplied when \code{L-BFGS-B} is used
  or \code{method} is supplied explicitly.
}
\value{
  An object of class \code{"mde"}, a list with two components:
  
  \item{estimate}{the parameter estimates, and}  
  \item{distance}{the distance.}
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (1998),
  \emph{Loss Models, From Data to Decisions}, Wiley.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Mathieu Pigeon
}
\examples{
## Individual data example
data(dental)
mde(dental, pexp, start = list(rate = 1/200), measure = "CvM")

## Example 2.21 of Klugman et al. (1998)
data(gdental)
mde(gdental, pexp, start = list(rate = 1/200), measure = "CvM")
mde(gdental, pexp, start = list(rate = 1/200), measure = "chi-square")
mde(gdental, levexp, start = list(rate = 1/200), measure = "LAS")

## Two-parameter distribution example
try(mde(gdental, ppareto, start = list(shape = 3, scale = 600),
        measure = "CvM")) # no convergence

## Working in log scale often solves the problem
pparetolog <- function(x, shape, scale)
    ppareto(x, exp(shape), exp(scale))

( p <- mde(gdental, pparetolog, start = list(shape = log(3),
           scale = log(600)), measure = "CvM") )
exp(p$estimate)
}
\keyword{distribution}
\keyword{htest}

\name{aggregateDist}
\alias{aggregateDist}
\alias{print.aggregateDist}
\alias{plot.aggregateDist}
\alias{summary.aggregateDist}
\alias{mean.aggregateDist}
\alias{diff.aggregateDist}
\title{Aggregate Claim Amount Distribution}
\description{
  Compute the aggregate claim amount cumulative distribution function of
  a portfolio over a period using one of five methods.
}
\usage{
aggregateDist(method = c("recursive", "convolution", "normal",
                         "npower", "simulation"),
              model.freq = NULL, model.sev = NULL, p0 = NULL,
              x.scale = 1, convolve = 0, moments, nb.simul, \dots,
              tol = 1e-06, maxit = 500, echo = FALSE)

\method{print}{aggregateDist}(x, \dots)

\method{plot}{aggregateDist}(x, xlim, ylab = expression(F[S](x)),
     main = "Aggregate Claim Amount Distribution",
     sub = comment(x), \dots)

\method{summary}{aggregateDist}(object, \dots)

\method{mean}{aggregateDist}(x, \dots)

\method{diff}{aggregateDist}(x, \dots)
}
\arguments{
  \item{method}{method to be used}
  \item{model.freq}{for \code{"recursive"} method: a character string
    giving the name of a distribution in the \eqn{(a, b, 0)} or \eqn{(a,
      b, 1)} families of distributions. For \code{"convolution"} method:
    a vector of claim number probabilities. For \code{"simulation"}
    method: a frequency simulation model (see \code{\link{rcomphierarc}} for
    details) or \code{NULL}. Ignored with \code{normal} and
    \code{npower} methods.}
  \item{model.sev}{for \code{"recursive"} and \code{"convolution"}
    methods: a vector of claim amount probabilities. For
    \code{"simulation"} method: a severity simulation model (see
    \code{\link{rcomphierarc}} for details) or \code{NULL}. Ignored with
    \code{normal} and \code{npower} methods.}
  \item{p0}{arbitrary probability at zero for the frequency
    distribution. Creates a zero-modified or zero-truncated
    distribution if not \code{NULL}. Used only with \code{"recursive"}
    method.}
  \item{x.scale}{value of an amount of 1 in the severity model (monetary
    unit). Used only with \code{"recursive"} and \code{"convolution"}
    methods.}
  \item{convolve}{number of times to convolve the resulting distribution
    with itself. Used only with \code{"recursive"} method.}
  \item{moments}{vector of the true moments of the aggregate claim
    amount distribution; required only by the \code{"normal"} or
    \code{"npower"} methods.}
  \item{nb.simul}{number of simulations for the \code{"simulation"} method.}
  \item{\dots}{parameters of the frequency distribution for the
    \code{"recursive"} method; further arguments to be passed to or
    from other methods otherwise.}
  \item{tol}{the resulting cumulative distribution in the
    \code{"recursive"} method will get less than \code{tol} away from
    1.}
  \item{maxit}{maximum number of recursions in the \code{"recursive"}
    method.}
  \item{echo}{logical; echo the recursions to screen in the
    \code{"recursive"} method.}
  \item{x, object}{an object of class \code{"aggregateDist"}.}
  \item{xlim}{numeric of length 2; the \eqn{x} limits of the plot.}
  \item{ylab}{label of the y axis.}
  \item{main}{main title.}
  \item{sub}{subtitle, defaulting to the calculation method.}
}
\details{
  \code{aggregateDist} returns a function to compute the cumulative
  distribution function (cdf) of the aggregate claim amount distribution
  in any point.

  The \code{"recursive"} method computes the cdf using the Panjer
  algorithm; the \code{"convolution"} method using convolutions; the
  \code{"normal"} method using a normal approximation; the
  \code{"npower"} method using the Normal Power 2 approximation; the
  \code{"simulation"} method using simulations. More details follow.
}
\section{Recursive method}{
  The frequency distribution must be a member of the \eqn{(a, b, 0)} or
  \eqn{(a, b, 1)} families of discrete distributions.

  To use a distribution from the \eqn{(a, b, 0)} family,
  \code{model.freq} must be one of
  \code{"binomial"},
  \code{"geometric"},
  \code{"negative binomial"} or
  \code{"poisson"},
  and \code{p0} must be \code{NULL}.

  To use a zero-truncated distribution from the \eqn{(a, b, 1)} family,
  \code{model.freq} may be one of the strings above together with
  \code{p0 = 0}. As a shortcut, \code{model.freq} may also be one of
  \code{"zero-truncated binomial"},
  \code{"zero-truncated geometric"},
  \code{"zero-truncated negative binomial"},
  \code{"zero-truncated poisson"} or
  \code{"logarithmic"},
  and \code{p0} is then ignored (with a warning if non \code{NULL}).

  (Note: since the logarithmic distribution is always zero-truncated.
  \code{model.freq = "logarithmic"} may be used with either \code{p0 =
  NULL} or \code{p0 = 0}.)

  To use a zero-modified distribution from the \eqn{(a, b, 1)} family,
  \code{model.freq} may be one of standard frequency distributions
  mentioned above with \code{p0} set to some probability that the
  distribution takes the value \eqn{0}. It is equivalent, but more
  explicit, to set \code{model.freq} to one of
  \code{"zero-modified binomial"},
  \code{"zero-modified geometric"},
  \code{"zero-modified negative binomial"},
  \code{"zero-modified poisson"} or
  \code{"zero-modified logarithmic"}.

  The parameters of the frequency distribution must be specified using
  names identical to the arguments of the appropriate function
  \code{\link{dbinom}}, \code{\link{dgeom}}, \code{\link{dnbinom}},
  \code{\link{dpois}} or \code{\link{dlogarithmic}}. In the latter case,
  do take note that the parametrization of \code{dlogarithmic} is
  different from Appendix B of Klugman et al. (2012).

  If the length of \code{p0} is greater than one, only the first element
  is used, with a warning.

  \code{model.sev} is a vector of the (discretized) claim amount
  distribution \eqn{X}; the first element \strong{must} be \eqn{f_X(0) =
    \Pr[X = 0]}{fx(0) = Pr[X = 0]}.

  The recursion will fail to start if the expected number of claims is
  too large. One may divide the appropriate parameter of the frequency
  distribution by \eqn{2^n} and convolve the resulting distribution
  \eqn{n =} \code{convolve} times.

  Failure to obtain a cumulative distribution function less than
  \code{tol} away from 1 within \code{maxit} iterations is often due
  to too coarse a discretization of the severity distribution.
}
\section{Convolution method}{
  The cumulative distribution function (cdf) \eqn{F_S(x)}{Fs(x)} of the
  aggregate claim amount of a portfolio in the collective risk model is
  \deqn{F_S(x) = \sum_{n = 0}^{\infty} F_X^{*n}(x) p_n,}{%
    Fs(x) = sum(n; Fx^\{*n\}(x) * pn)}
  for \eqn{x = 0, 1, \dots}; \eqn{p_n = \Pr[N = n]}{pn = Pr[N = n]} is
  the frequency probability mass function and
  \eqn{F_X^{*n}(x)}{Fx^\{*n\}(x)} is the cdf of the \eqn{n}th convolution of
  the (discrete) claim amount random variable.

  \code{model.freq} is vector \eqn{p_n}{pn} of the number of claims
  probabilities; the first element \strong{must} be \eqn{\Pr[N =
  0]}{Pr[N = 0]}.

  \code{model.sev} is vector \eqn{f_X(x)}{fx(x)} of the (discretized)
  claim amount distribution; the first element \strong{must} be
  \eqn{f_X(0)}{fx(0)}.
}
\section{Normal and Normal Power 2 methods}{
  The Normal approximation of a cumulative distribution function (cdf)
  \eqn{F(x)} with mean \eqn{\mu}{m} and standard deviation
  \eqn{\sigma}{s} is
  \deqn{F(x) \approx \Phi\left( \frac{x - \mu}{\sigma} \right).}{%
    F(x) ~= pnorm((x - m)/s).}

  The Normal Power 2 approximation of a cumulative distribution function (cdf)
  \eqn{F(x)} with mean \eqn{\mu}{m}, standard deviation \eqn{\sigma}{s}
  and skewness \eqn{\gamma}{g} is
  \deqn{F(x) \approx \Phi \left(%
    -\frac{3}{\gamma} + \sqrt{\frac{9}{\gamma^2} + 1 %
      + \frac{6}{\gamma} \frac{x - \mu}{\sigma}} \right).}{%
    F(x) ~= pnorm(-3/g + sqrt(9/g^2 + 1 + (6/g) * (x - m)/s)).}
  This formula is valid only for the right-hand tail of the distribution
  and skewness should not exceed unity.
}
\section{Simulation method}{
  This methods returns the empirical distribution function of a sample
  of size \code{nb.simul} of the aggregate claim amount distribution
  specified by \code{model.freq} and
  \code{model.sev}. \code{\link{rcomphierarc}} is used for the simulation of
  claim amounts, hence both the frequency and severity models can be
  mixtures of distributions.
}
\value{
  A function of class \code{"aggregateDist"}, inheriting from the
  \code{"function"} class when using normal and Normal Power
  approximations and additionally inheriting from the \code{"ecdf"} and
  \code{"stepfun"} classes when other methods are used.

  There are methods available to summarize (\code{summary}), represent
  (\code{print}), plot (\code{plot}), compute quantiles
  (\code{quantile}) and compute the mean (\code{mean}) of
  \code{"aggregateDist"} objects.

  For the \code{diff} method: a numeric vector of probabilities
  corresponding to the probability mass function evaluated
  at the knots of the distribution.
}
\seealso{
  \code{\link{discretize}} to discretize a severity distribution;
  \code{\link{mean.aggregateDist}} to compute the mean of the
  distribution;
  \code{\link{quantile.aggregateDist}} to compute the quantiles or the
  Value-at-Risk;
  \code{\link{CTE.aggregateDist}} to compute the Conditional Tail
  Expectation (or Tail Value-at-Risk);
  \code{\link{rcomphierarc}}.
}
\references{
  Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012),
  \emph{Loss Models, From Data to Decisions, Fourth Edition}, Wiley.

  Daykin, C.D., \enc{Pentikäinen}{Pentikainen}, T. and Pesonen, M.
  (1994), \emph{Practical Risk Theory for Actuaries}, Chapman & Hall.
}
\author{
  Vincent Goulet \email{vincent.goulet@act.ulaval.ca} and
  Louis-Philippe Pouliot
}
\examples{
## Convolution method (example 9.5 of Klugman et al. (2012))
fx <- c(0, 0.15, 0.2, 0.25, 0.125, 0.075,
        0.05, 0.05, 0.05, 0.025, 0.025)
pn <- c(0.05, 0.1, 0.15, 0.2, 0.25, 0.15, 0.06, 0.03, 0.01)
Fs <- aggregateDist("convolution", model.freq = pn,
                    model.sev = fx, x.scale = 25)
summary(Fs)
c(Fs(0), diff(Fs(25 * 0:21))) # probability mass function
plot(Fs)

## Recursive method (example 9.10 of Klugman et al. (2012))
fx <- c(0, crossprod(c(2, 1)/3,
                     matrix(c(0.6, 0.7, 0.4, 0, 0, 0.3), 2, 3)))
Fs <- aggregateDist("recursive", model.freq = "poisson",
                    model.sev = fx, lambda = 3)
plot(Fs)
Fs(knots(Fs))		      # cdf evaluated at its knots
diff(Fs)                      # probability mass function

## Recursive method (high frequency)
fx <- c(0, 0.15, 0.2, 0.25, 0.125, 0.075,
        0.05, 0.05, 0.05, 0.025, 0.025)
\dontrun{Fs <- aggregateDist("recursive", model.freq = "poisson",
                    model.sev = fx, lambda = 1000)}
Fs <- aggregateDist("recursive", model.freq = "poisson",
                    model.sev = fx, lambda = 250, convolve = 2, maxit = 1500)
plot(Fs)

## Recursive method (zero-modified distribution; example 9.11 of
## Klugman et al. (2012))
Fn <- aggregateDist("recursive", model.freq = "binomial",
                    model.sev = c(0.3, 0.5, 0.2), x.scale = 50,
                    p0 = 0.4, size = 3, prob = 0.3)
diff(Fn)

## Equivalent but more explicit call
aggregateDist("recursive", model.freq = "zero-modified binomial",
              model.sev = c(0.3, 0.5, 0.2), x.scale = 50,
              p0 = 0.4, size = 3, prob = 0.3)

## Recursive method (zero-truncated distribution). Using 'fx' above
## would mean that both Pr[N = 0] = 0 and Pr[X = 0] = 0, therefore
## Pr[S = 0] = 0 and recursions would not start.
fx <- discretize(pexp(x, 1), from = 0, to = 100, method = "upper")
fx[1L] # non zero
aggregateDist("recursive", model.freq = "zero-truncated poisson",
              model.sev = fx, lambda = 3, x.scale = 25, echo=TRUE)

## Normal Power approximation
Fs <- aggregateDist("npower", moments = c(200, 200, 0.5))
Fs(210)

## Simulation method
model.freq <- expression(data = rpois(3))
model.sev <- expression(data = rgamma(100, 2))
Fs <- aggregateDist("simulation", nb.simul = 1000,
                    model.freq, model.sev)
mean(Fs)
plot(Fs)

## Evaluation of ruin probabilities using Beekman's formula with
## Exponential(1) claim severity, Poisson(1) frequency and premium rate
## c = 1.2.
fx <- discretize(pexp(x, 1), from = 0, to = 100, method = "lower")
phi0 <- 0.2/1.2
Fs <- aggregateDist(method = "recursive", model.freq = "geometric",
                    model.sev = fx, prob = phi0)
1 - Fs(400)			# approximate ruin probability
u <- 0:100
plot(u, 1 - Fs(u), type = "l", main = "Ruin probability")
}
\keyword{distribution}
\keyword{models}
